!** filters.f is a driver for running Kalman covariance, U-D and SRIF filters.
!** and is called by program Tst_IMU_filters.
!** Filters is general-purpose and may be used with other models.
!** author B. Gibbs  12/2009

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!**   using Intel compiler:
!**        in ifort.cfg
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort filters.f
!**************************************

      module filterMod;
      use filterPrec, only: real_kind1;   !precision (4 or 8)
      !** contains global variables used in defining Kalman filter, UD and SRIF
      implicit none;
      integer(4) nstate;        !total number of states in model (set by model)
      integer(4) ncore;         !number of core states in model
      integer(4) nMarkov;       !number of markov states in model
      integer(4) nbias;         !number of bias states in model

      integer(4) nest;          !total number of estimated states
      integer(4) nConsider;     !total number of consider parameters
      integer(4) nUnadj;        !total number of unadjusted-analyze parameters
      integer(4) ncoreE;        !number of estimated core states
      integer(4) nMarkE;        !number of estimated Markov states
      integer(4) nbiasE;        !number of estimated bias states
      integer(4) nDynE;         !number of estimated core+markov states

      integer(4) nmeas;         !total number of vector meas in input file
      integer(4) :: method_udtime =1;!method used for ud time update: 1=wgs, 2=rank-1, 3=square phi*u*d*u^T*phi^T  ###
      integer(4) :: KFmethod =2;   !Kalman filter (covariance) measurement update method: 1=short, 2=Bierman;s Joseph, 3=full Joseph ###

      integer(4),parameter :: mmaxi =10;  !maximum number of allowed scalar measurements at one time point (10)

      integer(4),allocatable :: mapxe(:);!mapping from estimated states (nest) to total model states
      integer(4),allocatable :: mapxu(:);!mapping from unadjusted analyze parameter (nUnadj) to total model states
      integer(4),allocatable :: mapxei(:);!mapping from total model states (nstate) to estimated states
      integer(4),allocatable :: mapxui(:);!mapping from total model states (nstate) to estimated states

!      character(80) setupFile   !name of file containing data for structure xstate
!      character(80) measFile    !name of file containing measurement data
      character(20),allocatable :: xnames(:); !names of total states as set by model routines

      logical(4) :: sparse =.true.;  !true = use sparse multiplies  !###
      logical(4) :: debugp =.true.;  !generates debug print
      logical(4) :: first(3) =.true.;!initializes filters
      logical(4) :: runFilt(3) =.true.; !denotes whether Kalman covariance, U-D and SRIF filters are to be run

      real(8) tepoch;           !epoch time of x0 and p0
      real(8) tend;             !end time for filter processing
      real(8) rsum(3);          !residual sum-of-squares for 3 filters

      real(real_kind1),allocatable :: xfu(:),ud(:),xfk(:),pf(:,:),xfs(:);
      real(real_kind1),allocatable :: xfa(:),pfa(:,:),rsrif(:);
      real(8),allocatable :: dxsum(:,:),sigsum(:,:);

!      real(8) rnoise(mmaxi,mmaxi)           !meas noise covariance

      type states;
        integer(4) status;      !0=ignore, 1=unadjusted analyze, 2=consider, 3=estimate
        integer(4) xtype;       !1=core state, 2=1 or 2 order Markov, 3=bias
        character(20) xname;    !name of state that must match name in hard-coded model
        real(8) x0;             !a priori estimate at epoch time
        real(8) sig0;           !1-sigma uncertainty of x0 at epoch time
        real(8) qs;             !PSD of driving white noise
        real(8) const(2);       !known constants associated with this state in dynamic or measurement model
      end type;
      type (states),allocatable :: xstate(:);

      type meass;
        integer(4) mtype;       !meas type (must match types in model)
        integer(4) m;           !number of scalar meas included in this meas
        character(20) mname;    !name of meas for output (may include sensor name)
        real(8) t;              !meas time
        real(8) y(mmaxi);        !up to mmax scalar meas values
        real(8) sigr(mmaxi);     !1-sigma meas noise for mmax scalar meas
      end type;
!      type (meass),allocatable :: meas(:)

      contains;
!*****************************************************
      subroutine filters (phit,qt,ht,y,sigr,t,tlst,xtr,xfr,xsigr,       &
     &                     m,mdim,n,ndim,ndyn);
      !** filters is called from tst_tai to process each measurement set.
      !** standard Kalman, UD and SRIF filters are run in parallel and outputs
      !** are compared with a reference solution (from file) generated by real(8) UD filter.
      !** Filters can run in single or double precision (real_kind1)

      use filterPrec, only: real_kind1;   !precision (4 or 8)
      implicit none;
      integer(4),intent(in) :: m,mdim,n,ndim,ndyn;
      real(8),intent(in) :: t,tlst;
      real(8),intent(in) :: xtr(9);       !true (simulated) values of 9 states (for error print)
      real(8),intent(inout) :: xfr(n);    !accurate state estimate from double precision UD filter
      real(8),intent(inout) :: xsigr(n);  !accurate 1-sigmas from double precision UD filter
      real(real_kind1),intent(in) :: phit(ndim,n),qt(ndim,n);
      real(real_kind1),intent(in) :: sigr(m);
      real(8),intent(in) :: y(m),ht(mdim,n);

      integer(4) i;
      real(8) dt;
      real(real_kind1) tf,xsigk(n),xsigu(n),xsigs(n);
      character(40) pltFile;
      !_______________________________

      if (ndim /=nstate) call errhand ("filters error: ndim /= nstate");
      if (n /= nstate) call errhand ("filters error: n /= nstate");
      if (nest /= nstate) call errhand ("filters error: nest /= nstate");
      if (mdim > mmaxi) call errhand ("filters error: mdim > mmaxi");

      !** filter time update
      dt =t-tlst;
      tf =t;
      if (debugp) write (*,'(/"t =",g13.5,", y=",3g13.5)') t, y(:m);     !sigr =0.002, ht =I (3x3)

!*** standard Kalman covariance filter
      if (runFilt(1)) then;
        call KF (xsigk,  xfk,pf,                                        &
     &         phit,qt,ht,y,sigr,tf,xtr,m,mdim,n,ndim,ndyn,sparse);
        if (debugp) then;
          write (11,1000) t, 304.8*xfk(:6), xfk(7:9);
          write (12,1000) t, 304.8*(xfk(:6) -xfr(:6)),                  & !pos/vel in mm
     &                   xfk(7:9) -xfr(7:9);       !tilt in arc-sec
          write (14,1000) t, 304.8*xsigk(:6), xsigk(7:9);   !pos/vel in mm
        endif;
 1000   format (50g13.5)
        dxsum(:n,1) =dxsum(:n,1) +(xfk(:n) -xfr(:n))**2;    !pos/vel in mm
        sigsum(:n,1) =sigsum(:n,1) +(xsigk(:n) -xsigr(:n))**2;
      endif;

!*** UD filter
      if (runFilt(2)) then;
        call udKF (xsigu,  xfu,ud,                                      &
     &           phit,qt,ht,y,sigr,tf,m,mdim,n,ndim,ndyn,sparse);
        if (debugp) then;
          write (12,1000) t, 304.8*(xfu(:6) -xfr(:6)),                  & !pos/vel in mm
     &                   xfu(7:9) -xfr(7:9);       !tilt in arc-sec
          write (14,1000) t, 304.8*xsigu(:6), xsigu(7:9);   !pos/vel in mm
          do i=1,n;
            if (abs(xsigk(i)-xsigu(i)) > 1.d-9)                         &
     &        write (*,'(i3,", xsigk-xsigu =",g13.5)')                  &
     &        i,xsigk(i)-xsigu(i);
          enddo;
        endif;
!        write (*,'("xsigk-xsigu =",3(/10g13.5))') xsigk(:n)-xsigu(:n)
        dxsum(:n,2) =dxsum(:n,2) +(xfu(:n) -xfr(:n))**2;    !pos/vel in mm
        sigsum(:n,2) =sigsum(:n,2) +(xsigu(:n) -xsigr(:n))**2;
      endif;

!*** SRIF filter
      if (runFilt(3)) then;
        call srif (xsigs,   xfs,rsrif,                                  &
     &           phit,qt,ht,y,sigr,tf,m,mdim,n,ndim,ndyn,sparse);
        if (debugp) then;
          write (12,1000) t, 304.8*(xfs(:6) -xfr(:6)),                  & !pos/vel in mm
     &                   xfs(7:9) -xfr(7:9);       !tilt in arc-sec
          write (14,1000) t, 304.8*xsigs(:6), xsigs(7:9);   !pos/vel in mm
          write (*,'("xsigk-xsigs =",3(/10g13.5))') xsigk(:n)-xsigs(:n);
        endif;
        dxsum(:n,3) =dxsum(:n,3) +(xfs(:n) -xfr(:n))**2;    !pos/vel in mm
        sigsum(:n,3) =sigsum(:n,3) +(xsigs(:n) -xsigr(:n))**2;
      endif;

!      pltFile ="filterSig.plt"
!      call write_results (t,pltFile)

      return;
      end subroutine filters;
!*****************************************************
      subroutine write_results (t,pltFileName);

      implicit none;
      real(8),intent(in) :: t;   !meas time
      character(*),intent(in) :: pltFileName;

      integer(4) i,j;
      logical(4),save :: first =.true.;
      character(20) :: labels(5,nest);
      real(8) sigx(5,nest);
      !_______________________________

      if (first) then;
        open (10,file=pltFileName,status='unknown');
        first =.false.;
        !**
        do i=1,nest;
          j =mapxe(i);
          labels(1,i) ="sigx-"//xstate(j)%xname(:15);
        enddo;

        write (10,1000) "t (s)", (trim(labels(1,i)),i=1,nest);
 1000   format('variables =',200('"',a,'", '))
      endif;

      do i=1,nest;
        sigx(1,i) =sqrt(pf(i,i));
      enddo;
      sigx(1,1:3) =0.3048*1.d3*sigx(1,1:3);    !change units from ft to mm
      sigx(1,4:6) =0.3048*1.d3*sigx(1,4:6);    !change units from ft to microm/s
      write (10,1100) t,(sigx(1,i),i=1,nest);
 1100 format (200g13.5)

      return;
      end subroutine write_results;
!*****************************************************
      subroutine KF_init (setupFile,nstateM,ncoreM,nMarkovM,nbiasM,     &
     &                     sig0,xnameM);
      implicit none;
      integer(4),intent(in) :: nstateM,ncoreM,nMarkovM,nbiasM;
      character(*),intent(in) :: setupFile, xnameM(nstateM);
      real(8),intent(in) :: sig0(nstateM);

      integer(4) ier,i,j,k;
!      real(8) t
      !_______________________________

      !** get model size
      nstate =nstateM;
      ncore =ncoreM;
      nmarkov =nMarkovM;
      nbias =nbiasM;
      allocate (mapxe(nState),mapxei(nState),xstate(nstate),            &
     &          xnames(nstate),stat=ier);
!              mapxu(nState),mapxc(nState),mapxui(nState),mapxci(nState),
!     &          xtot(nstate),xtota(nstate)
      xnames(:) =xnameM(:);

      !** read setup file to allocate and initialize xstate, phi
      call read_setup (setupFile);

      allocate (xfu(nest),ud(nest*(nest+1)/2),xfk(nest),                &
     &          pf(nest,nest),xfs(nest),xfa(nest),dxsum(nest,3),        &
     &          sigsum(nest,3),pfa(nest,nest),rsrif(nest*(nest+3)/2),   &
     &          stat=ier);    !xrefUD8(2*ndim+1,1000),
      if (ier /= 0) call errhand ("KF_init error: allocate error");
!      xstate(:nstate)%sig0 =sig0(:nstate)    !### test reset from driver

      xfu(:) =0.d0;
      ud(:) =0.d0;
      xfk(:) =0.d0;
      pf(:,:) =0.d0;
      xfs(:) =0.d0;
      rsrif(:) =0.d0;
      dxsum(:,:) =0.d0;
      sigsum(:,:) =0.d0;
      rsum(:) =0.d0;

      j =0;
      do i=1,nest;
        j=j+i;
        k =mapxe(i);
        xfk(i) =xstate(k)%x0;
        xfu(i) =xstate(k)%x0;
        xfs(i) =xstate(k)%x0;
        ud(j) =xstate(k)%sig0**2;
        pf(i,i) =xstate(k)%sig0**2;
        rsrif(j) =1.d0/xstate(k)%sig0**2;
      enddo;

      call prtmx (xfk,1,1,nest,"xfk");
      write (6,'("sigx0 =",5(/10g13.5))') xstate(1:nest)%sig0;

      return;
      end subroutine KF_init;
!*****************************************************
      subroutine read_setup (setupFile);

      implicit none;
      character(*),intent(in) :: setupFile;

      integer(4) ns,ier;
!      real(8)
      character(15) keyword;
      character(80) aline;
      !_______________________________

      open (10,file=setupFile,status='old',iostat=ier);
      if (ier /= 0) call errhand ('error opening setup file');

      do;
        read (10,'(a12,a80)',err=999) keyword,aline;
!        write (6,'(1x,a15,a60)') keyword,aline;
        if (keyword(1:1) == '#') cycle;     !comment

        select case(keyword);
          case ('TEpoch');
            read (aline,*) tepoch;
          case ('TEnd');
            read (aline,*) tEnd;
          case ('StateInit');
            ns =0;
            nest =0;          !total number of estimated states
            nConsider =0;     !total number of consider parameters
            nUnadj =0;        !total number of unadjusted-analyze parameters
            ncoreE =0;       !total number of estimated core states
            nMarkE =0;       !total number of estimated Markov states
            nbiasE =0;        !total number of estimated bias states

            !** read info on states
            do;
              read (10,'(a)',err=999,end=999) aline;
              ns =ns+1;
              if (ns > nstate) exit;
              write (*,'(i3,1x,a)') ns,trim(aline);
              read (aline,*,err=999) xstate(ns)%status,                 &
     &             xstate(ns)%xtype, xstate(ns)%xname, xstate(ns)%x0,   &
     &             xstate(ns)%sig0, xstate(ns)%qs, xstate(ns)%const(2);
              if (xstate(ns)%xname /= xnames(ns)) then;
                call errhand                                            &
     &             ('read_setup error: states names do not match');
              endif;

!              if (xstate(ns)%status == 0) cycle  !skip ignore states
!              if (xstate(ns)%status == 1) then   !unadjusted-analyze state
!                if (ns <= ncore+NMarkov) call errhand
!     &    ('read_setup error: unadjusted-analyze states must be biases')
!                nUnadj =nUnadj +1
!                mapxu(nUnadj) =ns
!                mapxui(ns) =nUnadj
!!                if (xstate(ns)%xtype == 1) then  !core state  !or get from model ??
!              else if (xstate(ns)%status == 2) then   !consider state
!                if (ns <= ncore+NMarkov) call errhand
!    &             ('read_setup error: consider states must be biases')
!                nConsider =nConsider +1
!                mapxc(nConsider) =ns
!                mapxci(ns) =nConsider
!              else if (xstate(ns)%status == 3) then   !estimated state
                nest =nest +1;
                mapxe(nest) =ns;
                mapxei(ns) =nest;
                if (ns <= ncore+nMarkov) nDynE =nest;
                nbiasE =nest-nDynE;
!              endif

            enddo;  !___ end read states loop

            exit;
        end select;

      enddo;

      !** set values and allocate arrays

      close (10);
      write (6,'("nest,nbiase =",3i5)') nest,nbiase;
      return;

  999 call errhand ('read_setup error: read setup file');
      end subroutine read_setup;

      end module;
!************************************************************
      subroutine errhand (message);
      character(*),intent(in) :: message;
      write (6,'(a)') trim(message);
      write (*,'(a)') trim(message);
      stop "error";
      end subroutine errhand;

