!** Program Tst_IMU_filters tests the numeric accuracy and timing of the
!** covariance Kalman filter, U-D filter and square root information filter
!** (SRIF) on the 27-state INS example.  The runFilt(3) logical variables
!** determines whether the Kalman, U-D and SRIF are to be executed. [All may
!** operate in parallel if desired, but this is not used when timing the
!** algorithms.]  The KFmethod variable determines whether the short Kalman,
!** Bierman's version of the Joseph measurement update, or the full Joseph
!** covariance update are to be used for the covariance Kalman filter.

!** The input measurement file "taimes.dat" is generated for a simulated
!** missile flight using program sim_INS.   Text file taiState.ctl defines
!** the state names, initial values, initial 1-s uncertainty, process noise
!** and status (unadjusted, consider, adjusted).  Input file xestUD8_1.dat
!** (or similar name) contains the state estimates from a U-D filter that was
!** executed in double precision.  It is used as a reference for the other filters.

!** The output plot files are
!**   KFEst.plt (standard Kalman filter estimates versus time),
!**   KFErr.plt (error in filter estimates versus time for all filters executed),
!**   KFsig.plt (1-s uncertainty versus time for all filters executed).
!** Program Tst_IMU_filters is used for examples in Section 10.4

!** In addition to library subroutines, file filters.f must be linked: it is
!** the driver for the filters.

!** author B. Gibbs  12/2009

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /link /nodefaultlib:msvcrt.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\Chapter10\filter_smo\ifort.cfg
!**   ifort filterprec.f /c
!**   ifort *.f90 /c
!**   ifort tst_IMU_filters.f90 *.obj
!**   tst_IMU_filters >tst_IMU_filters.lis
!**************************************

      module modelDat;
      implicit none;
      integer(4),parameter :: ndim=27, n2dim=(ndim*(ndim+1))/2, mdim=3;

      integer(4) ndat;
      real(8) phi(ndim,ndim),q(n2dim),h(mdim,ndim),sigr(mdim),y(mdim);
      real(8) xf(ndim),xsigr(ndim),sig0(ndim),q1(45),q2(ndim,ndim);
      real(8) xt(9),at(3),f(3,ndim);
      real(8) xrefUD8(2*ndim+1,1000);

      type taidats;
        integer(4) nt,m,nq;
        real(8) t,y(mdim),sigr(mdim);
        real(8) xt(9),at(3),f(3,ndim);
        real(8) h(mdim,ndim);
        real(8) phi(ndim,ndim),q(n2dim);
      end type;
      type (taidats) taidat(200);

      end module modelDat;
!*****************************************************
      program tst_IMC_filtFact;
      use modelDat;
      use filterMod,only: filters,dxsum,sigsum,debugp,first,rsum,       &
     &                   method_udtime,KFmethod,runFilt;
      use filterPrec, only: real_kind1;   !precision (4 or 8)
      implicit none;

      integer(4),parameter :: maxjtim=90,maxjmp=20,nparm=1230,maxbias=30;
      logical(4) :: runKF =.false.;
      integer(4) nbias,nmeas;
      integer(4) i,j,k,n,nx,nt,m,nq,ndyn,nd,it,iter;
      integer(4) :: niter =1;     !###  1 or 100 (for timing only)
      real(8) :: tmax =800.d0;
      real(8) :: t,tlst,rsumt,t0,t1,secnds,dxtsum(9),sigtsum(9);
      real(8) xref(ndim);
      real(8) :: x0(ndim) =                                             & !values used in tai_sim (6/4/09)
     &  (/ 5.2059, -3.1082, 4.8691, 0.16132E-02, -0.52446E-03,          &
     &   -0.10638E-02, -201.54, -371.89, 928.61, -2.3146, 22.269,       &
     &   -31.094, 51.637, -38.059, -20.899, 0.63379E-01, -0.44826E-01,  &
     &   -0.14239, 118.96, 90.237, -120.81, -0.28707E-02, 0.11579,      &
     &   -0.37882E-01, -0.10443E-02, -0.46317E-02, -0.10848E-01 /);
      real(real_kind1) :: phif(ndim,ndim),q2f(ndim,ndim);
      real(real_kind1) :: hf(mdim,ndim),yf(mdim),sigrf(mdim);
      character(80) UDestRef;
      !_____________________________________________________________

      method_udtime =1;    !method used for ud time update: 1=wgs, 2=rank-1, 3=square phi*u*d*u^T*phi^T  ###
      KFmethod =2;         !Kalman filter (covariance) measurement update method: 1=short, 2=Bierman;s Joseph, 3=full Joseph ###
      if (niter > 1) then;
        runFilt(:) =.false.; !denotes whether Kalman covariance, U-D and SRIF filters are to be run
        runFilt(1) =.true.;  !###
      else;
        runFilt(:) =.true.; !denotes whether Kalman covariance, U-D and SRIF filters are to be run
      endif;

      open (11,file="KFEst.plt",status='unknown');
      write (11,1020);

      open (12,file="KFErr.plt",status='unknown');
      write (12,1020);
 1020 format ('variables ="time", "delP-x", "delP-y", "delP-z", '        &
     &    '"delV-x", "delV-y", "delV-z", "delT-x", "delT-y", "delT-z"')

      open (14,file="KFsig.plt",status='unknown');
      write (14,1020);

      !** open file of state estimates and sigmas for UD filter run in double precision ###
!      UDestRef ="xestUD8_1.dat"        !use when y(:) =taidat(it)%y(:)
      UDestRef ="xestUD4_refTrue.dat";  !use when y(:) =taidat(it)%y(:) -taidat(it)%xt(:3)
!      UDestRef ="xestUD8_?.dat"        !use when y(:) =taidat(it)%y(:) -xref(:3)
      open (13,file=UDestRef,form="unformatted",status='old');
      i =0;
      do;
        i=i+1;
        read (13,end=888) nd,xrefUD8(:nd,i);
      enddo;
  888 close (13);

!***        read and save measurement data (unit 30)
      call readTaiDat (ndyn,  tmax);
      if (niter > 1) debugp =.false.;

      write (6,'("start ")');
      t0 =0.d0;
      t0 =secnds(t0);

      !** iteration loop for timing
      do iter =1,niter;
        first(:) =.true.;
        nmeas =0;
        xref(:) =x0(:);

        do it=1,ndat;
          xref(:) =matmul(taidat(it)%phi(:,:),xref(:));
          t =taidat(it)%t;
          nt =taidat(it)%nt;
          m =taidat(it)%m;
          nq =taidat(it)%nq;
!          y(:) =taidat(it)%y(:)               !y =(60777., 64313., 13800.) ft (2d7 mm) at end
!          y(:) =taidat(it)%y(:) -xref(:3)    !differences are 1d7 at end
          y(:) =taidat(it)%y(:) -taidat(it)%xt(:3); !differences are approx 0.8 mm
          sigr(:) =taidat(it)%sigr(:);
          xt(:) =taidat(it)%xt(:);
          at(:) =taidat(it)%at(:);
          f(:,:) =taidat(it)%f(:,:);
          h(:,:) =taidat(it)%h(:,:);
          q(:) = taidat(it)%q(:);
          phi(:,:) =taidat(it)%phi(:,:);
          if (debugp) write (*,'("t =",g13.5,", y =",3g13.5)')          &
     &               t,taidat(it)%y(:);

          n =nt-nbias;
          nx =n;
          nbias =nt-nx;
          nmeas =nmeas+m;

          !** standard filter in double precision
          if (runKF) then;
            call filter(t,phi,q,h,sigr,y,sig0,n,nt,nx,m,ndim,mdim,rsumt);
          else;
            !** read estimates from double precision U-D filter
            if (xrefUD8(1,it) /= t) stop 'xrefUD8 time does not match';
            xf(:n) =xrefUD8(2:n+1,it);
            xsigr(:n) =xrefUD8(n+2:n+n+1,it);
            rsumt =0.d0;
          endif;

          !** test filters
          phif(:,:) =phi(:,:);
          hf(:,:) =h(:,:);
          yf(:) =y(:);
          sigrf(:) =sigr(:);
          k =0;
          do i=1,ndim;
            do j=1,i;
              k =k+1;
              q2f(i,j) =q(k);
              q2f(j,i) =q(k);
            enddo;
!            if (q(k) > 0.d0) ndyn =i
          enddo;
          call filters (phif,q2f,h,y,sigrf,t,tlst,xt,xf,xsigr,          & !pass double precision y
     &                  m,mdim,ndim,ndim,ndyn);
        enddo;   !___  end of filter loop
      enddo;

  999 continue;
      t1 =secnds(t0);
      write (*,'("total time/pass =",g13.5)') t1/niter;
      if (niter > 1) stop;
      write (*,'("rsum =",3g13.5)') rsum(:);

      write(6,'(" rsumt =",g12.5,", nmeas =",i4)')  rsumt,nmeas;
      dxsum(1:6,:) =(304.8**2)*dxsum(1:6,:);  !output pos/vel in mm
      sigsum(1:6,:) =(304.8**2)*sigsum(1:6,:);  !output pos/vel in mm

      write (*,'(/"total time steps =",i4)') it;
      write (*,'(/"RSS differences between KF and r*8 UD filter states",&
     &     3(/10g13.5))') sqrt(dxsum(:ndim,1)/it);
      do i=1,ndim,3;
        j =(i-1)/3+1;
        dxtsum(j) =sum(dxsum(i:i+2,1))/it;
        sigtsum(j) =sum(sigsum(i:i+2,1))/it;
      enddo;
      write (*,'("RSS differences between KF and r*8 UD by type"/       &
     &      9g13.5)') sqrt(dxtsum(:9));
      write (*,'("RSS sigma differences between KF and r*8 UD by type"/ &
     &      9g13.5)') sqrt(sigtsum(:9));

      write (*,'(/"RSS differences between UD and r*8 UD filter states",&
     &     3(/10g13.5))') sqrt(dxsum(:ndim,2)/it);
      do i=1,ndim,3;
        j =(i-1)/3+1;
        dxtsum(j) =sum(dxsum(i:i+2,2))/it;
        sigtsum(j) =sum(sigsum(i:i+2,2))/it;
      enddo;
      write (*,'("RSS differences between UD and r*8 UD by type"/       &
     &      9g13.5)') sqrt(dxtsum(:9));
      write (*,'("RSS sigma differences between UD and r*8 UD by type"/ &
     &      9g13.5)') sqrt(sigtsum(:9));

      write (*,                                                         &
     &   '(/"RSS differences between SRIF and r*8 UD filter states",    &
     &     3(/10g13.5))') sqrt(dxsum(:ndim,3)/it);
      do i=1,ndim,3;
        j =(i-1)/3+1;
        dxtsum(j) =sum(dxsum(i:i+2,3))/it;
        sigtsum(j) =sum(sigsum(i:i+2,3))/it;
      enddo;
      write (*,'("RSS differences between SRIF and r*8 UD by type"/     &
     &      9g13.5)') sqrt(dxtsum(:9));
      write(*,'("RSS sigma differences between SRIF and r*8 UD by type"/&
     &      9g13.5)') sqrt(sigtsum(:9));

      stop;
      end program;
! *****************************************************************
      subroutine filter (t,phi,q,h,r,y,sig0,n,nt,nx,m,ndim,mdim,  rsum);
!         kalman filter subroutine with auxilliary eqns to compute
!         sensitivity to jump states

      use modelDat, only: xf;
      implicit none;
      integer(4),intent(in) :: n,nt,nx,m,ndim,mdim;
      real(8),intent(in) :: t,sig0(n),phi(ndim,nt),q(n*(n+1)/2);
      real(8),intent(in) :: h(mdim,nt),r(m),y(m);
      real(8),intent(out) :: rsum;

      integer(4) i,j,l,im;
      integer(4),save :: ifirst =1;
      real(8) c,yp,dz;
      integer(4),parameter :: nn=27, n2dim=(nn*(nn+1))/2, mm=3;

      real(8),save :: p(n2dim);
      real(8),save :: d1(nn),d2(nn,nn),hp(nn),g(nn);
      !_____________________________________

      if (ifirst > 0) then;
        ifirst=0;
        rsum=0.;
        xf(:) =0.d0;
        p(:) =0.d0;
        !** initialize p using sig0
        l=0;
        do i=1,n;
          l=l+i;
          p(l)=sig0(i)**2;
        enddo;
      endif;

!         time update
      call smult (d1,  phi,xf, n,n,1, ndim,n,n);
      xf(:n) =d1(:n);

!         p=phi*p*phi**t+q
      do i=1,n;
        d1(:n) =phi(i,:n);
        call symult (d2(1,i),  p,d1,n);
      enddo;

      l=0;
      do i=1,n;
        do j=1,i;
          l=l+1;
          p(l) =q(l) +dot_product(phi(i,:n),d2(:n,j));
        enddo;
      enddo;

!          propagate state sensitivity to jump parameters
!          add new parameters at current time
!      call jumptime (phi,t,nt,nx,m,ndim)    !###

!         measurement update (scalar)
      do im=1,m;
        yp=0.;
        d1(:nt) =h(im,:nt);
        yp =dot_product(h(im,:n),xf(:n));
        dz =y(im)-yp;

        call symult(hp,  p,d1,n);
        c =dot_product(hp(:n),h(im,:n)) +r(im)**2;
        rsum =rsum+dz**2/c;
!        write(6,'('' t, rsum = '',f10.2,g12.5)') t, rsum

        l=0;
        do i=1,n;
          g(i) =hp(i)/c;
          xf(i) =xf(i)+g(i)*dz;
          do j=1,i;
            l=l+1;
            p(l) =p(l)-g(i)*hp(j);
          enddo;
        enddo;

!           update jump sensitivities
!        call jumpmes (d1,g,c,dz,im,nt,nx)  !###
      enddo;

      return;
      end subroutine filter;
!**********************************************************
      subroutine readTaiDat (ndyn,  tmax);
      use modelDat;
      use filterMod,only: kf_init;
      implicit none;
      real(8),intent(in) :: tmax;
      integer(4),intent(out) :: ndyn;

      integer(4) i,j,j1,j2,it,l,nt,m,nq;
      integer(4) ncore,nMarkov,nbias;
      real(8) t;
      character(80) setupFile;
      character(20) :: xname(ndim) = (/ "delPx", "delPy", "delPz",      &
     &  "delVx", "delVy", "delVz", "tiltx", "tilty", "tiltz",           &
     &  "accBx", "accBy","accBz", "accSFEx", "accSFEy", "accSFEz",      &
     &  "GBdrftx", "GBdrfty","GBdrftz", "GSFEx", "GSFEy", "GSFEz",      &
     &  "Gsenx", "Gseny", "Gsenz", "G2senx", "G2seny", "G2senz" /);
      !_______________________________________

!***        input file of measurement data
      open(30,file='taimes.dat',status='old', form='unformatted');
      phi(:,:) =0.d0;
      q(:) =0.d0;

!***      read header of initial standard deviations
      read(30) sig0;
      write (6,'("sigx0 =",5(/10g13.5))') sig0(:ndim);

      setupFile ="taiState.ctl";
      ncore =9;
      nMarkov =0;
      nbias =0;
      ndyn =ncore +nMarkov;
      call kf_init (setupFile,ndim,ncore,nMarkov,nbias,                 &
     &               sig0,xname);
      it =0;

!****     time loop for reading and filtering measurement data
      do while (t < tmax);
        read(30,end=999) t,nt,m,nq,y(1:m),sigr(1:m);
        it =it+1;
        if (nt > ndim .or. m > mdim) stop 'invalid nt or m';
        taidat(it)%t =t;
        taidat(it)%nt =nt;
        taidat(it)%m =m;
        taidat(it)%nq =nq;
        taidat(it)%y(:) =y(:);
        taidat(it)%sigr(:) =sigr(:);

        !** read true xt error state in ft, ft/s, arc-sec, ...
        !** at is accel error (ft/s^2), not total.
        !** f is dynamic partials of velocity WRT tilt error (arc-sec),
        !** accel bias (micro-g), accel sfe (ppm), gyro bias drift (arc-sec/sec),
        !** gyro sfe (ppm), g-sensitive (arc-sec/sec/g), g**2-sensitive (arc-sec/sec)/g**2)

        taidat(it)%f(:,:) =0.d0;
        read(30) taidat(it)%xt(:),taidat(it)%at(:),                     &
     &          (taidat(it)%f(1:3,j),j=7,ndim);
        h(:,:) =0.d0;
        do i=1,m;
          read(30) j1,j2,h(i,j1:j2);
          if (j1 < 1 .or. j1 > nt .or. j2 < 1 .or. j2 > nt)             &
     &        stop 'invalid h index';
        enddo;
        taidat(it)%h(:,:) =h(:,:);     ! =3 x 3 identity plus zeros

        q(:) =0.d0;
        do i=1,nq;
          read(30) l,q(l);
          if (l < 1 .or. l > n2dim) stop 'invalid q index';
        enddo;
        taidat(it)%q(:) =q(:);

        phi(:,:) =0.d0;
        do i=1,nt;
          read(30) j1,j2,(phi(i,j),j=j1,j2);
          if (j1 < 1 .or. j1 > nt .or. j2 < 1 .or. j2 > nt)             &
     &          stop 'invalid phi index';
        enddo;
        taidat(it)%phi(:,:) =phi(:,:);

!        write(6,*) 't =',t
!        call prtmx(y,1,1,m,'y       ')
!        call prtmx(r,1,1,m,'rsig    ')
!        call prtmx(h,mdim,m,n,'h       ')
!        call prtmx(phi,ndim,n,n,'phi     ')
!        call prtsmx(q,n,'q       ')
      enddo;   !___  end of tai data

  999 close (30);
      ndat =it;
      return;
      end subroutine readTaiDat;
! ***************************************************************
      real(8) function secnds(t0);
      real(4),intent(in) :: t0;

      integer(4) idatetime(8);

!       get today'S DATE/TIME
      call date_and_time (values=idatetime);

!        idatetime(1) = year (e.g. 1997)
!        idatetime(2) = month (1-12)
!        idatetime(3) = day (1-31)
!        idatetime(4) = utc time difference
!        idatetime(5) = hour (0-23)
!        idatetime(6) = minutes (0-59)
!        idatetime(7) = seconds (0-59)
!        idatetime(8) = milliseconds (0-999)

      secnds=idatetime(7)+60.*(idatetime(6)+60*idatetime(5))+           &
     &        idatetime(8)*0.001-t0;

      return;
      end;

