!** ACTrkFil_jump.f is a nonlinear filter that tracks an aircraft
!** Program ACTrkFil_jump is an EKF filter that tracks a simulated aircraft
!** using the jump detection/estimation algorithm described in Section 11.3.4
!** for detecting and adapting to maneuvers.  It uses the 8-state aircraft
!** model described in Section 3.2.2.

!** Aircraft maneuvers are defined in input file mnvs.dat.
!** Output plot files ACTrk.plt lists the true (simulated) states versus
!** time, Filt_Xerr.plt lists the filter state estimate errors and
!** Filt_sig.plt lists the 1-sigma uncertainty.  Module jumpCom contains all
!** jump detection/estimation algorithm, and must be linked with ACTrkFil_jump.
!** This is used in Example 11.3-2.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort jumpcom.f90 /c
!**   ifort acTrkFil_jump.f90 jumpcom.obj         !libkf.lib includes rnoruni, sinv, symult, prtmx,
!**   acTrkFil_jump >actrk.lis
!**************************************

      module modelDat;
      !** includes model info for funct
      integer(4),parameter :: nState =8,ndynam=8;
      integer(4),parameter :: mmax =3;
      integer(4) irecs;    !direct access record number for smoothing
      integer(4) nestF,nF3,nDynamEst;
      real(8) xt(nState),xtl(nState),xo(3);
      real(8) xp(nState),xa(nState),pa(nState,nState),pad(nState);
      real(8) xf(nState),pf(nState,nState),phi(nState,nState);
      real(8) q(nState,nState),qs(nState),tau(nstate);
      real(8) pfs(nState*(nState+1)/2);
      real(8) xref(nState);
      real(8) rn,rnoise(3),dt;
      real(8) :: gacc = 9.80d0;    !m/s^2
      logical(4) cartesian, debugP, markov;
      logical(4) :: firstIter =.false.;     !normal mode linearizing all iter about estimate ###
!      logical(4) :: firstIter =.true.      !test linearizing partials about meas for 1st iter ###

      integer(4) nMnvr;         !number of maneuvers
      type mnvrDats;
        real(8) tstart,tend;
        real(8) tanphi,accel,zdot;
      end type;
      type (mnvrDats) mnvrDat(10);

      end module modelDat;
!****************************************************
      program ACTrkFil;
      !** simulation of aircraft tracking with radar RAE tracking every 1 second
      !** 8 states are x,y,z position (m), x,y,z velocity (m/s), tan(bank), alongtrack acceleration (m/s^2)
      use jumpcom;
      use modelDat;
      implicit none;
      integer(4) i,n,m,ix,nrec,imnvr,lenrec,ierr;
      integer(4) :: method =1;  !1=EKF, 2=IEKF   !little smoothed difference between 1 and 2
      logical(4) IntegCov;
      real(8) r,y(3),yp(3),h(3,nState);
      real(8) tend,tstart,tstop,tanphi,accel,zdot,t,tLst;
      real(8),external :: rnor;
      !____________________________________________________

      m =mmax;
      n =nState;
      nestF =nState;
      nF3 =(nestF*(nestF+3))/2;
      nDynamEst =nestF;    !(no biases in this problem)
      ix =191919;
      r =rnor(ix);      !initialize randum number generator
      rnoise(:) =(/ 1., 0.2e-3, 0.2e-3 /);  !RAE noise

      !** maneuvers
      open (10,file="mnvs.dat",status='old');
      nmnvr =1;
      do;
        read (10,*,end=100,err=100) tstart,tstop,tanphi,accel,zdot;
        nmnvr =nmnvr +1;
        mnvrDat(nmnvr)%tstart =tstart;
        mnvrDat(nmnvr)%tend =tstop;
        mnvrDat(nmnvr)%tanphi =tanphi;
        mnvrDat(nmnvr)%accel =accel;
        mnvrDat(nmnvr)%zdot =zdot;
      enddo;

!   45., 55.0, -0.53, 0., 0.
!   76.1, 87.1, -0.47, 0., -4.
!   90., 95., 0., -1., -6.
!  125., 510., 0., 0., 0.

  100 close(10);
      dt =1.d0;  !       !sec
      tend =mnvrDat(nmnvr)%tstart -1.;
      cartesian =.true.;      !###

      !** true state, prior and observer position
      xt(:) =(/ 1200.d0, 0.d0, 300.d0, 0.d0, 32.d0, 0.d0, 0.d0, 0.d0 /);
      xo(:) =0.d0;          !own position

      xp(:) =(/ 1100.d0, 10.d0, 330.d0,  20.d0, 0.d0, 0.d0, 0.d0, 0.d0/);   !prior estimate
      pad(:) =(/ 100.d0, 100.d0, 100.d0, 20.d0, 20.d0, 6.d0,            &
     &           0.50d0, 1.0d0 /); !prior 1-sigma  (Acc-alongwas 0.63d0)

      markov =.false.;  ! ###
      tau(6:8) =(/ 5., 10., 10. /);
      pf(:,:) =0.d0;
      do i=1,nState;
        pf(i,i) =pad(i)**2;
      enddo;
      qs(:3) = 0.d0;
      qs(4:5) =0.001;
      qs(6) =0.08;
      qs(7:8) =(/ 0.01, 0.4 /);
      qs(6:8) =0.0003d0;       !when using jump detection ###
      if (markov) forall (i=6:8) qs(i) =2.d0*pad(i)**2/tau(i);

      xf(:) =xp(:);     !initialize estimate with prior
!      xf(:) =xt(:)      !###  test initialized at truth
      xref(:) =0.d0;    !reference state is 0

      debugp =.false.;   !###

      open (11,file='ACTrk.plt',status='unknown');
      write (11,1200);
 1200 format('VARIABLES ="t (s)", "x (m)", "y (m)", "z (m)", ',         &
     &   '"xdot (m/s)", "ydot (m/s)", "zdot (/s)", ',                   &
     &   '"tanphi", "a-acc (m/s2"')

      open (20,file='Filt_Xerr.plt',status='unknown');
      write (20,1100);
 1100 format('VARIABLES ="t", "x-error", "y-error", "z-error", ',       &
     &   '"xdot-error", "ydot-error", "zdot-error", "tanphi-error", ',  &
     &   '"accel-a-error"')
      open (21,file='Filt_sig.plt',status='unknown');
      write (21,1110);
 1110 format('VARIABLES ="t", "sig-x", "sig-y", "sig-z", "sig-xd", ',   &
     &   '"sig-yd", "sig-zd", "sig-tanphi", "sig-acca"')

      !** set jump parameters and initialize jump arrays
      call set_jump;
      call init_jump;

      t =0.d0;
      tLst =t;
      zdot =0.d0;
      imnvr =1;

      do while (t < tend);
        t =t+dt;
        tanphi =0.d0;
        accel =0.d0;
        xtl(:) =xt(:);

        !** add
        if (t >= mnvrDat(imnvr)%tend) imnvr =imnvr+1;
        if (t >= mnvrDat(imnvr)%tstart) then;
          xt(6) =mnvrDat(imnvr)%zdot;
          tanphi =mnvrDat(imnvr)%tanphi;
          accel =mnvrDat(imnvr)%accel;
        endif;
        xt(7) =tanphi;
        xt(8) =accel;

        !** propagate state over dt and compute phi,q
        call integx (xt,  xt,dt,n);
        write (11,'(9g13.5)') t,xt(:);

        !** get meas (r,az,el)
        call get_yh (yp,h,  xt,xo,n);
        do i=1,3;
          y(i) =yp(i) +rnoise(i)*rnor(0);
        enddo;

        !** process measurements in extended Kalman filter (with jump detection)
        call ekf (t,dt,y,rnoise,qs,method);     !extended Kalman filter

        tLst =t;
      enddo;

      stop;
      end program;
!******************************************************************
      subroutine ekf (t,dt,y,rnoise,qs,method);
      use jumpcom, only:jumptime,jumpmeas,estjmpj,nxp,xpt,ilxp;
      use modelDat, only:xref,xt,xtl,xo,xa,pa,xf,pf,phi,q,mmax,         &
     &                    firstIter,irecs,nestF,nstate,debugP,pfs       &
     &;
      implicit none;

      integer(4),intent(in) :: method;       !1=ekf, 2=iekf
      real(8),intent(in) :: t;               !time of current measurements
      real(8),intent(in) :: dt;              !time increment between this and last measurement
      real(8),intent(in) :: y(mmax);         !measurement vector
      real(8),intent(in) :: rnoise(mmax);    !measurement noise 1-sigma
      real(8),intent(in) :: qs(nestF);       !PSD of continuous process noise

      integer(4) j,i,k,n,im,lEdit;
      real(8) h(nestF),hp(nestF),g(nestF),xfi(nestF),xfil(nestF);
      real(8) h3(mmax,nestF),yp(mmax),dy,yvar,xfj(nestF);
      !_________________________________________

      n =nestF;

      if (dt > 0.d0) then;
        call get_phix (xa,phi,q,   xf,dt,qs,n);
        pa(:,:) =matmul(matmul(phi,pf),transpose(phi)) +q(:,:);
      else;
        xa(:) =xf(:);
        pa(:,:) =pf(:,:);
      endif;

!      irecs =irecs+1
!      xref(:) =xtl(:)
!      write (42,rec=irecs) t,t-dt,xref,xa,xf,pf,phi,q  !for smoother

      !** time update for jump variables
      irecs =nint(t);
      call jumptime (phi,t,nestf,nestf,nstate,irecs);

      !** now measurement update one-at-a-time
      do im=1,mmax;
        call get_yh (yp,h3,  xa,xo,n);
        h(:) =h3(im,:);
        hp(:) =matmul(h,pa);
        yvar =dot_product(h,hp) +rnoise(im)**2;
        g(:) =hp(:)/yvar;
        dy =y(im)-yp(im);     !/rn
        xf(:) =xa(:) +g(:)*dy;

        if (method == 2) then;
          !** iterated extended KF
          xfi(:) =xf(:);
          xfil(:) =1.d20;
          do k=1,100;
            call get_yh (yp,h3,  xf,xo,n);
            h(:) =h3(im,:);
            hp(:) =matmul(h,pa);
            yvar =dot_product(h,hp) +rnoise(im)**2;
            g(:) =hp(:)/yvar;
            dy =y(im) -yp(im) -dot_product(h,xa(:)-xfi(:));     !/rn
            xf(:) =xa(:) +g(:)*dy;
            if (sqrt(sum((xfil(:n) -xfi(:n))**2)) < 0.001) exit;
            xfil(:) =xfi(:);
!            write (6,'("iekf x =",g13.5)') xfi
          enddo;
          xf(:) =xfi(:);
        endif;

        !** covariance update
        k =0;
        do i=1,n;
          do j=1,i;
            k=k+1;
            pf(j,i) =pa(j,i) -g(j)*hp(i);
            pf(i,j) =pf(j,i);
            pfs(k) =pf(j,i);
          enddo;
        enddo;

        !** update jump variables for measurements
        lEdit =0;
        call jumpmeas (h,g,yvar,dy,im,n,lEdit);

        !** reset prior for next meas
        xa(:) =xf(:);
        pa(:,:) =pf(:,:);
      enddo;   !___ end measurement loop

      !** test for jumps and correct if any are detected
      !** xfj is temporary state vector including pending jumps
      !*** (use for prediciton, but not in filter)
      xfj(:) =xf(:);    !set as default
      call estjmpj (xfj,  xf,pfs,  n);

      if (nxp > 0) then;
        do i=1,nxp;
          write (6,'("jump ",g13.5," corrected for state ",i3,"/",i3)') &
     &            xpt(i), ilxp(:,i);
        enddo;
        nxp =0;    !reset jump counter
      endif;

      !** copy symmetric covariance to 2D
      k =0;
      do i=1,n;
        do j=1,i;
          k =k+1;
          pf(j,i) =pfs(k);
          pf(i,j) =pf(j,i);
        enddo;
      enddo;

      write (20,'(9g13.5)') t,xfj(:)-xt(:);
      write (21,'(9g13.5)') t,(sqrt(pf(i,i)),i=1,n);

      return;
      end subroutine ekf;
!*********************************************************
      subroutine get_yh (yp,h,  xt,xo,n);
!***    computes measurements range, az, el and partials wrt state
      use modelDat, only:nState;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xo(3);      !own cartesian position
      real(8),intent(in) :: xt(8);      !target cartesian position, velocity, tan phi, long accel
      real(8),intent(out) :: yp(3);      !range, az, el
      real(8),intent(out) :: h(3,n);     !partials of y wrt xt

      real(8) dx(3),r,r2,az,el,rh,rh2;
      !_______________________________________

      dx(:) =xt(:3) -xo(:);
      r2 =sum(dx(1:3)**2);
      r =sqrt(r2);
      rh2 =dx(1)**2 +dx(2)**2;
      rh =sqrt(rh2);
      az =atan2(dx(1),dx(2));
      el =asin(dx(3)/r);
      yp(:) =(/ r,az,el/);

      h(:,:) =0.d0;
      h(1,:3) =(/ dx(1), dx(2), dx(3) /)/r;
      h(2,:3) =(/ dx(2), -dx(1), 0.d0 /)/rh2;
      h(3,:3) =(/ -dx(1)*dx(3), -dx(2)*dx(3), rh2 /)/(rh*r2);

      return;
      end subroutine get_yh;
!*********************************************************
      subroutine get_phix (xa,phi,q,   xf,dt,qs,n);
!***    computes state transition and process noise matrix and integrates state
      use modelDat, only:nState,ndynam,gacc,debugP;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(in) :: dt;         !time step (s)
      real(8),intent(in) :: qs(n);      !state noise psd
      real(8),intent(out) :: xa(n);     !predicted state at t+dt
      real(8),intent(out) :: phi(nState,n); !state transition matrix
      real(8),intent(out) :: q(nState,n);   !state noise matrix

      integer(4) i;
      real(8) xdot(nState),f(n,n),phii(n,n),dti,t;
!      real(8) :: dtmax =1.1  !###
      real(8) :: dtmax =0.2;     !little difference between 0.1 and 0.2
      !_______________________________________

      if (n /= nState) stop 'get_phix error: n /= nState';

      t =0.d0;
      if (dt <= 0.) stop 'get_phix error: dt <= 0';
      xa(:) =xf(:);
      phi(:,:) =0.d0;
      forall (i=1:n) phi(i,i) =1.d0;
      q(:,:) =0.d0;

      do while (t < dt);
        dti =min(dtmax,dt-t);
        t =t+dti;
        call get_F (f,xdot,   xa,n);
        phii(:,:) =dti*f(:,:);
!        phii(:,:) =dti*(f(:,:) +0.5*dti*matmul(f,f))        !###
        forall (i=1:n) phii(i,i) =phii(i,i) +1.d0;
        phi(:,:) =matmul(phii(:,:),phi(:,:));
!        call prtmx (phi,n,n,n,"phii")
        q(:,:) =matmul(matmul(phii,q),transpose(phii));
        forall (i=1:n) q(i,i) =q(i,i) +qs(i)*dti;
        xa(:) =xa(:) +dti*xdot(:);
      enddo;

      if (debugP) then;
        write (6,'("get_phix t =",g13.5)') t;
        call prtmx (phi,n,n,n,"phi");
        call prtmx (q,n,n,n,"q");
      endif;

      return;
      end subroutine get_phix;
!*********************************************************
      subroutine integx (xa,  xf,dt,n);
!***    computes state transition and process noise matrix and integrates state
      use modelDat, only:nState;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(in) :: dt;         !time step (s)
      real(8),intent(out) :: xa(n);     !predicted state at t+dt

      real(8) xdot(nState),f(n,n),dti,t;
!      real(8) :: dtmax =1.1  !###
      real(8) :: dtmax =0.2;
      !_______________________________________

      if (n /= nState) stop 'integx error: n /= nState';

      t =0.d0;
      if (dt <= 0.) stop 'integx error: dt <= 0';
      xa(:) =xf(:);

      do while (t < dt);
        dti =min(dtmax,dt-t);
        t =t+dti;
        call get_F (f,xdot,   xa,n);
        xa(:) =xa(:) +dti*xdot(:);
      enddo;
!      write (6,'("integx t =",g13.5)') t

      return;
      end subroutine integx;
!**********************************************************************
      subroutine get_F (F,xdot,   xf,n);
!***    computes state transition and process noise matrix and integrates state.
!**     8 states are x,y,z position (m), x,y,z velocity (m/s), tan(bank), alongtrack acceleration (m/s^2)
      use modelDat, only:nState,ndynam,gacc,tau,markov;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(out) :: F(n,n); !state transition matrix
      real(8),intent(out) :: xdot(n);     !steta derivative

      integer(4) i;
      real(8) ac,aa,vh,vh2;
      !_______________________________________

      if (n /= nState) stop 'get_f error: n /= nState';

      ac =gacc*xf(7);   !crosstrack accel
      aa =xf(8);        !alongtrack accel
      vh2 =xf(4)**2 +xf(5)**2;
      vh =sqrt(vh2);
      xdot(:3) =xf(4:6);
      xdot(4) =(ac*xf(5) +aa*xf(4))/vh;
      xdot(5) =(-ac*xf(4) +aa*xf(5))/vh;
      xdot(6:8) =0.d0;
      if (markov) forall (i=6:8) xdot(i) =-xf(i)/tau(i);

      f(:,:) =0.d0;
      forall (i=1:3) f(i,i+3) =1.d0;
      f(4,4:8) =(/  aa -xdot(4)*xf(4)/vh, ac -xdot(4)*xf(5)/vh,         &
     &            0.d0,  gacc*xf(5), xf(4) /) /vh;
      f(5,4:8) =(/ -ac -xdot(5)*xf(4)/vh, aa -xdot(5)*xf(5)/vh,         &
     &            0.d0, -gacc*xf(4), xf(5) /) /vh;
      if (markov) forall (i=6:8) f(i,i) =-1.d0/tau(i);

      return;
      end subroutine get_F;
!**********************************************************************
      subroutine set_jump;
!***    allocates and initializes jump parameters and arrays
!***    Note: Before calling init_jump, the following variables must be set:
!***       f2rem,f2add,tolmin,nlag,njstat,jmpstat,sigxj,qmin,
!***       ndim,mdim,maxjtim,maxjmp,sigfac,nxjmp,xt,at,f.
!***       maxjmp should be set to nxjmp so that nparm =np
!***       Also jmpstat(njstat), sigxj(njstat), qmin(njstat), must be allocated.

      use jumpCom;
      use modelDat;
      implicit none;
      integer(4) i,k,njstat,ier,is,isat;
      !____________________________________________________

      f2rem =4.0; !0.50
      f2add =9.;  !36.     !9.0
      tolmin =0.01;
      nlag =1;         !set lag window for real-time jump detection ###
      write(6,'(a,1x,3f8.2,i8)') "f2rem,f2add,tolmin,nlag",             &
     &     f2rem,f2add,tolmin,nlag;

      njstat =3;     !# of states allowed to jump
      write(6,1100) "njstat",njstat;
 1100 format(a,10(/1x,20i4))
      allocate (jmpstat(njstat),sigxj(njstat),qmin(njstat),stat=ier);
      if (ier /= 0) stop 'allocate jmpstat,sigxj,qmin error';

      ndim =nstate;            !# filter states
      mdim =mmax;              !max # measurements
      maxjtim =12.d0/dt;      !max # times in moving window  ###
!      maxjmp =njstat           !max number of jump states at each time point
      sigfac =1.0;             !factor multiplying sigxj used as prior information
      jmpstat(:) =(/ 6,7,8 /);
      nxjmp =0;
      do i=1,njstat;
        if (jmpstat(i) <= 1 .or. jmpstat(i) > ndim)                     &
     &                stop 'invalid jmpstat';
        if (jmpstat(i) > 0) nxjmp =nxjmp+1;
        sigxj(i) =pad(jmpstat(i));
      enddo;
      maxjmp =nxjmp;

      write(6,1100) "jmpstat", jmpstat(1:njstat);
      write(6,1000) "sigxj", sigxj(1:njstat);
 1000 format(a,10(/1x,8f8.2))

!      if (nxjmp > maxjmp) stop 'jumptime error: nxjmp > maxjmp'
!      if (nx > ndim) stop 'jumptime error: nx > ndim'

      return;
      end subroutine set_jump;
