!**  tst_fis2.f tests the calculation of the Fisher information matrix for a 4-state problem.
!**  It can also run scoring iterations.

!**  Module modelMod contains the model-specific information and tst_fis is the driver for
!**  the filter calculations.  4-state model is two independent 1st order markov processes
!**  with integral.  Measurements may be either independent states 1 (x) and 3 (y) of
!**  integrated markov (mmod=2), or range and bearing (mmod=1).
!**  The 6 parameters to be estimated via MLE are: Markov constants (2), Markov process noise (2),
!**  and measurement noise (2).

!**  Search for "###" to find parameters that are typically changed

!**  If test_fis =.true., the gradient and Fisher are computed using numeric
!**  central-difference partial derivatives and the average Fisher is computed via
!**  1000-sample Monte Carlo simulation. If nsamp = 1, the numeric gradient and Fisher can be
!**  compared with the analytic gradient/Fisher (when test_fis =.false.) to verify coding.

!**  The filter/simulation is coded with the following assumptions:
!**    1) initial true state is (1,10,0,0)
!**    2) filter state is initialized at truth, with diagonal covariance
!**    3) phi is analytic, but Q is assumed diagonal because of high sampling rate
!**       compared to Markov time constants). random process noise is applied
!**       independently to the two channels
!**    4) measurement noise is assumed independent for 2 meas
!**    5) measurement model does not include unknown parameters (no del-y/del-theta)
!**    6) filter coding uses intrinsic matrix/vector operators with no provision for
!**       sparse matrices so this code should only be used for small problems.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tst_fis2.f90
!**   tst_fis2 >tstf_qs1_a1_xyScor_t100_dt25a.lis
!**************************************

      module modelMod;
      !** contains model-specific information for tst_fis
      integer(4),parameter :: n=4;        !filter state dimension
      integer(4),parameter :: np=6;       !number of model parameters to be estimated via MLE
      integer(4),parameter :: mmeas=2;    !number of measurements at one time
      integer(4),parameter :: mmax=3000;  !maximum total number of meas time samples

      real(8) phi(n,n);       !state transition matrix
      real(8) q(n,n);         !discrete state noise matrix
      real(8) dphi(n,n,np);   !partial phi wrt theta
      real(8) dq(n,n,np);     !partial q wrt theta
      real(8) drdt(np);       !partial meas noise variance wrt theta
      real(8) yts(mmeas,mmax);!simulated measurements
      real(8) xt0(n);         !true initial state
      real(8) :: dt =0.25d0;  !integration time sec (sec)   ### (also 0.1 or 0.2)
      real(8) :: tmax =100.d0;!maximum simulation time (sec)  ### (also 50)
      integer(4) :: nmeas;    !number of meas time steps

      !** control variables for model options
      integer(4) :: amod = 1;   !model for theta(1:2) (Markov coeff: 1=a, 2=log(a), 3=tau) ###   !3 doesn't work
      integer(4) :: qsmod = 1;  !model for theta(3:4) (q: 1=q, 2=log(q), 3=ss-sigma)   ###
      integer(4) :: mmod = 2;   !measurement model: 1=r,bearing, 2=x,y

      contains;
      !*********************************************************************
      subroutine get_phiq (theta);
      !** compute discrete phi, Q and partials dphi =del-phi/del-theta, dq =del-q/del-theta
      !** qsmod,amod are inputs.  4-state model is two independent 1st order markov processes
      !** with integral.
      implicit none;
      real(8),intent(in) :: theta(np); !model parameters to be estimated via MLE

      real(8) f(n,n),exp1,exp2,a(2),dadth(2);
      !_________________________________________________________

!      f(:,:) =0.d0
!      f(1,2) =1.d0
!      f(2,2) =-thetae(1)
!      f(3,4) =1.d0
!      f(4,4) =-thetae(2)

      if (amod == 1) then;
        a(:) =theta(1:2);     !theta contains a-coeff
        dadth(:) =1.d0;
      else if (amod == 2) then;
        a(:) =exp(theta(1:2));!theta contains log(a-coeff)
        dadth(:) =a(:);
      else if (amod == 3) then;
        a(:) =1.d0/theta(1:2);!theta contains 1/(a-coeff)
        dadth(:) =-a(:)**2;
      endif;

      exp1 =exp(-a(1)*dt);
      exp2 =exp(-a(2)*dt);
      phi(:,:) =0.d0;
      phi(1,1:2) =(/1.d0, (1.d0-exp1)/a(1) /);
      phi(2,2) =exp1;
      phi(3,3:4) =(/1.d0, (1.d0-exp2)/a(2) /);
      phi(4,4) =exp2;

      dphi(:,:,:) =0.d0;
      dphi(1,2,1) =(dt*exp1/a(1) -(1.d0-exp1)/a(1)**2)*dadth(1);
      dphi(2,2,1) =-dt*exp1*dadth(1);
      dphi(3,4,2) =(dt*exp2/a(2) -(1.d0-exp2)/a(2)**2)*dadth(2);
      dphi(4,4,2) =-dt*exp2*dadth(2);

      q(:,:) =0.d0;
      dq(:,:,:) =0.d0;
      if (qsmod == 1) then;
        q(2,2) =theta(3)*dt;        !qs
        q(4,4) =theta(4)*dt;
        dq(2,2,3) =dt;
        dq(4,4,4) =dt;
      else if (qsmod == 2) then;
        q(2,2) =exp(theta(3))*dt;     !log(qs)
        q(4,4) =exp(theta(4))*dt;
        dq(2,2,3) =q(2,2);
        dq(4,4,4) =q(4,4);
      else if (qsmod == 3) then;
        q(2,2) =2.d0*theta(1)*exp(theta(3))*dt;     !log(ss-variance)
        q(4,4) =2.d0*theta(2)*exp(theta(4))*dt;
        dq(2,2,1) =2.d0*exp(theta(3))*dt;
        dq(4,4,2) =2.d0*exp(theta(4))*dt;
        dq(2,2,3) =q(2,2);
        dq(4,4,4) =q(4,4);
      endif;

      return;
      end subroutine get_phiq;
!********************************************************************
      subroutine get_yh (yp,dy,h,r,  yt,x,theta,im);
      !** compute measurements, residuals, partials and measurement noise variance
      !** from given filter state vector, theta vector and actual measurement (if available)
      implicit none;
      integer(4),intent(in) :: im;     !meas index (1 or 2)
      real(8),intent(in) :: x(n);      !filter state vector
      real(8),intent(in) :: yt(mmeas); !actual measurements (may be 0)
      real(8),intent(in) :: theta(np); !model parameters (last 2 are meas variances)
      real(8),intent(out) :: yp;       !model-predicted measurement
      real(8),intent(out) :: dy;       !measurement residual
      real(8),intent(out) :: h(n);     !measurement partials wrt x
      real(8),intent(out) :: r;        !measurement variance

      real(8) r2;
      real(8) :: pi =3.141592653589793d0;
      !_______________________________________________

      if (mmod == 1) then;
        !** range and bearing
        r2 =x(1)**2 +x(3)**2;
        if (im == 1) then;
          yp =sqrt(r2);
          if (r2 > 0.d0) then;
            h(:) =(/ x(1)/yp, 0.d0, x(3)/yp, 0.d0 /);
          else;
            h(:) =(/ 0.707d0, 0.d0, 0.707d0, 0.d0 /);
          endif;
          dy =yt(im)-yp;
        else if (im == 2) then;
          yp =atan2(x(1),x(3));
          if (r2 > 0.d0) then;
            h(:) =(/ x(3), 0.d0, -x(1), 0.d0 /)/r2;
          else;
            h(:) =(/ 1.d0, 0.d0, 1.d0, 0.d0 /);
          endif;
          dy =yt(im)-yp;
          if (abs(dy) > pi) dy =dy -sign(2.d0*pi,dy);
        endif;

      else if (mmod == 2) then;
        !** x and y position
        yp =x(2*im-1);
        h(:) =0.d0;
        h(2*im-1) =1.d0;
        dy =yt(im)-yp;
      endif;

      r =theta(np-2+im);
      drdt(:) =0.d0;
      drdt(np-2+im) =1.d0;

      return;
      end subroutine get_yh;

      end module modelMod;
!**********************************************************************
      program tst_fis;
      !** tests the calculation of the fisher information for a 4-state problem
      use modelMod;
      implicit none;

      integer(4) :: nsamp=1;              !1000 ###
      integer(4) :: maxScore =10;         !max number of scoring iterations (use 1 to compute Fisher at truth ###)
      logical(4) :: test_fis =.false.;    !if true, make Monte Carlo evaluation of numeric Hessian and gradient ###
      integer(4) :: ix =191919;           !seed of random number generator
      real(8) :: theta(np) =(/ 0.1, 0.2, 0.8, 1.6, 0.5, 0.001 /);  !model parameters
      real(8) :: sig0(n) =(/2., 4., 3., 4. /);                     !initial filter state 1-sigma
      real(8) :: delth =1.d-5;            !perturbation for numeric partials

      integer(4) :: is,iscore,ierr,i,j;
      real(8) thetae(np),grad(np),fis(np,np),fis2(np,np),cost,csum;
      real(8) fsum(np,np),fsum2(np,np),dthet(np),rnor,converg;
      real(8) fis0(np,np),hes(np,np),cost1,cost2,cost3,cost4;
      real(8) grad1(np),grad2(np),gradn(np),gsum(np),theta1(np);
      !______________________________________________

      !** options for parameterizing the process noise
      if (qsmod == 1) then;
        theta(3) =2.d0*theta(1)*dt*sig0(2)**2;   !=qs(2)
        theta(4) =2.d0*theta(2)*dt*sig0(4)**2;   !=qs(4)
      else if (qsmod == 2) then;
        theta(3) =log(2.d0*theta(1)*dt*sig0(2)**2); !=log(qs(2))
        theta(4) =log(2.d0*theta(2)*dt*sig0(4)**2); !=log(qs(4))
      else if (qsmod == 3) then;
        theta(3) =log(sig0(2)**2);     !log(ss-variance)
        theta(4) =log(sig0(4)**2);
      endif;

      !** options for parameterizing the Markov feedback constant
      if (amod == 2) then;
        theta(1:2) =log(theta(1:2));   !log a
      else if (amod == 3) then;
        theta(1:2) =1.d0/theta(1:2);   !tau
      endif;

      !** options for the measurement model
      if (mmod == 2) then;
        theta(np) =theta(np-1);
      endif;
      write (6,'("dt =",g13.5,", tmax =",g13.5)') dt,tmax;
      write (6,'("theta =",6g13.5)') theta(:);

      cost =rnor(ix);   !initialize rnor seed
      fsum(:,:) =0.d0;
      fsum2(:,:) =0.d0;
      gsum(:) =0.d0;
      csum =0.d0;

      if (test_fis) then;
        !** Option 1: Monte Carlo simulation to compute the expected value of
        !**           the negative log likelihood gradient (WRT theta) and Hessian where
        !**           both are computed numerically using central differences.
        do is=1,nsamp;                    !about 30 min for 1000 samples
          call sim_meas (sig0,theta);     !simulate random measurements
          thetae(:) =theta(:);
          do i=1,np;
            thetae(i) =theta(i) +delth;
            theta1(:) =thetae(:);
            call filter (cost1,grad,fis,fis2,  sig0,thetae);
            do j=1,i;
              theta1(j) =thetae(j) +delth;
              call filter (cost3,grad,fis,fis2,  sig0,theta1);
              theta1(j) =thetae(j) -delth;
              call filter (cost4,grad,fis,fis2,  sig0,theta1);
              theta1(j) =thetae(j);
              grad1(j) =(cost3-cost4)/(2.d0*delth);
            enddo;

            thetae(i) =theta(i) -delth;
            theta1(:) =thetae(:);
            call filter (cost2,grad,fis,fis2,  sig0,thetae);
            gradn(i) =(cost1-cost2)/(2.d0*delth);
            do j=1,i;
              theta1(j) =thetae(j) +delth;
              call filter (cost3,grad,fis,fis2,  sig0,theta1);
              theta1(j) =thetae(j) -delth;
              call filter (cost4,grad,fis,fis2,  sig0,theta1);
              theta1(j) =thetae(j);
              grad2(j) =(cost3-cost4)/(2.d0*delth);
              hes(i,j) =(grad1(j)-grad2(j))/(2.d0*delth);
              hes(j,i) =hes(i,j);
            enddo;
            thetae(i) =theta(i);
          enddo;

          fsum(:,:) =fsum(:,:) +hes(:,:);
          gsum(:) =gsum(:) +gradn(:);
          call filter (cost,grad,fis,fis2,  sig0,thetae);
          fsum2(:,:) =fsum2(:,:)/nsamp;
          csum =csum +cost;
        enddo;    !___ end is loop

        fsum(:,:) =fsum(:,:)/nsamp;
        gsum(:) =gsum(:)/nsamp;
        write (6,'("nsamp =",i4,", avg cost =",g13.5)')                 &
     &         nsamp,2.d0*csum/nsamp;
        call prtmx (fsum,np,np,np,"favg");
        call prtmx (gsum,1,1,np,"gavg");
        stop;
      endif;

      !** Option 2: Scoring iterations: either single sample or Monte Carlo simulation to
      !** compute average Fisher information
      do is=1,nsamp;
        call sim_meas (sig0,theta);    !simulate measurements
        thetae(:) =theta(:);
        do iscore=1,maxScore;          !scoring iterations
          call filter (cost,grad,fis,fis2,  sig0,thetae);
          write (6,'(/"******************"/"iscore =",i5,               &
     &           ", nmeas =",i5,", cost =",g13.5)') iscore,nmeas,cost;
          call prtmx (fis,np,np,np,"fis");

          dthet(:) =grad(:);
          fis0(:,:) =fis(:,:);
          call matInvGJ (ierr,  fis,dthet,  np,np,1);
          if (ierr /= 0) stop 'matInvGJ error';
          converg =sqrt(dot_product(dthet(:),grad)/np);
          write (6,'("grad =",6g13.5)') grad(:);
!         d(:) =grad(:) -matmul(fis0,dthet);
!         write (6,'("grad-fis*dthet =",6g13.5)') d(:);
          if (iscore >= 10) stop 'scoring not converged';
          write (6,'("cost =",g13.5,", converg =",g13.5,                &
     &           ", dthet =",6g13.5)') cost,converg,-dthet(:);
          thetae(:) =thetae(:) -dthet(:);
          write (6,'("thetae =",6g13.5)') thetae(:);
          if (converg < 0.02) exit;
        enddo;

        fsum(:,:) =fsum(:,:) +fis0(:,:);
        fsum2(:,:) =fsum2(:,:) +fis2(:,:);
        csum =csum +cost;
      enddo;

      call prtmx (fis,np,np,np,"finv");
      write (6,'("1-sigma =",6g13.5)') (sqrt(fis(i,i)),i=1,np);
      fsum(:,:) =fsum(:,:)/nsamp;
      fsum2(:,:) =fsum2(:,:)/nsamp;

      write (6,'("nsamp =",i4,", avg 2*cost =",g13.5)')                 &
     &         nsamp,2.d0*csum/nsamp;
      call prtmx (fsum,np,np,np,"favg");
      call prtmx (fsum2,np,np,np,"f2avg");

      stop;
      end;
!***********************************************************
      subroutine filter (cost,grad,fis,fis2,  sig0,thetae);
      use modelMod, only: phi,q,dphi,dq,drdt,n,np,qsmod,amod,           &
     &                   mmeas,get_phiq,get_yh,nmeas,tmax,dt,yts,xt0;
      implicit none;
      real(8),intent(in) :: sig0(n);      !initial state 1-sigma
      real(8),intent(in) :: thetae(np);   !model parameters used in filter
      real(8),intent(out) :: cost;        !cost function (J = negative) log-likelihood)
      real(8),intent(out) :: grad(np);    !gradient of cost wrt thetae
      real(8),intent(out) :: fis(np,np);  !Fisher information matrix (expected hessian)
      real(8),intent(out) :: fis2(np,np); !part of Fisher information that should be 0

      integer(4) i,j,k,im,ii;
      real(8) xf(n),p(n,n);
      real(8) dxf(n,np),dp(n,n,np),xfa(n);
      real(8) yt(mmeas),yp,h(n),hp(n),dhp(n),g(n),dg(n),dc(np);
      real(8) phipt(n,n),dphip(n,n),phit(n,n),qt(n,n);
      real(8) t,dy,r,c,hdx(np);
      logical(4) :: debug=.false.;     !activates debug print
      character(10) pname;
      !_________________________________________________

      !** initialize variables
      cost =0.d0;
      grad(:) =0.d0;
      fis(:,:) =0.d0;
      fis2(:,:) =0.d0;

      xf(:) =xt0(:);     !initialize filter estimate to truth
      p(:,:) =0.d0;
      forall (i=1:n) p(i,i) =sig0(i)**2;
      dxf(:,:) =0.d0;
      dp(:,:,:) =0.d0;

      !** compute dynamic model phi, q, dphi, dq
      call get_phiq (thetae);   !returns phi,q,dphi,dq

      if (debug) then;
        call prtmx (thetae,1,1,np,"thetae");
        call prtmx (phi,n,n,n,"phi");
        call prtmx (phit,n,n,n,"phit");
        call prtmx (q,n,n,n,"q");
        call prtmx (qt,n,n,n,"qt");
        do i=1,np;
          write (pname,'("dphi-",i1)') i;
          call prtmx (dphi(:,:,i),n,n,n,trim(pname));
          write (pname,'("dq-",i1)') i;
          call prtmx (dq(:,:,i),n,n,n,trim(pname));
        enddo;
        stop;
      endif;

      !** time loop for measurements
      t =0.d0;
      do ii=1,nmeas;
        t =t+dt;
        if (t > tmax) exit;

        !** filter time update
        xfa(:) =xf(:);
        xf(:) =matmul(phi,xf);
        phipt(:,:) =matmul(p,transpose(phi));
        p(:,:) =matmul(phi,phipt) +q(:,:);

        do i=1,np;
          dxf(:,i) =matmul(phi,dxf(:,i)) +matmul(dphi(:,:,i),xfa(:));
          dphip(:,:) =matmul(dphi(:,:,i),phipt(:,:));
          dp(:,:,i) =matmul(phi,matmul(dp(:,:,i),transpose(phi)))       &
     &               +dphip(:,:) +transpose(dphip) +dq(:,:,i);
        enddo;

        !** filter meas update
        yt(:) =yts(:,ii);
        do im=1,mmeas;
          !** compute predicted meas, residual and partials
          call get_yh (yp,dy,h,r,  yt,xf,thetae,im);
          hp(:) =matmul(h,p);
          c =dot_product(h,hp) +r;
!         write (6,'("yt,yp,r =",3g13.5)') yt,yp,r;
!         call prtmx (xf,1,1,n,"xf");
          cost =cost +0.5d0*(dy**2/c +log(c) +1.83788d0);   !avg dy**2/c for 1000 samples =993.05
          g(:) =hp(:)/c;                                    !Kalman gain
          xf(:) =xf(:) +g(:)*dy;
!          forall (i=1:n) p(:,i) =p(:,i) -g(:)*hp(i)
          !** make p symmetric
          do i=1,n;
            do j=1,i;
              p(j,i) =p(j,i) -g(j)*hp(i);
              p(i,j) =p(j,i);
            enddo;
          enddo;

          !** filter partials
          do i=1,np;
            dhp(:) =matmul(h,dp(:,:,i));
            hdx(i) =dot_product(h,dxf(:,i));
            dc(i) =dot_product(h,dhp);
            dc(i) =dc(i) +drdt(i);
            dg(:) =(dhp(:) -g(:)*dc(i))/c;
            dxf(:,i) =dxf(:,i) -g(:)*hdx(i) +dg(:)*dy;
!            forall(j=1:n) dp(:,j,i) =dp(:,j,i) -g(:)*dhp(j) -dg(:)*hp(j)
            !** compute symmetric dp
            do j=1,n;
              do k=1,j;
                dp(k,j,i) =dp(k,j,i) -g(k)*dhp(j) -dg(k)*hp(j);
                dp(j,k,i) =dp(k,j,i);
              enddo;
            enddo;

            !** now compute gradient and Fisher information matrix
            grad(i) =grad(i) +0.5d0*dc(i)/c -0.5d0*dc(i)*(dy/c)**2      &
     &                 -dy*hdx(i)/c;
            do j=1,i;
              fis(j,i) =fis(j,i) +0.5d0*dc(i)*dc(j)/c**2+hdx(i)*hdx(j)/c;
              fis(i,j) =fis(j,i);
              fis2(j,i) =fis2(j,i) +hdx(i)*dc(j)*dy/c**2                &
     &                             +hdx(j)*dc(i)*dy/c**2;!+
              fis2(i,j) =fis2(j,i);
            enddo;
!           write (6,'(i2,", c,hdx,dc,grad =",4g13.5)')                 &
!    &               i,c,hdx(i),dc(i),grad(i);
          enddo;     !___ filter partial loop (i)
        enddo;    !___ im loop

      enddo;   !___ ii loop

      return;
      end subroutine filter;
!***********************************************************
      subroutine sim_meas (sig0,theta);
      !** sim_meas generates simulated random measurements using the
      !** model parameters defined by
      use modelMod, only: phi,q,n,np,dt,tmax,nmeas,yts,xt0,             &
     &                   mmeas,mmax,get_phiq,get_yh;
      implicit none;
      real(8),intent(in) :: sig0(n);      !initial state 1-sigma (not used but could be used to randomly initialize xt0)
      real(8),intent(in) :: theta(np);    !true model parameters

      integer(4) i,im;
      real(8) xt(n),yp,h(n),phit(n,n),qt(n,n);
      real(8) t,rnor,r,dy,yt(mmeas);
      !_________________________________________________

      xt0(:) =0.d0;
      xt0(1) =10.;     !set x-component to +10  ###
      xt(:) =xt0(:);
      yt(:) =0.d0;

      !** compute phi and q
      call get_phiq (theta);
      phit(:,:) =phi(:,:);
      qt(:,:) =q(:,:);
      nmeas =0;
      t =0.d0;

      !** time loop
      do;
        t =t+dt;
        if (t > tmax) exit;

        !** time update of state with added diagonal process noise
        xt(:) =matmul(phit,xt);
        do i=1,n;
          xt(i) =xt(i) +sqrt(qt(i,i))*rnor(0);
        enddo;

        nmeas =nmeas+1;
        if (nmeas > mmax) stop 'sim_meas error: nmeas > mmax';

        !** compute measurements and add measurement noise
        yts(:,nmeas) =0.d0;
        do im=1,mmeas;
          call get_yh (yp,dy,h,r,  yt,xt,theta,im);
          yts(im,nmeas) =yp +sqrt(r)*rnor(0);
        enddo;
      enddo;

      return;
      end subroutine sim_meas;

