!**  MLE_parameterID.f is a general-purpose program for computing the maximum-likelihood
!**  estimates of model parameters.  It uses a Kalman filter to compute partial derivatives of the
!**  of the filter measurement innovation (residual) and covariance with-respect-to (WRT)
!**  the model parameters.  This is used to compute the negative log likelihood, gradient
!**  WRT model parameters, and Fisher information matrix.  These variables are used in constrained
!**  scoring (Gauss-Newton) iterations to converge on the model parameters minimizing the negative
!**  log likelihood.

!**  Typical parameters that can be estimated via MLE include 1) initial state,
!**  3) measurement noise variances, 4) process noise PSD, 5) dynamic constants (e.g.,
!**  feedback constants of Markov processes).

!**  Module modelMod contains the model-specific information and MLE_parameterID is the driver for
!**  the filter calculations.

!** A "keyword-type" control input text file sets control parameters, sets initial state values
!** and defines parameters to be estimated via MLE.  The acceptable keyword options are:

!  1) MeasFile   - defines the input file name of measurements
!  2) TEpoch     - sets the start time of measurement processing
!  3) Tend       - sets the end time of measurement processing
!  4) Rvar       - defines the measurement noise variance(s) for a given sensor (first input #)
!  5) Bayesian   - if the keyword is present, the MLE solution will be Bayesian (prior
!                  estimates are weighted in solution)
!  6) Debug      - sets debug print flags for up to 10 different options (only index 1, 4, 5, and 6 are currently used)
!  7) optlSet    - sets the maximum number of scoring iterations and the convergence test
!  8) LMopt      - indicates that Levinburg-Marquard is to be used rather than simple scoring steps,
!                  and sets the initial and maximum allowed lambda values
!  9) StateInit  - denotes that the following records will input state initial values:
!                   index = state index
!                   name = state name as hard-coded in modelMod.
!                   x0 = initial state value at TEpoch
!                   sig0 = 1-sigma uncertainty
!                   Qs = power spectral density of process noise
!                   const = constant associated with the state (e.g, Markov process feedback constant)
! 10) endState   - terminates StateInit input
! 11) xParm      - denotes that the following records will input MLE parameter values:
!                   index = parameter index
!                   name = parameter name: The names MUST be of the following form:
!                     C-xx is a model constant where the number (C-1, C-2, C-3, ...)
!                          is defined in modelMod
!                     QS-xx is a model process noise power spectral density onstant where the number
!                          (QS-1, QS-2, QS-3, ...) is the state index (as defined in modelMod)
!                     R-xy is a model measurement noise variance where the first number (x) is
!                          the sensor index, and the 2nd number (y) is the index of the measurement
!                          for the sensor. e.g., R-12 is for sensor 1, measurement type 2.
!                     x0-xx is the initial value for state xx, i.e., x0-1, x0-2, ...
!                   xparm = initial estimate of parameter (will ovveride values from above)
!                   sig0 = 1-sigma uncertainty on parameter if Bayesian solution is specified.
! 12) endXparm    terminates xParm input

!Sample input file (sim_6state.cfg) for test program

!# state setup file for 6-state (x, xdot, y, ydot, bias1, bias2) tracking problem
!#  to be used for testing the MLE_parameterID program.
!#  The first 4 state consist of two independent 1st order markov velocity processes
!#  with integral.  Measurements are range and bearing.
!# Notes: 1) Lines beginning with # are ignored
!#        2) First 12 characters of lines (except stateInit data) are a keyword.
!#           Accepted values are TEpoch, TEnd, MeasFile, Rvar, Debug, optlSet, LMopt
!#           StateInit, xParm, endState or endXparm. Other inputs are read as freeform
!#           (blanks or comma separator).
!#
!#
!TEpoch       0.
!TEnd         101.
!MeasFile     sim_4state.txt
!Rvar          1   0.5   0.001
!Bayesian
!#Debug          T T T T T T T T T T
!optlSet        10   0.02
!#LMopt         1.d-4  10.d0
!StateInit    (index  name   x0    sig0    Qs      const)
!   1                 x      10.0   2.     0.       0.
!   2                 xdot   0.0    4.     0.8      0.1
!   3                 y      0.0    3.     0.       0.
!   4                 ydot   0.0    4.     1.6      0.2
!   5                 bias1  0.0    4.     0.      0.
!   6                 bias2  0.0    4.     0.      0.
!endState
!xParm        (index  name  xparm    sig0)
!   1                 C-2   0.1     10000.
!   2                 C-4   0.2     10000.
!   3                 QS-2  0.8     10000.
!   4                 QS-4  1.6     10000.
!   5                 R-11   0.5     10000.
!   6                 R-12   0.001   10000.
!   7                 x0-1   11.    10000.
!   8                 x0-2   1.     10000.
!   9                 x0-3   0.     10000.
!  10                 x0-4   0.     10000.
!endXparm

!**  Search for "###" to find parameters that are typically changed

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort modelMod.f90 /c
!**   ifort MLE_parameterID.f90 modelMod.obj
!**   MLE_parameterID  MLE_4state.cfg  >tst_MLE4.lis

!**     OR to include measurement biases
!**   ifort modelModB.f90 /c
!**   ifort MLE_parameterID.f90 modelMod.obj
!**   MLE_parameterID  MLE_6state.cfg  >tst_MLE6.lis
!**   MLE_parameterID  MLE_6stateP.cfg  >tst_MLE6P.lis

!**********************************************************************
      module MLopt
      !** contains global variables used in MLE_parameterID

      implicit none;
      logical(4) :: first =.true.;     !initializes information arrays
      logical(4) :: partials           !simulate measurements. Otherwise read from measFile
      logical(4) :: lprint(10) =.false. !controls debug print
      logical(4) :: udfilt =.false.    !MLE_parameterID cannot use udfilt at current time
      logical(4) :: bayesian =.false.  !if true, use add prior on estimated ML parameters

      real(8) tend;                    !end time for filter processing
      real(8) :: editThresh =5.        !Kalman filter editing threshold
      real(8) :: convtst =0.02         !scoring convergence test
      real(8) :: lambda0 =1.d-4        !Levinburg-Marquard initial lambda
      real(8) :: maxLambda =10.d0      !Levinburg-Marquard maximum allowed lambda
      integer(4) :: maxScore =10;      !max number of scoring iterations ###
      logical(4) :: LMopt =.false.     !default is simple scoring steps (not Levinburg-Marquard)

      real(8),allocatable :: grad(:)   !gradient of -2 *log(likelihood) WRT parameters
      real(8),allocatable :: fis(:)    !Fisher information matrix

      end module MLopt
!*****************************************************************************
      program MLE_parameterID;
      !** General-purpose maximum-likelihood parameter estimation program, using a
      !** Kalman filter to compute the likelihood function

      use modelMod, only:nxp,nxp3,xparm,t0,tmax,xparm0,get_meas;
      use MLopt, only:grad,fis,partials,convtst,lambda0,maxLambda,       &
     &                maxScore,lprint,LMopt;
      use filterPrec, only: real_kind1;   !precision (4 or 8)
      implicit none;

      integer(4) lenferr,nargs,ier,istat,nmeaso;
      character(80) line;
      character(80) setupFile;  !name of file containing control setup
      character(80) measFile    !name of file containing measurement data
      integer(4) :: i;
      real(8) cost;
      real(8),allocatable :: fisInv(:),dxp(:);
      !______________________________________________

      !** MLE_parameterID will only work in double precision
      if (real_kind1 /= 8) call errhand ('error: real_kind1 /= 8')

! ....  Read run control file
      lenferr =nargs()-1;         !get number of command line arguments
      if (lenferr == 0) then;
        write (6,'(" Input setup file:")');
        read (5,'(a40)') setupFile;
      else;
        call getarg(1,line,ier);    !get command line argument
        read (line,*) setupFile;
        if (ier < 0) call errhand ('ERROR IN RETRIEVING setupFile');
      endif;

      !** Read text control file to set the xstate structure.  Note that variables
      !** nEstf and xnames must have been set in module modelMod before calling setup.
      call read_setup (measFile,  setupFile);

      write (6,'("t0 =",g13.5,", tmax =",g13.5)') t0,tmax;
      write (6,'("xparm0 =",10g13.5)') xparm(:nxp);
      xparm0(:) =xparm(:);    !save initial parameters for prior

      nxp3 =nxp*(nxp+1)/2
      allocate (grad(nxp),fis(nxp3),fisInv(nxp3),dxp(nxp),              &
     &          stat=ier);
      if (ier /= 0) call errhand                                        &
     &   ("MLE_parameterID error: allocate grad, fis");

!**   get measurements and load into measDat structure
      call get_meas (measFile);

      !** Scoring iterations:
      if (maxScore > 0) then;
        if (LMopt) then;
          !** Levinburg-Marquard iterations
          call optl_LM (cost,nmeaso,istat,  xparm,fisInv,               &
     &                nxp,maxScore,convtst,lambda0,maxLambda,lprint(2));
        else
          !** simple Gauss-Newton scoring iterations
          call scoring (cost,fisInv,nmeaso,istat,  maxScore);
        endif;

        call prtsmx (fisInv,nxp,"fisInv");
        write (6,'("1-sigma =",10g13.5)')                               &
     &        (sqrt(fisInv(i*(i+1)/2)),i=1,nxp);
        write (6,'(/"******************"/"nmeas =",i5,", cost =",       &
     &         g13.5,", optimization status =",i4)') nmeaso,cost,istat;

      else
        !** just evaluate cost function (fisInv,fis,grad are not set)
        partials =.false.;
        call funct (cost,fisInv,fis,grad,dxp,nmeaso,                    &
     &              partials,xparm,nxp,.true.);
        write (6,'(/"******************"/                               &
     &         "nmeas =",i5,", cost =",g13.5)') nmeaso,cost;
      endif

      stop;
      end program;
!******************************************************************
      subroutine scoring (cost,fisInv,nmeas,istat,  maxScore);
!**     computes simple Gauss-Newton scoring iterations
      use modelMod, only:nxp,nxp3,xparm;
      use MLopt, only:grad,fis,partials,convtst;
      implicit none;

      integer(4),intent(in) :: maxScore;   !max number of scoring iterations
      real(8),intent(out) :: fisInv(nxp*(nxp+1)/2);
      real(8),intent(out) :: cost;
      integer(4),intent(out) :: nmeas;
      integer(4),intent(out) :: istat;

      integer(4) :: iscore,i;
      real(8) converg,dxp(nxp);
      !_________________________________________________

      do iscore =1,maxScore;          !scoring iterations
        !** execute Kalman filter on all measurments to compute log likelihood,
        !** gradient and Fisher information matrix
        partials =.true.
        call funct (cost,fisInv,fis,grad,dxp,nmeas,                     &
     &              partials,xparm,nxp,.true.);

        write (6,'(/5("**********")/"Iteration",i3/                     &
     &         "nmeas =",i5,", cost =",g13.5)') iscore,nmeas,cost;
        call prtsmx (fis,nxp,"Fisher Info Matrix");
        write (6,'("grad =",10g13.5,10(/6x,10g13.5))') grad(:);
!       d(:) =grad(:) -matmul(fis,dxp);
!       write (6,'("grad-fis*dxp =",10g13.5)') d(:);
        converg =sqrt(-dot_product(dxp(:),grad)/nxp);
        write (6,'("converg test=",g13.5,", scoring step =",            &
     &         10(/10g13.5))') converg,dxp(:);
        xparm(:nxp) =xparm(:nxp) +dxp(:);
        write (6,'("new xparm =",10g13.5,10(/6x,10g13.5))') xparm(:nxp);
        if (converg < convtst) exit;
      enddo;

      call prtsmx (fisInv,nxp,"fisInv");
      write (6,'("1-sigma =",10(/6x,10g13.5))')                         &
     &       (sqrt(fisInv(i*(i+1)/2)),i=1,nxp);
      istat =0;         !

      return;
      end subroutine scoring;
!*******************************************************************
      subroutine funct (logLike,fisInv,fiso,grado,dxp,nmeaso,            &
     &                  partial,xparm,nxp,debugp);
      !** execute Kalman filter on measurements to compute the log likelihood,
      !** gradient and fisher information matrix

      use modelMod, only:nmeas,mmeas,measDat,xf,pf,ud,rvar,t0,          &
     &                   init_filter,xparm0,xParSig,nEstf;
      use MLopt, only:grad,fis,bayesian,partials,udfilt;        !grad,fis are output
      use optlDat, only:fstat;
      implicit none;

      integer(4),intent(in) :: nxp;             !number of estimated parameters
      real(8),intent(in) :: xparm(nxp);         !parameter vector
      logical(4),intent(in) :: partial;         !if true, partial derivatives of cost (i.e.,
                                                ! au, bu, pf) are evaluated. Otherwise only cost and last bu.
      logical(4),intent(in) :: debugp;          !activates debug print

      real(8),intent(out) :: logLike;           !cost function (= -2*log(likelihood))
      real(8),intent(out) :: grado(nxp);        !cost gradient wrt xparm
      real(8),intent(out) :: fiso(nxp*(nxp+1)/2); !Fisher information matrix
      real(8),intent(out) :: fisInv(nxp*(nxp+1)/2); !inverse Fisher information matrix
                                                    ! (error covariance amtrix)
      real(8),intent(out) :: dxp(nxp);           !Gauss-Newton (scoring) step
      integer(4),intent(out) :: nmeaso;          !number of measurements processed

      !** local variables
      integer(4) i,k,im,isensor,ier,ibias;
      logical(4) ledit;
      real(8) tobs,tLst,z(mmeas),zsig,chiSq,rvar1,eps,errm;
      !_________________________________________________________________

      partials =partial;
      call init_filter (xparm,nxp);    !initialize the Kalman filter
      if (udfilt) then
        ud(:) =pf(:)
        call cov2ud_m (ud,nEstF);
      endif

      tLst =t0;
      ibias =0;
      grad(:) =0.d0;
      fis(:) =0.d0;
      logLike =0.d0;
      fstat%rrms =0.d0

      if (bayesian) then;
        !** set prior (for Bayesian solution - optional)
        logLike =0.5*sum((xparm(:)-xparm0(:))/xparSig(:))**2;
        grad(:) =(xparm(:)-xparm0(:))/xparSig(:)**2;
        k=0;
        do i=1,nxp;
          k =k+i;
          fis(k) =1.d0/xparSig(i)**2;
        enddo;
      endif;

      !** run Kalman filter with partial derivatives to compute negative log
      !** likelihood, gradient and fisher information matrix
      do i=1,nmeas;
        tobs =measDat(i)%time;
        z(:) =measDat(i)%y(:);
        isensor =measDat(i)%isensor;

        !** Kalman filter time update
        call KF_timeUp (xf,pf,  tobs,tLst);

        !** Kalman filter measurement update
        do im=1,mmeas;
          Rvar1 =rvar(im,isensor);
          call KF_measUP (zsig,chiSq,ledit,   logLike,xf,pf,          &
     &                     z(im),Rvar1,im,isensor,tobs,ibias);
        enddo;
        tLst =tobs;
      enddo;

      if (partials) then;
        !** invert Fisher information matrix
        eps =1.d-5;
        fisInv(:) =fis(:);
        call sinv (ier,errm,  fisInv,  nxp,eps,'false');
        if (ier /= 0)                                                     &
     &   call errhand ('cannot invert Fisher information matrix');

        !** Compute Gauss-Newton step: dxp =-inv(fis) *grad
        call symult (dxp,  fisInv,grad,nxp);
        dxp(:) =-dxp(:);

        !** set output
        grado(:) =grad(:);
        fiso(:) =fis(:);
      endif

      nmeaso =nmeas;
      fstat%rrms =sqrt(fstat%rrms/nmeas);

      return
      end subroutine funct
!*******************************************************************
      subroutine KF_timeUp (xf,pf,  tobs,tKLst);
!**    Kalman filter time update with added calculations for partial derivatives WRT MLE parameters
      use modelMod, only:ud,psym,nEstF,nF3,nd3,nDynam,                 &
     &             phi,qd,dphidxp,dqdxp,dxdxp,dpdxp,dqdxp,nxp,pname;
      use MLopt, only:lprint,udfilt,partials;
      implicit none;

      real(8),intent(in) :: tobs,tKLst;         !current & last time
      real(8),intent(inout) :: xf(nEstF),pf(nF3);

      !** local variables
      integer(4) i,j,k,l,ierr,ip;
      integer(4),parameter :: method =2;  !UD time update method  ###
      character(12) name
      real(8) delt;
      real(8),allocatable,save :: phip(:,:),xp(:),pp(:,:),phipt(:,:);
      real(8),allocatable,save :: d(:),d2(:,:),v(:),vd(:),w(:,:),dw(:);   !,xrefL(:)
      real(8),allocatable,save :: xfl(:),xref(:);
      !_______________________________________________________

      if (.not.allocated(phip)) then;
        write (6,'("KF_timeUp allocate: nestf =",i4,", ndynam =",i4,    &
     &              ", tobs =",g13.5)') nestf,ndynam,tobs
        allocate (phip(nEstF,nEstF),xp(nEstF),pp(nEstF,nEstF),          &
     &          phipt(nEstf,nEstf),d(nEstF),d2(ndynam,nEstF),           &
     &          v(2*nEstF),vd(nEstF),w(nEstF,2*nEstF),                  &
     &          dw(2*nEstF),xfl(nEstF),xref(nEstf),stat=ierr);
        if (ierr > 0) call errhand ('KF_timeUp allocate failed');
      endif;

      delt =tobs -tKlst;
      if (lprint(1)) write (6,'(/"****************"/"KF_timeUp: tobs =", &
     &          f15.4,", delt (sec) =",g13.5)') tobs, delt;
      xfl(:) =xf(:);        !save a posteriori from last step

      !***  time update

!***   Compute xp (predicted state), phi and q.
!***   If partials = .true., also compute dphidxp, dqdxp and update dxdxp.
!***   The following code (or equivalent) should appear in get_phiq

!      call smult (xp,  phi,xf, ndynam,nEstf,1, ndynam,nEstf,nEstF);   !xp =phi*xf
!      if (nEstf > ndynam) xp(ndynam+1:) =xf(ndynam+1:);

!      if (partials) then;
        !** dxdxp <= phi*dxdxp +dphidxp*xf
        !** This expression assumes linear system.  If nonlinear must integrate nonlinear eqns.
!        do ip=1,nxp;
!          call smult (d,  phi,dxdxp(:,ip),                             &  !d =phi*dxdxp
!     &                ndynam,nEstf,1, ndynam,nEstF,ndynam);
!          dxdxp(:ndynam,ip) =d(:ndynam);
!          if (pname(ip)(1:2) == "C-") then
!            call smult (d,  dphidxp(:,:,ip),xf(:),                     &  !xp =dphidxp*xf
!     &                  ndynam,nEstf,1, ndynam,nEstF,ndynam);
!            dxdxp(:ndynam,ip) =dxdxp(:ndynam,ip) +d(:ndynam);
!          endif;
!        enddo;
!      endif;

      call get_phiq (phi,qd,xp,   tKLst,tobs,xf,partials);
      xf(:) =xp(:);

      if (partials .and. lprint(5)) then;
        do ip=1,nxp;
          write (name,'("dphidxp-",i2.2)') ip;
          call prtmx (dphidxp(:,:,ip),ndynam,ndynam,nEstF,trim(name));
          write (name,'("dqdxp-",i2.2)') ip;
          call prtsmx (dqdxp(:,ip),ndynam,trim(name));
        enddo;
      endif;

      if (lprint(4)) then;
        call prtmx (phi,ndynam,ndynam,nEstF,"phi");
        call prtsmx (qd,ndynam,"qd");
        call prtmx (xf,1,1,nEstF,"xf_i-1");
!        call prtmx (xref,1,1,nState,"xref-");
        if (udFilt) then
          call ud2cov_m (pf,  ud,nEstF);
          call prtsmx (pf,nEstF,"pf_i-1");
        else
          call prtsmx (pf,nEstF,"pf_i-1");
        endif
      endif;

      if (udFilt) then;
        if (partials) call errhand ("udfilt does not support partials");

        !** UD filter time update
        call phiu_m (phip,  phi,ndynam,ndynam,nEstF, ud,nEstF,nEstf);

        if (nEstf > ndynam) then;
          phip(ndynam+1:nEstf,:) =0.d0;
          do i=ndynam+1,nEstf;
            k =(i*(i-1))/2 +ndynam;
            do j=ndynam+1,i;
              k=k+1;
              phip(j,i) =ud(k);
            enddo;
            phip(i,i) =1.d0;
          enddo;
        endif;

        l=0;
        do i=1,nEstF;
          l=l+i;
          d(i) =ud(l);
        enddo;

        if (method == 1) then;
          !** compute phi*u, square ud factors, add qd, and re-factor
          !** (squaring is not as accurate as WGS)
          call triphu (ud,v,  phip,vd,   nEstF,nEstF,d);
          call ud2cov_m (psym,  ud,nEstF);
          l =0;
          do i=1,nEstF;
            do j=1,i;
              l=l+1;
              if (i <= ndynam) psym(l) =psym(l) +qd(l);
              ud(l) =psym(l);
            enddo;
          enddo;
          call cov2ud_m (ud,nEstF);

        else;
          !** weighted gram-schmidt  (could modify to work with ndynam partition)
          w(:,1:nEstF) =phip(:,:);
          dw(1:nEstF) =d(:);
          psym(:) =0.d0;
          psym(:nd3) =qd(:);
!          call prtsmx (psym,nEstf,"sqrt-qd")

          call cov2ud_m (psym,nEstf);
          w(:,nEstF+1:2*nEstF) =0.d0;
          l =0;
          do i=1,nEstF;
            do j=1,i;
              l=l+1;
              w(j,i+nEstf) =psym(l);
            enddo;
            w(i,i+nEstf) =1.d0; !replaces diagonal
            dw(i+nEstf) =psym(l);
          enddo;
!          call prtmx (w,nEstF,nEstF,2*nEstF,"w")
!          call prtmx (dw,1,1,2*nEstF,"dw")

          !*** weighted gram-schmidt orthogonalization
          call wgs_m (ud,v,  w,  nEstF,nEstF,2*nEstF,dw,.true.);

!          call prtmx (w,nEstF,nEstF,2*nEstF,"wgs-w")
!          call prtsmx (ud,nEstf,"wgs-ud")
          call ud2cov_m (pf,  ud,nEstF);
        endif;

      else;
        !** Kalman filter covariance time update
        psym(:) =pf(:);     !save a posteriori from last step
        k =0;
        do i=1,nEstF;
          do j=1,i;
            k =k+1;
            pp(i,j) =pf(k);
            pp(j,i) =pf(k);
          enddo;
        enddo;

        !**  phip =phi * pp
        call smult (phip,  phi,pp,                                      &
     &              ndynam,nEstf,nEstF,  ndynam,nEstF,nEstF);
        !** phipt =pp * phi^T
        phipt(:nEstf,:ndynam) =transpose(phip(:ndynam,:nEstf));
        if (nEstF > ndynam) then;
          phipt(:nEstf,ndynam+1:nEstf)                                  &
     &                        =transpose(pp(ndynam+1:nEstf,:nEstf));
        endif
        !** w =phi * phipt = phi * pp * phi^T
        call smult (w,  phi,phipt, ndynam,nEstf,nEstf,                  &
     &                             ndynam,nEstF,nEstF);
        if (nEstF > ndynam) then;
          !** copy bias portion to w
          w(ndynam+1:nEstF,ndynam+1:nEstF) =                            &
     &               phipt(ndynam+1:nEstF,ndynam+1:nEstF);
        endif;

        !**   pf = phi * pp * phi^T +qd
        k =0;
        do i=1,nEstF;
          !** make symmetric using upper triangular partition of phip and qd
          do j=1,i;
            k =k+1;
            pf(k) =w(j,i);
            if (k <= nd3) pf(k) =pf(k) +qd(k);
          enddo;
        enddo;

        if (partials) then;
          do ip =1,nxp;
            if (pname(ip)(1:2) == "C-") then
!              read (pname(ip)(3:),*) j
              !** d2 =dphidxp * phipt = dphidxp * pp * phi^T
              call smult (d2,                                             &
     &                  dphidxp(:ndynam,:nEstF,ip),phipt(:nEstf,:nEstf),  &  !
     &                    ndynam,nEstF,nEstF, ndynam,nEstF,ndynam);
            else
              d2(:,:) =0.d0
            endif

            k =0;
            do i=1,nEstF;
              do j=1,i;
                k =k+1;
                pp(i,j) =dpdxp(k,ip);
                pp(j,i) =dpdxp(k,ip);
              enddo;
            enddo;

            !** compute phip =phi * pp * phi^T =phi * dpdxp * phi^T
            !** first, phip = phi * pp = phi * dpdxp
            call smult (phip,  phi,pp,                                   &
     &                  ndynam,nEstf,nEstF,  ndynam,nEstF,nEstF);
            !** w = phip^T = dpdxp * phi^T
            w(:nEstf,:ndynam) =transpose(phip(:ndynam,:nEstf));
            if (nEstF > ndynam) then;
              w(:nEstf,ndynam+1:nEstf)                                   &
     &                        =transpose(pp(ndynam+1:nEstf,:nEstf));
            endif

            !** phip =phi * w =phi * dpdxp * phi^T
            call smult (phip,  phi,w,                                       &
     &                  ndynam,nEstf,nEstf, ndynam,nEstf,nEstF);
            if (nEstF > ndynam) then;
              !** copy bias portion
              phip(ndynam+1:nEstF,ndynam+1:nEstF) =                      &
     &                     pp(ndynam+1:nEstF,ndynam+1:nEstF);
            endif;

            !**  dpdxp =phip + d2 +d2^T +dqdxp
            !**        =phi*dpdxp*phi^T +(dphidxp*pp*phi^T) +(phi*pp*dphidxp^T +dqdxp
            k =0;
            do i=1,nEstF;
              !** make symmetric using upper triangular partition of phip and qd
              do j=1,i;
                k =k+1;
                dpdxp(k,ip) =phip(j,i)
                if (j <= ndynam) then
                  dpdxp(k,ip) =dpdxp(k,ip) +d2(j,i);
                  if (i <= ndynam) then;
                    dpdxp(k,ip) =dpdxp(k,ip) +d2(i,j) +dQdxp(k,ip);
                  endif
                endif;
              enddo;
            enddo;

            if (lprint(5)) then
              write (name,'("dxdxp-",i2.2)') ip
              call prtmx (dxdxp(:,ip),1,1,nEstF,trim(name))
              write (name,'("dpdxp-",i2.2)') ip
              call prtsmx (dpdxp(:,ip),nEstF,trim(name))
            endif
          enddo;     !___ end ip loop
        endif;

      endif;

      if (lprint(4)) then;
        call prtmx (xf,1,1,nEstF,"xp-");
        if (udFilt) then
          call ud2cov_m (pf,  ud,nEstF);
          call prtsmx (pf,nEstF,"pp-");
        else
          call prtsmx (pf,nEstF,"pp-");
        endif
      endif;

      return;
      end subroutine KF_timeUp;
!*******************************************************************
      subroutine KF_measUP (zsig,chiSq,ledit,   logLike,xf,pf,          &
     &                     zmeas,Rvar,meastyp,isensor,tobs,ibias);
!****   Kalman filter scalar measurement update with added calculation of partials WRT MLE parameters

      use modelMod, only:ud,nEstF,nF3,phi,qd,dphidxp,dqdxp,dxdxp,dpdxp,  &
     &                 nxp,pname,get_yh;
      use MLopt, only:grad,fis,lprint,udfilt,partials,editThresh;
      use optlDat, only:fstat;
      implicit none;

      integer(4),intent(in) :: meastyp;    !measurement type for given sensor
      integer(4),intent(in) :: isensor;    !measurement sensor
      integer(4),intent(in) :: ibias;      !index in xf of meas bias (not used)
      real(8),intent(in) :: zmeas;         !measurement
      real(8),intent(in) :: Rvar;          !measurement noise variance
      real(8),intent(in) :: tobs;          !obs time
      real(8),intent(inout) :: xf(nEstF);  !filter state estimate
      real(8),intent(inout) :: pf(nF3);    !filter error covariance
      real(8),intent(inout) :: logLike;    !-log(likelihood)
      real(8),intent(out) :: zsig;         !1-sigma of measurement residual
      real(8),intent(out) :: chiSq;        !measurement residual Chi-squared
      integer(4),intent(out) :: ledit;     !measurement editing flag: 0=accept, 1=below elev cutoff, 2=chisq edit

      !** local variables
      character(10) rname
      character(12) name
      integer(4) i,j,k,im,ip,kp;
      logical(4) :: covar =.true.; !very slow (15 -> 100 sec)
      real(8) hp(nEstF),h(nEstF),kg(nEstF),hdp(nEstF),Pz;
      real(8) f(nEstf),g(nEstF),dxSig(nEstf),z,zp,dz,zres;
      real(8) dkdxp(nEstF),dPzdxp(nxp),dzresdxp(nxp);
      !_______________________________________________________

      im =meastyp;
      write (rname,'("R-",i2.2)') isensor*10+meastyp
      lEdit =0;   !default is accept meas

!     call prtmx (xf,1,1,nEstF,"xp-");
!     write (6,'(" obs time=",f17.7)') tobs;

!***   set predicted meas., residual and partials (h)
      call get_yh (zp,zres,h,  zmeas,xf,isensor,meastyp);

!***    measurement update
      lEdit =0;

      if (udFilt) then;
        !** UD filter measurement update
        z =0.;
        call udmes1 (f,g,Pz,dz,  ud,nEstf,Rvar,h,xf,z);

        chiSq =zres**2/Pz;
        if (chiSq > editThresh**2) then;
          write (6,1100) tobs, zres,chiSq;
          lEdit =2;
          return;
        endif;

        call udmes2 (g,ud,xf,  nEstF,Rvar,f,zres);
        kg(:) =g(:)/Pz;
        hp(:) =g(:);    !save for ??

        if (covar .or. lprint(4)) then;
          !** compute covariance (this is very slow computationally)
          call ud2cov_m (pf,  ud,nEstF);
          do i=1,nEstF;
            dxsig(i) =sqrt(kg(i)*hp(i));
          enddo;
        endif;

      else;
        !** covariance Kalman filter update
        call sysmult (hp,  pf,h,nEstF);         !sparse symmetric mult: hp =pf*h
        Pz =Rvar +dot_product(hp,h);
        kg(:) =hp(:)/Pz;
        chiSq =zres**2/Pz;
        fstat%rrms =fstat%rrms +chiSq;

        if (chiSq > editThresh**2) then;
          write (6,1100) tobs, isensor,meastyp,zres,chiSq;
 1100     format("large meas residual at ",g16.8,                           &
     &           ", sensor ",i3,", meas ",i3,", zres =",g13.5,              &
     &            ", chi-sq =",g13.5)
          lEdit =2;
          return;
        endif;

        xf(:) =xf(:) +kg(:)*zres;
        k =0;
        do i=1,nEstF;
          dxsig(i) =kg(i)*hp(i);
          if (dxsig(i) >= 0.d0) then;
            dxsig(i) =sqrt(dxsig(i));
          else;
            write (6,'("error: dxvar(",i4,") =",g13.5)') i,dxsig(i);
            dxsig(i) =0.d0;
          endif;
          do j=1,i;
            k =k+1;
            pf(k) =pf(k) -kg(i)*hp(j);
          enddo;
        enddo;

        if (partials) then;
          kp =0;
          do ip=1,nxp;
            dzresdxp(ip) =-dot_product(h,dxdxp(:,ip));
            call sysmult (hdp,  dpdxp(:,ip),h,nEstF);         !sparse symmetric mult hdp =dpdxp(:,ip)*h
            dPzdxp(ip) =dot_product(hdp,h);   !+drdxp(ip)
            if (pname(ip) == rname) then
              !** add drdxp(ip)
              dPzdxp(ip) =dPzdxp(ip) +1.d0
!              write (6,'("ip =",i2,", dPzdxp(ip) =",g13.5)') ip,dPzdxp(ip)
            endif

            !** accumulate gradient and Fisher information
            grad(ip) =grad(ip) +(dzresdxp(ip)*zres                      &
     &                 +0.5d0*dPzdxp(ip)*(1.d0 -zres**2/Pz)) /Pz;
            do j=1,ip;
              kp =kp+1;
              fis(kp) =fis(kp) +(0.5d0*dPzdxp(ip)*dPzdxp(j)/Pz          &
     &                               +dzresdxp(ip)*dzresdxp(j)) /Pz;
            enddo;
!            write (6,'("ip =",i3,", grad =",g13.5,", fis(i,i)=",g13.5)') &
!     &             ip,grad(ip),fis(kp)

            !** update partials of state and covariance
            dkdxp(:) =(hdp(:) -kg(:)*dPzdxp(ip))/Pz;
            dxdxp(:,ip) =dxdxp(:,ip) +kg(:)*dzresdxp(ip) +dkdxp(:)*zres;
            k=0;
            do i=1,nEstF;
              do j=1,i;
                k =k+1;
                dpdxp(k,ip) =dpdxp(k,ip) -kg(i)*hdp(j) -dkdxp(i)*hp(j);
              enddo;
            enddo;

            if (lprint(5)) then;
              write (name,'("dxdxp-",i2.2)') ip
              call prtmx (dxdxp(:,ip),1,1,nEstF,trim(name))
              write (name,'("dpdxp-",i2.2)') ip
              call prtsmx (dpdxp(:,ip),nEstF,trim(name))
            endif
          enddo;  !___ end ip loop
        endif;

      endif;   !___ end partials test

      zsig =sqrt(Pz);
      chiSq =zres**2/Pz;
      logLike =logLike +0.5d0*(chiSq +log(Pz) +1.837877d0);   !check ??  (1.837877 =ln(2*pi))

      if (lprint(1)) write (6,'(" meastyp ",i3,", residual =",g13.5,    &
     &         ", resid sigma =",g13.5,", Rsig=",g13.5)')    &
     &         isensor*10+meastyp, zres, zsig, sqrt(Rvar);
!     call prtmx (xf,1,1,nEstF,"xp+");
      if (lprint(6)) then;
        write (6,'("Dxsig ="5(/10g12.4))') dxsig(:);
!        write (6,'("sig-a_tot =",3g12.4,
!     &          (sqrt(cova(i,i)),i=1,3),
      endif;

      if (lprint(4)) then;
        call prtmx (h,1,1,nEstF,"h");
        call prtmx (xf,1,1,nEstF,"xf+");
        call prtsmx (pf,nEstF,"pf+");
      endif;

      return;
      end subroutine KF_measUP;
!*****************************************************
      subroutine read_setup (measFile, setupFile);
!***   Reads text file of model, state and parameter configuration data.
!**    Additional keywords may be added for additional features.

      use modelMod, only:nEstf,ndynam,nd3,nxp,nf3,xnames,t0,                  &
     &                   xstate,defaultp,dphidxp,dqdxp,dxdxp,dpdxp,rvar,      &
     &                  xparm,xparm0,pname,nxpmax,nsensor,mmeas,xparSig;
      use MLopt, only:lprint,udfilt,editThresh,tend,bayesian,                 &
     &                convtst,lambda0,maxLambda,maxScore,LMopt;
      implicit none;
      character(*),intent(in) :: setupFile; !name of input setup file
      character(*),intent(out) :: measFile;  !name of measurement file

      integer(4) i,j,ns,nst,ier,isensor;
      character(15) keyword;
      character(80) aline;
      !_______________________________

      open (10,file=setupFile,status='old',iostat=ier);
      if (ier /= 0) call errhand ('error opening setup file');
      measFile ='';
      nxp =0;
      lprint(:) =.false.
      editThresh =5.

      !** set default parameters for filter
      call defaultp

      do;
        read (10,'(a12,a80)',end=990,err=999) keyword,aline;
        if (keyword(1:1) == '#' .or. keyword == '') cycle;     !comment
        write (*,'(a12,a)') keyword,trim(aline);

        select case(keyword);
          case ('TEpoch');
            read (aline,*) t0;
          case ('TEnd');
            read (aline,*) tEnd;
          case ('MeasFile');
            read (aline,*) measFile;
          case ("Rvar")
            read (aline,*,err=999) isensor, rvar(:,isensor);
          case ("Debug");
            read (aline,*) lprint(:)
          case ("Bayesian")
            bayesian =.true.
          case ("optlSet")
            read (aline,*) maxScore,convtst;
          case ("LMopt")
            read (aline,*) lambda0,maxLambda;
            LMopt =.true.;
          case ("UDfilt")
            UDfilt =.true.;      !doesn;t work with scoring (partial derivatives)
          case ('StateInit');
            ns =0;
            write (6,'("Input state initialization from file ",a)')      &
     &         trim(setupFile);

            !** read info on states
            do;
              read (10,'(a)',err=999,end=997) aline;
              if (index(aline,"endState") > 0) exit;
              write (*,'(a)') trim(aline);
              ns =ns+1;
              if (ns > nEstF) exit;
              read (aline,*,err=999) i, xstate(ns)%xname, xstate(ns)%x0, &  !xstate(ns)%status, xtype
     &             xstate(ns)%sig0, xstate(ns)%Qs,xstate(ns)%const(1);
              if (xstate(ns)%xname /= xnames(ns)) then;
                call errhand                                            &
     &             ('read_setup error: states names do not match');
              endif;
            enddo;  !___ end read states loop

          case ('xParm');
            nxp =0;
            write (6,'("MLE parameters from file ",a)')                  &
     &          trim(setupFile);

            !** read parameter info
            do;
              read (10,'(a)',err=999,end=997) aline;
              write (*,'(a)') trim(aline);
              if (index(aline,"endXparm") > 0) exit;
              nxp =nxp+1;
              if (nxp > nxpmax) exit;
              read (aline,*,err=999) i, pname(nxp),xparm(nxp),            &
     &                     xparSig(nxp);
            enddo;  !___ end read states loop

          case default;
            write (6,'("invalid input name ",a)') keyword;
        end select;

      enddo;

  990 continue
      if (ns /= nEstf) then
        write (6,'(/"input error: number of input states =",i4,           &
     &         ", model has ",i4)') ns, nEstf
        call errhand ('read_setup error: # states wrong');
      else if (nst > 0 .and. nst /= nEstf) then
        write (6,'(/"input error: number of input true states =",i4,      &
     &         ", model has ",i4)') nst, nEstf
        call errhand ('read_setup error: # states wrong');
      endif

      !** check that parameter names and values are valid
      do i=1,nxp
        if (pname(i)(:2) == "C-") then             !feedback constant
          read (pname(i)(3:),*,err=998) j
          if (j < 1 .or. j > nEstf) then
            write (6,'("error: invalid parameter constant ",a)')pname(i)
            call errhand ("invalid parameter constant")
          endif
        else if (pname(i)(:3) == "QS-") then        !process noise PSD
          read (pname(i)(4:),*,err=998) j
          if (j < 1 .or. j > nEstf) then
            write (6,'("error: invalid parameter QS ",a)') pname(i)
            call errhand ("invalid parameter QS")
          endif
          if (xparm(i) < 0.d0) then
            write (6,'("error: invalid parameter QS value ",a)')pname(i)
            call errhand ("invalid parameter QS")
          endif
        else if (pname(i)(:3) == "x0-") then        !state initial value
          read (pname(i)(4:),*,err=998) j
          if (j < 1 .or. j > nEstf) then
            write (6,'("error: invalid parameter x0 ",a)') pname(i)
            call errhand ("invalid parameter x0")
          endif
        else if (pname(i)(:2) == "R-") then          !measurement  noise variance
          read (pname(i)(3:),*,err=998) j
          if (j/10 < 1 .or. j/10 > nsensor) then
            write (6,'("error: invalid parameter sensor: ",a)')pname(i)
            call errhand ("invalid parameter R")
          else if (mod(j,10) < 1 .or. mod(j,10) > mmeas) then
            write (6,'("error: invalid parameter sensor: ",a)')pname(i)
            call errhand ("invalid parameter R")
          endif
          if (xparm(i) < 0.d0) then
            write (6,'("error: invalid parameter R value ",a)')pname(i)
            call errhand ("invalid parameter R value")
          endif
        else
          write (6,'("error: invalid parameter name ",a)') pname(i)
          call errhand ("invalid parameter name")
        endif;
      enddo

      !** set values and allocate arrays
      allocate (dphidxp(ndynam,nEstf,nxp),dqdxp(nd3,nxp),                &
     &          dxdxp(nEstf,nxp),dpdxp(nf3,nxp),stat=ier);               !drdt(nxp),
      if (ier > 0) call errhand ('read_setup allocate failed');
      xparm0(:) =xparm(:);

  997 close (10);
      write (6,'(/"Finished reading setup file:"/                        &
     &  "# states =",i4,", # MLE parameters =",i4)') nestF,nxp;
      write (6,'(/)')
      return;

  998 write (6,'("error reading parameter index in ",a)') pname(i)
      call errhand ('read_setup error: reading parameter index');

  999 call errhand ('read_setup error: read setup file');
      end subroutine read_setup;