!**  Module modelMod contains the model-specific information for a 4-state (x, xdot, y, ydot)
!**  tracking problem to be used for testing the MLE_parameterID program.
!**  The 4-state model consists of two independent 1st order markov velocity processes
!**  with integral.  Measurements are range and bearing.
!**  The 6 parameters to be estimated via MLE are: Markov constants (2), Markov process noise (2),
!**  and measurement noise (2).

!**  Search for "###" to find parameters that are typically changed

!**  The filter/simulation is coded with the following assumptions:
!**    1) initial state is (1,10,0,0)
!**    2) filter state is initialized at truth, with diagonal covariance
!**    3) phi is analytic, but Q is assumed diagonal because of high sampling rate
!**       compared to Markov time constants). random process noise is applied
!**       independently to the two channels
!**    4) measurement noise is assumed independent for 2 meas
!**    5) measurement model does not include unknown parameters (no del-y/del-xparm)

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort modelMod.f90 /c
!**   ifort MLE_parameterID.f90 modelMod.obj
!**   MLE_parameterID >MLE.lis
!**************************************
      module modelMod;
      !** contains model-specific information for MLE_parameterID
      !**
      integer(4),parameter :: nEstf =4;    !number of filter states
      integer(4),parameter :: ndynam =4;   !number of dynamic filter states (must be located before biases)
      integer(4),parameter :: nf3 =(nEstf*(nEstf+1))/2;
      integer(4),parameter :: nd3 =(ndynam*(ndynam+1))/2;
      integer(4),parameter :: nxpMax=10;    !max allowed  number of model parameters to be estimated via MLE
      integer(4),parameter :: nsensor =2;  !number of possible sensors
      integer(4),parameter :: mmeas =2;    !number of measurements at one time
      integer(4),parameter :: mmax =3000;  !maximum total number of meas time samples

      real(8) phi(ndynam,nEstf);          !state transition matrix
      real(8) qd(ndynam*(ndynam+1)/2);    !discrete state noise matrix
      real(8) xf(nEstf);                  !filter state
      real(8) pf(nf3);                    !filter state covariance
      real(8) psym(nf3);                  !filter state covariance
      real(8) ud(nf3);                    !filter state covariance   ??? check if n*(n+3)/2 ??

!      real(8) drdxp(nxpMax);             !partial meas noise variance wrt xparm
      real(8) :: xparm(nxpMax)            !model parameters
      real(8) :: xparm0(nxpMax)           !initial model parameters
      real(8) :: xparSig(nxpMax)          !1-sigma for xparm0
      character(20) :: pname(nxpMax)      !name of parameters that must match names in hard-coded model

      real(8) xt0(nEstf);                 !prior estimate of initial state
      real(8) :: t0 =0.d0;                !initial state epoch (sec)
      real(8) :: tmax =100.d0;            !maximum simulation time (sec)

      integer(4) :: nmeas;                !actual number of measurements
      integer(4) :: nxp;                  !actual number of model parameters to be estimated via MLE
      integer(4) :: nxp3;                 != (nxp*(nxp+1))/2

      character(20) :: xnames(nEstf) =(/"x", "xdot", "y", "ydot"/);   !filter state names

      real(8) :: sig0(nEstf)              !initial filter state 1-sigma
      real(8) :: x0(nEstf)                !initial filter state estimate
      real(8) :: rvar(mmeas,nsensor)      !meas noise 1-sigma for range and bearing
      real(8) :: xsensor(2,nsensor)       !sensor positions
!      real(8) :: delth =1.d-5;            !perturbation for numeric partials

      type measDats
        integer(4) isensor       !antenna, instrument, etc
!        integer(4) mtype         !meas type for given sensor
        real(8)  time            !meas time
        real(8)  y(mmeas)        !meas value
        real(8)  sigm(mmeas)     !meas noise 1-sigma
      end type
      type (measDats) measDat(mmax)

      type states;
!        integer(4) xtype;       !1=core state, 2=bias  (not used in MLE_parameterID)
!        character(8) status;    !must be "estimate"  (not used in MLE_parameterID)
        character(20) xname;    !name of state that must match name in hard-coded model
        real(8) x0;             !a priori estimate at epoch time
        real(8) sig0;           !1-sigma uncertainty of x0 at epoch time
        real(8) Qs;             !power spectral density of process noise
        real(8) const(2);       !known constants associated with this state in dynamic or measurement model
      end type;
      type (states) :: xstate(nEstf);  !state data used for filter
      type (states) :: xstateTmp(nEstf); !temporary copy of state data

      real(8),allocatable :: dphidxp(:,:,:); !partial phi wrt xparm
      real(8),allocatable :: dqdxp(:,:);     !partial q wrt xparm
      real(8),allocatable :: dxdxp(:,:);     !partial of filter xf wrt xparm
      real(8),allocatable :: dpdxp(:,:);     !partial filter covariance wrt xparm

      contains;
!*********************************************************************
      subroutine defaultp
      !** sets default parameters for simulation and filter
      implicit none
      !_______________________________________________

      xsensor(:,1) = (/ 0.d0, 0.d0 /)    !sensor 1 position ###
      xsensor(:,2) = (/ 5.d0,  3.d0 /)   !sensor 2 position

      rvar(:,1) =(/ 0.5, 0.001 /)    !meas noise 1-sigma, sensor 1 ###
      rvar(:,2) =(/ 0.5, 0.001 /)    !meas noise 1-sigma, sensor 2

      !** set xstate values for this case ###
      xstate(:)%x0 =0.d0
      xstate(:)%sig0 =(/ 2., 4., 3., 4. /);
      xstate(:)%Qs =(/0.d0, 0.8d0, 0.d0, 1.6d0 /)
      xstate(:)%const(1) =(/0.d0, 0.1d0, 0.d0, 0.2d0 /)

      !** option for parameterizing the process noise
!      xparm(3) =2.d0*xparm(1)*dt*sig0(2)**2;   !=qs(2)
!      xparm(4) =2.d0*xparm(2)*dt*sig0(4)**2;   !=qs(4)

      return
      end subroutine defaultp
!*********************************************************************
      subroutine init_filter (xparm,nxp)
!**    Initializes initial state, covariance, QS and internal constants of Kalman filter.
!**    Where defined by parameter names (pname), variables are set from xparm.
!**    Output variables are xstate,dxdxp,dpdxp,px,xf

      implicit none

      integer(4),intent(in) :: nxp        !number of parameters
      real(8),intent(in) :: xparm(nxp)    !current parameter values

      integer(4) i,j,j1,j2
      !_________________________________________________________

      dxdxp(:,:) =0.d0
      dpdxp(:,:) =0.d0

      do i=1,nxp
        if (pname(i) == "C-2") then          !feedback constant for state 2
          xstate(2)%const(1) =xparm(i)
        else if (pname(i) == "C-4") then     !feedback constant for state 4
          xstate(4)%const(1) =xparm(i)
        else if (pname(i) == "QS-2") then    !process noise PSD for state 2
          xstate(2)%Qs =xparm(i)
        else if (pname(i) == "QS-4") then    !process noise PSD for state 4
          xstate(4)%Qs =xparm(i)
        else if (pname(i)(:3) == "x0-") then !initial state value
          read (pname(i)(4:),*,err=999) j
          xstate(j)%x0 =xparm(i)
          dxdxp(j,i) =1.d0
        else if (pname(i)(:2) == "R-") then  !measurement  noise variance
          read (pname(i)(3:),*,err=999) j
          j1 =j/10
          j2 = mod(j,10)
          rvar(j2,j1) =xparm(i)
        else
          write (6,'("error: invalid parameter name ",a)') pname(i)
          stop 'invalid parameter name'
        endif
      enddo

      !** initialize state and covariance
      pf(:) =0.d0
      do i=1,nEstf
        xf(i) =xstate(i)%x0
        pf(i*(i+1)/2) =xstate(i)%sig0**2
      enddo

      return
  999 write (6,'("error reading index of initial state in ",a)')pname(i)
      stop 'error reading index of initial state'

      end subroutine init_filter
!********************************************************************
      subroutine get_yh (yp,dy,h,  yt,x,isensor,im);
      !** compute measurements, residuals, partials and measurement noise variance
      !** from given filter state vector, xparm vector and actual measurement (if available)

      implicit none;
      integer(4),intent(in) :: im;     !meas index (1 or 2)
      integer(4),intent(in) :: isensor;!sensor (1 to 2)
      real(8),intent(in) :: x(nEstf);  !filter state vector
      real(8),intent(in) :: yt;        !actual measurements (may be 0)
      real(8),intent(out) :: yp;       !model-predicted measurement
      real(8),intent(out) :: dy;       !measurement residual
      real(8),intent(out) :: h(nEstf); !measurement partials wrt x

      real(8) r2,dx(2);
      real(8) :: pi =3.141592653589793d0;
      !_______________________________________________

      dx(:) =(/ x(1) -xsensor(1,isensor), x(3)-xsensor(2,isensor) /)

      !** range and bearing
      r2 =dx(1)**2 +dx(2)**2;
      h(:) =0.d0

      if (im == 1) then;
        !** range measurement
        yp =sqrt(r2);
        if (r2 > 0.d0) then;
          h(:4) =(/ dx(1)/yp, 0.d0, dx(2)/yp, 0.d0 /);
        else;
          h(:4) =(/ 0.707d0, 0.d0, 0.707d0, 0.d0 /);
        endif;
        dy =yt-yp;

      else if (im == 2) then;
        !** bearing measurement
        yp =atan2(dx(1),dx(2));
        if (r2 > 0.d0) then;
          h(:4) =(/ dx(2), 0.d0, -dx(1), 0.d0 /)/r2;
        else;
          h(:4) =(/ 1.d0, 0.d0, 1.d0, 0.d0 /);
        endif;
        dy =yt-yp;
        if (abs(dy) > pi) dy =dy -sign(2.d0*pi,dy);
      endif;

      return;
      end subroutine get_yh;
!***********************************************************
      subroutine get_meas (measFile);
!** reads measurements from file measFile and loads structure measDat
      implicit none
      character(*),intent(in) :: measFile;

      integer(4) isensor;
      real(8) t,yp(mmeas),rvari(mmeas);
      !____________________________________________

      open (10,file=measFile,status='old');
      nmeas =0;

      !** time loop
      do;
        read (10,*,end=999,err=999) t,isensor,yp(:),rvari(:);
        nmeas =nmeas+1;
        measDat(nmeas)%time =t;
        measDat(nmeas)%isensor =isensor;
        measDat(nmeas)%y(:) =yp(:);
        measDat(nmeas)%sigm(:) =sqrt(rvari(:));
      enddo;

  999 close (10);
      return;
      end subroutine get_meas;

      end module modelMod;
!*********************************************************************
      subroutine get_phiq (phi,qd,xp,   tKLst,tobs,xf,partials);

      !** get_phiq updates the filter state xf ftom tKlst to tobs, and computes discrete
      !** phi, Qd.  If partials =.true., dxdxp, dphidxp =del-phi/del-xparm and
      !** dqdxp =del-q/del-xparm are also computed.
      !** 4-state model is two independent 1st order markov processes with integral.

      use modelMod, only:xstate,nEstf,ndynam,nd3,dphidxp,dqdxp,dxdxp,       &
     &                   nxp,pname;
      implicit none;

      real(8),intent(in) :: tKLst         !time of last filter state update
      real(8),intent(in) :: tobs          !current measurement time
      real(8),intent(in) :: xf(nEstf)     !filter state at tKlst
      logical(4),intent(in) :: partials   !true if partials are computed

      real(8),intent(out) :: phi(ndynam,nEstf) !dynamic portion of state transition matrix
      real(8),intent(out) :: qd(nd3)       !state noise matrix
      real(8),intent(out) :: xp(nEstf)     !predicted state

      integer(4) i
      real(8) exp1,exp2,a(2),dadth,qs(ndynam),d(ndynam),dt;      !f(nEstf,nEstf),
      !_________________________________________________________

!      f(:,:) =0.d0
!      f(1,2) =1.d0
!      f(2,2) =-a(1)
!      f(3,4) =1.d0
!      f(4,4) =-a(2)

      dt =tobs-tKlst;
      a(1) =xstate(2)%const(1);
      a(2) =xstate(4)%const(1);
      exp1 =exp(-a(1)*dt);
      exp2 =exp(-a(2)*dt);
      phi(:,:) =0.d0;
      phi(1,1:2) =(/1.d0, (1.d0-exp1)/a(1) /);
      phi(2,2) =exp1;
      phi(3,3:4) =(/1.d0, (1.d0-exp2)/a(2) /);
      phi(4,4) =exp2;
!      call prtmx (phi,ndynam,ndynam,nestf,"phi")

      !** Assume that QD is diagonal if time interval is short compared to time
      !** constants in F.  Otherwise need to iterate over short intervals, i.e.,
      !** put get_phiq call in a loop using a short time interval until full interval is reached.
      qs(:) =xstate(:ndynam)%Qs
      qd(:) =0.d0;
      forall (i=1:ndynam) qd(i*(i+1)/2) =qs(i)*dt;

      !** xp =phi * xf  (only is system is linear)
      !** (only if system is linear. Otherwise integrate dxdxp)
      call smult (xp,  phi,xf, ndynam,nEstf,1, ndynam,nEstF,nEstF);
      if (nEstf > ndynam) xp(ndynam+1:) =xf(ndynam+1:);

      if (partials) then
        dphidxp(:,:,:) =0.d0;
        do i=1,nxp
          !** update dxdxp =phi*dxdxp +dphidxp*xf
          !** (only if system is linear. Otherwise integrate dxdxp)
          call smult (d,  phi,dxdxp(:,i), ndynam,nEstf,1,               &
     &                                   ndynam,nEstF,ndynam);   !d =phi*dxdxp
          dxdxp(:ndynam,i) =d(:ndynam)

          if (pname(i) == "C-2") then
            dadth =1.d0;
            dphidxp(1,2,i) =(dt*exp1/a(1) -(1.d0-exp1)/a(1)**2)*dadth;
            dphidxp(2,2,i) =-dt*exp1*dadth;
          else if (pname(i) == "C-4") then
            dadth =1.d0;
            dphidxp(3,4,i) =(dt*exp2/a(2) -(1.d0-exp2)/a(2)**2)*dadth;
            dphidxp(4,4,i) =-dt*exp2*dadth;
          else
            cycle
          endif

          !** add dxdxp =dxdxp +dphidxp * xf
          call smult (d,  dphidxp(:,:,i),xf,  ndynam,nEstf,1,  &
     &                                   ndynam,nEstf,ndynam);    !dxdxp =dphidxp*xf
          dxdxp(:ndynam,i) =dxdxp(:ndynam,i) +d(:ndynam)
        enddo

        dqdxp(:,:) =0.d0;
        do i=1,nxp
          if (pname(i) == "QS-2") then
            dqdxp(2*3/2,i) =dt
          else if (pname(i) == "QS-4") then
            dqdxp(4*5/2,i) =dt;
          endif
        enddo
      endif

      return;
      end subroutine get_phiq;