!** Program AdaptFilt_tank is an adaptive multiple-model tank tracking filter.
!** The five different models are defined in Section 3.2.1. The optimal
!** prediction during the projectile time-of-flight is obtained by integrating
!** the state vector from the filter with the greatest likelihood function.

!** The measurement data is obtained from the re-constructed ATMT Scout maneuver data listed
!** in input file scoutSim4-xo.plt.  Note that the Kalman filter and jump
!** detection algorithm are implemented in SINGLE PRECISION because a prototype
!** operational implementation used single precision.

!** The output report should be directed to a file such as afilts.lis.
!** Program AdaptFilt_tank is used in Example 11.4-1.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort adaptFilt_tank.f90      !only rnor needed from libkf_bnd_ifort.lib
!**   adaptFilt_tank >afilts.lis
!**************************************

      module fcomm;
      !** includes variables used in afilt
      implicit none;
      integer(4),parameter :: maxfilt =10, ndim =10;

      integer(4) :: l1 =0, l2=4, l3 =2;
      integer(4) :: l2s(2,maxfilt);
      data l2s /0,0, 1,1, 6*2, 10*0/;
      integer(4) :: np;
      integer(4) :: nf(maxfilt) =(/ 4,6,8,8,8, 0,0,0,0,0 /);

      integer(4) :: imd =1;
      integer(4) :: mm;
      integer(4) :: niter =1;
      integer(4) :: nfil =5;
      integer(4) :: nqi =2;
      real(4) :: ym(6,150);
      real(4) :: tstrt;
      real(4) :: tmax =10.;
      real(4) :: r(2) =(/ 2., 0.3 /);

      integer(4) nsampt,nsamp;
      real(4) hptot(3),hist(10,3),hp(3);

      end module fcomm;
!*****************************************************************
      program afilt;
      use fcomm;
      integer(4) ndata,im,istrt;

      real(4) xp(12,maxfilt),xown(2),qs(5,maxfilt),beta(6,maxfilt);
      real(4) u(55,maxfilt),x(ndim,maxfilt);
      real(4) y(3),yx(3),cn(2,maxfilt);
      !** new variables
      integer(4) :: ix =191919;
      real(4) ydat(4,6000),fer(maxfilt),az;
      real(4) :: rrange =2000;  !observer-target range (m) ###
      real(8) rnor;
      real(8) :: pi =3.141592653589793d0;

!      data tstart,xown /38630.,57700.,78000./   !for twister
      data x,beta/160*0./;       !default is 0
      !** model parameters (theta): order is
      !**   alongtrack denominator coeff. (1 or 2 defined by l2s(1,i)),
      !**   crosstrack denominator coeff. (1 or 2 defined by l2s(2,i)),
      !**   alongtrack/crosstrack process noise PSD (2)
      data xp /2.,2.,10*0.,                                             & !constant velocity model
     &     2.52,1.88, 2.03,1.27, 8*0.,                                  & !low maneuver
     &     2.51,1.58, .440,.193, 3.7,1.33, 6*0.,                        & !M60
     &     2.51,1.58, .628,.395, 23.7,4.19, 6*0.,                       & !Scout
     &     1.76,1.58, .980,.667, 15.2,17.1, 6*0.,  60*0./;  !Twister
      data cn /4*0., .38,.25, .31,.28, .50,.31, 10*0./; !alongtrack/crosstrack numerator coeff.
      data t,dt /0.,0.4/;
      data irlist /4/;
      data icase /0/;
      data ix /191919/;
      data tin /3.0/;
      data ifirst /0/;
      namelist /indata/icase,tstart,xown,tmax,tin,r,nfil,ix,niter,      &
     & nqi,irlist;
      !______________________________________________________________

      write(6,500);
  500 format(/5x,'a l o n g / c r o s s   t r a c k   f i l t e r'//)
      xstrt =111.;  !starting x-position for scout track

      nsampt =0.;
      hptot(:) =0.;
      hist(:,:) =0.;

      t0 =11.7;   !start time for scout data (x =110.44)
      open (11,file ="scoutSim4-xo.plt",status='old');     !twister track
      !** read tank data
      read (11,*);   !skip 2 header
      read (11,*);
      im =1;
      do;
        read (11,*,end=5,err=5) ydat(:,im);
        if (ydat(1,im) < t0) cycle;
        im =im+1;
        if (im > 6000) stop 'error: data records > 6000';
      enddo;

    5 ndata =im-1;
      tstart =ydat(1,1);
      istrt =0;
      az =rnor(ix);     !initialize random number generator with seed ix

      do;      !loop for 5 sec intervals of sampling data for trials
        tstrt =tstart;
        tstp =tstart+tmax;
        iy =0;
        t =tstart;
!        yx(:) =ydat(2:4,istrt+1)
        xown(:) =-2.d10;
        i =istrt;
        istrt =-10;     !reset so that a new value will be set in loop

        !** loop for processing tank data every 5 sec
        do;
          tl =t;
          i =i+1;
          if (i > ndata) go to 999;
          t =ydat(1,i);
          if (t > tstp+3.) exit;    !allow 3 sec projectile flight time
          yx(:) =ydat(2:4,i);
          if (t < tstart) cycle;

          if (istrt < 0) istrt =i;
!          if (yx(1) == 0. .or. yx(2) == 0.) cycle
          if (xown(1) < -1.d10) then;
            az =(pi/6.)*rnor(0) +0.44; !25 deg bias toward south
            xown(:) =yx(1:2) +rrange*(/ cos(az), sin(az) /);
          endif;
          dt =t-tl;

          dx =yx(1)-xown(1);
          dy =yx(2)-xown(2);
          rn =sqrt(dx**2+dy**2);
          ang =atan2(dx,dy)*1000.;
          call gauss(ix,r(1),rn,y(1));  !add meas noise (not in this sim)
          call gauss(ix,r(2),ang,y(2));
          iy =iy+1;
          ym(1,iy)=t;
          ym(2,iy)=y(1);   !range
          ym(3,iy)=y(2);   !azimuth (bearing)
          ym(4,iy)=yx(3);  !altitude
          ym(5,iy)=dx;
          ym(6,iy)=dy;
!          xx =y(1)*sin(y(2)/1000.)
!          yy =y(1)*cos(y(2)/1000.)
          if (irlist /= 1) cycle;
          write (6,110) t,y(1),y(2),dx,dy;
  110     format(1x,f8.2,4f12.2)
        enddo;

   16   ym(1,iy) =-2.e10;   !use -2e10 to denote last entry
   15   mm =2;
        np =l1+l2+l3;
        if (ifirst == 0) write(6,250);
  250   format(/' initialization'/2x,'ifl, xp, and p(i,i)')

        !** initialize filter variables beta,qs,x,u from xp,nf,l2s
        do if=1,nfil;
          if (ifirst == 0) write(6,'(i4)') if;
          call funct(xp(:,if),u(:,if),qs(:,if),beta(:,if),x(:,if),      &
     &               nf(if),l2s(:,if),ifirst);
        enddo;

        ifirst=1;
        write(6,400) tstart;
  400   format(/' start time ',g13.5/7x,'i',3x,'nms',4x,'sum2',5x,      &
     &         'plike',6x,'err',6x,'sige',7x,'hp')
        call filter (fer,  u,x,qs,beta,cn,  nf,irlist,tin);
        tstart =tstart +5.;     !start next 5 sec track
      enddo;

  999 continue;
      hist(:,:) =hist(:,:)/nsampt;
      hptot(:) =hptot(:)/nsampt;
      write (6,'(/"total samples =",i5)') nsampt;
      write (6,'(/"constant vel hp avg =",g13.5)') hptot(2);
      write (6,'("constant vel hist =",10f8.3)') hist(:,2);
      write (6,'(/"adaptive filt hp avg =",g13.5)') hptot(1);
      write (6,'("adaptive filt vel hist =",10f8.3)') hist(:,1);
      write (6,'(/"weighted filt hp avg =",g13.5)') hptot(3);
      write (6,'("weighted filt vel hist =",10f8.3)') hist(:,3);

      !** write data for histogram
      open (12,file='hitProb.plt',status='unknown');
      write (12,1000);
 1000 format ('variables ="hit probability", "frequency-adaptive", ',   &
     &       '"frequency-CV", "frequency-average"')
 1010 format (4g13.5)
      do i=1,10;
        write (12,1010) (i-1)*0.1, 0., 0., 0.;
        write (12,1010) (i-1)*0.1, hist(i,:);
        write (12,1010) i*0.1, hist(i,:);
      enddo;
      write (12,1010) 1.0, 0., 0., 0.;

      write (12,1010) hptot(1), 0.,  0., 0.;
      write (12,1010) hptot(1), 0.25, 0., 0.;
      write (12,1010) hptot(1), 0.,  0., 0.;
      write (12,1010) hptot(2), 0.,  0., 0.;
      write (12,1010) hptot(2), 0., 0.25, 0.;
      write (12,1010) hptot(2), 0.,  0., 0.;
      write (12,1010) hptot(3), 0.,  0., 0.;
      write (12,1010) hptot(3), 0.,  0., 0.25;
      write (12,1010) hptot(3), 0.,  0., 0.;

      stop;
      end;
!**************************************************************************
      subroutine filter (fer,  u,x,qs,beta,cn,  nf1,irlist,tin);
      !** parallel Kalman filters for tank tracking
      use fcomm;
      implicit none;
      integer(4),intent(in) :: nf1(maxfilt);    !number of states in each filter
      integer(4),intent(in) :: irlist;          !not used (was print option ?)
      real(4),intent(in) :: qs(5,maxfilt);      !alongtrack and crosstrack acceleration process noise PSD (3:5 are unused)
      real(4),intent(in) :: beta(6,maxfilt);    !acceleration Markov process denominator coefficients
                                                ! (1:3 are alongtrack, 4:6 are crosstrack)
      real(4),intent(in) :: cn(2,maxfilt);      !alongtrack and crosstrack accel. Markov process numerator coeff.
      real(4),intent(in) :: tin;                !filter initialization time (s)
      real(4),intent(inout) :: x(ndim,maxfilt); !state vector at tin for each filter
      real(4),intent(inout) :: u(55,maxfilt);   !initial U-D arrays at tin for each filter
      real(4),intent(out) :: fer(maxfilt);      !predition error (m) for each filter

      integer(4) nms(maxfilt),irl1,irl2,m,npts,ioht,nx,iy;
      integer(4) ifl,n,i,j,im,irsamp;
      real(4) phi(ndim,ndim),y(3);
      real(4) h(ndim),g(ndim),f(ndim),sum1(maxfilt),sum2(maxfilt);
      real(4) az0,t,told,dt,dt0,t0,tlst,delx,dely,rn2,dy,ang,rvar,c,rn;
      !______________________________________________________________

      h(:) =0.;
      g(:) =0.;
      irl1 =irlist/10;
      irl2 =irlist-10*irl1;
      m =2;
      npts =0;
      ioht =0;
      sum1(:) =0.d0;
      sum2(:) =0.d0;
      nms(:) =0;
      iy =1;
      t0 =ym(1,iy);
      y(1) =ym(2,iy);   !range
      y(2) =ym(3,iy);   !azimuth
      az0 =y(2);
      x(1:4,1) =(/ 0., y(1), 1., 0. /);   !filter coordinates are local (y =to target)
      t =t0;
      do i=2,nfil;
        x(1:4,i) =x(1:4,1);
      enddo;
      n =nf1(1);
      nx =4;
      tlst =0.;
      hp(:) =0.;
      nsamp =0;
      irsamp =0;

      !** start of time loop
      do;
        told =t;
        iy =iy+1;
        t =ym(1,iy);
        if (t-t0 > tmax .or. t < -1.e10) exit;

        y(1) =ym(2,iy);
        y(2) =ym(3,iy)-az0;
        dt =t-told;
        dt0 =t-t0;
        if (dt < 0.) then;
          t =told;
          cycle;
        endif;

        do ifl =1,nfil;
          n =nf1(ifl);
          call dynam (x(1,ifl),phi,n,ndim,beta(1,ifl),cn(1,ifl),dt,     &
     &                l2s(1,ifl));
          call udupd1 (phi,u(1,ifl),qs(1,ifl),n,nx,ndim,dt,l2s(1,ifl));

          do im=1,mm;
            delx =x(1,ifl);
            dely =x(2,ifl);
            rn2 =delx**2+dely**2;
            h(1:2)=0.;

            if (im == 1) then;
              !** range
!              if (mod(irsamp,10) > 0) cycle   !only range meas every 1.0 sec  ###
              rn =sqrt(rn2);
              dy =y(1)-rn;
              h(1) =delx/rn;
              h(2) =dely/rn;
            else;
              !** azimuth
              ang =atan2(delx,dely)*1000.;
              dy =y(2)-ang;
              h(1) =(dely/rn2)*1000.;
              h(2) =-(delx/rn2)*1000.;
            endif;
            rvar=r(im)**2;
            call udmeas(x(1,ifl),u(1,ifl),n,rvar,h,f,g,c,dy);

            if (c <= 0.) then;
              write(6,1222)t,ifl,im,dy,c;
 1222         format(1x,'**',f10.2,2i5,4g12.5)
              exit;
            endif;
            sum1(ifl) =sum1(ifl) +alog(c);
            sum2(ifl) =sum2(ifl) +dy**2/c;
            nms(ifl) =nms(ifl)+1;
          enddo;    !___ im loop

          irsamp =irsamp+1;
        enddo;

        if (dt0 > tin .and. mod(dt0+.04,1.) < 0.10) then;
          call lead_err (fer,  x,nf1,beta,cn,sum1,sum2,iy,t,az0);
        endif;
      enddo;   !___ end time loop

  999 continue;
      hp(:) =hp(:)/nsamp;
      hptot(:) =hptot(:) +hp(:);
      do j=1,3;
        i =floor(hp(j)/0.1) +1;
        hist(i,j) =hist(i,j)+1;
      enddo;
      nsampt =nsampt +1;

      return;
      end subroutine filter;
!***********************************************************
      subroutine lead_err (fer,  x,nf1,beta,cn,sum1,sum2,iy,t,az0);
      use fcomm;
      implicit none;
      integer(4),intent(in) :: nf1(maxfilt);    !state order for each filter
      integer(4),intent(in) :: iy;              !sample index in ym array at time t
      real(4),intent(in) :: beta(6,maxfilt);    !filter denominator coefficients
      real(4),intent(in) :: cn(2,maxfilt);      !filter numerator coefficients
      real(4),intent(in) :: sum1(maxfilt);      !sum of
      real(4),intent(in) :: sum2(maxfilt);
      real(4),intent(in) :: x(ndim,maxfilt);    !filter state vector at time t
      real(4),intent(in) :: t;                  !time of filter state
      real(4),intent(in) :: az0;                !initial azimuth used to set filter coordinate system
      real(4),intent(out) :: fer(maxfilt);      !lead prediction error (m) for each filter, nfilt+1 is averaged filter state

      integer(4) ipp,imax,n,i,j,ic;
      real(4) dt,phi(ndim,ndim),xf(ndim),xavg(ndim),plike,acca,accc;
      real(4) plikef(maxfilt),prob(maxfilt),probsum,weight(maxfilt);
      real(4) range,dtp,plm,tt,fact,xt,yt,a,saz0,caz0,xx,yy,err;
      !_____________________________________

      ipp =0;
      range =sqrt(ym(5,iy)**2 +ym(6,iy)**2);
      dtp =t+range/1500.;        !time of impact
      plm=1.e10;

      !**  find true samples near impact time
      j =4+iy;
      do while (ym(1,j) < dtp);
        j =j+1;
!        if (j > 30+iy) then
        if (ym(1,j) < -1.e10) then;
          write (6,'("time ",g13.5,", dtp =",g13.5,", iy =",i3,         &
     &            " cannot find impact time")') t,dtp,iy;
          return;
        endif;
      enddo;

      !** interpolate true position at impact time
      fact =(dtp-ym(1,j-1))/(ym(1,j)-ym(1,j-1));
      xt =ym(5,j-1) +fact*(ym(5,j)-ym(5,j-1));
      yt =ym(6,j-1) +fact*(ym(6,j)-ym(6,j-1));
      a =sqrt(xt**2+yt**2);
      saz0 =sin(az0*1.e-3);
      caz0 =cos(az0*1.e-3);
!     write (6,'("t =",g13.5,", ym(1,j) =",g13.5,", range =",g13.5,     &
!    &       ", dt =",g13.5)') t,ym(1,j),a,range/1500.;

      plikef(:nfil) =sum1(:nfil)+sum2(:nfil);
      plm =minval(plikef(:nfil));
      imax =minloc(plikef(:nfil),1);
      plikef(:nfil) = plikef(:nfil) -plm;    !bias -log likelihood so that minimum is 0
      prob(:) =0.d0;
      do i=1,nfil;
        if (abs(plikef(i)) < 20.) prob(i) =exp(-plikef(i));     !biased probability density
      enddo;
      probsum =sum(prob(:nfil));
      weight(:nfil) =prob(:nfil)/probsum;     !probability-based weights
      xavg(:) =0.d0;

      do i=1,nfil;   !loop for filters
        n =nf1(i);
        xf(:n) =x(:n,i);
!       write (6,'("xf(",i2,") =",10g13.5)') i,xf(:n);
        xavg(:4) =xavg(:4) +weight(i)*xf(:4);       !compute weighted sum of states
        ic =5;
        if (l2s(1,i) == 1) then;
          xavg(5) =xavg(5) +weight(i)*xf(5);
          ic =6;
        else if (l2s(1,i) == 2) then;
          acca =xf(5) +cn(1,i)*xf(6);
          xavg(5) =xavg(5) +weight(i)*acca;
          ic =7;
        endif;
        if (l2s(2,i) == 1) then;
          xavg(6) =xavg(6) +weight(i)*xf(ic);
        else if (l2s(2,i) == 2) then;
          accc =xf(ic) +cn(2,i)*xf(ic+1);
          xavg(6) =xavg(6) +weight(i)*accc;
        endif;
        tt =t;

        !** integrate to impact in 0.2 s steps
        do while (tt < dtp);
          dt =min(0.2,dtp-tt);
          tt =tt+dt;
          call dynam (xf,phi,n,ndim,beta(1,i),cn(1,i),dt,l2s(1,i));
        enddo;

        !**   calculate cross bearing error
        !** compute cross product of filter-true difference and true vector to target
        xx = xf(1)*caz0+xf(2)*saz0;
        yy =-xf(1)*saz0+xf(2)*caz0;
        err =(yt*(xx-xt)-xt*(yy-yt))/a;
        fer(i) =err;
      enddo;

      !** now compute lead prediction error for weighted averaged filter state
      !** Note: states 5:n are meaningless as an average of different models,
      !** but we will try using 1st order markov model (#2) with averaged states 5:6
      xavg(:6) =xavg(:6)/nfil;
      xavg(7:) =0.;

      i =2;    !use model #2 (6-state 1st-order markov)
      tt =t;
      !** integrate to impact in 0.2 s steps
      do while (tt < dtp);
        dt =min(0.2,dtp-tt);
        tt =tt+dt;
        call dynam (xavg,phi,n,ndim,beta(1,i),cn(1,i),dt,l2s(1,i));
      enddo;

      !**   calculate cross bearing error
      !** compute cross product of filter-true difference and true vector to target
      xx = xf(1)*caz0+xf(2)*saz0;
      yy =-xf(1)*saz0+xf(2)*caz0;
      err =(yt*(xx-xt)-xt*(yy-yt))/a;
      fer(nfil+1) =err;

      write (6,1022) t,imax,(fer(i),i=1,nfil+1);
 1022 format(1x,f10.2,i3,10f8.3)
      if (abs(fer(imax)) < 1.15) hp(1) =hp(1)+1;
      if (abs(fer(1)) < 1.15) hp(2) =hp(2)+1;
      if (abs(fer(nfil+1)) < 1.15) hp(3) =hp(3)+1;
      nsamp =nsamp+1;

      return;
      end subroutine lead_err;
!**************************************************************************
      subroutine funct (xp,u,qs,beta,x,nf,l2s,ifirst);
      !** sets filter qs, beta, x and u-d from given xp vector
      implicit none;
      integer(4),intent(in) :: nf,l2s(2),ifirst;
      real(4),intent(in) :: xp(12);
      real(4),intent(out) :: u(nf*(nf+1)/2),x(nf),qs(5),beta(6);

      integer(4) i,k,l2,i0,i1,j1;
      real(4) a0,a1,a2,a3,qvar,dum;
      !______________________________________________________________

      k=0;
      x(:nf) =0.d0;
      u(:) =0.d0;    !

      l2 =l2s(1)+l2s(2);
      qs(1) =xp(l2+1);     !PSD of process noise: alongtrack
      qs(2) =xp(l2+2);     !PSD of process noise: crosstrack
      if (l2 == 0) go to 35;

      do i=1,l2;
        i0 =0;
        if (i > l2s(1)) i0 =3-l2s(1);
        beta(i+i0) =xp(i);
      enddo;

      do i=1,2;      !alongtrack and crosstrack
        i0 =l2s(1)*(i-1)+1;
        i1 =i0;
        j1 =4+i0;
        k =(j1*(j1+1))/2;
        a0 =1.;
        a1 =xp(i1);
        a2 =xp(i1+1);
        a3 =xp(i1+2);
        qvar =qs(i);
        if (a1 == 0.) cycle;

        if (l2s(i) == 1) then;
          !** first order
          u(k) =0.5*qvar/(a0*a1);
        else if (l2s(i) == 2) then;
          !** second order
          dum =0.5*qvar/(a0*a1);
          u(k) =dum;
          k =k+j1+1;
          u(k) =a0*dum/a2;
        else;
          !**   third order
          dum =0.5*qvar/((-a0*a3+a1*a2)*a0);
          u(k) =a2*dum;
          k =k+j1+1;
          u(k) =a0*dum;
          k =k+j1+2;
          u(k) =a0*a1*dum/a3;
        endif;
      enddo;

   35 u(1) =10.;
      u(3) =10.;
      u(6) =1000.;
      u(10) =1000.;
      if (nf >= 5)then;
        k =(4*5)/2;
        do i=5,nf;
          k =k+i;
          if (u(k) > 10000.) u(k)=0.;
        enddo;
      endif;

      if (ifirst == 1) return;
      write(6,100) xp(1:l2+2);
  100 format(5x,6g12.5)
      write(6,'(/"beta =",10g13.5)') beta(1:6);
      k =0;
      do i=1,nf;
        write (6,'("u-row ",i3," =",10g13.5)') i,u(k+1:k+i);
        k =k+i;
      enddo;

      return;
      end subroutine funct;
!**************************************************************************
      subroutine dynam (x,phi,n,nn,beta,cn,dt,l2s);
!        this subroutine computes phi and q and updates x and p
!        from to to t+dt
      use fcomm, only:ndim;
      implicit none;
      integer(4),intent(in) :: l2s(2),n,nn;
      real(4),intent(in) :: beta(6),cn(2),dt;
      real(4),intent(inout) :: x(nn);
      real(4),intent(out) :: phi(nn,nn);

      integer(4) i,j,ii,ic,ic1,iv,im;
      real(4) cna,cnc,v2,v,a1,a2,d1,d2,d3,d5,acca,accc;
      real(4) d6,d8,d9,xdt2,ydt2,d18,d19,dt2,x5,x6,x7;
      real(4) f(ndim,ndim),g(ndim,ndim),s(ndim,ndim);
!      data f /100*0./
!      data s /100*0./
      !______________________________________________________________

      f(:,:) =0.;
      s(:,:) =0.;

      if (n == 4) then;
        !** constant velocity filter
        x(1)=x(1) +x(3)*dt;
        x(2)=x(2) +x(4)*dt;
        phi(:n,:n) =0.;
        forall (i=1:n) phi(i,i) =1.;
        phi(1,3)=dt;
        phi(2,4)=dt;
        return;
      endif;

!        compute f (xdot=(f/dt)*x)
      cna =cn(1);
      cnc =cn(2);

      !** iterate once to refine f
      do ii=1,2;
!    4   ii =ii+1
        v2 =x(3)**2+x(4)**2;
        v =sqrt(v2);
        a1 =x(3)/v;
        a2 =x(4)/v;
        d1 =a2**2;
        d2 =a1*a2;
        d3 =a1**2;
        d5 =dt/v;
        f(1,3) =dt;
        f(2,4) =dt;
        ic =5+l2s(1);
        ic1 =ic+1;
        iv =1;
        acca =x(5) +cna*x(6);       !cna/cnc should be 0 for 1st-order markov
        accc =x(ic) +cnc*x(ic1);    !
        if (v <= 0.20) then;
          iv=0;
          d5=0.;
          d6=0.;
          acca=0.;
          accc=0.;
        endif;

        f(3,3) =(d1*acca-d2*accc)*d5;
        f(3,4) =(-d2*acca+d3*accc)*d5;
        f(3,5) =x(3)*d5;
        f(3,ic) =x(4)*d5;
        f(4,3) =-(d2*acca+d1*accc)*d5;
        f(4,4) =(d3*acca+d2*accc)*d5;
        f(4,5) =f(3,ic);
        f(4,ic) =-f(3,5);

        im =l2s(1);
        d8 =-beta(1)*x(5);
        f(5,5) =-beta(1)*dt;

        if (l2s(1) > 1) then;
          !** alongtrack 2nd-order markov
          d8 =d8-beta(2)*x(6);
          f(5,6) =-beta(2)*dt;
          f(6,5) =dt;
          f(3,6) =f(3,5)*cna;
          f(4,6) =f(4,5)*cna;
        endif;

        d9 =-beta(4)*x(ic);
        f(ic,ic) =-beta(4)*dt;

        if (l2s(2) > 1) then;
          !** crosstrack 2nd-order markov
          d9 =d9-beta(5)*x(ic1);
          f(ic,ic1) =-beta(5)*dt;
          f(ic1,ic) =dt;
          f(3,ic1) =f(3,ic)*cnc;
          f(4,ic1) =f(4,ic)*cnc;
        endif;
        xdt2 =a1*acca+a2*accc;
        ydt2 =a2*acca-a1*accc;
        d18 =d8+cna*x(5);
        d19 =d9+cnc*x(ic);

        if (ii > 1) exit;
        s(:,:) =f(:,:);

        !**    update x
        dt2 =dt/2.;
        x(1) =x(1) +x(3)*dt +xdt2*dt2*dt;
        x(2) =x(2) +x(4)*dt +ydt2*dt2*dt;
        x(3) =x(3) +xdt2*dt +(f(3,5)*d18 +f(3,ic)*d19 +accc*ydt2*d5)*dt2;
        x(4) =x(4) +ydt2*dt +(f(4,5)*d18 +f(4,ic)*d19 -accc*xdt2*d5)*dt2;
        x5 =x(5) +(d8+ ((beta(1)**2-beta(2))*x(5)+                      &
     &        (beta(1)*beta(2))*x(6))*dt2)*dt;
        x7 =x(ic) +(d9+ ((beta(4)**2-beta(5))*x(ic)+                    &
     &        (beta(4)*beta(5))*x(ic1))*dt2)*dt;
        x6 =x(6) +(x(5) +d8*dt2)*dt;
        x(ic1) =x(ic1) +(x(ic) +d9*dt2)*dt;
        x(5) =x5;
        x(ic) =x7;
        if (l2s(1) > 1) x(6) =x6;
      enddo;

   21 call fxb(g,  f,s,n,ndim,l2s);
      do i=1,n;
        do j=1,n;
          f(j,i) =0.5*(f(j,i)+s(j,i));
          phi(j,i) =f(j,i) +0.5*g(j,i);
        enddo;
        phi(i,i) =phi(i,i)+1.;
      enddo;

      return;
      end subroutine dynam;
!**************************************************************************
      subroutine fxb (c,  f,b,n,nn,l2s);
      !** computes matrix multiplication C =F * B where F is the Jacobian
      !** F = del-xdot/ del-x
      implicit none;
      integer(4),intent(in) :: n,nn,l2s(2);
      real(4),intent(in) :: f(nn,nn),b(nn,nn);
      real(4),intent(out) :: c(nn,nn);

      integer(4) i,ic,ic1,ic2;
      !______________________________________________________________

!        c=f*b
      ic =5+l2s(1);
      ic1 =ic+1;
      ic2 =ic+2;
      do i=1,n;
        c(1,i)=f(1,3)*b(3,i);
        c(2,i)=f(2,4)*b(4,i);
        c(3,i)=f(3,3)*b(3,i)+f(3,4)*b(4,i)+f(3,5)*b(5,i)+f(3,ic)*b(ic,i);
        c(4,i)=f(4,3)*b(3,i)+f(4,4)*b(4,i)+f(4,5)*b(5,i)+f(4,ic)*b(ic,i);
        c(5,i)=f(5,5)*b(5,i);
        c(ic,i)=f(ic,ic)*b(ic,i);

        if (l2s(1) /= 1) then;
          c(5,i)=c(5,i)+f(5,6)*b(6,i);
          c(6,i)=f(6,5)*b(5,i);
          c(3,i)=c(3,i)+f(3,6)*b(6,i);
          c(4,i)=c(4,i)+f(4,6)*b(6,i);
        endif;

        if (l2s(2) /= 1) then;
          c(ic,i)=c(ic,i)+f(ic,ic1)*b(ic1,i);
          c(ic1,i)=f(ic1,ic)*b(ic,i);
          c(3,i)=c(3,i)+f(3,ic1)*b(ic1,i);
          c(4,i)=c(4,i)+f(4,ic1)*b(ic1,i);
        endif;
      enddo;

      return;
      end subroutine fxb;
!**************************************************************************
      subroutine rank1 (uin,uout,n,c,v);
!
!     stable u-d factor rank 1 update
!
!      (uout)*dout*(uout)**t=(uin)*din*(uin**t+c*v*v**t
!
!     uin(n*(n+1)/2)  input vector stored positive semi-definite u-d
!                     array, with d elements stored on the diagonal
!     uout(n*(n+1)/2)  output vector stored positive (possibly) semi-
!                     definite u-d result. uout=uin is permitted
!     n               matrix dimension, n.ge.2
!     c               input scalar. should be non-negative
!                     c  is destroyed during the process
!     v(n)            input vector for rank one modification.
!
!     cognizant persons: g.j.bierman/m.w.nead   (jpl,sept.1977)
!
      implicit none;
      integer(4),intent(in) :: n;
      real(4),intent(in) :: uin(n*(n+1)/2);
      real(4),intent(inout) :: v(n),c;
      real(4),intent(out) :: uout(n*(n+1)/2);

      integer(4) i,j,k,jj,np2,l,jm1;
      real(4) z,s,alpha,beta,d;

      real(4) :: eps =0.;
      real(4) :: tst =.0625;
!      in single precision epsilon is machine accuracy
!     tst=1/16 is used for rank1 algorithm switching
      !______________________________________________________________
!
      z =0.0;
      jj =n*(n+1)/2;
      if (c <= z) then;
        uout(:jj) =uin(:jj);
        return;
      endif;

      np2 =n+2;
      do l=2,n;
        j =np2-l;
        s =v(j);
        beta =c*s;
        d =uin(jj)+beta*s;
        if (d <= eps) then;
          if (d < z) then;
            write(6,100);
            return;
          endif;
          jj =jj-j;
          write(6,110);
          uout(jj+1:jj+j) =z;
          cycle;
        endif;
        beta =beta/d;
        alpha =uin(jj)/d;
        c =alpha*c;
        uout(jj) =d;
        jj =jj-j;
        jm1 =j-1;
        if (alpha >= tst) then;
          do i=1,jm1;
            v(i) =v(i) -s*uin(jj+i);
            uout(jj+i) =beta*v(i) +uin(jj+i);
          enddo;
        else;
          do i=1,jm1;
            d =v(i) -s*uin(jj+i);
            uout(jj+i) =alpha*uin(jj+i) +beta*v(i);
            v(i) =d;
          enddo;
        endif;
      enddo;

      uout(1) =uin(1) +c*v(1)**2;
      return;

  100 format (/10x,'* * * error return due to a computed negative dia   &
     &gonal in rank1 * * *')
  110 format (/10x,'* * * note: u-d result is singular * * *')
      end subroutine rank1;
!**************************************************************************
      subroutine wgs (w,imaxw,iw,jw,dw,u,v);
!     modified gramm-schmidt algorithm for reducing wdw(**t) to udu(**t)
!     form where u is a vector stored triangular matrix with the
!     resulting d elements store on the diagonal
!
!
!     w(iw,jw)     input matrix to be reduced to triangular form.
!                  iw.le.imaxw.and.iw.gt.1
!     imaxw        row dimension of w matrix
!     iw           no. rows of w matrix, dimension of u
!     jw           no. cols of w matrix
!     dw(jw)       vector of non-negative weights for the
!                  orthogonalization process. the d)s are unchanged
!                  by the calculation.
!     u(iw*(iw+1)/2) output upper triangular vector stored output
!     v(jw)          work vector
!
!                    (see book:
!     ) factorization methods for discrete sequential estimation ),
!                     by g.j.bierman)
!
!     cognizant persons: g.j.bierman/m.w.nead   (jpl, feb.1978)
!
      implicit none;
      integer(4),intent(in) :: imaxw,iw,jw;
      real(4),intent(in) :: dw(jw);
      real(4),intent(inout) :: w(imaxw,jw);
      real(4),intent(out) :: u(iw*(iw+1)/2),v(jw);

      integer(4) i,j,k,l,ij,jm1,iwp2;
      real(4) z,one,suma,dinv;
      !______________________________________________________________

      z =0.0;
      one =1.0;
      iwp2 =iw+2;

      do l=2,iw;
        j =iwp2-l;
        suma =z;
        do k=1,jw;
          v(k) =w(j,k);
          u(k) =dw(k)*v(k);
          suma =suma +v(k)*u(k);
        enddo;
        w(j,j) =suma;
        dinv =suma;
        jm1 =j-1;
        if (suma <= z) then;
          !** w(j,.)=0. when dinv=0 (dinv=norm(w(j,.)**2))
          w(j,1:jm1) =z;
          cycle;
        endif;

        do k=1,jm1;
          suma =z;
          do i=1,jw;
            suma =w(k,i)*u(i)+suma;
          enddo;
          suma =suma/dinv;
          !**  divide here (in place of reciprocal multiply) to avoid
          !**  possible overflow
          do i=1,jw;
            w(k,i) =w(k,i)-suma*v(i);
          enddo;
          w(j,k)=suma;
        enddo;
      enddo;

!**     the lower part of w is u transpose
      suma =z;
      do k=1,jw;
        suma =suma +dw(k)*w(1,k)**2;
      enddo;
      u(1) =suma;
      ij =1;
      do j=2,iw;
        do i=1,j;
          ij =ij+1;
          u(ij) =w(j,i);
        enddo;
      enddo;

      return;
      end subroutine wgs;
!**************************************************************************
      subroutine udmeas (x,u,n,r,a,f,g,alpha,dy);
!
!     computes estimate and u-d measurement updated
!     covariance, p=u*d*u**t
!
!     *** input ***
!     u      upper triangular matrix, with d elements stored as the
!            diagonal. u is vector stored and corresponds to the
!            a priori covariance.
!     n      dimension of the state estimate. n.gt.1
!     r      measurement variance
!     a      vector of measurement coefficients,
!
!     *** outputs ***
!
!     u      updated, vector stored factors
!
!     alpha  innovations variance of the measurement residual
!     g      vector of unweighted kalman gains. the kalman
!           gain k is equal to g/alpha
!     f      contains u**t*a
!            one can have f overwrite a to save storage
!
!        cognizant persons: g.j. bierman/m.w. nead (jpl. feb.1978)
      implicit none;
      integer(4),intent(in) :: n;
      real(4),intent(in) :: r,a(n),dy;
      real(4),intent(inout) :: x(n),u(n*(n+1)/2);
      real(4),intent(out) :: f(n),g(n),alpha;

      integer(4) np1,np2,jjn,j,k,jj,jm1,kj,ntot,l;
      real(4) suma,c,dyc,gamma,beta,temp,s,dum,p;
      !______________________________________________________________

      np1 =n+1;
      np2 =n+2;
      ntot =n*np1/2;

      jjn =ntot;
      do l=2,n;
          j=np2-l;
          jj=jjn-j;
          jm1=j-1;
          suma =a(j) +dot_product(u(jj+1:jj+jm1),a(1:jm1));
        f(j) =suma;
        g(j) =suma*u(jjn);
        jjn =jj;
      enddo;
      f(1) =a(1);
      g(1) =u(1)*f(1);
!**     f=u**t*a and g=d*(u**t*a)

      suma =r+g(1)*f(1);
      c =suma +dot_product(g(:n),f(:n));
      dyc =dy**2/c;
      if (dyc > 50) go to 40;
      gamma=1./suma;
      if (f(1) /= 0.) u(1)=u(1)*r*gamma;

      kj=2;
      do j=2,n;
        beta =suma;
        temp =g(j);
        suma =suma+temp*f(j);
        p =-f(j)*gamma;
        jm1 =j-1;
        do k=1,jm1;
          s =u(kj);
          u(kj) =s+p*g(k);
          g(k) =g(k)+temp*s;
          kj =kj+1;
        enddo;
        gamma =1./suma;
        u(kj) =u(kj)*beta*gamma;
        kj =kj+1;
      enddo;
      alpha =suma;

      dum =dy*gamma;
      x(:n)=x(:n)+g(:n)*dum;
      return;

   40 alpha= -c;
      return;
      end subroutine udmeas;
!**************************************************
      subroutine gauss(ix,s,am,v);
      real(8) rnor;
      !** set output v to mean am + Gaussian random number
      v =am +s*rnor(0);
!      v =am       !mean only: no random number ###
      return;
      end subroutine gauss;
!**************************************************************************
      subroutine udupd1 (phi,u,qs,n,nx,nn,dt,l2s);
!  subroutine udupd performs the time update on u-d arrays
!  This version differs from upupd above in that dinv is used in loop 50
!  (no difference in results)
      dimension phi(nn,1),u(1),qs(1),t1(4,10),t2(4,4),d1(10),t3(10);
      dimension t4(4,4),u3(15),l2s(2);
      !______________________________________________________________

      np=n-nx;
!        t1=phi*u
      ll=0;
      do i=1,nx;
        ll=ll+i;
        d1(i)=u(ll);
        t1(i,1)=phi(i,1);
        l=0;
        do j=2,n;
          j1=j-1;
          suma =phi(i,j);
          l=l+1;
          do k=1,j1;
            l=l+1;
            suma =suma+phi(i,k)*u(l);
          enddo;
          t1(i,j)=suma;
        enddo;
      enddo;

      if(np.eq.0)go to 64;

!        t2=phi(3)*u(3)
      i1=nx+1;
      i2=i1+1;
      nxt=(nx*(nx+1))/2;
      ll=nxt;
      do i=i1,n;
        l=nxt+i1;
        ix=i-nx;
        t2(ix,1)=phi(i,i1);
        ll=ll+i;
        d1(i)=u(ll);
        do j=i2,n;
          j1=j-1;
          l=l+i1;
          suma=phi(i,j);
          do k=i1,j1;
            suma=suma+phi(i,k)*u(l);
            l=l+1;
          enddo;
          t2(ix,j-nx)=suma;
        enddo;
      enddo;

!        t4=t1(2)*d1*t2**t
      do i=1,nx;
        do j=1,np;
          j1=j+nx;
          t3(j)=t1(i,j1)*d1(j1);
        enddo;
        do j=1,np;
          suma=0.0;
          do k=1,np;
            suma=suma+t3(k)*t2(j,k);
          enddo;
          t4(i,j)=suma;
        enddo;
      enddo;

!        factor t2 into u - d
      call wgs(t2,4,np,np,d1(nx+1),u3,t3);

!        store u3 in u,t3
      l=nxt;
      m=0;
      do i=1,np;
        l=l+nx;
        do j=1,i;
          m=m+1;
          l=l+1;
          u(l)=u3(m);
        enddo;
        t3(i)=u3(m);
        u3(m)=1.0;
      enddo;

!        invert u3
      u3(1)=1./u3(1);
      jj=1;
      do j=2,np;
        jjold =jj;
        jj =jj+j;
        dinv =1./u3(jj);
        u3(jj) =dinv;    !this is different from other version
        ii =0;
        ik =1;
        jm1 =j-1;
        do i=1,jm1;
          ii =ii+i;
          ik =ii;
          suma =0.0;
          do k=i,jm1;
            suma =suma+u3(ik)*u3(jjold+k);
            ik=ik+k;
          enddo;
          u3(jjold+i)=-suma*dinv;    !this is different
        enddo;
      enddo;

!        u(2)=t4*u3**(-t)*t3**(-1)
      do i=1,nx;
        ii=0;
        ij=nxt+i;
        do j=1,np;
          ii=ii+j;
          l=ii;
          suma=0.0;
          do k=j,np;
            suma=suma+t4(i,k)*u3(l);
            l=l+k;
          enddo;
          u(ij)=suma/t3(j);
          ij=ij+nx+j;
        enddo;
      enddo;

!        factor t(1) into u - d
   64 call wgs(t1,4,nx,nx,d1,u,t3);
      t3(:n) =0.d0;

!        add in state noise using rank1 update
      nxq =nx+1;
      if (np == 0) nxq=nx-1;
      t3(nxq) =1.0;
      c =qs(1)*dt;
      call rank1(u,u,n,c,t3);
      t3(:n) =0.d0;
      nxq =nx+l2s(1)+1;
      if (np == 0) nxq=nx;
      t3(nxq) =1.0;
      c =qs(2)*dt;
      call rank1(u,u,n,c,t3);

      return;
      end subroutine udupd1;

