!**  tst_cva3.f tests cva operaiton on a 4th-order arma model with loosely damped 2-nd order

!** Program tst_cva3 tests four different approaches for spectral estimation and
!** ARMA modeling.  The four approaches are 1) periodogram, 2) BT correlogram,
!** 3) Yule-Walker MEM, 4) Burgs MEM, and 5) CVA.  Simulated measurements are
!** generated using a 4th-order ARMA model consisting of two loosely damped 2nd
!** order Markov processes where measurements y1 are a function of only one 2nd
!** order section and the y2  measurements are a function of both sections.
!** A control signal of u1=0.8sin(2*pi*t/25) feeds the second 2nd order section
!** and y2. The undamped natural frequencies of the two sections are 0.1 and 0.2 Hz.
!** A total of 500 measurements at a sampling interval of 1.0 seconds are simulated.

!** Program Tst_cva3 links in subroutines cva, memopt, spect and BT_spect from
!** the library.  Note that these modules are general purpose and may be used
!** for alternate models.

!** The output plot files are
!** 1. ARMAdata.plt contains the simulated measurements and control versus time,
!** 2. PSD_fft.plt contains the power spectrum computed by the FFT periodogram,
!** 3. PSD_BT.plt contains the power spectrum computed by BT correlogram,
!** 4. PSD_YW.plt contains the power spectrum computed by the Yule-Walker method,
!** 5. PSD_mem.plt contains the power spectrum computed by Burgs MEM,
!** 6. ypsd.plt contains the true power spectrum of the simulated data,
!** 7. ypsd_cva.plt contains the power spectrum computed by thre CVA state-space model
!** 8. ypsd_corr1.plt contains the power spectrum computed from the correlation matrices generated by CVA.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.

!**************************************
!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib \library\lapack-3.1.1\lapack_win32.lib \library\lapack-3.1.1\blas_win32.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tst_cva3.f90   !libkf_bnd_ifort.lib and lapack should be linked by default
!**                        !libkf_bnd_ifort.lib contains cva.obj memopt.obj spect.obj BT_spect.obj YW_spect.obj
!**   tst_cva3 >tst_cva.lis

!**       using Compaq compiler
!**   set lib=%lib%;\kf_book\library\libkf_bnd_ifort.lib;\library\lapack-3.1.1\lapack_win32.lib;\library\lapack-3.1.1\blas_win32.lib
!**   df tst_cva3.f cva.obj dsvdc.obj blasd.obj ..\library\libkf_bnd.lib \library\lapack-3.1.1\lapack.lib
!**************************************

      implicit none;
      integer(4),parameter :: ny =2;         !number of measurements at each time
      integer(4),parameter :: nu =1;         !number of controls at each time
      integer(4),parameter :: nyu =ny+nu;
      integer(4),parameter :: ndim =10;      !row dimensioning of phi,q
      integer(4),parameter :: mmax =500;     !# meas times
      integer(4),parameter :: nlag =20;      !number of lags used to compute cva model. nlag*ny must
                                             ! be greater than the actual number of states in system,
                                             ! but very large (> 20) values degrade model.
      integer(4),parameter :: nd =4;         !true model order

      integer(4),parameter :: lworkc =4*ndim;
      integer(4),parameter :: nud =max(nu,1);
      integer(4)           :: minlag =10;    !for example ARMAX (4,4) problem, CVA selects 3 if allowed, but CVA model is still reasonable

      integer(4) i,j,m,ix,info,lwork,nplot,nord,nBT,nYW;
      integer(4) iwork(10*ndim);             !LAPACK SVD (only need 8*nyunlg)
      integer(4) iunit,junit,npts,nlag1,npmin,npmax,nplt,nplt2,iWindow;
      integer(4) nseg,n3;
      logical(4) :: directUfeed =.true.;      !assume direct feed of control u to measurement y (A) ###

      real(8) dt,a(nd),b(nd),x(nd),xd(nd),yu(nyu,mmax),w(nd),eps,rsig,c1;
      real(8) phi(ndim,ndim),q(ndim,ndim),h(ny,ndim),rn(ny,ny);
      real(8) gu(ndim,nud),u(nud),ac(ny,nud),uc(nud,mmax);
!      real(8) gu(ndim,0:nu),u(0:nu),ac(ny,0:nu),uc(0:nu,mmax)
      real(8) bw(ny,ndim),sigy(ny*(ny+1)/2);
      real(8) rad,theta(2),ar(2),ai(2);
      real(8) :: pi=3.141592653589793d0;
      complex phic(ndim,ndim),workc(lworkc),wc(ndim);
      complex vl(ndim,ndim),vr(ndim,ndim);
      real(4) rwork(2*ndim),dt1,datai(mmax);
      real(4) psd(mmax,ny),psd2(mmax,ny),psd3(mmax,ny),psd4(mmax,ny);
      real(4) delf,delf2,delf3,delf4,g(mmax);
      real(8),allocatable :: work(:),sv(:),us(:,:),vt(:,:),d1(:,:);
      real(8),external :: rnor;
      !_________________________________________________

      ix =191919;     !###
      dt =rnor(ix);   !initialize seed
      dt =1.;

      rad =0.90;
      theta(1) =0.20*pi;
      theta(2) =0.40*pi;
!      theta(1) =0.10*pi
!      theta(2) =0.20*pi
      ar(1) =rad*cos(theta(1));
      ai(1) =rad*sin(theta(1));
      ar(2) =rad*cos(theta(2));
      ai(2) =rad*sin(theta(2));
      a(1:2) =(/ ar(1)**2+ai(1)**2, -2.d0*ar(1) /);  !denominator coefficents
      a(3:4) =(/ ar(2)**2+ai(2)**2, -2.d0*ar(2) /);
      b(1:2) =(/ 0.5d0,  1.0d0 /);     !numerator coefficients
      b(3:4) =(/ 0.5d0,  1.0d0 /);     !numerator coefficients
!      c1 =0.0        !###
      c1 =0.3;
      xd(:) =0.d0;
      rsig =0.5;
!        x(1) =xd(2)
!        x(2) =-a(1)*xd(1) -a(2)*xd(2) +w(2)
!        x(3) =(b(1)*xd(1) +b(2)*xd(2) +u(1)) -a(3)*xd(3) +w(3)
      phi(:,:) =0.d0;
      phi(1,:nd) =(/ 0.d0, 1.d0, 0.d0, 0.d0 /);
      phi(2,:nd) =(/ -a(1), -a(2), 0.d0, 0.d0 /);
      phi(3,:nd) =(/ 0.d0, 0.d0, 0.d0, 1.d0 /);
      phi(4,:nd) =(/ 0.d0, 0.d0, -a(3), -a(4) /);
!** eigenvalues are  (0.??), (0.??)

      !**  compute eigenvalues of phi: output is s, u and v
      lwork =nd*(10*nd+4);   !n*(7*n+4)
        ! 3*min(M,N)*min(M,N) + max(max(M,N),4*min(M,N)*min(M,N)+4*min(M,N))
      allocate (work(lwork),sv(nd),us(nd,nd),vt(nd,nd),d1(nd,nd),       &
     &           stat=info);
      if (info /= 0) stop 'allocate work,... failed';
      d1(:nd,:nd) =phi(:nd,:nd);
      call DGESDD("N", nd, nd, d1, nd, sv, Us, nd,                      & !"N" specifies that only singular values are computed
     &               vt, nd, work, lwork, iwork, info );
      if (info /= 0) stop 'error computing svd of phi';
      call prtmx (sv,1,1,nd,"phi-singular values");
      deallocate (work,sv,us,vt,d1);

      !** compute complex right eigenvalues
      phic(:,:) =phi(:,:);
      call CGEEV("N", "V", Nd, phic, ndim, wc, VL, ndim, VR, ndim,      &
     &                  WORKc, LWORKc, RWORK, INFO );
      write (6,'("returned from cgeev: ",i3)') info;
      write (6,'("phi complex eigenvalues",10(/2g13.5))') wc(:nd);

      q(:,:) =0.d0;
      q(1,1) =1.d-20;
      q(2,2) =1.;
      q(3,3) =1.d-20;
      q(4,4) =8.;

      h(:,:) =0.d0;
      h(1,1:2) =(/ b(1)-a(1), b(2)-a(2) /);
!      h(1,1:2) =(/ b(1)-c1*a(1), b(2)-c1*a(2) /)      used before 10/25
      h(2,1:2) =c1*h(1,1:2);              !c1*y1 is added to y2
      h(2,3:4) =(/ b(3), b(4) /);
      bw(:,:) =0.d0;
      bw(1,2) =1.d0;
      bw(2,2) =c1*bw(1,2);
      rn(:,:) =0.d0;
      rn(1,1) =rsig**2;
      rn(2,2) =rsig**2;
      ac(:,:) =0.d0;
      gu(:,:) =0.d0;
      if (nu > 0) then;
        gu(4,:) =1.d0;
        if (directUfeed) then;
          ac(2,:) =0.2;
        endif;
      endif;
      sigy(:) =0.d0;
      w(:) =0.d0;
      u(:) =0.d0;

      open (11,file="ARMAdata.plt",status='unknown');
      if (nu ==1) then;
        write (11,2000);
 2000   format ('variables ="time (s)", "y1", "y2", "u1"')
      else;
        write (11,2010);
 2010   format ('variables ="time (s)", "y1", "y2", "u1", "u2"')
      endif;

!**  Monte Carlo simulation using
!**      y(i) =Ac*u(i) +H*x(i) +Bw*w(i) +v(i)
!**      x(i+1) =Phi*x(i) +Gu*u(i) +w(i)

      do i=1,mmax;
        do j=1,nd;
          w(j) =sqrt(q(j,j))*rnor(0);
        enddo;
        !** measurement/control
        yu(:ny,i) =matmul(h(:,:nd),xd(:nd)) +matmul(bw(:,:nd),w(:nd));
        yu(1,i) =yu(1,i) +rsig*rnor(0);
        yu(2,i) =yu(2,i) +rsig*rnor(0);
        if (nu > 0) then;
!          u(1) =rnor(0)    !test ###
          u(1) =0.8*sin(6.283*i/25.);
          if (nu > 1) u(nu) =10.;           !### test effect of bias
          uc(:,i) =u(:);
          yu(:ny,i) = yu(:ny,i) +matmul(ac(:ny,:),u(:));
          yu(ny+1:nyu,i) =u(:nu);
        endif;
        !** state update
        x(:nd) =matmul(phi(:nd,:nd),xd(:nd)) +w(:nd);
        if (nu > 0) x(:nd) =x(:nd) +matmul(gu(:nd,:nu),u(:nu));

        xd(:) =x(:);       !save x(i+1) for next time
        sigy(1) =sigy(1) +yu(1,i)**2;
        sigy(2) =sigy(2) +yu(1,i)*yu(2,i);
        sigy(3) =sigy(3) +yu(2,i)**2;
        write (11,'(i4,5g13.5)') i,yu(:,i);
      enddo;

      close (11);
      write (6,'("sigy =",3g13.5)') sigy(:)/mmax;

      write (6,'("true model")');
      call prtmx (phi,ndim,nd,nd,"phi");
      call prtmx (q,ndim,nd,nd,"q");
      call prtmx (h,ny,ny,nd,"h");
      call prtmx (rn,ny,ny,ny,"rn");
      call prtmx (bw,ny,ny,nd,"bw");
      call prtmx (ac,ny,ny,nu,"ac");
      call prtmx (gu,ndim,nd,nu,"nu");

      nplot =100;
      call plotPsd2 (nplot,nd,ndim,ny,nu,phi,q,h,rn,bw,gu,ac,uc,mmax,   &
     &            dt,'ypsd.plt');

      !** perform canonical variate analysis
!        calling sequence -
!          yu(nyu,m) - input data and control versus time (m)
!          ny - number of elements in measurement vector at each time
!          nyu - sum of number of elements in measurement and control
!                vectors (nyu=ny+nu)
!          m - number of time points
!          nlag - number of lags used to compute cva model. nlag must
!                 be greater than the actual number of states in system.
!          eps - tolerance used to determine system rank (ex. eps=1d-12)
      eps =1.d-12;
      m =mmax;
      call cva (phi,h,q,rn,bw,gu,ac,nord,                               &
     &          ny,nyu,m,minlag,nlag,ndim,yu,eps,dt,directUfeed);

      write (6,'("cva n =",i3)') nord;
      call prtmx (phi,ndim,nord,nord,"phi");
      call prtmx (q,ndim,nord,nord,"q");
      call prtmx (h,ny,ny,nord,"h");
      call prtmx (rn,ny,ny,ny,"rn");
      call prtmx (bw,ny,ny,nord,"bw");
      if (nyu > ny) then;
        call prtmx (gu,ndim,nord,nyu-ny,"gu");
        call prtmx (ac,ny,ny,nu,"ac");
      endif;

      iunit =6;
      junit =10;
      dt1 =dt;
      npts =mmax;
      nplt =mmax/2+1;
      nplt2 =101;

      do i=1,ny;
        datai(:mmax) =yu(i,:mmax);
        !** c is the input data vector of length n
        !** d is the power spectrum of c (length n/2)
        !** nplt is the number of points of d which are to be plotted on iunit
        nseg =10;         !break FFT into segments for power averaging
        iWindow =3;       !window function: 0=none, 1=10% cosine taper, 2=Hamming, 3=Welch 50% overlap windows
        call spect(psd(:,i),delf,n3,                                    &
     &             datai(:),npts-1,nplt,0,dt1,nseg,iWindow);

        !** Blackman-Tukey (correlogram) PSD
        nlag1 =npts/10;
        iWindow =1;     !1=Bartlett, 2=Hamming (unreliable scaling)
        call BT_spect (psd3(:,i),delf3,nBT,                             &
     &                 datai(:),npts,dt1,nlag1,iWindow,nplt2);

        !** yule-walker spectrum
        nlag1 =npts/2;
        call YW_spect (psd4(:,i),delf4,g, datai(:),npts,dt1,nlag1,nplt2);

        !** maximum entropy method
        nlag1 =npts/2;
        npmin =2;
        npmax =nlag1;
!        npmax =30    !###
        call memopt (delf2,psd2(:,i),  datai(:),npts-1,npmin,npmax,     &
     &               nlag1,dt1,iunit,junit+i+2,nplt2);
      enddo;

      open (10,file="PSD_fft.plt",status='unknown');
      write (10,1000);
 1000 format ('variables ="frequency", "y1-PSD (dB)", "y2-PSD (dB)"')

      do i=2,n3;
        write (10,'(3g13.5)') (i-1)*delf,psd(i,1:2);
      enddo;
      close (10);

      open (10,file="PSD_BT.plt",status='unknown');
      write (10,1000);
      do i=1,nBT;
        write (10,'(3g13.5)') (i-1)*delf3, psd3(i,1:2);
      enddo;
      close (10);

      open (10,file="PSD_YW.plt",status='unknown');
      write (10,1000);
      do i=1,nplt2;
        write (10,'(3g13.5)') (i-1)*delf4, psd4(i,1:2);
      enddo;
      close (10);

      open (10,file="PSD_mem.plt",status='unknown');
      write (10,1000);
      do i=1,nplt2;
        write (10,'(3g13.5)') (i-1)*delf2, psd2(i,1:2);
      enddo;
      close (10);

      stop;
      end;
!****************************************************************
      subroutine plotPsd (nplot,n,ndim,ny,nu,phi,q,h,rn,bw,gu,ac,uc,mu, &
     &                    plotFile);
!**      generate a plot file of measurement power spectral density for
!**      white process noise input: given the state transition,
!**      state noise, measurementsensitivity matrix.
!**      Also plot contribution from control inputs ?

!**      State space Model is
!**         y(i) =Ac*u(i) +H*x(i) +Bw*w(i) +v(i)
!**         x(i+1) =Phi*x(i) +Gu*u(i) +w(i)
!**      where
!**         u(i) is a known control
!**         w(i) is state noise
!**         v(i) is measurement noise

      implicit none;
      character(*),intent(in) :: plotFile;   !plot file name
      integer(4),intent(in) :: ny;           !# of measurements at each time
      integer(4),intent(in) :: nu;           !# of controls at each time
      integer(4),intent(in) :: nplot;        !# of points to be plotted (e.g., 100)
      integer(4),intent(in) :: n;            !state order
      integer(4),intent(in) :: ndim;         !row dimension of phi,q (ndim >= n)
      integer(4),intent(in) :: mu;           !# of time samples in uc
      real(8),intent(in) :: phi(ndim,n);     !state transition matrix
      real(8),intent(in) :: q(ndim,n);       !process noise covariance
      real(8),intent(in) :: h(ny,n);         !meas sensitivity matrix
      real(8),intent(in) :: rn(ny,ny);       !meas noise covariance
      real(8),intent(in) :: bw(ny,ndim);     !state noise mapping to measurements
      real(8),intent(in) :: gu(ndim,nu);     !control mapping to states
      real(8),intent(in) :: ac(ny,nu);       !control mapping to measurements
      real(8),intent(in) :: uc(nu,mu);       !time samples of the controls

      integer(4) ip,i,j,k,info,nrhs,iwork(n);
      real(8) ypsd(ny),eps1,sq(n*(n+1)/2),delF,pmag(nu),psum(nu),scalu;
      real(8) :: pi=3.141592653589793d0;
      real(4) theta;
      complex vl(n,n),vr(n,n),zimphi(n,n),z,z1,uz(nu);
      complex hx(ny,n),pv(ny,nu),pw(ny,n);
      !____________________________________

      open (10,file=plotfile,status='unknown');
      write (10,1000);
 1000 format('variables ="frequency", "y1-psd", "y2-psd"')
      delF =1.d0/(2*nplot);   !for dt =1.0
      scalu =sqrt(1.d0/mu);
      psum =0.d0;

      do ip=0,nplot;
        theta =ip*2.*pi*delF;
        z =cmplx(cos(theta),sin(theta));

        if (nu > 0) then;
          !** compute z-transform of control data u
          z1 =cmplx(1.,0.);
          uz(:) =0.;
          do i=mu,1,-1;
            uz(:) =uz(:) +uc(:,i)*conjg(z1);
            z1 =z1*z;
          enddo;
          uz(:) =uz(:)*scalu;
          pmag =real(uz(:))**2+aimag(uz(:))**2;
          psum =psum +pmag;
!          write (6,'("f =",g13.5,", uz =",2g13.5)')
!          write (6,'(5g13.5)') ip*delF, pmag   !###
        endif;

        zimphi(:n,:n) =-phi(:n,:n);
        forall (j=1:n) zimphi(j,j) =zimphi(j,j) +z;
        !** invert z*I-phi
        nrhs =n;
        vl(:,:) =0.;
        do j=1,n;
          vl(j,j) =cmplx(1.,0.);
        enddo;
        vr(:,:) =zimphi(:,:);
        !** vl = complex matrix inversion of (zI-phi)
        call CGESV( n,nrhs, zimphi, n, iwork, vl, n, info);
        if (info /= 0) stop 'error inverting z*I-phi';

        !** check
!       vr(:,:) =matmul(vr,vl);
!       do i=1,n;
!         vr(i,i) =vr(i,i) -1.0;
!         do j=1,n;
!           if (real(vr(i,j)) > 1.d-6 .or. aimag(vr(i,j)) > 1.d-6) then;
!             write (6,'("error: ",a," (",2i4," =",2g13.5)')            &
!    &           "zimphi*inv",i,j,vr(i,j);
!           endif;
!         enddo;
!       enddo;

        !** cholesky factor q as L*L^T where L is lower triangular
        k =0;
        do i=1,n;
          do j=1,i;
            k =k+1;
            sq(k) =q(j,i);
          enddo;
        enddo;
        eps1 =1.d-4;
        call cfactor (info,  sq,n,eps1);
        if (info < 0) stop 'plotPSD error: Q is singular';
        vr(:,:) =0.;
        k =0;
        do i=1,n;
          do j=1,i;
            k =k+1;
            vr(i,j) =sq(k);   !save lower triangular part in complex array
          enddo;
        enddo;

        !** transfer function is hx =h*(zI-phi)^(-1)
        !** power spectrum is magnitude squared of (hx+bw)*sqrt(q)
        hx(:ny,:) =matmul(h(:ny,:n),vl(:,:));
        pw(:ny,:) =matmul((hx(:ny,:)+bw(:ny,:)),vr(:,:));
        forall (j=1:nu)                                                 &
     &     pv(:ny,j) =(matmul(hx(:ny,:n),gu(:n,j)) +ac(:ny,j)) *uz(j);
        do j=1,ny;
!          ypsd(j) =sum(real(hx(j,:n))**2 +aimag(hx(j,:n))**2)
          ypsd(j) =sum(pw(j,:)*conjg(pw(j,:)));
!          do i=1,n
!            ypsd(j) =ypsd(j) +pw(j,i)*conjg(pw(j,i))
!          enddo
          if (nu > 0 .and. mu > 0)                                      &
     &             ypsd(j) =ypsd(j) +sum(pv(j,:)*conjg(pv(j,:)));
          ypsd(j) =10.*log10(ypsd(j) +rn(j,j));
        enddo;
        write (10,'(4g13.5)') ip*delf, ypsd(:ny);
      enddo;

      write (6,'("average U power =",4g13.5)') psum/nplot;
      close(10);
      return;
      end subroutine plotPSD;

