!** jumpCom.for includes the global variables and subroutines required to implement jump
!** detection/estimation in a Kalman filter as described in
!**   Gibbs, B.P. (1992), Jump detection/estimation using stepwise regression,
!**   IEEE Trans. on Aerospace and Electronic Sys., 28 (4), 1105-1118, Oct. 1992

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you  agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

      module jumpCom;
      save;
!***    the jumpCom module includes global variables used in jump detection

      !** dimension variables
      integer(4) :: mdim;        !max # measurements
      integer(4) :: nparm;       !max # parameters
      integer(4) :: n2par;       !max size of jump covariance matrix =(nparm*(nparm+1))/2
      integer(4) :: npmax;       !max allowable number of "accepted" jump parm (<= nparm)
      integer(4) :: np2max;      !=(npmax*(npmax+1))/2

      integer(4) np;             !total number of candidate jump parameters ( <=nparm)

      real(8) f2rem;             !value of f-statistic used as a minimum
                                 ! threshold when removing a parameter from the
                                 ! stepwise regression. (suggest 0.5)
      real(8) f2add;             !value of f-statistic used as a maximum
                                 ! threshold when adding a parameter to the
                                 ! stepwise regression. (suggest 9.0)
      real(8) tolmin;            !minimum value of tolerance test (variance ratio) when deciding
                                 ! whether to add a parameter to the regression.
                                 ! the tolerance is a measure of the independence
                                 ! of the parameter from other parameters in the
                                 ! regression.  (suggest tolmin=0.01).

      real(8),allocatable :: sigxj(:); !1-sigma prior uncertainty on initial value
                                       ! of states allowed to jump. a scaled amount
                                       ! of sigxj is used as the prior uncertainty
                                       ! on the jumps in routine estjmp.
      integer(4),allocatable :: mapx(:);!mapping from estimated parm to total parameters

      !** buffer variables
      integer(4) :: nxp;               !number of accepted "accepted" jump parameters
      character(15),allocatable :: xlab(:);  !jump parameter labels
!      logical(4),allocatable :: isel(:)   !jump parameter selection flag: 1=selected  not used
      character(15),allocatable :: xlxp(:);!"accepted" jump parameter labels
      real(8) sigfac;                  !factor multiplying sigxj used as prior
                                       ! information on regression states
      real(8) :: rssum;                !corrected residual sum-of-squares
      real(8),allocatable :: a12c(:,:);!cross-information matrix used by corresr
      real(8),allocatable :: xpt(:);   !"accepted" jump parameter values

      contains;
!**********************************************************************
      subroutine init_jump (mdim1,nparm1,xlab1);
!***    allocates and initializes jump parameters and arrays.

!***    Note: Before calling init_jump, the following variables must be set:
!***       f2rem,f2add,tolmin,sigxj

      implicit none;
      integer(4),intent(in) :: mdim1,nparm1;
      character(*),intent(in) :: xlab1(nparm1);
      integer(4) ier;
      !____________________________________________________

      f2rem =2.;  !0.5
      f2add =4.;     !f2add - f-statistic to add a parameter (e.g., 16.0 for 4-sigma)
      tolmin =0.001;  !minimum tolerance required to include a parameter
      mdim =mdim1;

      nparm =nparm1;     !max possible # of total  states
      n2par =(nparm*(nparm+1))/2;!max size of state covariance matrix
      npmax =nparm;          !or smaller ??
      np2max =(npmax*(npmax+1))/2;
      np =nparm;

      !** jump variables
      allocate (xlab(nparm),xlxp(nparm),a12c(npmax,nparm),              &
     &          xpt(nparm),sigxj(nparm),mapx(nparm),stat=ier);
      if (ier /= 0) call errhand ('allocate jbuff,... error');

      rssum =0.d0;
      xlab(:nparm) =xlab1(:nparm);
!      isel(:) =.false.
      sigxj(:) =1.d20;
      sigfac =1.;

      return;
      end subroutine init_jump;
! *****************************************************************
      subroutine estjmpj (xp,njAdd,rrms,  isel,  dz,dzdp,np,m,mdim1);
!         subroutine estjmp estimates the jump states and times using
!         the kalman filter residuals as data for least squares.
!         forward stepwise regression is used to determine significant jump states
!         note: prior information (sigxj) is used

!           inputs:
!             dz(m) - measurements
!             dzdp(mdim,np) - meas partials
!             xlab(np) - labels of parameters
!             sigxj(np) - 1-sigma uncertainty on jump states
!             np - maximum number of parameters estimated

!             f2rem - f-statistic to remove a parameter (e.g., 0.5)
!             f2add - f-statistic to add a parameter (e.g., 16.0 for 4-sigma)
!             tolmin - minimum tolerance required to include a parameter
!                      (e.g., 0.01)
!             sigfac - factor multiplying sigxj used as
!                     prior information on regression states

      implicit none;

      integer(4),intent(in) :: np;             !number of states in model
      integer(4),intent(in) :: m,mdim1;         !number of measurements
      real(8),intent(in) :: dz(m),dzdp(mdim1,np); !meas and partials
      logical(4),intent(inout) :: isel(np);    !jump parameter selection flag: 1=selected  not used
      integer(4),intent(out) :: njAdd;         !number parameters estimated
      real(8),intent(out) :: xp(np);           !estimated parameters
      real(8),intent(out) :: rrms;             !fit residual rms

      !** local
      integer(4) iter,i,j,l,l1,mt,ip,im,jp;
      integer(4) ladd,ier;
      integer(4) :: nbias =0;    !no bias states in regression
      real(8) rss,dzc,eps,xpm,rss1,rsos,epsm;
      real(8),allocatable :: a(:), b(:), d1(:);
      real(8),allocatable :: a1(:),b1(:),a12(:);
      !____________________________________________________

      allocate (a(n2par), b(np), d1(np),                                &
     &          a1(np2max),b1(npmax),a12(npmax),stat=ier);
      if (ier /= 0) call errhand ('estjmpj allocate error');

      iter =0;
!      maxit =5    !max iterations
      rss =0.;

!***     add prior on jump magnitudes
      b(:) =0.d0;
      a(:) =0.d0;
      l=0;
      do i=1,np;
        l =l+i;
        b(i) =0.d0;        !prior nominal is 0
        a(l) =1.d0/sigxj(i)**2;
      enddo;
!      call prtsmx(a,j,'a0      ')

      !** form information arrays
      rss =0.d0;
      do im=1,m;
        rss =rss +dz(im)**2;
        !**   compute information arrays (assumes non-zero intercept)
        l =0;
        do ip=1,np;
          dzc =dzdp(im,ip);
          b(ip) =b(ip)+dz(im)*dzc;
          do jp=1,ip;
            l =l+1;
            a(l) =a(l)+dzdp(im,jp)*dzc;
          enddo;
        enddo;
      enddo;

      write (6,'(" estjmpj: ",i6," meas, residual sos =",g12.5)') m,rss;
!      np2 =np
      mt =m;

!***    start of stepwise regression loop:
!***    form information matrix & vector for estimated states
      ladd =1;
      eps =1.d-5;

      do while (ladd /= 0);
        iter =iter+1;
        l1=0;
        njAdd =0;
        do i=1,np;
          if (isel(i)) then;
            njAdd =njAdd+1;
            if (njAdd > npmax)                                          &
     &               call errhand ('estjmpj error: njAdd > npmax');
            mapx(njAdd) =i;
            b1(njAdd) =b(i);
            l =(i*(i-1))/2;
            do j=1,i;
              if (isel(j)) then;
                l1 =l1+1;
                a1(l1) =a(l+j);
              endif;
            enddo;
          endif;
        enddo;

        rsos =rss;

        if (njAdd > 0) then;
!          call prtsmx (a1,njAdd,'a info  ')
          !**    compute least squares estimate
          call sinv (ier,epsm,  a1,  njAdd,eps,.false.);
!          call prtsmx (a1,njAdd,'a inv   ')
          if (ier < 0) call errhand ('matrix inversion error');
          call symult (xp,  a1,b1,njAdd);

          l=0;
          do i=1,njAdd;
            l =l+i;
            d1(i) =sqrt(a1(l));
          enddo;
          if (iter == 1) then;
            call prtmx (xp,1,1,njAdd,'xp      ');
            call prtmx (d1,1,1,njAdd,'xp sig  ');
          endif;

          rsos =rsos -dot_product(b1(1:njAdd),xp(1:njAdd));
          write(6,2200) iter,rss,rsos,mt,njAdd;
 2200     format(/'**************'/                                     &
     &     ' at iteration ',i3,', data sum-of-squares =',g12.5,         &
     &     ', resid. sos =',g12.5,', # meas=',i5,', nstate =',i3/)

        endif;

        !**   check f-statistic to add or remove states from regression.
        !**   Output is ladd and isel
        call addremr (ladd,  isel,  a,b,a1,xp,np,njAdd,mt,rsos,nbias);

        !** if states added or removed, recompute regression
      enddo;   !___ end do while loop

      rrms =sqrt(rsos/m);

      if (njAdd > 0) then;
        !**  Note: individual rss1 below may be bigger or smaller than rsos above,
        !**  but sum over all jump parameters matches

        do jp=1,njAdd;
          xpm =xp(jp);
          ip =mapx(jp);
          xlxp(jp) =xlab(ip);
!          xvar =a1((jp*(jp+1))/2)
          !**  parameter estimated - correct state and residuals
          rss1 =(xpm/sigxj(ip))**2;
          do im=1,m;
            dzc =dzdp(im,ip)*xpm;
            rss1 =rss1 +(dz(im)-dzc)**2;
          enddo;
          write (6,'(" Parameter ",a15,g13.5,                           &
     &          ": original and corrected sos =",2g13.5)')              &
     &          xlab(ip),xpm,rss,rss1;
        enddo;    !___ end jp loop

        !** calculate the corrected residuals and
        !** residual partials given the estimated bias parameters (uses jbuff(i)%dzdp)
!        call corresr (dzdp,  a,a1,xp,dzdp,dz,isel,njAdd,np,nxjmp,m)
      endif;   !___ end if (njAdd > 0)

      deallocate (a, b, d1, a1,b1,a12);

      return;
      end subroutine estjmpj;
! **********************************************************************
      subroutine addremr (ladd,  isel,  a,b,a1,xp,np,njAdd,mt,rsos,     &
     &                     nbias);
!                         f2add,f2rem,tolmin,xlab,mapx  from global
!***     subroutine addrem tests the f-statistic to determine whether
!***     states should be added or removed from the regression

      implicit none;

      integer(4),intent(in) :: np;        !number of jump parameters in a and b information arrays
      integer(4),intent(in) :: njAdd;     !number of current "accepted" jump parm
      integer(4),intent(in) :: mt;        !total number of (scalar) meas
      integer(4),intent(in) :: nbias;     !number of bias states in regression
!      character(15),intent(in) :: xlab(np ) !labels of parameters
!      integer(4),intent(in) :: mapx(npmax) !mapping from accepted states to np2 index
      real(8),intent(in) :: a(np*(np+1)/2); !information matrix for candadate jump parameters
      real(8),intent(in) :: b(np);        !information vector for candadate jump parameters
      real(8),intent(in) :: a1(np2max);   !information array for "accepted" jump parameters
      real(8),intent(in) :: xp(npmax);    !estimates of jump parameters
!      real(8),intent(in) :: f2add         !regression f-test threshold to add state (e.g., 16.0)
!      real(8),intent(in) :: f2rem         !regression f-test threshold to remove state (e.g., 0.5)
!      real(8),intent(in) :: tolmin        !minimum variance ratio required to add a state (e.g. 0.01)
      real(8),intent(in) :: rsos;         !residual sum-of-squares before regression
      logical(4),intent(inout) :: isel(np); !jump parameter selection flag for all parm
      integer(4),intent(out) :: ladd;     !number of parameters added to regression

      !** local
      integer(4) i,j,j1,l,l1,imin,imax,iflag;
      real(8) a12(np),c12(np),xsig,f,sms,fmin,fmax;
      real(8) a22,b2,c22,tol,ri,fi,df,tolmax,rmax;
      !____________________________________________________

      ladd =0;
      df =mt -njAdd;!+njAdd  if prior is used  ###
      sms =rsos/df;

      if (njAdd > 0) then;
        !**   check f-to-remove to remove states from regression
        l =0;
        iflag =0;
        fmin =f2rem;
        imin =0;
        do i=1,njAdd;
          l =l+i;
          xsig =sqrt(a1(l));
          f =xp(i)**2/(a1(l)*sms);
          write(6,2300) mapx(i),xp(i),xsig,f;
 2300     format(' parameter ',i4,', xp =',g12.5,', xsig =',g12.5,      &
     &           ', ftest =',g12.5)
          if (f < fmin) then;
            fmin =f;
            imin =i;
          endif;
        enddo;

        if (imin > 0) then;
          isel(mapx(imin)) =.false.;
          l =(imin*(imin+1))/2;
          write(6,2000) imin,mapx(imin),xp(imin),a1(l);
 2000     format(' **** estimated state ',i4,'(',i4,')',                &
     &         ' removed from regression. xp = ',g12.5,', cov =',g12.5)
          ladd =-1;
          return;
        endif;
      endif;

!***     check f-to-add to add states
      imax=0;
      fmax =f2add;
      write(6,'("addremr: mt,njAdd =",2i6)') mt,njAdd;
      do i=1,np;
        if (.not.isel(i)) then;
          l =(i*(i+1))/2;
          a22 =a(l);
          b2 =b(i);
          j1 =0;
          do j=1,np;
            if (isel(j)) then;
              j1=j1+1;
              if (j <= i) then;
                l1 =l-i+j;
              else;
                l1 =(j*(j-1))/2+i;
              endif;
              a12(j1) =a(l1);
            endif;
          enddo;

          c22 =a22;
          if (njAdd > 0) then;
            call symult(c12,  a1,a12,njAdd); !symult(a1,a12,c12,njAdd)
            do j=1,njAdd;
              b2 =b2-xp(j)*a12(j);
              c22 =c22-c12(j)*a12(j);
            enddo;
            if (c22 <= 0.d0) c22 =1.d-20;
          endif;

!         write(6,'(i3,", a22,c22,rsos,b2 =",4g13.5)')                  &
!    &         i,a22,c22,rsos,b2;
          if (a22 > 0.d0) then;
            tol =c22/a22;
            df =mt-(njAdd+1);    !+(njAdd+1) if prior is used
            ri =b2/sqrt(c22*rsos);
            fi =(b2**2*df)/(c22*rsos-b2**2);
            write(6,'(i4,", tol,fi =",2g13.5)') i,tol,fi;

            if (tol > tolmin .and. fi > fmax) then;
              imax =i;
              fmax =fi;
              tolmax =tol;
              rmax =ri;
            endif;
          endif;

        endif;    !___ end test for .not.isel
      enddo;   !___ end i-loop

      if (imax > 0) then;
        isel(imax) =.true.;
        write(6,2100) imax, fmax, tolmax,rmax, df;
 2100   format(/'**** state ',i4,' added to regression.'/               &
     &         ' fmax =',g13.5,', tolmax =',g13.5,', rmax =',g13.5,     &
     &         ', deg-of-freedom =',f6.0)
        ladd=1;
        return;
      endif;

      return;
      end subroutine addremr;
! ****************************************************************
      subroutine corresr (dzdp,dz, a,a1,xp,isel,njAdd,np,m);
!***      subroutine corres computes the corrected residuals and
!***      residual partials given the estimated bias parameters
!***      Uses structure jbuff data.
      implicit none;

      integer(4),intent(in) :: njAdd;        !current number of "accepted" jump states
      integer(4),intent(in) :: np,m;          !number of possible jump parameters
      logical(4),intent(in) :: isel(np);     !jump parameter selection flag: 1=selected
      real(8),intent(in) :: a(np*(np+1)/2);  !information matrix for all possible jump parameters
      real(8),intent(in) :: a1(njAdd*(njAdd+1)/2);!covariance matrix for "accepted" jump parameters
      real(8),intent(in) :: xp(njAdd);       !estimates of jump parameters
      real(8),intent(inout) :: dzdp(mdim,np),dz(m);!partial of meas residual wrt jump parameters

      !** local
      integer(4) i,j,j1,l,l1,im,ip;
      real(8) dum(njAdd),h1(njAdd),yp,ha;
      !____________________________________________________


!***    save cross information between adjusted and all other parameters
      do i=1,np;
        l=(i*(i+1))/2;
        j1=0;
        do j=1,np;
          if (isel(j)) then;
            j1 =j1+1;
            if (j <= i) then;
              l1 =l-i+j;
            else;
              l1 =(j*(j-1))/2+i;
            endif;
            a12c(j1,i) =a(l1);
          endif;
        enddo;
        if (j1 /= njAdd) call errhand ('corresr error: j1 /= njAdd');
      enddo;

!***     save inv(a1)*a12c in a12c
      do i=1,np;
        call symult (dum,  a1,a12c(:,i),njAdd); !(a1,a12c(:,i),dum,njAdd)
        a12c(1:njAdd,i) =dum(1:njAdd);
      enddo;

!***     now form corrected residuals and residual partials

      do im=1,m;
        yp=0.d0;
        j=0;
        h1(:) =0.d0;
        do i=1,np;
          if (isel(i)) then;
            j=j+1;
            h1(j) =dzdp(im,i);
            yp =yp+dzdp(im,i)*xp(j);
          endif;
        enddo;
        dz(im) =dz(im)-yp;    !wbuf(1,im,ibuf)

        !**  compute corrected partials
        do ip=1,np;
!          call smult (h1,a12c(1,ip),ha, 1,njAdd,1, 1,npmax,1)
          ha =dot_product(h1(:njadd),a12c(:njadd,ip));
          dzdp(im,ip) =dzdp(im,ip) -ha;
        enddo;
      enddo;

      return;
      end subroutine corresr;
!*************************************************************
      subroutine errhand(string);
      character(*),intent(in) :: string;

      write (6,'("error: ",a)') string;
      write (0,'("error: ",a)') string;
      stop;
      end subroutine errhand;
!******************************************************************
      end module jumpCom;
