!** LS_cov.f performs linear least-squares covariance error analysis.
!** LS_cov is general-purpose and may be used with other models (defined
!** by module LSmodel.f90).  A driver routine must first first call
!** read_setup before calling LS_cov.  See run_LScov.f90 for an example.

!** author B. Gibbs  1/2011

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!**   using Intel compiler:
!**        in ifort.cfg
!**   /check:bounds /traceback /include:\kf_book\distribSW\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSW\library\libkf_bnd_ifort.lib

!**   set ifortcfg=\kf_book\distribSW\ifort.cfg
!**   ifort LSmodel.f90 /c
!**   ifort LS_cov.f90 /c
!**   ifort run_LScov.f90 LSmodel.obj LS_cov.obj
!**   run_LScov poly.cfg >ls_covp.lis

!****************************************************************************
      module LScovMod;
      use filterPrec, only: real_kind1;   !precision (4 or 8)
      use LSmodel, only:ntotal            !number of states in total model
      !** contains global variables used in defining least squares covariance analysis

      implicit none;
      integer(4) nstate;        !total number of states in model (set by model)
!      integer(4) ncore          !number of core states in model
!      integer(4) nbias          !number of bias states in model

      integer(4) nest;          !total number of estimated states
      integer(4) nUnadj;        !total number of unadjusted-analyze parameters
      integer(4) nEU;           !=nest+nUnadj

      integer(4) nmeas;         !total number of vector meas in input file

      integer(4),allocatable :: mapxe(:); !mapping from estimated states (nest) to total model states
      integer(4),allocatable :: mapxu(:); !mapping from unadjusted analyze parameter (nUnadj) to total model states
      integer(4),allocatable :: mapx(:);  !mapping from estimated or unadjusted states (nEU) to total model states,
                                          ! with unadjusted states last
      integer(4),allocatable :: mapxei(:);!mapping from total model states (nstate) to estimated states
      integer(4),allocatable :: mapxui(:);!mapping from total model states (nstate) to estimated states
      integer(4),allocatable :: mapxi(:); !mapping from total model states (nstate) to estimated/unadjusted states

!      character(80) setupFile   !name of file containing data for structure xstate
!      character(80) measFile    !name of file containing measurement data

      logical(4) :: sparse =.true.;  !true = use sparse multiplies  !###
      logical(4) :: debugp =.true.;  !generates debug print
      logical(4) :: first =.true.;   !initializes information arrays

      real(8) t0;               !epoch time of x0
      real(8) tend;             !end time for filter processing
      real(8) rsum;             !residual sum-of-squares

      real(real_kind1),allocatable :: xf(:)

!      real(8) rnoise(mmaxi,mmaxi)           !meas noise covariance

      type states;
        integer(4) xtype;       !1=core state, 2=bias  (not used in LS_cov)
        character(8) status;    !must be "estimate", "analyze" or "ignore"
        character(20) xname;    !name of state that must match name in hard-coded model
        real(8) x0;             !a priori estimate at epoch time
        real(8) sig0;           !1-sigma uncertainty of x0 at epoch time
        real(8) const(2);       !known constants associated with this state in dynamic or measurement model
      end type;
      type (states) :: xstate(ntotal);

      contains;
!*****************************************************
      subroutine LS_cov
      !** LS_cov is a general subroutine for linear least squares covariance analysis
      !** when some parameters may be unadjusted.
      !** LS_cov can execute in single or double precision (real_kind1)

      !** IMPORTANT: The calling routine must first call read_setup before calling LS_cov.

      use filterPrec, only: real_kind1;   !precision (4 or 8)

      !** module LSmodel contains the model information and also computes the
      !** full set of measurements (y) and partial derivatives (ht) using
      !** subroutines get_yh and get_yTrue
      use LSmodel, only:ntotal,xnames,tepoch,get_yh,get_yMeas,mmax;

      implicit none;
      integer(4) i,j,k,kk,ier,n1,nn2;
      integer(4),allocatable :: mtype(:);
      character(40) :: pltFile ="residualDat.plt"

      real(real_kind1) eps,errm,dof,rsos0,rsos;
      real(real_kind1),allocatable :: px(:,:),bu(:),au1(:);
      real(real_kind1),allocatable :: hf(:,:),ha(:,:),dyResdx2(:,:);
      real(real_kind1),allocatable :: h1(:,:),h2(:,:),htt(:,:),ht(:,:);
      real(real_kind1),allocatable :: pyra(:),pyrp(:),pyrpn(:);
      real(real_kind1),allocatable :: a2(:,:),a3(:,:),pa(:);
      real(real_kind1),allocatable :: xe(:),x0(:),sigx(:),sigxn(:);
      real(real_kind1),allocatable :: yres(:),yp(:),yt(:),ytime(:);
      real(real_kind1),allocatable :: sigr(:),hs(:),e(:,:),c(:,:);
      !_______________________________

      allocate (x0(ntotal),pa(ntotal),stat=ier);
      if (ier > 0) call errhand ("error in allocating state vector");

      !** check that state labels of input file and LSmodel match
      nstate =ntotal;      !set in LSmodel
      ier =0;
      do i=1,nstate;
        x0(i) =xstate(i)%x0;              !Note: the order of states in x0 may be reset below
        if (xstate(i)%xname /= xnames(i)) then;
          write (6,'("Input error for state ",i4,", label ",a,           &
     &       " does not match model label ",a)')                         &
     &       i,xstate(i)%xname, xnames(i);
          ier =ier+1;
        endif;
      enddo;
      if (ier > 0) stop 'input error';

      !** allocate measurement-related arrays using mmax
      allocate (yt(mmax),ytime(mmax),mtype(mmax),stat=ier);
      if (ier > 0) call errhand ("error in allocating yt & ytime");

!** Call get_yMeas to retrieve/compute "measurement data" nmeas,yt,ytime, and mtype.
!** get_yMeas will usually read measurements from a file.
      call get_yMeas (nmeas,yt,ytime,mtype);

      !** allocate measurement-related arrays using nmeas
      allocate (yp(nmeas),sigr(nmeas),yres(nmeas),ht(nmeas,ntotal),      &
     &          h1(nmeas,nest),h2(nmeas,nUnadj),dyResdx2(nmeas,nUnadj),  &
     &          hf(nmeas,nEU),ha(nmeas,nEU),hs(nmeas),htt(nmeas,nEU),    &
     &          pyra(nmeas),pyrp(nmeas),pyrpn(nmeas),                    &
     &          stat=ier);
      if (ier > 0) call errhand ("error in allocating meas. arrays");

!** Now call get_yh to compute "measurement data" ht,y and sigr at measurements times ytime.
!** State vector at time t0 is initialized as x0(nstate) for use when model is nonlinear.
!** Note: this approach should be acceptable when mtotal <=10000, but for very large problems,
!** it may be better to directly accumulate the au and bu arrays in get_yh so that ht need not be stored.

      !** nmeas = total number of measurements
      !** y(nmeas) =measurements (or residual if nonlinear model)
      !** yp(:) =matmul(ht(:,:),x0(:));    !predicted (a priori) measurements from initial
                                          ! state (including unadjusted)
      !** ht(nmeas,ntotal) =partial derivatives of measurements WRT states
      !** sigr(nmeas) =measurement noise standard deviations (noise assumed to be white)

      call get_yh (yp,sigr,ht,  t0,x0,ytime,mtype,nmeas);

      nn2 =(nest*(nest+1))/2;
      allocate (px(nEU,nEU),xe(nest),a2(nest,nUnadj),                   &
     &          a3(nEU,nEU),e(nest,nest),c(nest,nmeas),                 &
     &          bu(nest),au1(nn2),sigxn(nest),sigx(nest), stat=ier);
      if (ier > 0) call errhand ("error in allocating scratch arrays");

!** Form Bayesian least squares information arrays (au1 and bu).  Adjusted states are ordered
!** first and unadjusted last, with ignored states not included.  This code uses the normal
!** equation arrays, but the square-root information arrays of the QR algorithm could also
!** be used by calling TDHHT.

      bu(:) =0.d0;
      au1(:) =0.d0;

      !** Form measurement partials (h1) WRT estimated states,
      !** and WRT unadjusted states (h2 matrix).
      !** Both are scaled by inverse 1-sigma measurement noise.

      do j=1,nest
        i =mapxe(j)
        h1(:,j) =ht(:,i)/sigr(:);
        x0(j) =xstate(i)%x0;
        pa(j) =xstate(i)%sig0;
      enddo;
      hf(:,:nest) =h1(:,:nest)
      do j=1,nUnadj
        i =mapxu(j)
        h2(:,j) =ht(:,i)/sigr(:);
        hf(:,nest+j) =h2(:,j)
        x0(nest+j) =xstate(i)%x0;
        pa(nest+j) =xstate(i)%sig0;
      enddo;
!      call prtmx (hf, nmeas,nmeas,nEU,"hf")

      !** reset predicted measurement to ignore prior estimate effects of ignored parameters
      yp(:) =matmul(h1(:,:nest),x0(:nest))
      if (nUnadj > 0) yp(:) =yp(:)+matmul(h2(:,:nUnadj),x0(nest+1:nEU));
      yres(:) =yt(:nmeas)/sigr(:) -yp(:)      !normalized a priori residual
      rsos0 =sqrt(sum(yres(:)**2));

      !** Now accumulate Bayesian weighted least information arrays
      kk =0;
      do j=1,nest;
        hs(:) =h1(:,j);
        bu(j) =dot_product(yres(:),hs(:));
        do k=1,j;
          kk =kk+1;
          au1(kk) =dot_product(hs,h1(:,k));    !information matrix only (no prior)
        enddo;
        do k=1,nUnadj;
          a2(j,k) =dot_product(hs,h2(:,k));
        enddo;

        !** add prior
        bu(j) =bu(j) +x0(j)/pa(j)**2;
        au1(kk) =au1(kk) +1.d0/pa(j)**2;
      enddo;
!      call prtsmx (au1,nest,"information matrix")

      !** scale a2 by prior 1-sigma: a2 =transpose(hf1)*hf2*pa^(1/2)
      if (nUnadj > 0) forall (i=1:nUnadj) a2(:,i) =a2(:,i)*pa(i+nest);

      !*** solve normal equations using Cholesky factorization
      !** first invert information array for estimated parameters
      eps =1.d-3;
      call sinv (ier,errm,  au1,  nest,eps,.false.);
      if (ier < 0) then;
        write (6,'("*** sinv error")');
        stop 'sinv error';
      endif;

      !** set px1 =au1
      kk =0;
      do j=1,nest;
        do k=1,j;
          kk =kk+1;
          px(k,j) =au1(kk);
          px(j,k) =au1(kk);
        enddo;
      enddo;

      !** compute state estimate and a posteriori measurement residuals
      call symult (xe,  au1,bu,nest);
      xe(:nest) =xe(:nest) +x0(:nest);
      yres(:) =yt(:nmeas) -sigr(:)*matmul(h1(:,:nest),xe(:nest));
      if (nUnadj > 0)                                                    &
     &   yres(:) =yres(:) -sigr(:)*matmul(h2(:,:nUnadj),x0(nest+1:nEU));
      rsos =sqrt(sum(yres(:)**2));
      write (6,'(/"For ",i5," measurements, a priori measurement RSS =", &
     &   g13.5,", fit residual RSS =",g13.5)') nmeas,rsos0,rsos

      write (6,'(/"noise-only a posteriori 1-sigma and ",                &
     &   "correlations for estimated states"/                            &
     &   4x,"name          1-sigma    correlations")');
      do i=1,nest;
        sigxn(i) =sqrt(px(i,i));
        write (6,1200) i,xstate(mapx(i))%xname,sigxn(i),                 &
     &    (px(i,j)/(sigxn(i)*sigxn(j)),j=1,i);
 1200   format(i3,1x,a10,g13.5,10f8.4,50(/27x,10f8.4))
      enddo;

      !** compute a posteriori NORMALIZED measurement residual covariance for
      !** adjusted states only.  Start by forming pyrp =hf1*px1*hf1^T

      ha(:nmeas,:nest) =matmul(h1(:nmeas,:nest),px(:nest,:nest));   !=hf1*px1
!      ha(nmeas+1:,:n) =matmul(hp(:nmeas,:n),px(:n,:n))             !=hp1*px1  (when including predicted measurements)
      forall (i=1:nmeas) pyrp(i) =dot_product(h1(i,:nest),ha(i,:nest));
      forall (i=1:nEU) htt(:nmeas,i) =hf(:nmeas,i)*pa(i);           !=hf*pa^1/2

      !** compute multiplier used in get_ln_like  (Not used in LS_cov)
      !** e =pa1^(-1/2)*px
      forall (i=1:nest) e(:nest,i) =px(i,:nest)/pa(i);
      !** c =pa1^(-1/2)*px1*hf^T
      c(:nest,:nmeas)=matmul(e(:nest,:nest),transpose(hf(:nmeas,:nest)));

      !** a posteriori measurement residual covariance is pyrp =r -hf1*px*hf1^T
      dof =nest;
      do i=1,nmeas;
        pyrp(i) =1.d0 -pyrp(i);           !R=I because hf1 was already normalized by 1/sigr
        pyrpn(i) =sigr(i)*sqrt(pyrp(i))   !unscaled noise-only part of residual covariance
        dof =dof +pyrp(i);
      enddo;
      write (6,'(/"cost function degrees-of-freedom =",g13.5)') dof

      if (nUnadj > 0) then;
        !** compute cross term (a3) used in x1,x2 error covariance and in meas
        !** residual covariance:  a3 =px1*a2 =px1*transpose(hf1)*hf2*pa^(1/2)
        a3(:nest,:nUnadj) =matmul(px(:nest,:nest),a2(:nest,:nUnadj));

        !** compute x1 error covariance: px11 =px1 +px1*a2*transpose(px1*a2)
        px(:nest,:nest) =px(:nest,:nest)                                 &
     &           +matmul(a3(:nest,:nUnadj),transpose(a3(:nest,:nUnadj)));
        write (6,'(/"total a posteriori state 1-sigma and correlations", &
     &     " due to measurement noise and unadjusted parameters:"/       &
     &     4x,"name          1-sigma    correlations")');
        do i=1,nest;
          sigx(i) =sqrt(px(i,i));
          write (6,1200) i,xstate(mapx(i))%xname,sigx(i),                &
     &                  (px(i,j)/(sigx(i)*sigx(j)),j=1,i);
        enddo;

        n1 =nest+1
        !** px12 =-px1*a2*pa^(1/2) =-px1*transpose(hf1)*hf2*pa
        px(:nest,n1:nEU) =-a3(:nest,:nUnadj);
        forall (i=n1:nEU) px(:nest,i) =px(:nest,i)*pa(i);
        px(n1:nEU,:nest) =transpose(px(:nest,n1:nEU));

        !** px22 =pa
        px(n1:nEU,n1:nEU) =0.d0;
        forall (i=n1:nEU) px(i,i) =pa(i)**2;

        !** compute 1-sigma perturbation of a posteriori measurement residual
        !**    (yt-Hf1*xest) due x2 states:
        !** dyResdx2 =hf2*pa^(1/2) -hf1*a3
        !**          =[I -hf1*px1*transpose(hf1)] *hf2*pa^(1/2)
        forall (i=n1:nEU) dyResdx2(:nmeas,i-nest) =hf(:nmeas,i)*pa(i);
        dyResdx2(:nmeas,:nUnadj) =dyResdx2(:nmeas,:nUnadj)               &
     &                     -matmul(hf(:nmeas,:nest),a3(:nest,:nUnadj));

        !** now compute total measurement residual covariance including both
        !** measurement noise, prior x1 estimate errors and effects of
        !** unadjusted parameters x2.  (Note: initial pyrp is noise-only covariance)
        forall (i=1:nmeas) pyrp(i) =pyrp(i) +sum(dyResdx2(i,:nUnadj)**2)

        write (6,'(/"1-sigma error in estimated states due to:"/         &
     &    "  name",6x,"meas/prior noise   total       ",100(a10,3x))')   &
     &     (xstate(mapx(j+nest))%xname,j=1,nUnadj);
        do i=1,nest;
          write (6,'(i3,1x,a10,100g13.5)') i,xstate(mapx(i))%xname,      &
     &           sigxn(i),sqrt(px(i,i)),a3(i,1:nUnadj);
        enddo;

        !***** compute 1-sigma perturbation of predicted meas (yp) due x2 states   ###
        !**  dyResdx22 =hp2*pa^(1/2) -hp1*a3
!        forall (i=n+1:nEU) dyResdx2(m+1:,i-n) =hp(:m,i)*pa(i)
!        dyResdx2(m+1:,:nUnadj) =dyResdx2(m+1:,:nUnadj) -matmul(hp(:m,:n),a3(:n,:nUnadj))
      endif;

      !*** Compute a priori measurement residual covariance (including x2 states)
      !*** used in computing a priori state error for ln(like):
      !**  pya =hf*pa1*hf^T +hp*pa2*hp^T +r

      do i=1,nmeas
        pyra(i) =sigr(i)*sqrt(1.d0 +sum(htt(i,:nEU)**2));
      enddo

      !** Print a priori measurement residual, a priori residual 1-sigma,
      !** and a posteriori residual 1-sigma

      write(6,'(/"measurement a-posteriori residual, a-priori "          &
     &   "residual 1-sigma, and a-posteriori residual 1-sigma"/5x,       &
     &   "time      mtype  residual   a-priori 1-sig  noise-post 1-sig", &
     &   "   a-post 1-sig")')
      do i=1,nmeas
        write (6,1000) ytime(i),mtype(i),yres(i),                        &
     &         pyra(i),pyrpn(i),sigr(i)*sqrt(pyrp(i))
      enddo
 1000 format (g15.7,i5,4g15.7)

      !** Write plot file containing a priori measurement residual,
      !** a posteriori residual 1-sigma, and effect of 1-sigma error in
      !** unadjusted parameters on residuals

      open (10,file=pltFile,status='unknown')
      write (10,1100) (trim(xnames(mapxu(i))),i=1,nUnadj)
 1100 format ('Title ="A-posteriori measurement residual, 1-sigma and ', &
     & 'effect of 1-sigma error in unadjusted parameters on residuals"'  &
     & /'variables ="Time", "mType", "yRes", "yResSigma"',               &
     &        20(', "ResDeriv-',a,'"'));
      do i=1,nmeas
        write (10,'(g13.5,i5,10g13.5)') ytime(i),mtype(i),yres(i),       &
     &              sigr(i)*sqrt(pyrp(i)),sigr(i)*dyresdx2(i,:nUnadj)
      enddo
      close (10)
!      call write_results (t0,a,pltFile)

      return;
      end subroutine LS_cov;
!*****************************************************
      subroutine read_setup (setupFile);
!***   reads text file of model and state configuration data.
!**    Additional keywords may be added for additional features

      use LSmodel, only:ntotal,xnames         !number of states in full model
      implicit none;
      character(*),intent(in) :: setupFile; !name of input setup file

      integer(4) i,ns,ier;
      character(15) keyword;
      character(80) aline;
      !_______________________________

      allocate (mapxe(ntotal),mapxu(ntotal),mapx(ntotal),               &
     &          mapxei(ntotal),mapxui(ntotal),mapxi(ntotal),stat=ier); !allocate mapping arrays
      if (ier /= 0) call errhand ('error allocating mapping arrays');

      open (10,file=setupFile,status='old',iostat=ier);
      if (ier /= 0) call errhand ('error opening setup file');

      do;
        read (10,'(a12,a80)',err=999) keyword,aline;
        if (keyword(1:1) == '#') cycle;     !comment

        mapxi(:) =0;
        mapxui(:) =0;
        mapxei(:) =0;

        select case(keyword);
          case ('TEpoch');
            read (aline,*) t0;
          case ('TEnd');
            read (aline,*) tEnd;
          case ('StateInit');
            ns =0;
            nest =0;          !total number of estimated states
            nUnadj =0;        !total number of unadjusted-analyze parameters
            nEU =0;
            write (6,                                                      &
     &      '("Listing of input state initialization from file ",a)')      &
     &         trim(setupFile);
            write (*,'(9x,a)') trim(aline);

            !** read info on states
            do;
              read (10,'(a)',err=999,end=998) aline;
              ns =ns+1;
              if (ns > ntotal) exit;
              write (*,'(a)') trim(aline);
              read (aline,*,err=999) i,xstate(ns)%status,                 &
     &             xstate(ns)%xtype, xstate(ns)%xname, xstate(ns)%x0,   &
     &             xstate(ns)%sig0, xstate(ns)%const(:2);
              if (xstate(ns)%xname /= xnames(ns)) then;
                call errhand                                            &
     &             ('read_setup error: states names do not match');
              endif;

              if (xstate(ns)%status == "ignore") cycle; !skip ignore states

              nEU =nEU+1;
              mapx(nEU) =ns;
              mapxi(ns) =nEU;

              if (xstate(ns)%status == "analyze") then;  !unadjusted-analyze state
                nUnadj =nUnadj +1;
                mapxu(nUnadj) =ns;
                mapxui(ns) =nUnadj;
              else if (xstate(ns)%status == "estimate") then;  !estimated state
                nest =nest +1;
                mapxe(nest) =ns;
                mapxei(ns) =nest;
              else
                write (6,'("invalid input state # ",i3," status ",a)')     &
     &                  ns,xstate(ns)%status
                call errhand("invalid input state status")
              endif;
            enddo;  !___ end read states loop

            exit;
        end select;

      enddo;

      !** set values and allocate arrays

  998 close (10);
      write (6,'(/"Finished reading setup file:"/                       &
     &  " # estimated states =",t34,i5/                                 &
     &  " # unadjusted analyze states =",t34,i5/                        &
     &  " # ignored states =",t34,i5/)') nest,nUnadj,ntotal-nEU;
      write (6,'("mapping of estimated states =",t34,20i5)')mapxe(:nest)
      write (6,'("mapping of unadj-analyze states =",t34,20i5)')        &
     &       mapxu(:nUnadj)
      write (6,'("mapping of all active states =",t34,20i5)') mapx(:nEU)
      write (6,'(/)')
      return;

  999 call errhand ('read_setup error: read setup file');
      end subroutine read_setup;

      end module LScovMod;

