      module LSmodel;
      !** module LSmodel contains the model information for the LS_cov covariance error analysis.
      !** It also computes the full set of measurements (y) and partial derivatives (ht).
      !** Note that xnames must also be set because input names in the setup file are verified.

      use filterPrec, only: real_kind1;   !precision (4 or 8)

      integer(4),parameter :: ntotal =6;                    !total number of states in full model
      integer(4)  :: mmax =101;                             !maximum number of measurements (row dimension of y,ytime,sigr,h)
                                                            !  must be >= actual number of meas
      character(20) :: xnames(ntotal) =                                  &
     &                  (/"c0", "c1", "c2", "c3", "c4", "c5"/);   !state names for the full model
      real(real_kind1) :: tepoch = 0.d0;                    !state epoch time
      real(real_kind1) :: xt(ntotal)                                     &
     &                    =(/ 10., 0.1, -0.004, 0.02, 0.0006, 1e-6 /);  !true state values at tepoch

      contains;
!*********************************************************
      subroutine get_yh (y,sigr,ht,  t0,xi,ytime,mtype,nmeas);
      !** computes the measurements and partial derivatives.
      !** For more general problems, get_yh could read measurements (ytime and y)
      !** from a file, and only compute ht and sigr.

      implicit none;
      real(real_kind1),intent(in) :: t0;                 !epoch time of state
      real(real_kind1),intent(in) :: xi(ntotal);         !state vector at t0
      real(real_kind1),intent(in) :: ytime(nmeas);       !measurement times
      integer(4),intent(in) :: mtype(mmax);              !measurement types
      integer(4),intent(in) :: nmeas                     !number of measurements

      real(real_kind1),intent(out) :: y(nmeas);          !computed measurement vector
      real(real_kind1),intent(out) :: ht(nmeas,ntotal);  !partial derivative of each measurement
                                                         ! with-respect-to the full model states
      real(real_kind1),intent(out) :: sigr(nmeas);       !1-sigma measurement noise

      integer(4) i;
      real(real_kind1) dt,t;
      !______________________________________________

!      a =rnor(ix)         !initialize random number generator

      !** generate predicted measurements from input xi and measurement model ht.
      !** Note: measurement model doesn't necessarily match true model

      do i=1,nmeas;
        t =ytime(i)
        ht(i,:) =(/ 1.d0, t, t**2, t**3, t**4, t**5 /);
        sigr(i) =2.d0;
        y(i) =dot_product(ht(i,:),xi);    !for noiseless measurements
      enddo;

      return;
      end subroutine get_yh;
!*********************************************************
      subroutine get_yMeas (nmeas,y,ytime,mtype);
      !** computes the "true" measurements.  For more general problems,
      !** get_yTrue will read measurements (ytime, y and mtype) from a file.

      implicit none;
      real(real_kind1),intent(out) :: y(mmax);           !measurement vector
      real(real_kind1),intent(out) :: ytime(mmax);       !measurement times
      integer(4),intent(out) :: mtype(mmax);             !measurement types
      integer(4),intent(out) :: nmeas                    !number of measurements

      integer(4) i;
      integer(4) :: iseed =19191919;      !seed of random number generator (set to
                                          ! any desired value, or read from input)
      real(real_kind1) dt,t,delt,a,ht(ntotal);

!      real(8),external :: rnor
      !______________________________________________

!      a =rnor(ix)         !initialize random number generator
!      t0 =tepoch;
!      t =t0;
      dt =0.1d0;           !time step between measurements
      nmeas =mmax;         !could be less

      do i=1,nmeas;
        t =(i-1)*dt;
        delt =t-tepoch
        ytime(i) =delt;
        mtype(i) =1;
        ht(:) =(/ 1.d0, delt, delt**2, delt**3, delt**4, delt**5 /);
        y(i) =dot_product(ht(:),xt);    !for noiseless measurements
!        y(i) =y(i) +sigr(i)*rnor(0);     !can add measurement noise using rnor function
      enddo;

      return;
      end subroutine get_yMeas;

      end module;

