!** Program montePoly_constr tests the numerical accuracy of different methods
!** for constrained least squares estimation.  The Monte Carlo simulation uses
!** an 11th-order polynomial model.  The RMS errors are listed in Example 7.1-1
!** for the following constrained least squares methods:

!**  1. normal equations with constraints as measurements,
!**  2. unconstrained MGS QR using Lagrange multipliers,
!**  3. MGS QR with constraints as measurements,
!**  4. LAPACK GSVD (generalized SVD)
!**  5. LAPACK GSVD with explicit coding of constraints.

!** Note: montePoly_constr.f uses libraries libkf_bnd_ifort.lib (rnoruni, sinv, symult, prtmx) and lapack.

!**  Author B. Gibbs, 12/2009

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib \library\lapack-3.1.1\lapack_win32.lib \library\lapack-3.1.1\blas_win32.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort monte_poly_constr.f90    (above libraries should be linked by default)
!**   monte_poly_constr >constr.lis
!**************************************

      implicit none;
      integer(4),parameter :: nd =11, mmax=101, ncr=2;
      integer(4),parameter :: nsamp =1000;
      integer(4) i,j,k,n,m,kk,ier,ix,nn2,isamp;
      real(8) au(nd*(nd+1)/2),r(nd*(nd+3)/2);
      real(8) x(nd),xa(nd),hf(mmax,nd);
      real(8) y(mmax),bb(nd),xe(nd),pa(nd),pai(nd),z(nd);
      real(8) dt,eps,errm,det,sumstat(2,nd,5),kscale,rsos;
      real(8) c(ncr,nd),d(ncr),e(ncr,nd),f(ncr,ncr),f1(ncr,ncr);
      real(8) ef(nd,ncr),ff(nd);     !constraints
      real(8),external :: rnor;
      !____________________________________________________

      !** initialize random number generator
      ix =191919;
      rsos =rnor(ix);

      !** set constraints
      c(:,:) =0.d0;
!      c(1,8) =1.d0
!      c(1,8:9) =(/ 1.d0, -1.d0 /)
      c(1,6:8) =(/ 1.d0, 0.d0, -1.d0 /);
      c(2,10) =1.d0;
      d(:) =0.d0;    !will be reset below

      m =mmax;
      n =nd;
      nn2 =(n*(n+1))/2;
      pa(:) =1.d20;       !###
      pai(:) =1.d0;
      sumstat(:,:,:) =0.d0;

      !** measurement model
      dt =1.d0/(mmax-1);
      call get_haPoly (hf,au,mmax,nd,nd,dt,0.d0);      !polynomial span 0 to 1

      !** set prior (mean) state values
      xa(:) =0.d0;

!***    start sampling loop
      do isamp =1,nsamp;
        do i=1,n;
          x(i) =xa(i) +pai(i)*rnor(0);      !add random component for 1:nt states
        enddo;

        !** set constraint right-hand-side
        d(:) =matmul(c(:,:n),x(:n));

        !** create measurements
        y(:m) =matmul(hf(:m,:n),x(:n));
        do i=1,m;
!          y(i) =y(i) +rnor(0)       !add meas noise ? ###
        enddo;

        !************ method 1: normal equations with constraints as meas
        r(:) =0.d0;
        kscale =1.d5;       !need 1d-6 to work with c(1,6:8) =(1, 0, -1)
                            !kscale =1d-10 will work with c(1,9) =1
        bb(:n) =matmul(y(:m),hf(:m,:n)) +matmul(d(:),c(:,:n))*kscale;

        kk =0;
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            r(kk) =r(kk) +dot_product(hf(:m,j),hf(:m,k))                &
     &                   +dot_product(c(:ncr,j),c(:ncr,k))*kscale;
          enddo;
          bb(j) =bb(j) +xa(j)/pa(j)**2; !add prior
          r(kk) =r(kk) +1.d0/pa(j)**2;
        enddo;

        eps =1.d-3;
        call sinv (ier,errm,  r,  n,eps,.false.);
        if (ier < 0) then;
          write (6,'("*** sinv error ",i3)') ier;
          xe(:) =0.d0;
        else;
          if (isamp == 1) then;
            call prtsmx (r,n,"p1");
            call prtmx (bb,1,1,n,"bb");
          endif;
          call symult(xe,  r,bb,n);  !unconstrained
        endif;
        sumstat(1,:n,1) =sumstat(1,:n,1) +(xe(:n)-x(:n));
        sumstat(2,:n,1) =sumstat(2,:n,1) +(xe(:n)-x(:n))**2;

        !************ method 2: unconstrained srif with added processing
        call srif (xe,r,  hf,y(:mmax),pa,xa,c,d,mmax,n,0);
        z(:n) =r(nn2+1:nn2+n);
        do i=1,ncr;
          call rtinz_m (e(i,:n),ier,  r,n,c(i,:n));
          if (ier /= 0) stop 'rtinz error';
        enddo;
        f(:,:) =matmul(e,transpose(e));
        !** f1 =inverse(f)
        det =1.d0/(f(1,1)*f(2,2)-f(1,2)*f(2,1));
        f1(1,:) =(/ f(2,2), -f(1,2) /)*det;
        f1(2,:) =(/ -f(2,1), f(1,1) /)*det;
!        f1(1,2) =0.5d0*(f1(1,2)+f1(2,1))     !make symmetric (no difference)
!        f1(2,1) =f1(1,2)
        ef(:n,:) =transpose(matmul(f1,e(:,:n)));    !=e^T * f1
        ff(:n) =z(:n) +matmul(ef(:n,:),(d(:)-matmul(e(:,:n),z(:n))));
        call rinz_m (xe(:n),ier,  r,n,ff(:n));
        if (ier /= 0) stop 'rinz error';
        sumstat(1,:n,2) =sumstat(1,:n,2) +(xe(:n)-x(:n));
        sumstat(2,:n,2) =sumstat(2,:n,2) +(xe(:n)-x(:n))**2;

        !************ method 3: srif with constraints as meas
        call srif (xe,r,  hf,y(:mmax),pa,xa,c,d,mmax,n,ncr);
        sumstat(1,:n,3) =sumstat(1,:n,3) +(xe(:n)-x(:n));
        sumstat(2,:n,3) =sumstat(2,:n,3) +(xe(:n)-x(:n))**2;

        !************ method 4: LAPACK GSVD
        call LSGSVD (xe,  hf,y,pa,xa,c,d,mmax,n,ncr);
        sumstat(1,:n,4) =sumstat(1,:n,4) +(xe(:n)-x(:n));
        sumstat(2,:n,4) =sumstat(2,:n,4) +(xe(:n)-x(:n))**2;

        !************ method 5: LAPACK GSVD with explicit coding of constraints
        call LSGSVD2 (xe,  hf,y,pa,xa,c,d,mmax,n,ncr);
        sumstat(1,:n,5) =sumstat(1,:n,5) +(xe(:n)-x(:n));
        sumstat(2,:n,5) =sumstat(2,:n,5) +(xe(:n)-x(:n))**2;
      enddo;   !___ end isamp loop

      write (6,                                                         &
     &   '(/"Mean and RMS error for 5 constraint approaches:")');
      do i=1,5;
        rsos =sqrt(sum(sumstat(2,:n,i))/(nsamp*n));
        write (6,'(/i3,13g13.5)') i,sumstat(1,:n,i)/nsamp;
        write (6,'(3x,13g13.5)') sqrt(sumstat(2,:n,i)/nsamp),rsos;
      enddo;

      stop;
      end program;
!*******************************************************************
      subroutine get_haPoly (h,au,  mmax,nd,n,dt,t0);
      implicit none;

      integer(4),intent(in) :: mmax,nd,n;
      real(8),intent(in) :: dt,t0;
      real(8),intent(out) :: h(mmax,nd),au(nd*(nd+1)/2);

      integer(4) it,j,k,kk;
      real(8) dtn,t,hnorm(mmax);
      !_________________________________________

      au(:) =0.d0;
      h(:,:) =0.d0;

      do it=1,mmax;
        t =(it-1)*dt +t0;
        dtn =1.d0;
        kk =0;
        do j=1,n;
          h(it,j) =dtn;
          dtn =dtn*t;
          do k=1,j;
            kk =kk+1;
            au(kk) =au(kk) +h(it,j)*h(it,k);
          enddo;
        enddo;
        hnorm(it) =sqrt(sum(h(it,1:n)**2));
      enddo;    !end time loop

      write (*,'("n =",i3,", hnorm =",10(/10g13.5))') n,hnorm(1:mmax);
      return;
      end subroutine get_haPoly;
!********************************************
      subroutine srif (xe,r,  hf,y,pa,xa,c,d,mmax,n,ncr);
      !*** solve ls using SRIF (QR)
      implicit none;
      integer(4),intent(in) :: mmax,n,ncr;
      real(8),intent(in) :: y(mmax),hf(mmax,n),pa(n),xa(n);
      real(8),intent(in) :: c(ncr,n),d(ncr);
      real(8),intent(out) :: xe(n),r(n*(n+3)/2);

      integer(4) i,m,nstrt,nn2,ier;
      real(8) rsos,h1(mmax+n+ncr,n+1),y1(mmax+n+ncr),kscale;
      !______________________________________________

      kscale =1.d20;    !do not use kscale =1.d15: all precision is lost, but 1d10 works as well as 1d20
      nn2 =(n*(n+1))/2;
      r(:) =0.d0;
      h1(:,:) =0.d0;
      h1(:mmax,:n) =hf(:,:);     !already normalized to unit meas noise
      h1(:mmax,n+1) =y(:mmax);
      do i=1,n;
        h1(mmax+i,i) =1.d0/pa(i);
        h1(mmax+i,n+1) =xa(i)/pa(i);
      enddo;
      if (ncr > 0) then;
        do i=1,ncr;
          h1(mmax+n+i,:n) =c(i,:n)*kscale;
          h1(mmax+n+i,n+1) =d(i)*kscale;
        enddo;
      endif;

      nstrt =1;
      rsos =0.d0;    !computes RHS
      m =mmax+n+ncr;
      call thhc_m (r,h1,rsos,  n,mmax+n+ncr,m,nstrt,.false.); !was thhc (r,n,h1,mmax+n+ncr,m,rsos,nstrt)

      y1(:n) =r(nn2+1:nn2+n);
      call rinz_m (xe,ier,  r,n,y1);
      if (ier /= 0) stop "rinz error";

      return;
      end subroutine srif;
! *******************************************************
      subroutine LSGSVD (xe,  hf,y,pa,xa,c,d,mmax,n,ncr);
      !*** solve ls using GSVD
      implicit none;
      integer(4),intent(in) :: mmax,n,ncr;
      real(8),intent(in) :: y(mmax),hf(mmax,n),pa(n),xa(n);
      real(8),intent(in) :: c(ncr,n),d(ncr);
      real(8),intent(out) :: xe(n);

      integer(4) i,nn2,ier;
      integer(4),parameter :: lwork =1000;
      real(8) h1(mmax+n,n),y1(mmax+n),work(lwork);
      real(8) c1(ncr,n),d1(ncr);
      !______________________________________________

      nn2 =(n*(n+1))/2;
      h1(:,:) =0.d0;
      h1(:mmax,:n) =hf(:,:);
      y1(:mmax) =y(:mmax);
      do i=1,n;
        h1(mmax+i,i) =1.d0/pa(i);
        y1(mmax+i) =xa(i)/pa(i);
      enddo;
      c1(:,:) =c(:,:);
      d1(:) =d(:);
      xe(:) =0.d0;

      !** LAPACK SVD - SDD is divide and conquer
      call DGGLSE (mmax+n,n,ncr, h1,mmax+n, c1,ncr, y1,d1,xe,           &
     &                  work, lwork, ier );
      if (ier /= 0) stop 'dgglse error';
      !DGGLSE ( M, N, P, A, LDA, B, LDB, C, D, X, WORK, LWORK, INFO )

      return;
      end subroutine LSGSVD;
!*********************************************************
      subroutine LSGSVD2 (xe,  hf,y,pa,xa,c,d,mmax,n,ncr);
      !*** solve ls using GSVD with explicit
      implicit none;
      integer(4),intent(in) :: mmax,n,ncr;
      real(8),intent(in) :: y(mmax),hf(mmax,n),pa(n),xa(n);
      real(8),intent(in) :: c(ncr,n),d(ncr);
      real(8),intent(out) :: xe(n);

      integer(4),parameter :: lwork =1000;
      integer(4) i,j,k,l,m,kk,nn2,ier,iwork(n);
      real(8) h1(mmax+n,n),y1(mmax+n),work(lwork);
      real(8) c1(ncr,n),d1(ncr),alpha(mmax),beta(mmax);
      real(8) u(mmax+n,mmax+n),v(ncr,ncr),q(n,n),z(n),w(n);
      real(8) t1(mmax),t2(ncr),r(n*(n+1)/2);
      !______________________________________________

      nn2 =(n*(n+1))/2;
      h1(:,:) =0.d0;
      h1(:mmax,:n) =hf(:,:);
      y1(:mmax) =y(:mmax);
      do i=1,n;
        h1(mmax+i,i) =1.d0/pa(i);
        y1(mmax+i) =xa(i)/pa(i);
      enddo;
      c1(:,:) =c(:,:);
      d1(:) =d(:);
      xe(:) =0.d0;

      !** LAPACK GSVD
      m =mmax+n;
      call DGGSVD( "U", "V", "Q", M, N, ncr, K, L, h1, m, c1,           &
     &                   ncr, ALPHA, BETA, U, m, V, ncr, Q, n, WORK,    &
     &                   IWORK, ier );
      if (ier /= 0) stop 'dggsvd error';

!      write (6,'("k,l,n,mmax,ncr =",5i5)') k,l,n,mmax,ncr
      if (k /= n-ncr) stop 'k /= n-ncr';
      if (l /= ncr) stop 'l /= ncr';

      t1(:) =matmul(y1(:m),u(:,:));    !=U^T * y1
      t2(:) =matmul(d(:ncr),v(:,:));   !=V^T * d
      z(:k) =t1(:k);
      z(k+1:n) =t2(:ncr)/beta(k+1:n);

      kk =0;
      do i=1,n;
        do j=1,i;
          kk =kk+1;
          r(kk) =h1(j,i);
        enddo;
      enddo;
      call rinz_m (w,ier,  r,n,z);
      xe(:) =matmul(q,w);

      return;
      end subroutine LSGSVD2;

