!** Program nonlinearBall simulates range tracking of a falling ball with
!** unknown initial location to compare the convergence of Gauss-Newton
!** and Newton iterations for nonlinear least squares solutions.
!** It generates plot files "problem.plt" (problem geometry) and "cost.plt"
!** (cost versus x and y) for use in Example 7.3-2.  The output report file is called "ball.lis"

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort nonlinearBall.f90    (libkf_bnd_ifort.lib should be linked by default)
!**   nonlinearBall >ball.lis
!**************************************

      implicit none;
      integer(4),parameter :: nd =3, mmax=61;
      integer(4) i,j,kk,n,m,ier,ix,ixx,iyy,iter,npltx,nplty;
      real(8) au(nd*(nd+1)/2),au0(nd*(nd+1)/2),r(nd*(nd+1)/2);
      real(8) bu(nd),bu0(nd),grad(nd);
      real(8) xt(nd),xe(nd),xp(nd),xs(nd),dx(nd),pa(nd),delx(nd);
      real(8) dyr(mmax),xo(2),dcdy1(nd),dcdy2(nd),hes(nd*(nd+1)/2);
      real(8) dt,eps,cnb,errm,lnLike,cost,cost0,cost1,cost2,cost3,cost4;
      real(8) rn,delxx;
      real(8),external :: rnor;
      !____________________________________________________

      m =mmax;
      n =nd;
      ix =191919;
      cnb =rnor(ix);

!      pa(:) =1.d10        !###
!      pa(:) =(/ 10.d0, 10.d0, 1.d0 /)
      pa(:) =(/ 10.d0, 10.d0, 0.0001d0 /);
      rn =4.d0;
      dt =0.05;    !sec (4.7 sec total)
      !** true state, prior and observer position
      xt(:) =(/ 30.d0, 110.d0, 0.d0 /);
      xp(:) =(/ 20.d0, 100.d0, 0.d0 /);   !prior estimate
      xo(:) =0.d0;
      xe(:) =xt(:);
      do i=1,m;
        dyr(i) =rn*rnor(0);    !meas noise
      enddo;

      open (11,file='problem.plt',status='unknown');
      write (11,1200);
 1200 format('VARIABLES ="x (m)", "y (m)"')

      eps =1.d-3;
      do iter =1,5;
        call funct (cost0,au0,bu0,  xt,xe,xo,xp,pa,dyr,mmax,n,dt,rn);
        r(:) =au0(:);
        call sinv (ier,errm,  r,  n,eps,.true.);
        if (ier < 0) stop '"*** sinv error';
        call symult(dx,  r,bu0,n);
        xe(:) =xe(:) +dx(:);
        write (6,'("iter",i2,", c=",g16.8,", x=",2f8.3,                 &
     &         ", dx =",2g13.5,", convrg =",g13.5)')                    &
     &      iter,cost0,xe(:2),dx(:2), sqrt(dot_product(dx,bu0)/n);
      enddo;
      write (6,'("bu0 =",3g13.5,", xe =",3g13.5,", bu0*dx =",g13.5)')   &
     &       bu0(:),xe(:),dot_product(bu0,dx);

      !** compute hessian
      delxx =0.1;
      hes(:) =0.d0;
      kk =0;
      do i=1,n;
        xs(:) =xe(:);
        xs(i) =xe(i) +delxx;
        call funct (cost3,au,bu,  xt,xs,xo,xp,pa,dyr,mmax,n,dt,rn);
        do j=1,i;
          xs(j) =xs(j) +delxx;
          call funct (cost1,au,bu,  xt,xs,xo,xp,pa,dyr,mmax,n,dt,rn);
          xs(j) =xs(j) -delxx;
          call funct (cost2,au,bu,  xt,xs,xo,xp,pa,dyr,mmax,n,dt,rn);
          dcdy1(j) =(cost1-cost2)/(2.d0*delxx);
        enddo;
        xs(:) =xe(:);
        xs(i) =xe(i) -delxx;
        call funct (cost4,au,bu,  xt,xs,xo,xp,pa,dyr,mmax,n,dt,rn);
        grad(i) =(cost3-cost4)/(2.d0*delxx);
        do j=1,i;
          xs(j) =xs(j) +delxx;
          call funct (cost1,au,bu,  xt,xs,xo,xp,pa,dyr,mmax,n,dt,rn);
          xs(j) =xs(j) -delxx;
          call funct (cost2,au,bu,  xt,xs,xo,xp,pa,dyr,mmax,n,dt,rn);
          dcdy2(j) =(cost1-cost2)/(2.d0*delxx);

          kk =kk+1;
          hes(kk) =hes(kk) +(dcdy1(j) -dcdy2(j))/(2.d0*delxx);
        enddo;
      enddo;
      write (6,'("grad =",3g13.5)') grad(:);

      npltx =60;
      nplty =40;
      open (10,file='cost.plt',status='unknown');
      write (10,1100) 2*nplty+1, 2*npltx+1;
 1100 format('VARIABLES ="x-xt", "y-yt", "true-cost", "fisher-cost", ', &
     &      '"Hessian-cost"'/                                           &
     &      'ZONE T="ball-cost", I=',i3,', J=',i3,', F=POINT ')

      xs(:) =xe(:);

!***    compute cost funct on grid and compare with fisher and hessian models
      do ixx =-npltx,npltx;
        xs(1) =xe(1) +ixx*0.25d0;
        do iyy =-nplty,nplty;
          xs(2) =xe(2) +iyy*0.25d0;
!          xs(3) =0.d0
          !** evaluate cost at this
          call funct (cost,au,bu,  xt,xs,xo,xp,pa,dyr,mmax,n,dt,rn);
          delx(:) =xs(:)-xe(:);
          call symult(dx,  au0,delx,n);
!          cost1 =cost0 +0.5d0*dot_product(delx,dx)
          cost1 =cost0 -dot_product(bu0,delx)+0.5d0*dot_product(delx,dx);!fisher model
          call symult(dx,  hes,delx,n);
          cost2 =cost0 -dot_product(bu0,delx)+0.5d0*dot_product(delx,dx);!hessian model
          write (10,'(5g14.6)') xs(:2)-xt(:2),cost,cost1,cost2;

          !** sum log likelihood
!          lnLike =0 !n*log(2.d0*pi) +2.d0*sum(log(pa(:n)))
!          lnLike =sum(yr(:m)**2) +sum(((xe(:n)-0.d0)/pa(:n))**2)
        enddo;   !___ end iyy loop
      enddo;   !___ end ixx loop

      stop;
      end program;
!*******************************************************************
      subroutine funct (cost,au,bu,  xt,xe,xo,xp,pa,dyr,mmax,n,dt,rn);
      implicit none;

      integer(4),intent(in) :: mmax,n;
      real(8),intent(in) :: dt,xt(n),xe(n),xo(2),xp(n),pa(n);
      real(8),intent(in) :: dyr(mmax),rn;
      real(8),intent(out) :: au(n*(n+1)/2),bu(n),cost;

      integer(4) it,j,k,kk,i,itp;
      real(8) t,hnorm(mmax),phi(n,n),xti(n),xei(n),dxt(2),dxe(2);
      real(8) y(mmax),ye(mmax),dydx(n),h(n),dy,xpl(2),d2ydx2(n,n);
      real(8) xtp(2,40);
      real(8) :: g =9.8d0;    !m/s^2
      logical(4),save :: first =.true.;
      !_________________________________________

      if (n /= 3) stop 'n /= 3';
      au(:) =0.d0;
      bu(:) =0.d0;
      cost =0.d0;
      !** add prior
      kk =0;
      do i=1,n;
        cost =cost +0.5d0*((xp(i)-xe(i))/pa(i))**2;
        bu(i) =bu(i) +(xp(i)-xe(i))/pa(i)**2;
        kk =kk+i;
        au(kk) =1.d0/pa(i)**2;
      enddo;

      xti(:) =xt(:);
      xei(:) =xe(:);
      itp =0;
      if (first) then;
        write (11,1000) 3*(mmax)/5+3;    !xx
 1000   format ('ZONE T="own_targ", I=',i4,', F=POINT')
        write (11,'(3g13.5)') xo(:);
      endif;

      do it=1,mmax;
        t =(it-1)*dt;
        xti(:) =xt(:) +(/ 0.d0, t*(xt(3) -0.5d0*t*g), -t*g /);     !true model
        xei(:) =xe(:) +(/ 0.d0, t*(xe(3) -0.5d0*t*g), -t*g /);     !estimate model
        phi(1,:) =(/ 1.d0, 0.d0, 0.d0 /);
        phi(2,:) =(/ 0.d0, 1.d0, t /);
        phi(3,:) =(/ 0.d0, 0.d0, 1.d0 /);

        dxt(:) =xti(:2) -xo(:);
        y(it) =sqrt(sum(dxt(:)**2)) +dyr(it);

        dxe(:) =xei(:2) -xo(:);
        ye(it) =sqrt(sum(dxe(:)**2));
        dydx(:) =(/ dxe(1)/ye(it), dxe(2)/ye(it), 0.d0 /);
        d2ydx2(1,:) =(/ dxe(2)**2, -dxe(1)*dxe(2), 0.d0 /)/ye(it)**3;
        d2ydx2(2,:) =(/ -dxe(1)*dxe(2), dxe(1)**2, 0.d0 /)/ye(it)**3;
        d2ydx2(3,:) =0.d0;
        h(:) =matmul(dydx,phi)/rn;
        dy =(y(it)-ye(it))/rn;
!        h2(:) =(/ dxe(2)**2, dxe(1)**2, 0.d0 /)/ye(it**3

        cost =cost +0.5d0*dy**2;
        bu(:) =bu(:) +h(:)*dy;
        kk =0;
        do j=1,n;
          do k=1,j;
            kk =kk+1;
            au(kk) =au(kk) +h(j)*h(k);
!     &             -(dy/rn**2)*d2ydx2(j,k)  !### Hessian
          enddo;
        enddo;

        hnorm(it) =sqrt(sum(h(1:n)**2));
        if (first) then;
          write (*,'(i3,5g13.5)') it,y(it),y(it)-ye(it),h(:);
          if (mod(it-1,5) == 0) then;
            if (it > 1) then;
              write (11,'(3g13.5)') xpl(:);
            endif;
            write (11,'(3g13.5)') xti(:2);
            write (11,'(3g13.5)') xo(:);
            xpl(:) =xti(:2);
            itp =itp+1;
            xtp(:,itp) =xti(:2);
          endif;
        endif;
      enddo;    !end time loop

      if (first) then;
        write (11,1010) itp;
 1010   format ('ZONE T="targ", I=',i4,', F=POINT')
        do it=1,itp;
          write (11,'(4g13.5)') xtp(:,it);
        enddo;
      endif;

      first =.false.;

      return;
      end subroutine funct;
