!** Program passTrk tests the convergence of four nonlinear least squares
!** solution techniques for passive ship tracking, and it computes a table
!** grid of the true least squares cost function, the Fisher information
!** (quadratic model) cost function, and the Hessian quadratic model of the
!** cost function versus the four epoch states of position and velocity.

!** The four user-selectable optimization algorithms are Levenberg-Marquard,
!** backtracking, backtracking followed by secant iterations, and the NL2SOL
!** algorithm.  PassTrk generates three output plot files for use in Example
!** 7.3-3: "passTrk.plt" is the "own ship" and "target ship" position versus
!** time, "costp.plt" is the true, Fisher and Hessian cost versus position given
!** the least squares estimate of velocity, and "covtv.plt" is true, Fisher and
!** Hessian cost versus velocity given the least squares estimate of position.
!** The plot files also contain, as a separate zone, the position and velocity
!** states for each step of the selected algorithm.

!**    Modification: 4/10/09: B. Gibbs bearing definition changed to atan(dx/dy) rather than atan(dy/dx) used in book
!**    Modification 12/20/09: B. Gibbs, made method an input

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort passTrk.f90    (libkf_bnd_ifort.lib should be linked by default for sinv, symult, prtmx, condest, rnoruni)
!**   passTrk 1 >ptrk_lm2.lis
!**   passTrk 2 >ptrk_back2.lis
!**   passTrk 3 >ptrk_backSec.lis
!**   passTrk 4 >ptrk_nl2sol.lis
!**************************************

      module modelDat;
      !** includes model info for funct
      integer(4),parameter :: nd =4;
      integer(4),parameter :: mmax =121;    !61
      real(8) xt(4),xo(4,2),tturn,h(mmax+nd,nd),res(mmax+nd);
      real(8) xp(nd),pa(nd),dyr(mmax),dt,rn;
      logical(4) optlPath,cartesian;
      logical(4) :: firstIter =.false.;     !normal mode linearizing all iter about estimate ###
!      logical(4) :: firstIter =.true.      !test linearizing partials about meas for 1st iter ###
      logical(4) :: XoverY =.false.;     !bearing measurement is atan(dy/dx) as used in book ###
!      logical(4) :: XoverY =.true.       !bearing measurement is atan(dx/dy)
      end module modelDat;
!****************************************************
      program passTrk;
      !** tests nonlinear least squares solution of passive ship tracking
      use modelDat;
      use optlDat, only:fstat,srif;
      implicit none;
      integer(4) i,j,kk,n,m,mn,ier,ix,ixx,iyy,npltx,nplty;
      integer(4) nmeas,niter,lprnt_o;
      integer(4) :: method =2;  !1=LM, 2=backtrack, 3=backtrack with "secant", 4=NL2SOL  !###
      logical(4) partial,debugp;
      integer(4) lenferr,nargs;
      character(10) line;
      real(8) au(nd*(nd+1)/2),au0(nd*(nd+1)/2),r(nd*(nd+3)/2);
      real(8) pf(nd*(nd+1)/2),bu(nd),bu0(nd),dxn(nd),sigx(nd);
      real(8) xe(nd),xs(nd),delx(nd),grad(nd),xmin(nd),xmax(nd);
      real(8) dcdy1(nd),dcdy2(nd),hes(nd*(nd+1)/2),delxx,convtst;
      real(8) eps,cnb,cost,cost0,cost1,cost2,cost3,cost4;
      real(8) lambda0,maxLambda,h1(mmax+nd,nd+1),rsos;
!      real(8) r1((mmax+nd)*(mmax+nd+3)/2),rsos
      real(8),external :: rnor,condest;
      !** for nl2sol
      integer(4) iv(60+nd+20), uiparm(1);
      real(8) v(93+mmax*nd+3*mmax+nd*(3*nd+33)/2+2000), urparm(1);
      external calcr, calcj, ufparm;
      !____________________________________________________

      lenferr =nargs()-1;        !get number of command line arguments

! ....  Get run control file name
      if (lenferr == 0) then;
        write (6,'(" Input run optimization method:"/                   &
     &           "1=LM, 2=backtrack, 3=backtrk +secant, 4=NL2SOL:")');
        read (5,'(a40)') method;
      else;
        call getarg(1,line,ier);   !get command line argument
        read (line,*) method;
        if (ier < 0) stop 'ERROR IN RETRIEVING method';
      endif;

      m =mmax;
      n =nd;
      ix =191919;
      cnb =rnor(ix);

      rn =1.5d0/57.296;      !1.5 deg 1-sigma
      dt =1.d0/240.d0;       !0.25 min samples, 15 min total
      cartesian =.true.;     !###

      !** true state, prior and observer position
      xt(:) =(/ -2.d0, 20.d0, 17.d0, -8.d0 /);
      xo(:,1) =(/ 0.d0, 0.d0, 0.d0, 20.d0 /);
      xo(:,2) =(/ 0.d0, 5.d0, -10.0d0, 17.32d0 /);
!      xo(:,2) =(/ 0.d0, 5.d0, -14.14d0, 14.14d0 /)   !### optional

      if (cartesian) then;
        xp(:) =(/ -3.d0, 30.d0, 0.d0, 0.d0 /);   !prior estimate
        pa(:) =(/ 15.d0, 15.d0, 20.d0, 20.d0 /); !prior 1-sigma
      else;
        !** range, az, vel, heading
        xp(:) =(/ 30.d0, 0.d0, 15.d0, 0.d0 /);   !prior estimate
        pa(:) =(/ 10.d0, 2.d0, 8.d0, 100.d0 /);    !prior 1-sigma
      endif;
!      xe(:) =xt(:)      !###  test initialized at truth
      xe(:) =xp(:);     !initialize estimate with prior
      tturn =0.25;      !own ship turn after 1/4 hour
      srif =.false.;

      do i=1,m;
        dyr(i) =rn*rnor(0);    !meas noise
      enddo;

      open (11,file='passTrk.plt',status='unknown');
      write (11,1200);
 1200 format('VARIABLES ="x (nmi)", "y (nmi)"')

      eps =1.d-3;
      partial =.true.;
      debugp =.false.;

      open (10,file='costp.plt',status='unknown');
      open (12,file='costv.plt',status='unknown');
      write (10,1100);
      write (12,1100);
 1100 format('VARIABLES ="x-xt", "y-yt", "xd-xtd", "yd-ytd", ',         &
     &      '"true-cost", "fisher-cost", "Hessian-cost"'/               &
     &      'ZONE T="optl-cost", F=POINT ')

      !*** test nl2sol interface routines caldr, calcdj ###
!      call checkN2lsol (xe,n)

      if (firstIter) then;
        !** take two GN iterations to initialize using linearizing partials
        optlPath =.true.;
        do i=1,2;
          call funct (cost,pf,au,bu,dxn,nmeas,                          &
     &                    partial,xe,n,debugp);
          xe(:n) =xe(:n) +dxn(:n);
        enddo;
      endif;

      firstIter =.false.;
      optlPath =.true.;
      niter =15;
      convtst =0.01;
      lambda0 =1.d-4;
      maxLambda =1000.;
      lprnt_o =1;
      xmin(:) =-1.d20;
      xmax(:) =1.d20;
      if (method == 1) then;
        call optl_LM (cost,nmeas,ier,  xe,pf,                           & !Levenberg-Marquard  ###
     &                       n,niter,convtst,lambda0,maxLambda,lprnt_o);
      else if (method == 2) then;
        call optl_back (cost,r,bu,nmeas,ier,  xe,                       & !backtracking
     &                      n,niter,xmin,xmax,convtst,lprnt_o);
      else if (method == 3) then;
        call optl_back_sec (cost,r,bu,nmeas,ier,  xe,                   & !backtracking followed by "secant" iterations
     &                      n,niter,convtst,lprnt_o);
      else if (method == 4) then;
!        call dfault(iv,v)
!        v(38) =1.d0     !initialize d to 1
!        iv(1) =12      !fresh start but default has already been called
        iv(1) =0;     !fresh start: use nl2sol defaults
!        call nl2sno (mmax+n,n,xe, calcr,iv,v,uiparm,urparm,ufparm)      !numerical jacobian
        !** disable nl2sol call to allow link without including nl2sol library
!        call nl2sol (mmax+n,n,xe, calcr,calcj,iv,v,uiparm,urparm,ufparm)  !calcj for analytic jacobian
        ier =0;
        if (iv(1) > 6) ier =iv(1);
      endif;

      if (ier /= 0) then;
        write (6,'("optimization ier =",i3)') ier;
        stop 'optl: ier /= 0';
      endif;

!      write (10,1400)
! 1400 format('ZONE T="pTrk-hess", F=POINT ')
      optlPath =.false.;
      call funct (cost0,pf,au0,bu0,dxn,nmeas,                           &
     &                  partial,xe,n,debugp);

!      do iter =1,5
!        call funct (cost0,pf,au0,bu0,dxn,nmeas,
!     &                  partial,xe,n,debugp)
!        xe(:) =xe(:) +dxn(:)
!        write (6,'("iter",i2,", c=",g16.8,", x=",4f8.3,
!     &         ", dxn =",4g13.5)') iter,cost0,xe(:4),dxn(:4)
!      enddo
      write (6,'("bu0 =",4g13.5/"xe =",4g13.5,", bu0*dxn =",g13.5)')    &
     &       bu0(:),xe(:),dot_product(bu0,dxn);
      write (6,'(/"sigma and correlations")');
      kk=0;
      do i=1,n;
        kk =kk+i;
        sigx(i) =sqrt(pf(kk));
        write (6,'(i3,10g13.5)')                                        &
     &         i,sigx(i),(pf(kk-i+j)/(sigx(i)*sigx(j)),j=1,i);
      enddo;

      !** use SRIF to get condition #
      r(:) =0.d0;
      mn =m+n;
      h1(:mn,:n) =h(:mn,:n);
      rsos =-1.d0;
      write (6,'("m,n,mmax,nd =",5i4)') m,n,mmax,nd,mn;
      call thhc_m (r,h1,rsos,  n,mmax+nd,mn,1,.false.);
!      call thhc (r,n,h1,mmax+nd,mn,rsos,1,n)
      write (6,'(/"condition number =",g13.5)') condest(r,n);

      !** compute hessian
      delxx =0.1;
      hes(:) =0.d0;
      kk =0;
      do i=1,n;
        xs(:) =xe(:);
        xs(i) =xe(i) +delxx;
        call funct (cost3,pf,au,bu,dxn,nmeas,                           &
     &                  partial,xs,n,debugp);
        do j=1,i;
          xs(j) =xs(j) +delxx;
          call funct (cost1,pf,au,bu,dxn,nmeas,                         &
     &                  partial,xs,n,debugp);
          xs(j) =xs(j) -delxx;
          call funct (cost2,pf,au,bu,dxn,nmeas,                         &
     &                  partial,xs,n,debugp);
          dcdy1(j) =(cost1-cost2)/(2.d0*delxx);
        enddo;
        xs(:) =xe(:);
        xs(i) =xe(i) -delxx;
        call funct (cost4,pf,au,bu,dxn,nmeas,                           &
     &                  partial,xs,n,debugp);
        grad(i) =(cost3-cost4)/(2.d0*delxx);
        do j=1,i;
          xs(j) =xs(j) +delxx;
          call funct (cost1,pf,au,bu,dxn,nmeas,                         &
     &                  partial,xs,n,debugp);
          xs(j) =xs(j) -delxx;
          call funct (cost2,pf,au,bu,dxn,nmeas,                         &
     &                  partial,xs,n,debugp);
          dcdy2(j) =(cost1-cost2)/(2.d0*delxx);

          kk =kk+1;
          hes(kk) =hes(kk) +(dcdy1(j) -dcdy2(j))/(2.d0*delxx);
        enddo;
      enddo;
      write (6,'("grad =",4g13.5)') grad(:);

      npltx =60;
      nplty =60;
      write (10,1300) 2*nplty+1, 2*npltx+1;
      write (12,1300) 2*nplty+1, 2*npltx+1;
 1300 format('ZONE T="pTrk-cost", I=',i3,', J=',i3,', F=POINT ')

      xs(:) =xe(:);

!***    compute cost funct on grid and compare with fisher and hessian models
      do ixx =-npltx,npltx;
        xs(1) =xe(1) +ixx*0.35d0;
        do iyy =-nplty,nplty;
          xs(2) =xe(2) +iyy*0.35d0;
          !** evaluate cost at this
          call funct (cost,pf,au,bu,dxn,nmeas,                          &
     &                  partial,xs,n,debugp);
          delx(:) =xs(:)-xe(:);
          call symult(dxn,  au0,delx,n);
!          cost1 =cost0 +0.5d0*dot_product(delx,dx)
          cost1=cost0 -dot_product(bu0,delx)+0.5d0*dot_product(delx,dxn);!fisher model
          call symult(dxn,  hes,delx,n);
          cost2=cost0 -dot_product(bu0,delx)+0.5d0*dot_product(delx,dxn);!hessian model
          write (10,'(7g14.6)') xs(:)-xt(:),cost,cost1,cost2;

          !** sum log likelihood
!          lnLike =0 !n*log(2.d0*pi) +2.d0*sum(log(pa(:n)))
!          lnLike =sum(yr(:m)**2) +sum(((xe(:n)-0.d0)/pa(:n))**2)
        enddo;   !___ end iyy loop
      enddo;   !___ end ixx loop

      xs(:) =xe(:);
!***    compute cost funct on grid and compare with fisher and hessian models
      do ixx =-npltx,npltx;
        xs(3) =xe(3) +ixx*0.40d0;   !###
        do iyy =-nplty,nplty;
          xs(4) =xe(4) +iyy*0.40d0; !###
          !** evaluate cost at this point
          call funct (cost,pf,au,bu,dxn,nmeas,                          &
     &                  partial,xs,n,debugp);
          delx(:) =xs(:)-xe(:);
          call symult(dxn,  au0,delx,n);
!          cost1 =cost0 +0.5d0*dot_product(delx,dx)
          cost1=cost0 -dot_product(bu0,delx)+0.5d0*dot_product(delx,dxn);!fisher model
          call symult(dxn,  hes,delx,n);
          cost2=cost0 -dot_product(bu0,delx)+0.5d0*dot_product(delx,dxn);!hessian model
          write (12,'(7g14.6)') xs(:)-xt(:),cost,cost1,cost2;

          !** sum log likelihood
!          lnLike =0 !n*log(2.d0*pi) +2.d0*sum(log(pa(:n)))
!          lnLike =sum(yr(:m)**2) +sum(((xe(:n)-0.d0)/pa(:n))**2)
        enddo;   !___ end iyy loop
      enddo;   !___ end ixx loop

      stop;
      end program;
!*******************************************************************
      subroutine funct (cost,pf,au,bu,dxn,nmeas,                        &
     &                  partial,xest,nest,debugp); !(cost,au,bu,  xe,n)
      use modelDat, only:xt,xo,tturn,xp,pa,dyr,mmax,dt,rn,optlPath,     &
     &              firstIter,h,res,XoverY;
      use optlDat,only:fstat;
      implicit none;

      integer(4),intent(in) :: nest;               !number of estimated parameters
      logical(4),intent(in) :: partial;            !if true, partial derivatives of cost (i.e.,
                                                   ! au, bu, pf) are evaluated. Otherwise only cost and last bu.
      logical(4),intent(in) :: debugp;             !activates debug print
      real(8),intent(in) :: xest(nest);            !input state vector
      real(8),intent(out) :: cost;                 !least squares cost function
      real(8),intent(out) :: pf(nest*(nest+1)/2);  !computed covariance matrix
      real(8),intent(out) :: au(nest*(nest+1)/2);  !information matrix
      real(8),intent(out) :: bu(nest),dxn(nest);   !cost gradient (= negative information vector)
      integer(4),intent(out) :: nmeas;             !number of measurements processed

      integer(4) it,j,k,kk,i,ier,itp,n;
      real(8) t,hnorm(mmax),phi(nest,nest),xti(nest),xei(nest);
      real(8) dxt(2),dxe(2),eps,errm,xop(2,40),xtp(2,40);
      real(8) y(mmax),ye(mmax),dydx(nest),dy,xpl(2),xoi(4),r2;    !d2ydx2(n)
      logical(4),save :: first =.true.;
      !_________________________________________

      if (nest /= 4) stop 'nest /= 4';
      if (partial) au(:) =0.d0;
      bu(:) =0.d0;
      h(:,:) =0.d0;
      cost =0.d0;
      n =nest;

      !** add prior
      kk =0;
      do i=1,nest;
        cost =cost +0.5d0*((xp(i)-xest(i))/pa(i))**2;
        bu(i) =bu(i) +(xp(i)-xest(i))/pa(i)**2;
        res(i) =(xp(i)-xest(i))/pa(i);
        h(i,i) =1.d0/pa(i);
        if (partial) then;
          kk =kk+i;
          au(kk) =1.d0/pa(i)**2;
        endif;
      enddo;

      xti(:) =xt(:);
      xei(:) =xest(:);
      xoi(:) =xo(:,1);
      if (first) then;
        write (11,1000) 3*(mmax)/10+4;    !
 1000   format ('ZONE T="own_targ", I=',i4,', F=POINT')
        write (11,'(4g13.5)') xoi(:2);
      endif;
      itp =0;

      do it=1,mmax;
        t =(it-1)*dt;
        xti(:) =xt(:) +(/ t*xt(3), t*xt(4), 0.d0,  0.d0 /);     !true model
        xei(:) =xest(:) +(/ t*xest(3), t*xest(4), 0.d0, 0.d0 /);     !estimate model
        if (t <= tturn) then;
          xoi(3:4) =xo(3:4,1);
        else;
          xoi(3:4) =xo(3:4,2);
        endif;
        xoi(1:2) =xoi(1:2) +dt*xoi(3:4);
        phi(1,:) =(/ 1.d0, 0.d0, t, 0.d0 /);
        phi(2,:) =(/ 0.d0, 1.d0, 0.d0, t /);
        phi(3,:) =(/ 0.d0, 0.d0, 1.d0, 0.d0 /);
        phi(4,:) =(/ 0.d0, 0.d0, 0.d0, 1.d0 /);

        dxt(:) =xti(:2) -xoi(:2);
        dxe(:) =xei(:2) -xoi(:2);
        r2 =sum(dxe(:2)**2);

        if (XoverY) then;
          !** bearing measurement is atan(dx/dy)
          y(it) =atan2(dxt(1),dxt(2)) +dyr(it);!bearing         !sqrt(sum(dxt(:)**2)) +dyr(it)  ###
          ye(it) =atan2(dxe(1),dxe(2));
          if (firstIter) then;
            !### alternate using bearing: 2nd iteration is almost as good as 4th iter above, but then never converges
            dydx(:) =(/ cos(y(it)), -sin(y(it)), 0.d0, 0.d0 /)/sqrt(r2);
          else;
            dydx(:) =(/ dxe(2), -dxe(1), 0.d0, 0.d0 /)/r2;
          endif;
        else;
          !** bearing measurement is atan(dy/dx)
          y(it) =atan2(dxt(2),dxt(1)) +dyr(it);!bearing
          ye(it) =atan2(dxe(2),dxe(1));
          if (firstIter) then;
            !### alternate using bearing: 2nd iteration is almost as good as 4th iter above, but then never converges
            dydx(:) =(/ -sin(y(it)), cos(y(it)), 0.d0, 0.d0 /)/sqrt(r2);
          else;
            dydx(:) =(/ -dxe(2), dxe(1), 0.d0, 0.d0 /)/r2;
          endif;
        endif;

        dy =(y(it)-ye(it))/rn;
        res(it+n) =dy;
        cost =cost +0.5d0*dy**2;

        if (partial) then;
          h(it+n,:nest) =matmul(dydx,phi)/rn;
          bu(:) =bu(:) +h(it+n,:nest)*dy;
          kk =0;
          do j=1,nest;
            do k=1,j;
              kk =kk+1;
              au(kk) =au(kk) +h(it+n,j)*h(it+n,k);
            enddo;
          enddo;

          hnorm(it) =sqrt(sum(h(it+n,1:nest)**2));
          if (first) then;
            write (*,'("meas ",i3,", y =",g13.5,", yres =",g13.5,       &
     &          ", h ="4g13.5)') it,y(it),y(it)-ye(it),h(it+n,:nest);
            if (mod(it-1,10) == 0) then;
              if (it > 1) then;
                write (11,'(4g13.5)') xpl(:2);
              endif;
              write (11,'(3g13.5)') xti(:2);
              write (11,'(3g13.5)') xoi(:2);
              xpl(:) =xti(:2);
              itp =itp+1;
              xop(:,itp) =xoi(:2);
              xtp(:,itp) =xti(:2);
            endif;
          endif;
        else;
          bu(:) =bu(:) +h(it+n,:nest)*dy;    !use last GN iteration value of h
        endif;
      enddo;    !end time loop

      if (first) then;
        write (11,'(3g13.5)') xtp(:,itp);
        write (11,1010) itp;
 1010   format ('ZONE T="ownp", I=',i4,', F=POINT')
        do it=1,itp;
          write (11,'(4g13.5)') xop(:,it);
        enddo;
        write (11,1020) itp;
 1020   format ('ZONE T="target", I=',i4,', F=POINT')
        do it=1,itp;
          write (11,'(4g13.5)') xtp(:,it);
        enddo;
      endif;

      first =.false.;
      nmeas =mmax;

      if (partial) then;
        !** compute GN iteration
        pf(:) =au(:);
        eps =1.d-4;
        call sinv (ier,errm,  pf,  nest,eps,.false.);
        if (ier < 0) stop '"*** sinv error';
        call symult(dxn,  pf,bu,nest);
        fstat%testc =sqrt(dot_product(bu,dxn)/nest);
        if (optlPath) then;
          write (10,'(7g14.6)') xest(:)-xt(:),cost, fstat%testc, 0.;
          write (12,'(7g14.6)') xest(:)-xt(:),cost, fstat%testc, 0.;
        endif;
      else;
        fstat%testc =0.;
!        if (optlPath)     !normally not printed for path plots ###
!     &    write (10,'(7g14.6)') xest(:)-xt(:),cost, fstat%testc, 0.
      endif;
      bu(:) =-bu(:);    !make gradient

      fstat%rrms =sqrt(2.d0*cost/mmax);
      fstat%cost =cost;
      fstat%nmeas =nmeas;

      return;
      end subroutine funct;
!********************************************************
      subroutine calcr ( m, n, xe, nf, r, uiparm, urparm, ufparm );
      use modelDat, only:res;
      implicit none;

      integer(4),intent(in) :: m;         !# measurements in res
      integer(4),intent(in) :: n;         !# states in xe
      integer(4),intent(inout) :: uiparm; !user input/output parm (not used)
      integer(4),intent(inout) :: nf;     !on input, total # of calls to calcr.  set to 0 if funct fails
      real(8),intent(in) :: xe(n);        !state vector
      real(8),intent(inout) :: urparm;    !user input/output parm (not used)
      real(8),intent(out) :: r(m);        !measurement residuals
      external ufparm;                    !user function (not used)

      integer(4) i,nmeas;
      real(8) cost,pf(n*(n+1)/2),au(n*(n+1)/2),bu(n),dxn(n);
      logical(4) partial,debugp;
      !_____________________________________________

!      partial =.true.
      partial =.false.;!no print on 1st iter
      debugp =.false.;
      call funct (cost,pf,au,bu,dxn,nmeas,                              &
     &                  partial,xe,n,debugp);
      r(:m) =res(:m);
      if (cost > 1.d10) nf =0;      !indicates that funct failed

      return;
      end subroutine calcr;
!********************************************************
      subroutine calcj (m,n,xe,nf,hj,uiparm,urparm,ufparm);
      use modelDat, only:h,mmax;
      implicit none;

      integer(4),intent(in) :: m;         !# measurements in res
      integer(4),intent(in) :: n;         !# states in xe
      integer(4),intent(inout) :: uiparm; !user input/output parm (not used)
      integer(4),intent(inout) :: nf;     !on input, total # of calls to calcr.  set to 0 if funct fails
      real(8),intent(in) :: xe(n);        !state vector
      real(8),intent(inout) :: urparm;    !user input/output parm (not used)
      real(8),intent(out) :: hj(m,n);     !measurement residual jacobian wrt states
      external ufparm;                    !user function (not used)

      integer(4) i,nmeas;
      real(8) cost,pf(n*(n+1)/2),au(n*(n+1)/2),bu(n),dxn(n);
      logical(4) partial,debugp;
      !_____________________________________________

      partial =.true.;
      debugp =.false.;
      call funct (cost,pf,au,bu,dxn,nmeas,                              &
     &                  partial,xe,n,debugp);
      hj(:m,:n) =-h(:m,:n);
      if (cost > 1.d10) nf =0;      !indicates that funct failed
      if (m /= mmax+n) stop 'calcj m /= mmax+n';

      return;
      end subroutine calcj;
!******************************************************
      subroutine checkN2lsol (xe,n);
      use modelDat, only:res,h,mmax;
      implicit none;

      integer(4),intent(in) :: n;         !# states in xe
      real(8),intent(in) :: xe(n);        !state vector

      integer(4) i,j,k,it,ier,nmeas,m,uiparm,nf;
      real(8) cost,pf(n*(n+1)/2),au(n*(n+1)/2),bu(n),dxn(n);
      real(8) au1(n*(n+1)/2),bu1(n),dxn1(n),x1(n);
      real(8) hj(mmax+n,n),r(mmax+n),urparm,eps,errm,erms,delx;
      real(8) r1(mmax+n),r2(mmax+n),hj1(mmax+n,n);
      logical(4) partial,debugp;
      external ufparm;
      !_____________________________________________

      m =mmax+n;
      nf =1;
      call calcr (m,n,xe,nf,r, uiparm,urparm,ufparm);
      call calcj (m,n,xe,nf,hj, uiparm,urparm,ufparm);

      !** check jacobian
      x1(:) =xe(:);
      delx =1.d-5;
      do i=1,n;
        x1(i) =xe(i)+delx;
        call calcr (m,n,x1,nf,r1, uiparm,urparm,ufparm);
        x1(i) =xe(i)-delx;
        call calcr (m,n,x1,nf,r2, uiparm,urparm,ufparm);
        x1(i) =xe(i);
        hj1(:m,i) =(r1(:)-r2(:))/(2.d0*delx);
        erms =sqrt(sum((hj(:m,i)-hj1(:m,i))**2)/m);
        write (6,'("state ",i2,", rms h error =",g13.5)') i,erms;
        write (6,'("hj1(",i2,") =",15(/10g13.5))') i,hj1(:m,i);
        write (6,'("hj1-hj(",i2,") =",15(/10g13.5))')                   &
     &            i,hj1(:m,i)-hj(:m,i);
      enddo;

      bu(:n) =-matmul(r(:m),hj(:m,:n));
      au(:) =0.d0;
      do it=1,m;
        k =0;
        do i=1,n;
          do j=1,i;
            k =k+1;
            au(k) =au(k) +hj(it,i)*hj(it,j);
          enddo;
        enddo;
      enddo;

      eps =1.d-4;
      call sinv (ier,errm,  au,  n,eps,.true.);
      if (ier /= 0) stop 'sinv error';
      call symult (dxn,  au,bu,n);
!      x1(:) =xe(:) +dxn(:)

      partial =.true.;
      debugp =.false.;
      call funct (cost,pf,au1,bu1,dxn1,nmeas,                           &
     &                  partial,xe,n,debugp);
      if (nmeas /= m-n) stop 'nmeas /= m-n';
      write (6,'("dxn =",10g13.5)') dxn(:);
      write (6,'("dxn1 =",10g13.5)') dxn1(:);
      write (6,'("dxn-dxn1 =",10g13.5)') dxn(:)-dxn1(:); !checks

      x1(:) =xe(:) +dxn1(:);
      partial =.false.;
      call funct (cost,pf,au1,bu1,dxn1,nmeas,                           &
     &                  partial,x1,n,debugp);
      write (6,'("cost at xe+dxn1 =",g13.5)') cost;

      stop;
      end subroutine checkN2lsol;
!*****************************************************
      subroutine ufparm;
      !** dummy
      end subroutine ufparm;
!*******************************************************
      subroutine optl_back_sec (cost,r,g,nmeas,ier,  xe,                &
     &                      n,niter,convtst,lprnt_o);
      !** computes optimum least squares solution using Gauss-Newton steps
      !** with backtracking followed by "secant" iterations
      use optlDat, only:fstat,srif;
      implicit none;

      integer(4),intent(in) :: n,lprnt_o,niter;
      real(8),intent(in) :: convtst;
      real(8),intent(inout) :: xe(n);
      integer(4),intent(out) :: ier,nmeas;
      real(8),intent(out) :: cost,r(n*(n+3)/2),g(n);

      !***  local variables
      integer(4) npnp1,jback,nback,npnp2,nhalf,iter,niters;
      logical(4) partial;
      real(8) sigx(n),xl(n),xnewt(n);
      real(8) delx(n),scalenewton;
      real(8) pf(n*(n+1)/2),rtmp(n*(n+3)/2),gtmp(n);
      real(8) cost0,costl,cost1,cost2,step,step1,step2,testc,djds;
      !________________________________________________________

      costl =1.d20;
      ier =-1;
      npnp2 =(n*(n+3))/2;
      npnp1 =(n*(n+1))/2;
      jback =0;

!****   iteration loop
      do iter=1,niter;
        fstat%iter =iter;
        nmeas =0;
        nback =0;
        nhalf =0;
        xl(:) =xe(:);      !save last state used to compute grad, fis

        write (6,1010) iter;
 1010   format(//1x,6('**********')/                                    &
     &   ' ITERATION ',i3/' CALLING FUNCTION WITH PARTIALS')

        !**  evaluate cost function and (optionally) srif R-matrix)
        partial=.true.;
        call funct (cost,pf,r,g,delx,nmeas,  partial,xe,n,.true.);
        cost0 =fstat%cost;
        if (iter == 1) costl =fstat%cost;

        !**  compute fraction of newton step based on residual rms
        scalenewton =20.;    !3.0
!        step=min(scalenewton/sqrt(fstat%rrms),1.d0)  !use scalenewton=3
        step=1.d0;
        write (6,1400) iter,fstat%testc,fstat%cost;
        write (0,1400) iter,fstat%testc,fstat%cost;
!        write (1,1400) iter,fstat%testc,fstat%cost
 1400   format(/' AT G-N ITERATION ',i3,', CONVERG. TEST =',g15.5,      &
     &        ', COST =',g12.5)
        write (6,'(" NEWTON STEP FACTOR =",F8.3/)') step;
        testc =fstat%testc;

        !**  compute newton point and limited newton step and
        xnewt(:) =xl(:)+delx(:); !1:n
        delx(:) =step*delx(:);
        xe(:) =xe(:)+delx(:);

        !_____________________
   10   continue;              !start of back-tracking logic loop
        !***  back-tracking logic

        !**  compute dot product of gradient and step
        !**  (xe-xl may not equal delx because of constraints)
        if (nback == 0 .and. nhalf == 0)                                &
     &      djds =dot_product(xe(:)-xl(:),g(:));

        if (lprnt_o > 1) call prtmx (xe,1,1,n,'xe    ');

        !** compute cost function at new point
        !** (funct must not set delx or g !!!!!!!!!)
        partial =.false.;
        call funct (cost,pf,rtmp,gtmp,delx,nmeas,                       &
     &               partial,xe,n,.true.);

        write (6,1420) iter,nback,step,fstat%cost,fstat%cost;
        write (0,1410) iter,nback,step,fstat%cost;
 1410   format(' AT G-N ITERATION ',i3,', BACK-TRACKING #',i2,          &
     &        ' STEP=',f8.5,', COST =',g12.5)
 1420   format(/' AT G-N ITERATION ',i3,', BACK-TRACKING #',i2,         &
     &        ' STEP=',f8.5,', COST =',g12.5)
        write (6,'("x =",3(10g13.5))') xe(:n);

        !**  test for reduction in cost function
        if (fstat%rrms > 1.d15 .or. fstat%cost > cost0) then;

          if (fstat%rrms > 1.d15)then;
            !** if get invalid model, cut interval in half
            step =0.5d0*step;
            nhalf =nhalf+1;

            if (nhalf > 9) then;
              ier =2;
              write (0,1440);
!              write (1,1440)
              write (6,1440);
 1440         format(' EXIT OPTIMIZATION: MORE THAN 9 STEP-HALVINGS ',  &
     &                  'DUE TO INVALID MODEL')
              go to 995;
            endif;

          else if (fstat%cost > cost0) then;
            !** cost function is worse
            !** accept step if converged and cost is only slightly high
            if (testc < convtst .and. fstat%cost < 1.01*cost0) then;
              ier =0;
              write (6,                                                 &
     &         '("converged: accept step even though slightly large")');
              go to 995;
            endif;

            !**  if cost function is worse, perform line minimization
            !**  assuming quadratic or cubic with gradient at xl known.
            !**  (cost= cost0+ djdx*dx+ dj2dx2*dx**2 where dx=1 for full step
            cost1 =cost;
            step1 =step;
            call backtrack (step1,step2,cost2,nback,                    & !            call stephalf (step1,step2,cost2,nback,
     &                      cost0,cost1,djds,lprnt_o);
            step=step1;
          endif;

          write (6,'(/" BACK-TRACKING NUMBER ",I3)') nback;

          if (nback == 2) then;
            jback =jback+1;
          else if (step < 0.01 .and. jback >= 3) then;
            write (6,'(//" !!!!!!!!! OPTIMIZATION IS STUCK: "/           &
     &        " ACCEPT STEP AND ALLOW INCREASE IN COST FUNCTION")');
            jback =0;
            go to 20;
          endif;

          if (nback > 5) then;
            ier =2;
            write (0,1430);
!            write (1,1430)
            write (6,1430);
 1430       format(' EXIT OPTIMIZATION: MORE THAN 5 BACK-TRACKS')
            go to 995;
          endif;

          xe(:) =xl(:) +step*delx(:);   !1:n
          go to 10;
        endif;

   20   costl =fstat%cost;
        if (nback == 0) jback =0;

        if (testc < convtst) then;
          ier =0;
          write (6,'("converged")');
          go to 995;
        endif;

        niters =5;
!        write (6,'("secant xe =",3(/10g13.5))') xe(:)
        call secant_iter (cost,nmeas,ier,  xe,                          &
     &                     pf,n,niters,convtst,lprnt_o);

      enddo;!_______end of iteration loop

      ier =1;

  995 continue;   !end of iteration loop
      fstat%testc =testc;

      return;
      end subroutine optl_back_sec;
!**********************************************************************
      subroutine secant_iter (cost,nmeas,ier,  xe,                      &
     &                        pf,n,niter,convtst,lprnt_o);
      !** computes optimum least squares solution using "secant" iterations

      use optlDat, only:fstat,srif;
      implicit none;

      integer(4),intent(in) :: n,niter,lprnt_o;
      real(8),intent(in) :: convtst,pf(n*(n+1)/2);
      real(8),intent(inout) :: xe(n);
      integer(4),intent(out) :: ier,nmeas;
      real(8),intent(out) :: cost;

      !***  local variables
      integer(4) iter;
      logical(4) partial;
      real(8) delx(n),pf1(n*(n+1)/2),rtmp(n*(n+3)/2),gtmp(n),testc;
      real(8) costl,xel(n);
      !________________________________________________________

      partial =.false.;
      ier =0;
      write (6,'(/"start of secant iterations")');
      write (6,'("Initial xe =",3(/10g13.5))') xe(:);
      costl =1.d10;
      xel(:) =xe(:);

      do iter=1,niter;
        call funct (cost,pf1,rtmp,gtmp,delx,nmeas,                      &
     &              partial,xe,n,.true.);
        call symult(delx,  pf,gtmp,n);
        testc =dot_product(gtmp,delx);
        if (lprnt_o) write (6,1410) iter,cost,testc,xe;
 1410   format(' SECANT ITERATION ',i3,', COST =',g12.5,                &
     &     ', TESTC =',g13.5,', xe =',3(/10g13.5))

        if (cost > costl) then;
          xe(:) =xel(:);
          cost =costl;
          exit;
        endif;
        xel(:) =xe(:);
        xe(:) =xe(:) -delx(:);
        costl =cost;
      enddo;

      write (6,'(/"Secant final cost =",g13.5,", xe =",3(/10g13.5))')   &
     &           cost,xe(:);

      return;
      end subroutine secant_iter;
