!** Program Tst_KF2cont simulates measurements for a 2-state CONTINUOUS system,
!** and computes the a posteriori variance and filter estimates for a 2-state
!** CONTINUOUS Kalman filter.  Process noise drives only the second state of the
!** continuous system, and three different levels (0.1, 1.0, 10) of process noise
!** PSD are tested.  State #1 is the integral of the second, and only this state
!** is measured.  The measurement noise PSD is 100.  The output plot file is
!** "kf2cont.plt".  This is used for example 8.3-1.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tst_KF2cont.f90 odeint.f         (Note: odeint.f is a Numerical Recipes adaptive ODE integration routine,
                                             !  but the version called by tst_KF2cont has slightly different arguments)
!**   tst_KF2cont
!**************************************

      module dercom;
      integer(4),parameter :: nplt=3, m=100;
      integer(4),parameter :: ninteg =5;
      integer(4) icase,ncall;
      real(8) r,xt(2),ym(0:m);
      real(8) :: qi(nplt) = (/ 0.1, 1.0, 10. /);
      end module;
!**************************************
      use dercom;
      implicit none;
      integer(4) i,j,ix,nok,nbad,ier,ncall0;
      integer(4) :: model =2;   !1=pdot, 2=sdot  ###
      real(8) pdot(2,2),pf(2,2),xf(2),xi(ninteg);
      real(8) dt,g(2),dy,a(2,2),b(2),d(2);
      real(8) fout(4,0:m,nplt),yn(m),qn(m),tlst,t;
      real(8) epsint,delt1,hmin,xscale(ninteg);
      real(8),external :: rnor,derivs,derivsSq,derivst,bsstep;
      !_______________________________

      xscale(:) =10.d0;
      epsint =1.d-8;    !###
      delt1 =1.;
      hmin =1.d-4;

      dt =1.d0;
      r =100.d0;
      ix =19191919;
      xf(1) =rnor(ix);
      !** generate meas noise
      do i=1,m;
        yn(i) =sqrt(r)*rnor(0);
        qn(i) =rnor(0);
      enddo;

      do icase =1,nplt;
        pf(:,:) =0.d0;
        pf(1,1) =10.*r;
        pf(2,2) =10.*r;
        xf(:) =0.d0;
        xt(:) =0.d0;
        ym(0) =0.d0;
        fout(:2,0,icase) =xf(:)-xt(:);
        fout(3:,0,icase) =(/ sqrt(pf(1,1)), sqrt(pf(2,2)) /);
        if (model == 1) then;
          xi(1:3) =(/ pf(1,1), pf(1,2), pf(2,2) /);
        else;
          xi(1:3) =(/ sqrt(pf(1,1)), 0.d0, sqrt(pf(2,2)) /);
        endif;
        xi(4:5) =xf(:);
        ncall =0;
        ncall0 =0;
        do i=1,m;
          tlst =(i-1)*dt;
          t =i*dt;
          xt(1) =xt(1) +xt(2)*dt;
          ym(i) =xt(1) +yn(i);
          if (model == 1) then;
            call odeint (nok,nbad,ier,  xi,                             &
     &         ninteg,tlst,t,epsint,delt1,hmin,xscale, derivs,bsstep);
!     &         ninteg,tlst,t,epsint,delt1,hmin,xscale, derivst,bsstep);  !###
            pf(1,:) =(/ xi(1), xi(2) /);
            pf(2,:) =(/ xi(2), xi(3) /);
          else;   !use s = sqrt(p)
            call odeint (nok,nbad,ier,  xi,                             &
     &         ninteg,tlst,t,epsint,delt1,hmin,xscale, derivsSq,bsstep);
            pf(1,1) =xi(1)**2+xi(2)**2;
            pf(1,2) =xi(2)*xi(3);
            pf(2,1) =pf(1,2);
            pf(2,2) =xi(3)**2;
          endif;
          xf(:) =(/ xi(4), xi(5) /);
          fout(:2,i,icase) =xf(:)-xt(:);
          fout(3:,i,icase) =(/ sqrt(pf(1,1)), sqrt(pf(2,2)) /);
          write (*,'(f10.2,2g13.5,i5)') t,fout(3:,i,icase), ncall-ncall0;
          ncall0 =ncall;
        enddo;
      enddo;

!      write (6,'("steady-state Kalman gains =",3g13.5)')
!     &   fout(2,m,:)**2/(r+fout(2,m,:)**2)

      open (10,file='kf2cont.plt',status='unknown');
      write (10,1000);
1000  format('variables ="t", '                                         &
     &   '"xe1-qp1", "xe2-qp1", "sigx1-qp1", "sigx2-qp1", ',            &
     &   '"xe1-q1",  "xe2-q1",  "sigx1-q1",  "sigx2-q1", ',             &
     &   '"xe1-q10", "xe2-q10", "sigx1-q10", "sigx2-q10"')
      do i=0,m;
        write (10,'(13g13.5)') i*dt,(fout(1:4,i,j),j=1,nplt);
      enddo;

      stop;
      end;
!****************************************************
      subroutine derivs (xidot,  xi,t);
!**      computes matrix ricatti eqn and state derivative
      use dercom;
      implicit none;
      real(8),intent(in) :: xi(ninteg),t;
      real(8),intent(out) :: xidot(ninteg);

      integer(4) i;
      real(8) pf(2,2),x(2),hp(2),g(2),a(2,2),pdot(2,2),yr;
      real(8) f(2,2),xd(2);
      !___________________________________________________

!      if (ninteg /= 5) stop ' ninteg /= 5)

      pf(1,:) =(/ xi(1), xi(2) /);
      pf(2,:) =(/ xi(2), xi(3) /);
      x(:) =(/ xi(4), xi(5) /);

      f(:,:) =0.d0;
      f(1,2) =1.d0;
      a(:,:) =matmul(f,pf);
      hp(:) =pf(1,:);
      g(:) =hp(:)/r;
      forall (i=1:2) pdot(:,i) =a(:,i) +a(i,:) -r*g(:)*g(i);
      pdot(2,2) =pdot(2,2) +qi(icase);
      xidot(1:3) =(/pdot(1,1), pdot(1,2), pdot(2,2) /);
      yr =xt(1) -x(1);!+yn(i) (problem with noise
      xd(:) =matmul(f,x) -g(:)*yr;
      xd(:) =0.d0;   !### for testing
      xidot(4:5) =xd(:);
!      xt(2) =xt(2) +sqrt(qi(icase))*qn(i)

      ncall =ncall+1;

      return;
      end subroutine derivs;
!****************************************************
      subroutine derivssq (xidot,  xi,t);
!**      computes square root matrix ricatti eqn and state derivative
      use dercom;
      implicit none;
      real(8),intent(in) :: xi(ninteg),t;
      real(8),intent(out) :: xidot(ninteg);

      integer(4) i;
      real(8) pf(2,2),x(2),hp(2),g(2),a(2,2),pdot(2,2),yr;
      real(8) f(2,2),xd(2),s(3),si(2,2),mu(2,2),sdot(3);
      !___________________________________________________

!      if (ninteg /= 5) stop ' ninteg /= 5)

      s(1:3) =xi(1:3);
      si(1,1) =1.d0/s(1);
      si(2,2) =1.d0/s(3);
      si(1,2) =-s(2)*si(2,2)/s(1);
      si(2,1) =0.d0;

      pf(1,1) =s(1)**2+s(2)**2;
      pf(1,2) =s(2)*s(3);
      pf(2,1) =pf(1,2);
      pf(2,2) =s(3)**2;

      x(:) =(/ xi(4), xi(5) /);

      f(:,:) =0.d0;
      f(1,2) =1.d0;
      a(:,:) =matmul(f,pf);
      hp(:) =pf(1,:);
      g(:) =hp(:)/r;
      forall (i=1:2) pdot(:,i) =a(:,i) +a(i,:) -r*g(:)*g(i);
      pdot(2,2) =pdot(2,2) +qi(icase);

      mu(:,:) =matmul(matmul(si,pdot),transpose(si));
      mu(1,1) =0.5d0*mu(1,1);
      mu(2,2) =0.5d0*mu(2,2);
      sdot(1) =s(1)*mu(1,1);
      sdot(2) =s(1)*mu(1,2) +s(2)*mu(2,2);
      sdot(3) =s(3)*mu(2,2);

      xidot(1:3) =(/sdot(1), sdot(2), sdot(3) /);
      yr =xt(1) -x(1);!+yn(i) (problem with noise
      xd(:) =matmul(f,x) -g(:)*yr;
      xd(:) =0.d0;   !### for testing
      xidot(4:5) =xd(:);
!      xt(2) =xt(2) +sqrt(qi(icase))*qn(i)

      ncall =ncall+1;

      return;
      end subroutine derivssq;
!****************************************************
      subroutine derivsT (xidot,  xi,t);
!**      tests equations used for square root matrix ricatti eqn and state derivative
      use dercom;
      implicit none;
      real(8),intent(in) :: xi(ninteg),t;
      real(8),intent(out) :: xidot(ninteg);

      integer(4) i;
      real(8) pf(2,2),x(2),hp(2),g(2),a(2,2),pdot(2,2),yr;
      real(8) f(2,2),xd(2),s(3),si(2,2),mu(2,2),sdot(3);
      !___________________________________________________

!      if (ninteg /= 5) stop ' ninteg /= 5)

      pf(1,:) =(/ xi(1), xi(2) /);
      pf(2,:) =(/ xi(2), xi(3) /);
      x(:) =(/ xi(4), xi(5) /);

      s(3) =pf(2,2);
      if (s(3) >= 0.d0) then;
        s(3) =sqrt(s(3));
      else;
        write (*,'("pf(2,2) =",g13.5)') s(3);
        s(3) =1.d-30;
      endif;
      s(2) =pf(1,2)/s(3);
      s(1) =pf(1,1) -s(2)**2;
      if (s(1) >= 0.d0) then;
        s(1) =sqrt(s(1));
      else;
        write (*,'("pf(1,1) -s(2)**2 =",g13.5)') s(1);
        s(1) =1.d-30;
      endif;
      s(1) =max(s(1),0.d0);

      si(1,1) =1.d0/s(1);
      si(2,2) =1.d0/s(3);
      si(1,2) =-s(2)*si(2,2)/s(1);
      si(2,1) =0.d0;

      f(:,:) =0.d0;
      f(1,2) =1.d0;
      a(:,:) =matmul(f,pf);
      hp(:) =pf(1,:);
      g(:) =hp(:)/r;
      forall (i=1:2) pdot(:,i) =a(:,i) +a(i,:) -r*g(:)*g(i);
      pdot(2,2) =pdot(2,2) +qi(icase);

      mu(:,:) =matmul(matmul(si,pdot),transpose(si));
      mu(1,1) =0.5d0*mu(1,1);
      mu(2,2) =0.5d0*mu(2,2);
      sdot(1) =s(1)*mu(1,1);
      sdot(2) =s(1)*mu(1,2) +s(2)*mu(2,2);
      sdot(3) =s(3)*mu(2,2);

      !** re-compute pdot from sdot & s
      pdot(1,1) =2.d0*(s(1)*sdot(1) +s(2)*sdot(2));
      pdot(1,2) =s(2)*sdot(3) +s(3)*sdot(2);
      pdot(2,1) =pdot(1,2);
      pdot(2,2) =2.d0*(s(3)*sdot(3));
      xidot(1:3) =(/pdot(1,1), pdot(1,2), pdot(2,2) /);

      yr =xt(1) -x(1);!+yn(i) (problem with noise
      xd(:) =matmul(f,x) -g(:)*yr;
      xd(:) =0.d0;   !### for testing
      xidot(4:5) =xd(:);
!      xt(2) =xt(2) +sqrt(qi(icase))*qn(i)

      ncall =ncall+1;

      return;
      end subroutine derivsT;
