!** Program Tst_KF2corr simulates time-correlated (correlation = 0.5) measurements
!** for a 2-state system, and computes the a posteriori variance and state estimates
!** using different methods for handling time-correlated measurements in the Kalman
!** filter.  The four methods are the standard 2-state Kalman filter, a 3-state
!** Kalman filter with the added state modeling the correlation, a 3-state Kalman
!** filter where the added state is a delay-state, and a 2-state Bryson-Hendrikson
!** approach.  Process noise with PSD of 1.0 drives only the second state of the
!** continuous system.  State #1 is the integral of the second, and only this state
!** is measured.  The measurement noise PSD is 100.  The output plot file is
!** "kf2corr.plt".  This is used for example 8.2-1.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort tst_KF2cor.f90
!**   tst_KF2cor
!**************************************

      implicit none;
      integer(4) i,j,ix;
      integer(4),parameter :: nplt=4, m=250;
      real(8) xf(2),phi(2,2),xt(2);
      real(8) r,dt,tau,yd;
      real(8) fout(4,0:m,nplt),yn(m),qn(m),vi,ynl,x1,p1,y(0:m);
      real(8) t,yi;
      real(8) :: qi = 1.0;
      real(8) :: corr =0.5;
      real(8),external :: rnor;
      !_______________________________

      dt =0.2d0;
!      dt =1.d0
      r =100.d0*1.d0/dt;
      tau =10.d0;
      phi(1,:) =(/ 1.d0, dt /);
      phi(2,:) =(/ 0.d0, 1.d0 /);   !=0.81873
!      phi(2,:) =(/ 0.d0, exp(-dt/tau) /)    !=0.81873
      ix =19191919;
      xf(1) =rnor(ix);

      !** generate correlated meas noise
      ynl =0.d0;
      do i=1,m;
        vi =sqrt(r*(1.d0-corr**2))*rnor(0);
        yn(i) =corr*ynl +vi;
        ynl =yn(i);
        qn(i) =rnor(0);
      enddo;

      !** initialize filters
      t =0.d0;
      xt(:) =0.d0;
      y(0) =0.d0;
      yi =y(0);
      call kf (x1,p1,  yi,t,dt,r,corr,phi,qi);
      call kf_aug (x1,p1,  yi,t,dt,r,corr,phi,qi,0);
      call kf_delay (x1,p1,  yi,t,dt,r,corr,phi,qi,0);
      call kf_bry (x1,p1,  yi,t,dt,r,corr,phi,qi,0);

      fout(1,0,1:4) =0.d0;
      fout(2,0,1:4) =sqrt(10.*r*dt);
      do i=1,m;
        t =i*dt;
        xt =matmul(phi,xt);
        xt(2) =xt(2) +sqrt(qi)*qn(i);
        y(i) =xt(1)+yn(i);
        yd =y(i) -corr*y(i-1);

        call kf (x1,p1,  y(i),t,dt,r,corr,phi,qi);
        fout(1,i,1) =x1-xt(1);
        fout(2,i,1) =p1;

        call kf_aug (x1,p1,  y(i),t,dt,r,corr,phi,qi,i);
        fout(1,i,2) =x1-xt(1);
        fout(2,i,2) =p1;

        if (i == 1) then;
          call kf_delay (x1,p1,  y(i),t,dt,r,corr,phi,qi,i);
        else;
          call kf_delay (x1,p1,  yd,t,dt,r,corr,phi,qi,i);
        endif;
        fout(1,i,3) =x1-xt(1);
        fout(2,i,3) =p1;

        if (i == 1) then;
          call kf_bry (x1,p1,  y(i),t,dt,r,corr,phi,qi,i);
        else;
          call kf_bry (x1,p1,  yd,t,dt,r,corr,phi,qi,i);
        endif;
        fout(1,i,4) =x1-xt(1);
        fout(2,i,4) =p1;
      enddo;

!      write (6,'("steady-state Kalman gains =",3g13.5)')
!     &   fout(2,m,:)**2/(r+fout(2,m,:)**2)

      open (10,file='kf2corr.plt',status='unknown');
      write (10,1000);
1000  format('variables ="t", '                                         &
     &   '"xe1-kf", "sigx1-kf", "xe1-aug",  "sigx1-aug", ',             &
     &   '"xe1-delay", "sigx1-delay", "xe1-bry", "sigx1-bry"')
      do i=0,m;
        write (10,'(13g13.5)') i*dt,(fout(1:2,i,j),j=1,nplt);
      enddo;

      stop;
      end;
!**********************************************
      subroutine kf (x1,p1,  yi,t,dt,r,corr,phi2,qi);
      !** standard 2-state kalman filter
      implicit none;

      real(8),intent(in) :: yi,t,dt,r,corr,phi2(2,2),qi;
      real(8),intent(out) :: x1,p1;

!      integer(4) i
      integer(4),parameter :: n =2;
      real(8) pa(n,n),pf(n,n),xf(n),xa(n),phi(n,n);
      real(8) g(n),dy,a(n,n),hp(n),h(n);
      !__________________________________________

      if (t == 0.d0) then;
        pf(:,:) =0.d0;
        pf(1,1) =10.*r*dt;
        pf(2,2) =10.*r*dt;
        xf(:) =0.d0;
        return;
      endif;

      phi(:2,:2) =phi2(:,:);

      xa =matmul(phi,xf);
      !** p =phi*p*phi^T
      a(:,:) =matmul(phi,pf);
      pa(:,:) =matmul(a,transpose(phi));

      !** add Q
      pa(1,1) =pa(1,1) +qi*dt**3/3.d0;
      pa(1,2) =pa(1,2) +qi*dt**2/2.d0;
      pa(2,1) =pa(1,2);   !pa(2,1) =pa(2,1) +qi*dt**2/2.d0
      pa(2,2) =pa(2,2) +qi*dt;

      !** meas update
      h(:) =(/ 1.d0, 0.d0 /);
      dy =yi -dot_product(h,xa);
      hp(:) =matmul(h,pa);
      g(:) =hp(:)/(dot_product(h,hp) +r);
      xf(:) =xa(:) +g(:)*dy;
      pf(1,:) =pa(1,:) -g(1)*hp(:);
!      pf(2,:) =pa(2,:) -g(2)*hp(:)    !###
      pf(2,:) =(/ pf(1,2), pa(2,2)-g(2)*hp(2) /);

      x1 =xf(1);
      p1 =sqrt(pf(1,1));

      return;
      end subroutine kf;
!*******************************************************
      subroutine kf_aug (x1,p1,  yi,t,dt,r,corr,phi2,qi,imeas);
      !** 3-state kalman filter using added state for correlated meas noise
      implicit none;

      integer(4),intent(in) :: imeas;
      real(8),intent(in) :: yi,t,dt,r,corr,phi2(2,2),qi;
      real(8),intent(out) :: x1,p1;

      integer(4) i,j;
      integer(4),parameter :: n =3;
      real(8) pa(n,n),pf(n,n),xf(n),xa(n),phi(n,n);
      real(8) g(n),dy,a(n,n),hp(n),h(n);
      !__________________________________________

      if (t == 0.d0) then;
        pf(:,:) =0.d0;
        pf(1,1) =10.*r*dt;
        pf(2,2) =10.*r*dt;
        pf(3,3) =r;
        xf(:) =0.d0;
        return;
      endif;

      phi(:,:) =0.d0;
      phi(:2,:2) =phi2(:,:);
      phi(3,3) =corr;

      xa =matmul(phi,xf);
      !** p =phi*p*phi^T
      a(:,:) =matmul(phi,pf);
      pa(:,:) =matmul(a,transpose(phi));

      !** add Q
      pa(1,1) =pa(1,1) +qi*dt**3/3.d0;
      pa(1,2) =pa(1,2) +qi*dt**2/2.d0;
      pa(2,1) =pa(1,2);   !pa(2,1) =pa(2,1) +qi*dt**2/2.d0
      pa(2,2) =pa(2,2) +qi*dt;
      pa(3,3) =pa(3,3) +r*(1.d0-corr**2);

      !** meas update
      h(:) =(/ 1.d0, 0.d0, 1.d0 /);
      dy =yi -dot_product(h,xa);
      hp(:) =matmul(h,pa);
      g(:) =hp(:)/(dot_product(h,hp));      !*** no r !!!
      xf(:) =xa(:) +g(:)*dy;
      forall (i=1:n) pf(i,:) =pa(i,:) -g(i)*hp(:);

      !** make symmetric
      do i=2,n;
        do j=1,i-1;
          pf(i,j) =0.5d0*(pf(i,j)+pf(j,i));
          pf(j,i) =pf(i,j);
        enddo;
      enddo;

      x1 =xf(1);
      p1 =sqrt(pf(1,1));

      return;
      end subroutine kf_aug;
!*******************************************************
      subroutine kf_delay (x1,p1,  yi,t,dt,r,corr,phi2,qi,imeas);
      !** 3-state kalman filter using added delay state
      implicit none;

      integer(4),intent(in) :: imeas;
      real(8),intent(in) :: yi,t,dt,r,corr,phi2(2,2),qi;
      real(8),intent(out) :: x1,p1;

      integer(4) i,j;
      integer(4),parameter :: n =3;
      real(8) pa(n,n),pf(n,n),xf(n),xa(n),phi(n,n);
      real(8) g(n),dy,a(n,n),hp(n),h(n),corr1;
      !__________________________________________

      if (t == 0.d0) then;
        pf(:,:) =0.d0;
        pf(1,1) =10.*r*dt;
        pf(2,2) =10.*r*dt;
        pf(3,3) =0.;    !not used anyway
        xf(:) =0.d0;
        return;
      endif;

      phi(:,:) =0.d0;
      phi(:2,:2) =phi2(:,:);
      phi(3,1) =1.d0;

      xa =matmul(phi,xf);
      !** p =phi*p*phi^T
      a(:,:) =matmul(phi,pf);
      pa(:,:) =matmul(a,transpose(phi));

      !** add Q
      pa(1,1) =pa(1,1) +qi*dt**3/3.d0;
      pa(1,2) =pa(1,2) +qi*dt**2/2.d0;
      pa(2,1) =pa(1,2);   !pa(2,1) =pa(2,1) +qi*dt**2/2.d0
      pa(2,2) =pa(2,2) +qi*dt;

      !** meas update
      corr1 =corr;
      if (imeas == 1) corr1 =0.d0;
      h(:) =(/ 1.d0, 0.d0, -corr1 /);
      dy =yi -dot_product(h,xa);
      hp(:) =matmul(h,pa);
      g(:) =hp(:)/(dot_product(h,hp) +r*(1.d0-corr1**2));
      xf(:) =xa(:) +g(:)*dy;
      forall (i=1:n) pf(i,:) =pa(i,:) -g(i)*hp(:);

      !** make symmetric
      do i=2,n;
        do j=1,i-1;
          pf(i,j) =0.5d0*(pf(i,j)+pf(j,i));
          pf(j,i) =pf(i,j);
        enddo;
      enddo;

      x1 =xf(1);
      p1 =sqrt(pf(1,1));

      return;
      end subroutine kf_delay;
!*******************************************************
      subroutine kf_bry (x1,p1,  yi,t,dt,r,corr,phi2,qi,imeas);
      !** 2-state kalman filter using bryson-hendrikson method for handling correlated noise
      implicit none;

      integer(4),intent(in) :: imeas;
      real(8),intent(in) :: yi,t,dt,r,corr,phi2(2,2),qi;
      real(8),intent(out) :: x1,p1;

      integer(4) i;
      integer(4),parameter :: n =2;
      real(8) pa(n,n),pf(n,n),xf(n),xa(n),phi(n,n);
      real(8) g(n),dy,a(n,n),hp(n),h(n),q(n,n);
      real(8) rr,hr(n),xfl(n),pfl(n,n),s(n),b(n),c;
      !__________________________________________

      if (t == 0.d0) then;
        pf(:,:) =0.d0;
        pf(1,1) =10.*r*dt;
        pf(2,2) =10.*r*dt;
        xf(:) =0.d0;
        return;
      endif;

      phi(:,:) =0.d0;
      phi(:2,:2) =phi2(:,:);

      xa =matmul(phi,xf);
      !** p =phi*p*phi^T
      a(:,:) =matmul(phi,pf);
      pa(:,:) =matmul(a,transpose(phi));

      !** add Q
      q(:,:) =0.d0;
      q(1,1) =qi*dt**3/3.d0;
      q(1,2) =qi*dt**2/2.d0;
      q(2,1) =q(1,2);
      q(2,2) =qi*dt;
      pa(:,:) =pa(:,:) +q(:,:);

      !** meas update
      h(:) =(/ 1.d0, 0.d0 /);

      if (imeas == 1) then;
        dy =yi -dot_product(h,xa);
        hp(:) =matmul(h,pa);
        g(:) =hp(:)/(dot_product(h,hp) +r);
        xf(:) =xa(:) +g(:)*dy;
        forall (i=1:n) pf(i,:) =pa(i,:) -g(i)*hp(:);
      else;
        hr(:) =matmul(h,phi) -corr*h(:);     !h does not change
        dy =yi -dot_product(hr,xf);
        hp(:) =matmul(hr,pf);                !uses hr and pf (not pa)
        s(:) =matmul(h,q);
        rr =r*(1-corr**2) +dot_product(s,h);
        c =1.d0/(dot_product(hr,hp) +rr);
        g(:) =hp(:)*c;

        xfl(:) =xf(:) +g(:)*dy;
        xf(:) =matmul(phi,xfl) +s(:)*c*dy;

        forall (i=1:n) a(i,:) =-g(i)*hr(:);
        forall (i=1:n) a(i,i) =a(i,i) +1.d0;
        pfl(:,:) =matmul(matmul(a,pf),transpose(a));
        forall (i=1:n) pfl(i,:) =pfl(i,:) +rr*g(i)*g(:);

        b(:) =matmul(phi,g);
        pf(:,:) =matmul(matmul(phi,pfl),transpose(phi)) +q(:,:);
        forall (i=1:n) pf(i,:)=pf(i,:) +s(i)*c*s(:) -b(i)*s(:)-s(i)*b(:);
      endif;

      x1 =xf(1);
      p1 =sqrt(pf(1,1));

      return;
      end subroutine kf_bry;
