!** Program steady_stateKF6 computes the solution to steady-state continuous
!** and discrete Kalman filters for a 6-state position-velocity-acceleration
!** problem.  The different solution methods are: 1) executing a standard KF
!** for a "long" time, 2) using Schur decomposition of a Hamiltonian matrix,
!** or 3) using the matrix sign function to compute eigenvalues of a Hamiltonian
!** matrix (dricss).  The output file is "ssKF6.lis".  The results are used in Example 8.5-1.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib \library\lapack-3.1.1\lapack_win32.lib \library\lapack-3.1.1\blas_win32.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort steady_stateKF6.f90
!**   steady_stateKF6 >ssKF6.lis
!**************************************

      implicit none;
      integer(4) i,j,k,ier,it;
      integer(4),parameter :: nd =6, nd2 =2*nd, nd1 =nd+1, mmax=2;
      integer(4),parameter :: lwork=10000, liwork =10000;    !1 +6*nd2 +2*nd2**2
      integer(4) IWORK(liwork),ia(nd2);

      real(8) delta,t;
      real(8) h(mmax,nd),hp(mmax,nd),g(nd,mmax),c(mmax,mmax);
      real(8) rn(mmax,mmax);
      real(8) sigx0(nd),p(nd,nd),pa(nd,nd),phi(nd,nd),qs(nd),q(nd,nd);
      real(8) hrht(nd,nd),f(nd,nd),d1(nd,nd),d2(nd,nd),xf(nd),xa(nd);

      real(8) t3(nd,nd2),t4(nd,nd2);

      real(8) a(nd2,nd2),b(nd2,nd2),WORK(lwork),d(nd2);   !LAPACK
      real(8) e(nd2,nd2),t1(nd2,nd2),s(nd2,nd2);

      real(8) :: dt =1.d0, tmax =1000.;
      data sigx0 / 2*10., 2*5., 2*5.  /;
      data qs /4*0., 2*1e-6/;
      real(8) :: sigr =1.;
      !_______________________________________________

      f(:,:) =0.d0;
      forall (i=1:4) f(i,i+2) =1.d0;
      rn(:,:) =0.d0;
      forall (i=1:mmax) rn(i,i) =sigr**2;

      h(:,:) =0.d0;
      hrht(:,:) =0.d0;
      forall (i=1:mmax) h(i,i) =1.d0;
      forall (i=1:mmax) hrht(i,i) =1.d0/sigr**2;

      a(:,:) =0.d0;
!      a(:nd,:nd) =-hrht(:,:)
!      a(nd1:nd2,:nd) =f(:,:)
!      a(:nd,nd1:nd2) =transpose(f(:,:))
!      forall (i=1:nd) a(nd+i,nd+i) =qs(i)*dt
      a(:nd,:nd) =f(:,:);
      a(nd1:nd2,:nd) =-hrht(:,:);
      a(nd1:nd2,nd1:nd2) =-transpose(f(:,:));
      forall (i=1:nd) a(i,nd+i) =qs(i)*dt;

      call prtmx (a,nd2,nd2,nd2,"a");
      !** apply eigenvalue shift
!      do i=1,nd
!        a(i,i) =a(i,i) -1.d-5
!        a(i+nd,i+nd) =a(i+nd,i+nd) +1.d-5
!      enddo

      !** check eigenvalues of H^T*H
      b(:,:) =a(:,:);
      call DSYEVD( "V", "U", nd2, b, nd2, d, WORK, lwork,               &
     &           IWORK, liwork, ier );
      if (ier /= 0) then;
        write (6,'("DSYEVD info =",i4)') ier;
        stop 'DSYEVD failure';
      endif;
      call prtmx (d,1,1,nd2,"eigenval");
      call prtmx (b,nd2,nd2,nd2,"eigenvec");

      !** check
!      forall (i=1:nd2) e(:,i) =b(:,i)*d(i)
!      s(:,:) =matmul(e,transpose(b)) -a(:,:)
!      call prtmx (s,nd2,nd2,nd2,"MAMT-A")

      !** sort eigenvalues in abs value
      forall (i=1:nd2) ia(i) =i;
!  ....  bubble sort pointers so that d(ia(1)) is largest
      do i=1,nd2;
        do j=i+1,nd2;
           if (abs(d(ia(j))) > abs(d(ia(i))) ) then;
             k =ia(j);
             ia(j) =ia(i);
             ia(i) =k;
           endif;
        enddo;
        write (6,'("sorted d",i3,g13.5)') i,d(ia(i));
      enddo;

      do i=1,nd;
        d2(:,i) =b(:nd,nd1-i);    !use only positive values
        d1(:,i) =b(nd1:nd2,nd1-i);
!        d1(:,i) =b(:nd,i)     !use only negative values
!        d2(:,i) =b(nd1:nd2,i)
!        d1(:,i) =b(:nd,ia(i))
!        d2(:,i) =b(nd1:nd2,ia(i))
      enddo;

      !** P =-inv(M11)*M12
!      d1(:,:) =a(:nd,:nd)
!      d2(:,:) =a(:nd,nd1:nd2)
      t1(:,:) =0.d0;
      call matInvGJ (ier,  d1,t1,  nd,nd,0);
      if (ier == 0) then;
        p(:,:) =-matmul(d2,d1);
        call prtmx (p,nd,nd,nd,"p1");
      else;
        write (6,'(/"m11 inversion failed"/)');
      endif;

      write (6,'(/"Compare solutions of Discrete Algebraic Ricatti",    &
     &    " Equation (DARE)")');

      !** KF
      write (6,'(/"start of Discrete Kalman Filter iterations for s"/)');
      p(:,:) =0.d0;
      forall (i=1:nd) p(i,i) =sigx0(i)**2;
      xf(:) =0.d0;
      call get_phiq (xa,phi,q,   xf,f,dt,qs,nd,nd);

      do it=1,100000;
        t =dt*it;
        if (t > tmax) exit;
        pa(:,:) =matmul(matmul(phi,p),transpose(phi)) +q(:,:);
        hp(:,:) =matmul(h,pa);
        c(:,:) =matmul(hp,transpose(h));
        forall (i=1:mmax) c(i,i) =c(i,i) +sigr**2;
        t1(:,:) =0.d0;
        call matInvGJ (ier,  c,d2,  mmax,mmax,0);
        g(:,:) =matmul(transpose(hp),c);
        p(:,:) =pa(:,:) -matmul(g,hp);
        do i=2,nd;
          do j=1,i-1;
            p(j,i) =0.5*(p(j,i)+p(i,j));
            p(i,j) =p(j,i);
          enddo;
        enddo;
        if (mod(it,10) == 0) then;
!          call prtmx(p,nd,nd,nd,"pi")
          write (6,'("t =",g13.5,", sigp =",6(/10g13.5))')              &
     &           it*dt,(sqrt(p(j,j)),j=1,nd);
        endif;
      enddo;
      call prtmx(pa,nd,nd,nd,"pi-");
      write (6,'("t =",g13.5,", sigpa =",6(/10g13.5))')                 &
     &           it*dt,(sqrt(pa(j,j)),j=1,nd);
      call prtmx(p,nd,nd,nd,"pi+");

      !** check p
      t3(:,:) =0.d0;
      t3(:nd,:nd) =p(:,:);
      forall (i=1:nd) t3(i,i+nd) =1.d0;
      !** check using a
      t4(:,:) =matmul(t3,a);
      d1(:,:) =matmul(t4,transpose(t3));
      call prtmx (d1,nd,nd,nd,"Pi*A*Pit");
      !** check using eigen
      t4(:,:) =matmul(t3,b);
      forall (i=1:nd2) t3(:,i) =t4(:,i)*d(i);
      d1(:,:) =matmul(t3,transpose(t4));
      call prtmx (d1,nd,nd,nd,"PiM*L*MtPit");

      !** solve continuous and discrete algebraic riccati eqns using schur factorization
      write (6,'(/"Compute DARE and CARE solutions using Schur ",       &
     &             "factorization")');
      call schur (f,qs,phi,q,hrht,h,rn,dt,nd,nd2,mmax);

      write (6,'(/"Compute CARE solution using matrix sign function")');
!       solve the steady-state continuous algebraic ricatti equation (CARE)
!          Pdot =G*Q*G^T +F*P +P*F^T -P*H^T*inv(R)*H*P
!
!          EC = | -F              -G*Q*G^T |   (Hamiltonian matrix
!               | -H^T*inv(R)*H    F^T     |

      p(:,:) =0.d0;
      b(:,:) =0.d0;
      !** B&D
      b(:nd,:nd) =-f(:,:);
      b(nd1:nd2,:nd) =-hrht(:,:);
      b(nd1:nd2,nd1:nd2) =transpose(f(:,:));
      forall (i=1:nd) b(i,nd+i) =-qs(i)*dt;

      !** Iterate using matrix sign function
      call cricss (p,e,t1,s,  b,nd2,nd,nd,1);
      call prtmx (p,nd,nd,nd,"p-cricss");

      stop;
      end;
!*********************************************************
      subroutine get_phiq (xa,phi,q,   xf,f,dt,qs,n,ndim);
!***    computes state transition and process noise matrix and integrates state
      implicit none;
      integer(4),intent(in) :: n,ndim;     !state dimension
      real(8),intent(in) :: xf(n);     !initial state
      real(8),intent(in) :: dt;         !time step (s)
      real(8),intent(in) :: qs(n);      !state noise psd
      real(8),intent(in) :: f(ndim,n);      !dynamics matrix
      real(8),intent(out) :: xa(n);     !predicted state at t+dt
      real(8),intent(out) :: phi(ndim,n); !state transition matrix
      real(8),intent(out) :: q(ndim,n);   !state noise matrix

      integer(4) i;
      real(8) xdot(n),phii(n,n),dti,t;
      real(8) :: dtmax =0.2;     !little difference between 0.1 and 0.2
      logical(4) :: debugp =.false.;
      !_______________________________________


      t =0.d0;
      if (dt <= 0.) stop 'get_phiq error: dt <= 0';
      xa(:) =xf(:);
      phi(:,:) =0.d0;
      forall (i=1:n) phi(i,i) =1.d0;
      q(:,:) =0.d0;

      do while (t < dt);
        dti =min(dtmax,dt-t);
        t =t+dti;
        phii(:,:) =dti*f(:,:);
!        phii(:,:) =dti*(f(:,:) +0.5*dti*matmul(f,f))        !add 2nd-order term ###
        forall (i=1:n) phii(i,i) =phii(i,i) +1.d0;
        phi(:,:) =matmul(phii(:,:),phi(:,:));
!        call prtmx (phi,n,n,n,"phii")
        q(:,:) =matmul(matmul(phii,q),transpose(phii));
        forall (i=1:n) q(i,i) =q(i,i) +qs(i)*dti;
        xa(:) =xa(:) +dti*xdot(:);
      enddo;

      if (debugP) then;
        write (6,'("get_phix t =",g13.5)') t;
        call prtmx (phi,n,n,n,"phi");
        call prtmx (q,n,n,n,"q");
      endif;

      return;
      end subroutine get_phiq;
!**********************************************************
      subroutine schur (f,qs,phi,q,hrht,h,rn,dt,nd,nd2,mmax);
!** solve continuous and discrete algebraic riccati eqns using schur factorization
      implicit none;
      integer(4),intent(in) :: nd,nd2,mmax;
      real(8),intent(in) :: f(nd,nd),qs(nd),phi(nd,nd),dt;
      real(8),intent(in) :: q(nd,nd),hrht(nd,nd);
      real(8),intent(in) :: h(mmax,nd),rn(mmax,mmax);

      integer(4),parameter :: lwork=10000, liwork =10000;    !1 +6*nd2 +2*nd2**2

!**   dgeesx variables
!     .. Scalar Arguments ..
      CHARACTER(80)      JOBVS, SENSE, SORT;
      INTEGER(4)         info,sdim;
      INTEGER(4) LDA, LDVS;     !, LIWORK =nd*(nd2-nd)+1000
!      INTEGER(4),parameter ::  LWORK =max(1,3*Nd2)
      real(8) ::           RCONDE, RCONDV;
!     .. Array Arguments ..
      LOGICAL(4)         BWORK(nd2);
      INTEGER(4)         IWORK(liwork);
      real(8)            VS(nd2, nd2);    !,  WORK(lwork),  A(nd2, nd2 ),
      real(8)            WR(nd2), WI(nd2);

!*     .. Function Arguments ..
      LOGICAL(4),external ::  SELECT1,select2;

      integer(4) i,j,ier,nd1;
      real(8) p(nd,nd),hp(mmax,nd),c(mmax,mmax);
      real(8) d1(nd,nd),d2(nd,nd),d3(nd,nd),g(nd,mmax);
      real(8) b(nd2,nd2),t1(nd2,nd2),WORK(lwork);   !LAPACK a(nd2,nd2),
      !______________________________________________________

      LDA =nd2;
      LDVS=nd2;
      nd1 =nd+1;

      !** continuous schur factorization
      p(:,:) =0.d0;
      b(:,:) =0.d0;
      !** B&D
      b(:nd,:nd) =transpose(f(:,:));
      b(:nd,nd1:nd2) =-hrht(:,:);
      forall (i=1:nd) b(nd+i,i) =-qs(i)*dt;
      b(nd1:nd2,nd1:nd2) =-f(:,:);
      write (6,'(/"*** Schur CARE ***")');

      !** schur factorization using LAPACK routine
      jobvs ='V';
      sort ='S';
      sense ='E';
      call DGEESX ('V', sort, SELECT1, SENSE, Nd2, b, LDA, SDIM,        &
     &                   WR, WI, VS, LDVS, RCONDE, RCONDV, WORK, LWORK, &
     &                   IWORK, LIWORK, BWORK, INFO );
      if (info /= 0) then;
        write (6,'("DGEESX info =",i5)') info;
        stop 'gdeesx failure';
      endif;
      call prtmx (b,nd2,nd2,nd2,'schur form');
      call prtmx (vs,nd2,nd2,nd2,'vs');
      call prtmx (wr,1,1,nd2,'wr');
      call prtmx (wi,1,1,nd2,'wi');
      write (6,'("RCONDE, RCONDV =",2g13.5,", sdim =",i4)')             &
     &          RCONDE, RCONDV,sdim;
      !** P =inv(M11)*M12
      d1(:,:) =vs(:nd,:nd);
      d2(:,:) =vs(nd1:nd2,:nd);
!      d2(:,:) =vs(:nd,nd1:nd2)
      t1(:,:) =0.d0;
      call matInvGJ (ier,  d1,t1,  nd,nd,0);  !was gaussj (ier,d1,nd,nd,t1,0,0)
      if (ier == 0) then;
        p(:,:) =matmul(d2,d1);
        call prtmx (p,nd,nd,nd,"p1");
        write (6,'("sigp =",6(/10g13.5))') (sqrt(p(j,j)),j=1,nd);
      else;
        write (6,'(/"u11 inversion failed"/)');
      endif;

      !** check
      d3(:,:) =matmul(f,p) +matmul(p,transpose(f)) +q(:,:)              &
     &         -matmul(matmul(p,hrht),p);
      call prtmx (d3,nd,nd,nd,"pdot");

      !** discrete schur factorization
      write (6,'(/"*** Schur DARE ***")');
      p(:,:) =0.d0;
      b(:,:) =0.d0;
      d1(:,:) =phi(:,:);
      d2(:,:) =q(:,:);
      !**  compute d1 =inv(phi), d2 =inv(phi)*q
      call matInvGJ (ier,  d1,d2,  nd,nd,nd);  !was gaussj (ier,d1,nd,nd,d2,nd,nd)
      if (ier /= 0) stop 'matInvGJ error on phi';
      p(:,:) =matmul(d2,d1);

      b(:nd,:nd) =transpose(phi(:,:)) +matmul(hrht,d2);
      b(:nd,nd1:nd2) =-matmul(hrht(:,:),d1);
      b(nd1:nd2,:nd) =-d2(:,:);
      b(nd1:nd2,nd1:nd2) =d1(:,:);

      !** schur factorization
      jobvs ='V';
      sort ='S';
      sense ='E';
      call DGEESX ('V', sort, SELECT2, SENSE, Nd2, b, LDA, SDIM,        &
     &                   WR, WI, VS, LDVS, RCONDE, RCONDV, WORK, LWORK, &
     &                   IWORK, LIWORK, BWORK, INFO );
      if (info /= 0) then;
        write (6,'("DGEESX info =",i5)') info;
        stop 'gdeesx failure';
      endif;
      call prtmx (b,nd2,nd2,nd2,'schur form');
      call prtmx (vs,nd2,nd2,nd2,'vs');
      call prtmx (wr,1,1,nd2,'wr');
      call prtmx (wi,1,1,nd2,'wi');
      write (6,'("RCONDE, RCONDV =",2g13.5,", sdim =",i4)')             &
     &          RCONDE, RCONDV,sdim;
      !** P =inv(M11)*M12
      d1(:,:) =vs(:nd,:nd);
      d2(:,:) =vs(nd1:nd2,:nd);
!      d2(:,:) =vs(:nd,nd1:nd2)
      t1(:,:) =0.d0;
      call matInvGJ (ier,  d1,t1, nd,nd,0);  !was gaussj (ier,d1,nd,nd,t1,0,0)
      if (ier == 0) then;
        p(:,:) =matmul(d2,d1);
        call prtmx (p,nd,nd,nd,"p1");
        write (6,'("sigp =",6(/10g13.5))') (sqrt(p(j,j)),j=1,nd);
      else;
        write (6,'(/"u11 inversion failed"/)');
      endif;
      !** check
      do i=2,nd;
        do j=1,i-1;
          p(j,i) =0.5*(p(j,i)+p(i,j));
          p(i,j) =p(j,i);
        enddo;
      enddo;
      hp(:,:) =matmul(h,p);
      c(:,:) =matmul(hp,transpose(h)) +rn(:,:);
      t1(:,:) =0.d0;
      call matInvGJ (ier,  c,t1,  mmax,mmax,0);  !was gaussj (ier,c,mmax,mmax,t1,0,0)
      g(:,:) =matmul(transpose(hp),c);
      d2(:,:) =p(:,:) -matmul(g,hp);
      d1(:,:) =matmul(matmul(phi,d2),transpose(phi)) +q(:,:);
      call prtmx (d1-p,nd,nd,nd,"pi(-)-p");
      write (6,'(/"*** End Schur ARE ***")');

      return;
      end subroutine schur;
!*****************************************************
      logical(4) function SELECT1 (wr,wi);
      real(8),intent(in) :: wr,wi;  !real and imaginary part of eigenvalue
      if (wr < 0.) then;
        select1 =.true.;
      else;
        select1 =.false.;
      endif;
      return;
      end function select1;
!*****************************************************
      logical(4) function SELECT2 (wr,wi);
      real(8),intent(in) :: wr,wi;  !real and imaginary part of eigenvalue
      if (wr**2+wi**2 < 1.d0) then;
        select2 =.true.;
      else;
        select2 =.false.;
      endif;
      return;
      end function select2;
