!** Program INS_sim generates simulated strapdown inertial navigation system
!** position error measurements as described in Section 3.3.  It also
!** simulates a missile trajectory.  The output data files are "taimes.dat"
!** (position measurement data), and "specficForce.plt" (specific force versus time).
!** These files are used in program Tst_INS to test Friedlands
!** bias-free-bias-restoring filter and in Example 10.xx to demonstrate a
!** jump detection algorithm

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort INS_sim.f90
!**   INS_sim >ins_sim.lis
!**************************************

      implicit real*8 (a-h,o-z);
      real*8 isp1, isp2;
      dimension r(3),v(3),vdot(3),tsg(3,3),tsgl(3,3),omegs(3);
      dimension fs(3),fl(3),fdir(3),tsl(3,3),r0(3),dr(3),as(3);
      dimension x(27),sigx0(27),qs(9),xjmp(3,20);

      data rearth,gm /6378155.,3.986013d14/;    !m and m^3/s^2
      data pi /3.141592653589793d0/;
      data xlat,xlon,v0 /40.,0.,100./;
      data dt,tmax /0.2,1000./;

      data vmass0 /1360./;
!         first stage data (vmassi1 is 1st stage dry mass, vmassp1 is propellant mass)
      data taz1,tel1,timth1 /0.,90.,70./;             !timth1 is thrust time in sec
      data vmassi1,vmassp1,isp1 /2030.,20600.,240./;  !mass in kg ?
!         second stage data
      data taz2,tel2,timth2 /0.,30.,110./;
      data vmassi2,vmassp2,isp2 /900.,11700.,290./;

!      data taz1,tel1,timth1 /0.,90.,27./
!      data thrust1,vmass1,vmdot1 /780000.,15400.,305./
!         second stage data
!      data taz2,tel2,timth2 /0.,30.,27./
!      data thrust2,vmass2,vmdot2 /470000.,6400.,170./

!         tai initial condition errors  (these numbers match the IEEE paper: from tai_sim1.f)
!         reduce tai tilt uncertainty to match imu
!      data sigx0 /3*13.12, 3*.002, 1.,1.,15.,
      data sigx0 /3*13.12, 3*.002, 3*1440.,                             & !ft, ft/s, arc-sec
     &            3*60., 3*60., 3*.2, 3*200., 3*.2, 3*.012/;
      data qs /3*0., 3*5.40d-9, 3*.004/;
      data sigr /.002/;    !1-sigma meas error in ft (0.61 mm)

      !**           jumps are    y-gyro sfe at 25 sec
      data xjmp /25.,20.,-60.,                                          & !y-gyro g-sen at 50
     &          50.,23.,.10,                                            & !z-gyro bias drift at 100
     &          100.,18.,-.10,                                          & !x accel sfe at 125
     &          125.,13.,1.0,                                           & !x-gyro bias drift at 300
     &          300.,16.,.10,                                           & !y accel bias at 400
     &          400.,11.,-4.,                                           &
     &          42*10000./;
      data n /27/;
      data ix /19191919/;
      !__________________________________________________________

      xjmp(:,:) =1.d20;         !### disable jumps

      open (20,file='taimes.dat',status='unknown',form='unformatted');
      write(20) sigx0;
      open (22,file="specificForce.plt",status='unknown');
      write (22,1030);
 1030 format ('variables ="time (s)", "accel-x", "accel-y", "accel-z"')

!         initialize tai state errors randomly
      xx =rnor(ix);
      do i=1,n;
        x(i)=sigx0(i)*rnor(0);
      enddo;
      write(6,1400) (x(i),i=1,n);
 1400 format('0xo =',3(/5x,10g12.5))

!         calculate booster parameters
      vmdot1 =vmassp1/timth1;       !
      thrust1 =isp1*9.8*vmdot1;     !nt
      vmdot2 =vmassp2/timth2;
      thrust2 =isp2*9.8*vmdot2;
      vmass2 =vmass0+vmassi2+vmassp2;
      vmass1 =vmass2+vmassi1+vmassp1;

!         calculate initial pos, vel, thrust
      d2rad = pi/180.d0;
      xlatr =xlat*d2rad;
      xlonr =xlon*d2rad;
      clat =cos(xlatr);
      slat =sin(xlatr);
      clon =cos(xlonr);
      slon =sin(xlonr);

      r(1) =rearth*clat*clon;
      r(2) =rearth*clat*slon;
      r(3) =rearth*slat;
      rmag =sqrt(r(1)**2+r(2)**2+r(3)**2);
!         initial velocity is vertical (slight offset to allow computation
!          of r cross v)
      v(1) =v0*r(1)/rmag;
      v(2) =v0*r(2)/rmag;
      v(3) =v0*r(3)/rmag + 1.;

!         compute thrust vector in local coordinates and transform to eci
!         (guidance frame)
      tazr =taz1*d2rad;
      telr =tel1*d2rad;
      telr2 =tel2*d2rad;
      fl(1) =cos(telr)*sin(tazr);
      fl(2) =cos(telr)*cos(tazr);
      fl(3) =sin(telr);
!         compute transformation from local (x=east, y=north, z=up) to eci
      tsl(1,1) =-slon;
      tsl(1,2) =-clon*slat;
      tsl(1,3) =clon*clat;
      tsl(2,1) =clon;
      tsl(2,2) =-slon*slat;
      tsl(2,3) =slon*clat;
      tsl(3,1) =0.;
      tsl(3,2) =clat;
      tsl(3,3) =slat;

!         transform to eci
!      do i=1,3
!        r0(i)=r(i)
!        fdir(i)=tsl(i,1)*fl(1)+tsl(i,2)*fl(2)+tsl(i,3)*fl(3)
!      enddo
      r0(:) =r(:);
      fdir(:) =matmul(tsl,fl);
      write(6,1300) fdir;
 1300 format('0fdir =',3g12.5)

!         compute transformation from guidance to system, using r and v
      call transg(tsg,r,v);

      t=0.;
      vmass=vmass1;
      vmdot=vmdot1;
      thrust=thrust1;
      alt=0.;
      flag2=0;
      itmax=tmax/dt+0.5;

!         time loop
      do 100 it=1,itmax;
        t=t+dt;

        !**  compute thrust acceleration (m/s^2) in eci coord.
        if (t.le.timth1) then;
          fs(1)=thrust*fdir(1);
          fs(2)=thrust*fdir(2);
          fs(3)=thrust*fdir(3);

        else if(t-timth1.le.timth2) then;
          thrust=thrust2;
          vmdot=vmdot2;
          fs(1)=thrust*fdir(1);
          fs(2)=thrust*fdir(2);
          fs(3)=thrust*fdir(3);
          if (flag2.eq.0) then;
            flag2=1;
            tazr=taz2*d2rad;
            vmass=vmass2;
          endif;

        else;
          vmass=vmass0;
          !**  after boost phase, add artificial force to simulate bus maneuvers
          !**   (15 second sinusoids, .07g acceleration)
          if (mod(t,100.d0) <= 15.) then;
            fbus=1000.;
            omegt=2.*pi*t/15.;
            fs(1)=fbus*sin(omegt+0.);
            fs(2)=fbus*sin(omegt+1.57);
            fs(3)=fbus*sin(omegt+3.14);
          else;
            fs(:) =0.d0;
          endif;
        endif;

!          total acceleration is sum of gravity and thrust (also add drag)
        rmag =sqrt(r(1)**2+r(2)**2+r(3)**2);
        dum=gm/rmag**3;
        vdot(1)=-dum*r(1)+fs(1)/vmass;    !m/s^2
        vdot(2)=-dum*r(2)+fs(2)/vmass;
        vdot(3)=-dum*r(3)+fs(3)/vmass;
        !**   As is specific force (acceleration without grav - eci frame) in ft/sec**2
        !**   for use in error model
        as(1)=(fs(1)/vmass)/.3048;
        as(2)=(fs(2)/vmass)/.3048;
        as(3)=(fs(3)/vmass)/.3048;
        write (22,'(4g13.5)') t,as(:)*0.3048;    !accel in m/s^2 for plot ###

!         save old tsg
        tsgl(:,:) =tsg(:,:);

        !**   integrate (euler integ)
        do i=1,3;
          r(i) =r(i) +v(i)*dt +0.5*vdot(i)*dt**2;  !m
          v(i) =v(i) +vdot(i)*dt;                  !m/s
        enddo;

!            compute boost thrust angle and mass
        if (t.lt.timth1+timth2) then;
          vmass=vmass-vmdot*dt;
          vmag=sqrt(v(1)**2+v(2)**2+v(3)**2);
          rmag=sqrt(r(1)**2+r(2)**2+r(3)**2);
          vgam=acos((r(1)*v(1)+r(2)*v(2)+r(3)*v(3))/(rmag*vmag));
          vgam=90.-vgam/d2rad;
          teld=telr/d2rad;
          write(6,1350) teld,vgam,teld-vgam;
 1350     format(' thrust ang =',f7.1,', flight path ang =',f7.1,       &
     &             ', dif =',f7.2)
          if (vmag.gt.200. .and. dabs(telr2-telr).gt.0.01) then;
!               if above 300 m/sec, compute 1.0 deg/sec pitchover
!            telr=telr+sign(1.0*d2rad*dt,telr2-telr)
            telr=(vgam-5.)*d2rad;
!              compute thrust vector in local coordinates and transform
!              to eci (guidance frame)
            fl(1)=cos(telr)*sin(tazr);
            fl(2)=cos(telr)*cos(tazr);
            fl(3)=sin(telr);
!             transform to eci
            do 12 i=1,3;
   12         fdir(i)=tsl(i,1)*fl(1)+tsl(i,2)*fl(2)+tsl(i,3)*fl(3);
            write(6,1300) fdir;
         endif;
        endif;

        write(6,1000) t,vmass,r,v,as;
 1000   format('0time ',f10.1,' vmass=',g12.5/' r =',3g12.5,            &
     &         ', v =',3g12.5/' as = ',3g12.5)

!           compute new tsg transformation
        call transg (tsg,r,v);
!           compute body rates omegs (rad/sec) in body coordinates
        call brate (r,v,vdot,omegs,tsg,tsgl,dt);

        !**  generate and write to taimes.dat simulated tai position error meas in ft.
        !**  Note position/velocity integration is in m, m/s but TAI error
        !**  model is in f, ft/s (it uses "as" in f/s^2 as input)
        call taimes(x,qs,t,dt,as,omegs,r,v,sigr,n,xjmp);

        do 40 i=1,3;
   40     dr(i)=r(i)-r0(i);
        range=sqrt(dr(1)**2+dr(2)**2+dr(3)**2);
        rmag=sqrt(r(1)**2+r(2)**2+r(3)**2);
        vmag=sqrt(v(1)**2+v(2)**2+v(3)**2);
        vgam=acos((r(1)*v(1)+r(2)*v(2)+r(3)*v(3))/(rmag*vmag));
        vgam=90.-vgam/d2rad;
        alt=rmag-rearth;
        write(6,1100) range,alt,vgam;
 1100   format(' range =',g12.5,', alt =',g12.5,', vgam =',g12.5)
        if (alt < -1.) stop;

!        write(10) t,vmass,r,v,as,omegs

  100 continue;

      stop;
      end;
! *********************************************************************
      subroutine transg (tsg,r,v);
!        subroutine transg computes the transformaton from strapdown
!        system (x= velocity vector, y=local horizontal, z=orthogonal)
!        to guidance computational (eci).   note that this definition
!        of strapdown system assumes zero missile roll.

!        calling sequence -
!          r(3) = eci position vector (input)
!          v(3) = eci velocity vector (input)
!          tsg(3,3) = output transformation from strapdown to guidance
!                     computational

      implicit real*8 (a-h,o-z);
      dimension tsg(3,3),r(3),v(3),d(3);
      !_____________________________________________

      vmag=sqrt(v(1)**2+v(2)**2+v(3)**2);
      tsg(:,1)=v(:)/vmag;
!         d=r cross v
      call xcross(r,v,d);
      dmag =sqrt(d(1)**2+d(2)**2+d(3)**2);
      tsg(:,2)=d(:)/dmag;
!        column three is cross product
      call xcross(tsg(1,1),tsg(1,2),tsg(1,3));
!         check magnitude
      dmag =sqrt(tsg(1,3)**2+tsg(2,3)**2+tsg(3,3)**2);
      if (dmag-1.d0 > 1.d-12) stop 'transg error: dmag > 1d-12';

      return;
      end subroutine transg;
! *****************************************************************
      subroutine brate (r,v,vdot,omegs,tsg2,tsg1,dt);
!         subroutine brate computes the body rates (omegs) in
!         strapdown coordinates
      implicit real*8 (a-h,o-z);
      dimension r(3),v(3),vdot(3),omegs(3),tsg2(3,3),tsg1(3,3);
      dimension omegg(3),t(3,3),omegs1(3);

      vmag2=(v(1)**2+v(2)**2+v(3)**2);
!         compute body rates in eci coordinates as v cross vdot/v**2
      call xcross(v,vdot,omegg);
      do 5 i=1,3;
    5   omegg(i)=omegg(i)/vmag2;
!      omegg(1)=(v(2)*vdot(3)-v(3)*vdot(2))/vmag2
!      omegg(2)=(v(3)*vdot(1)-v(1)*vdot(3))/vmag2
!      omegg(3)=(v(1)*vdot(2)-v(2)*vdot(1))/vmag2
!         transform to strapdown coord.  (omegs = tsg2**t*omegg)
      do 10 i=1,3;
   10   omegs(i)=tsg2(1,i)*omegg(1)                                     &
     &          +tsg2(2,i)*omegg(2)                                     &
     &          +tsg2(3,i)*omegg(3);

      write(6,1000) omegs;
 1000 format('0omegs =',3g12.5)

!         alternate calculation using (tsg2**t*tsg1-i)
      do 25 i=1,3;
        do 20 j=1,3;
          t(i,j)=tsg2(1,i)*tsg1(1,j)+                                   &
     &           tsg2(2,i)*tsg1(2,j)+                                   &
     &           tsg2(3,i)*tsg1(3,j);
   20   continue;
        t(i,i)=t(i,i)-1.d0;
!        write(6,1100) (t(i,j),j=1,3)
 1100   format(1x,3g12.5)
   25 continue;

      omegs1(1)=t(2,3)/dt;
      omegs1(2)=-t(1,3)/dt;
      omegs1(3)=t(1,2)/dt;
      write(6,1200) omegs1;
 1200 format(' omegs1 =',3g12.5)

      return;
      end;
! *****************************************************************
      subroutine taimes (x,qs,t,dt,as,omegs,r,v,sigr,n,xjmp);
!         subroutine taimes simulates the tai error equations using
!         a reduced 27-state model. the 27 states are:
!           pos error (3) - ft
!           vel error (3) - ft/sec
!           tilt error(3) - arc sec
!           accel bias (3) - micro-g
!           accel sfe (3)  - ppm
!           gyro bias drift (3) - arc sec/hour
!           gyro sfe (3)   - ppm
!           g-sensitive (3) - (arc sec/sec)/g
!           g**2-sensitive (3) - (arc sec/sec)/g**2

!         calling sequence -
!           inputs -
!             x(27) - state vector at time t-dt
!             qs(9) - psd of process noise on first 9 states
!             t -     current time (sec)
!             dt -    time step (sec)
!             as(3) - specific acceleration (ft/sec**2) in guidance
!                     computational (eci) coordinates
!             omegs(3)- body rates (rad/sec) in strapdown coordinates
!             r(3) -  position vector (eci in m)
!             v(3) -  velocity vector (eci in m/s)
!             sigr -  measurement noise sigma on tai position (ft)
!             n    -  number of states (27)
!             xjmp(3,20) - array describing jumps in states for 20 jumps.
!                     xjmp(i,j) describes jump j.
!                         i=1 is time of jump.
!                         i=2 is state number which jumps.
!                         i=3 is magnitude of jump.
!           outputs -
!             x(27) - state vector at time t

!           also note that t,z,phi,q,h,rm are written to unit 20 at
!           10 second increments

      implicit real*8 (a-h,o-z);
      dimension x(n),qs(9),as(3),omegs(3),r(3),v(3),xjmp(3,20);
      dimension f(27,27),phi(27,27),q(45),h(3,27),rm(3),tsg(3,3);
      dimension x1(27),s(9,9),z(3),phit(27,27),qt(45),dum(27,27);
      real(8) dums(45);

      data ix /191919/;
      data ijmp /1/;
      data dtmes,tnext /10., 10./;
      data ifirst /1/;
      !__________________________________________

      if (ifirst.ne.0) then;
!          initialize phit, qt
        ifirst=0;
        phit(:,:) =0.d0;
        qt(:) =0.d0;
        forall (i=1:n) phit(i,i) =1.d0;
      endif;

      call transg(tsg,r,v);

!         compute f matrix
      call ftai(f,tsg,as,omegs);    !f has units of ft, ft/s, arc-sec, micro-G, ...

      !**  integrate to get phi  (i+f*dt)
      do i=1,n;
        do j=1,n;
          phi(j,i)=f(j,i)*dt;
        enddo;
        phi(i,i) =phi(i,i)+1.d0;
      enddo;

!        propagate x = phi*x
      call smult (x1,  phi,x, n,n,1, 27,27,27);
      x(:n) =x1(:n);

      !**  add random process noise
      do i=1,9;
!        call gauss(ix,sqrt(qs(i)*dt),0.d0,vx)
        x(i)=x(i)+sqrt(qs(i)*dt)*rnor(0);
      enddo;

      !**  add jumps (if appropriate)
      if (t > xjmp(1,ijmp)) then;
        j=xjmp(2,ijmp);
        if (j <= 0 .or. j > n) stop 20;
        x(j)=x(j)+xjmp(3,ijmp);
        ijmp=ijmp+1;
      endif;

      q(:) =0.d0;
      q(1)=qs(1)*dt;
      q(3)=qs(2)*dt;
      q(6)=qs(3)*dt;
      q(10)=qs(4)*dt;
      q(15)=qs(5)*dt;
      q(21)=qs(6)*dt;
      q(28)=qs(7)*dt;
      q(36)=qs(8)*dt;
      q(45)=qs(9)*dt;

!          accululate phit = phi*phit, qt=phi*qt*phit**t+q
      call smult (dum,  phi,phit, n,n,n, 27,27,27);
      phit(:,:) =dum(:,:);

      call smop (dums,  phi,qt,9,9,27,27);
      do i=1,45;
        qt(i) =dums(i)+q(i);
      enddo;

      if (t+.001.gt.tnext) then;
        tnext=tnext+dtmes;
        !**  form position measurement error (ft) every 10 seconds
        h(:,:) =0.d0;
        do 40 i=1,3;
          rm(i)=sigr;
          h(i,i)=1.;
!          call gauss(ix,sigr,x(i),z(i))
          z(i)=x(i)+sigr*rnor(0);      !(sigr =0.002 ft)
!            modification 1/6/91 to delete measurement noise
!          z(i)=x(i)
   40   continue;

        write(6,1000) t,z,x;        !z in ft
 1000   format('0t=',g12.5,', z = ',3g12.5,', x =',3(/5x,10g12.5))

!           write t,y,phi,q,h,r to unit 20
        m=3;
        nq=0;
        n2=45;
        do 41 i=1,n2;
          if (qt(i).ne.0.) nq=nq+1;
   41   continue;

        write(20) t,n,m,nq,(z(i),i=1,m),(rm(i),i=1,m);   !z and rm in ft
!         compute velocity, accel error: x1 =f*x
        call smult (x1,  f,x, 6,n,1, 27,27,27);
        write(20) x(1:9), x1(4:6), (f(4:6,i),i=7,27);    !x,x1 and f in ft,ft/s,arc-sec, ...

        do 42 i=1,m;
          j1=0;
          do 43 j=1,n;
            if (h(i,j).ne.0.) then;
              j2=j;
              if (j1.eq.0) j1=j;
            endif;
   43     continue;
   42     write(20) j1,j2,(h(i,j),j=j1,j2);

        do 44 i=1,45;
          if (qt(i).ne.0.)    write(20) i,qt(i);
   44   continue;

        do 50 i=1,n;
          j1=0;
          do 45 j=1,n;
            if (phit(i,j).ne.0.) then;
               j2=j;
               if (j1.eq.0) j1=j;
            endif;
   45     continue;
   50     write(20) j1,j2,(phit(i,j),j=j1,j2);

        !**  reset phit, qt
        phit(:,:) =0.d0;
        qt(:) =0.d0;
        forall (i=1:n) phit(i,i) =1.d0;
      endif;

      return;
      end;
! *****************************************************************
      subroutine ftai (f,tsg,as,omegs);
!         subroutine ftai computes the f matrix (xdot=f*x+q) given
!         tsg (strapdown to guidance computational) transformation,
!         as (specific acceleration in guidance coord.) and omegs
!         (body rate in strapdown coord.)
      implicit real*8 (a-h,o-z);
      dimension f(27,27),tsg(3,3),as(3),omegs(3);
      dimension asst(3),apsiep(3,12);

      data c2,c3,c4,c5 /.03108, 32.174d-6, 1.d-6, 4.8481d-6/;    !english
!      data c2,c3,c4,c5 /.10197, 9.8066d-6, 1.d-6, 4.8481d-6/      !metric
      !__________________________________________________

      f(:,:) =0.d0;
      c2 =1.d-6/c3;

!         compute specific acceleration in strapdown frame
!         (asst = tsg**t*as)
      asst(:) =matmul(as,tsg);

!         f(dr,dv)
      f(1,4)=1.;
      f(2,5)=1.;
      f(3,6)=1.;
!         f(dv,dr) is ignored because tai meas are differenced with imu
!             (gravity error is not observable)

!         f(dv,psi) = -{f x}.  c5 converts arc sec to radians
!                    (note that right-hand rule is used for cross product)
      f(4,7)=0.;
      f(4,8)=c5*as(3);
      f(4,9)=-c5*as(2);

      f(5,7)=-c5*as(3);
      f(5,8)=0.;
      f(5,9)=c5*as(1);

      f(6,7)=c5*as(2);
      f(6,8)=-c5*as(1);
      f(6,9)=0.;

!         f(dv,da)      (accel bias - c3 converts micro-g to ft/sec**2)
      ida=9;
      f(4,ida+1)=c3*tsg(1,1);
      f(5,ida+1)=c3*tsg(2,1);
      f(6,ida+1)=c3*tsg(3,1);

      f(4,ida+2)=c3*tsg(1,2);
      f(5,ida+2)=c3*tsg(2,2);
      f(6,ida+2)=c3*tsg(3,2);

      f(4,ida+3)=c3*tsg(1,3);
      f(5,ida+3)=c3*tsg(2,3);
      f(6,ida+3)=c3*tsg(3,3);

!                       (scale factor - c4 converts ppm to factor)
      f(4,ida+4)=c4*tsg(1,1)*asst(1);
      f(5,ida+4)=c4*tsg(2,1)*asst(1);
      f(6,ida+4)=c4*tsg(3,1)*asst(1);

      f(4,ida+5)=c4*tsg(1,2)*asst(2);
      f(5,ida+5)=c4*tsg(2,2)*asst(2);
      f(6,ida+5)=c4*tsg(3,2)*asst(2);

      f(4,ida+6)=c4*tsg(1,3)*asst(3);
      f(5,ida+6)=c4*tsg(2,3)*asst(3);
      f(6,ida+6)=c4*tsg(3,3)*asst(3);

!         f(psi,eps)
      ieps=15;
!                        (bias drift - arc-sec/sec)
      apsiep(1,1)=1.;
      apsiep(2,2)=1.;
      apsiep(3,3)=1.;
!            (scale factor - c4/c5 converts ppm-rad/sec to arc-sec/sec)
      apsiep(1,4)=(c4/c5)*omegs(1);
      apsiep(2,5)=(c4/c5)*omegs(2);
      apsiep(3,6)=(c4/c5)*omegs(3);
!         (g-sensitive - c2 converts ft/sec**2 to g)
!                                 (error is arc-sec/sec/g)
      apsiep(1,7)=c2*asst(2);
      apsiep(2,8)=c2*asst(1);
      apsiep(3,9)=c2*asst(1);
!          (g**2 sensitive )
      apsiep(1,10)=c2**2*asst(2)**2;
      apsiep(2,11)=c2**2*asst(1)**2;
      apsiep(3,12)=c2**2*asst(1)**2;

!        transform from strapdown to guidance frame
      do 30 i=1,3;
        do 30 j=1,12;
          f(6+i,ieps+j)=tsg(i,1)*apsiep(1,j)                            &
     &                 +tsg(i,2)*apsiep(2,j)                            &
     &                 +tsg(i,3)*apsiep(3,j);
   30 continue;

      return;
      end;
! **********************************************************************
      subroutine xcross (a,b,c);
!         computes cross product c = a x b using right-hand rule
      real *8 a(3),b(3),c(3);

      c(1)=a(2)*b(3)-a(3)*b(2);
      c(2)=a(3)*b(1)-a(1)*b(3);
      c(3)=a(1)*b(2)-a(2)*b(1);

      return;
      end;


