!** Program wiener.f computes the Wiener filter and steady-state Kalman filter
!** solution for a 1-state Markov process problem.  The output file is "wkf.lis."
!** The results are used to validate equations in Example 8.6-1.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort wiener.f90
!**   wiener >wkf28_1.lis
!**************************************

      implicit none;
      integer(4) i,j,k,ier,it,jmin,nsum,nxsum,imax;
      integer(4) :: ix =19191919;
      integer(4),parameter :: kmax=1000, mmax=10000;

      real(8) delta,t,xrsum;
      real(8) h,g,rs,a,b,aa(0:kmax);
      real(8) f,d1,xf,x,xw,p,phi,qs,y(mmax);
      real(8) rd,qd,bd,cd,pa,pd,kd,xfd,xa,xta,rsum(2,5),res(5);
      real(8) :: sigx0 =2.d0, tau =5.d0;
      real(8) :: sigr =2.80d0;
      real(8) :: dt =0.1d0, tmax =1000.;
      real(8) :: pi =3.141592653589793d0;
      real(8),external :: rnor;
      !_______________________________________________________

      write (6,'("Compare Wiener and steady-state Kalman-Bucy filters"/ &
     &    "sigx0 =",g13.5,", tau =",g13.5,", sigr =",g13.5,             &
     &    ", dt =",g13.5)') sigx0,tau,sigr,dt;

      qs =2.d0*sigx0**2/tau;
      write (6,'("qs =",g13.5)') qs;
      rs =sigr**2;
      a =1.d0/tau;
      b =sqrt(a**2 +qs/rs);
      write (6,'("a,b =",2g13.5)') a,b;
      f =-a;
      phi =exp(f*dt);

      !** continuous KF
!      p =rs*(-a +sqrt(a**2 +qs/rs))   !steady-state Kalman-Bucy covariance
      p =rs*(-a+b);
      xf =0.d0;
      xfd =0.d0;
      x =0.d0;
      g =p/rs;        !steady-state Kalman-Bucy gain
      write (6,'("check riccati eqn ",g13.5,", p=",g13.5)')             &
     &     2.d0*f*p +qs -rs*g**2, p;

      !** discrete kf
      rd =rs/dt;
      qd =qs*dt;
      bd =rd*(1.d0-phi**2) -qd;
      cd =4.d0*qd*rd;
      pa =(-bd+sqrt(bd**2+cd))/2.d0;   !steady-state a priori cov
      kd =pa/(pa+rd);
      pd =pa*(1.d0-kd);
      write (6,'("pa-ss =",2g13.5,", pa-ss err =",g13.5)')              &
     &      pa, pd, phi**2*(pa-pa**2/(pa+rd))+qd-pa;

      !** Wiener filter weighting
      d1 =dt*qs/((a+b)*rs); !*dt ??
      do i=0,kmax;
        if (abs(b*i*dt) > 90.d0) exit;
        imax =i;
        aa(i) =d1*exp(-b*i*dt);
      enddo;
      write (6,'("imax =",i4)') imax;     !250
      h =rnor(ix);
      xrsum =0.d0;
      nxsum =0;
      nsum =0;
      rsum(:,:) =0.d0;

      do it=1,mmax;
        t =dt*it;
        if (t > tmax) exit;
        xta =x +dt*f*x;
        x =x +dt*f*x +sqrt(qs*dt)*rnor(0);
        y(it) =x +(sigr/sqrt(dt))*rnor(0);!??
        xf =xf +dt*(f*xf +g*(y(it)-xf));     !continuous Kalman-Bucy
        xa =phi*xfd;                         !discrete Kalman a priori
        xfd =xa +kd*(y(it)-xa);              !discrete Kalman a posteriori

        !** wiener filter
        jmin =min(max(it-1,0),imax);
        xw =0.d0;
        do i=0,jmin;
          xw =xw +aa(i)*y(it-i);             !wiener
        enddo;

        if (it > imax) then;
          nsum =nsum+1;
          res(:) =(/ xf-x, xa-xta, xfd-x, xw-x, xw-xf /);
          rsum(1,:) =rsum(1,:) +res(:);
          rsum(2,:) =rsum(2,:) +res(:)**2;
        endif;

        if (abs(xf/xw) < 4.) then;
          xrsum =xrsum +abs(xf/xw);
          nxsum =nxsum+1;
        else;
          write (6,'("t =",g13.5,", xf =",g13.5,", xw =",g13.5,         &
     &     ", xf/xw =",g13.5)')   it*dt,xf,xw,xf/xw;
        endif;

        if (mod(it-1,10) == 0) then;
          write (6,'("t =",g13.5,", xf =",g13.5,", xw =",g13.5,         &
     &     ", xw-xf =",g13.5,", xf/xw =",g13.5,", xfd/xf =",g13.5)')    &
     &       it*dt,xf,xw,xw-xf,xf/xw,xfd/xf;
        endif;
      enddo;

      write (6,'("average xf/xw =",g13.5,", nxsum =",i5)')              &
     &    xrsum/nxsum,nxsum;

      write (6,'(/"# samples =",i5)') nsum;
      write (6,'("mean errors =",5g13.5)') rsum(1,:)/nsum;
      write (6,'("rms errors =",5g13.5)') sqrt(rsum(2,:)/nsum);

      stop;
      end;
