!** Program PassTrk_Fil.f tests the linearized Kalman filter, extended Kalman
!** filter (EKF) and IEKF solutions for passive ship tracking.  The output
!** plot files are passTrk.plt (own and target ship positions versus time),
!** Xerr_ekf.plt (estimate errors for EKF), Xerr_lkf.plt (estimate errors
!** for linearized KF), and Xerr_iekf.plt (estimate errors for IEKF).
!** Results are discussed in Section 9.1.2, but results are not shown because
!** all methods performed poorly.
!** Modification: 4/10/09: bearing definition changed to atan(dx/dy) rather than atan(dy/dx) used in book

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort passTrk_Fil.f90    (libkf_bnd_ifort.lib should be linked by default for sinv, symult, prtmx, rnoruni)
!**   passTrk_Fil >passt.lis
!**************************************

      module modelDat;
      !** includes model info for funct
      integer(4),parameter :: nd =4;
      integer(4),parameter :: mmax =181;    !61
      real(8) xt(4),xo(4,2),tturn,h(mmax+nd,nd),res(mmax+nd);
      real(8) xp(nd),pad(nd),dyr(mmax),dt,rn;
      logical(4) optlPath,cartesian;
!      logical(4) :: extended =.true.       !true for extended, false for linearized
      logical(4) :: firstIter =.false.;     !normal mode linearizing all iter about estimate ###
!      logical(4) :: firstIter =.true.      !test linearizing partials about meas for 1st iter ###
      end module modelDat;
!****************************************************
      program passTrk;
      !** tests nonlinear least squares solution of passive ship tracking
      use modelDat;
      use optlDat, only:fstat;
      implicit none;
      integer(4) i,n,m,ix;
      integer(4) nmeas,nmeas1;
      integer(4) :: method =1;  !1=EKF, 2=IEKF  ###
      logical(4) partial,debugp;
      real(8) au(nd*(nd+1)/2),xe(nd);
      real(8) pfs(nd*(nd+1)/2),bu(nd),dxn(nd);
      real(8) cnb,cost;
      real(8),external :: rnor;
      !____________________________________________________

      m =mmax;
      n =nd;
      ix =191919;
      cnb =rnor(ix);

      rn =1.0d0/57.296;      !1. deg 1-sigma  (was 1.5 for LS)
      dt =1.d0/360.d0;       !10 min samples, 15 min total  (was 15 min for LS
      cartesian =.true.;

      !** true state, prior and observer position
      xt(:) =(/ -2.d0, 20.d0, 17.d0, -8.d0 /);
      xo(:,1) =(/ 0.d0, 0.d0, 0.d0, 20.d0 /);
      xo(:,2) =(/ 0.d0, 5.d0, -10.0d0, 17.32d0 /);
!      xo(:,2) =(/ 0.d0, 5.d0, -14.14d0, 14.14d0 /)   !### optional

      xp(:) =(/ -4.d0, 25.d0, 25.d0, -2.d0 /);   !prior estimate
!      xp(:) =(/ -3.d0, 30.d0, 0.d0, 0.d0 /)    !prior estimate for batch LS
      pad(:) =(/ 5.d0, 10.d0, 10.d0, 10.d0 /); !prior 1-sigma  !with "good" initialization ###
!      pad(:) =(/ 15.d0, 15.d0, 20.d0, 20.d0 /)  !prior 1-sigma

      xp(:) =xt(:);     !###  test initialized at truth
      xe(:) =xp(:);     !initialize estimate with prior
      tturn =0.25;      !own ship turn after 1/4 hour

      do i=1,m;
        dyr(i) =rn*rnor(0);    !meas noise
      enddo;

      open (11,file='passTrk.plt',status='unknown');
      write (11,1200);
 1200 format('VARIABLES ="x (nmi)", "y (nmi)"')

      partial =.true.;
      debugp =.false.;
      write (*,'("meas#, meas, meas-resid, H(1:4)")');

      open (10,file='Xerr_ekf.plt',status='unknown');
      write (10,1100);
 1100 format('VARIABLES ="time (hours)", "East-error (nmi)", ',         &
     &       '"North-error (nmi)", "East-vel-error (kn)", ',            &
     &      '"North-vel-error (kn)"')
!     &      'ZONE T="optl-cost", F=POINT ')

      if (firstIter) then;
        !** take two GN iterations to initialize using linearizing partials
        optlPath =.true.;
        nmeas1 =10;
        do i=1,2;
          call funct (cost,pfs,au,bu,dxn,nmeas,                         &
     &                    nmeas1,partial,xe,n,debugp);
          xe(:n) =xe(:n) +dxn(:n);
        enddo;
      endif;

      method =1;  !EKF
      call ekf (cost,nmeas,   xe,pfs,  n,debugp,method);
      close (10);

      open (10,file='Xerr_lkf.plt',status='unknown');
      write (10,1100);
!      extended =.false.
      method =3;     !linearized filter
      do i=1,2; !iterate twice to get better initial estimate for linearization
        xe(1:2) =xe(1:2) -xe(3:4)*dt*mmax;
        call ekf (cost,nmeas,   xe,pfs,  n,debugp,method);
      enddo;
      close (10);

      open (10,file='Xerr_iekf.plt',status='unknown');
      write (10,1100);
      method =2;     !IEKF
      xe(:) =xp(:);     !initialize estimate with prior
      call ekf (cost,nmeas,   xe,pfs,  n,debugp,method);

      stop;
      end program;
!******************************************************************
      subroutine ekf (cost,nmeas,   xf,pfs,  nest,debugp,method);
      use modelDat, only:xt,xo,tturn,pad,dyr,mmax,dt,rn,optlPath,       &
     &              firstIter,res;
      use optlDat,only:fstat;
      implicit none;

      integer(4),intent(in) :: nest,method;
      logical(4),intent(in) :: debugp;
      real(8),intent(inout) :: xf(nest);
      real(8),intent(inout) :: pfs(nest*(nest+1)/2);
      real(8),intent(out) :: cost;
      integer(4),intent(out) :: nmeas;

      integer(4) it,j,kk,i,k,itp,n;
      real(8) t,phi(nest,nest),xti(nest),xei(nest),ytmp;
      real(8) xop(2,40),xtp(2,40),xref(nest);
      real(8) pf(nest,nest),q(nest,nest),xa(nest),pa(nest,nest);
      real(8) y(mmax),ye(mmax),dy,xpl(2),xoi(4);    !d2ydx2(n)
      real(8) h(nest),hp(nest),g(nest),xfi(nest),xfil(nest);
      logical(4),save :: first =.true.;
      !_________________________________________

      if (nest /= 4) stop 'nest /= 4';
      cost =0.d0;
      n =nest;

      !** add prior
      pf(:,:) =0.d0;
      q(:,:) =0.d0;
      q(3,3) =0.001;
      q(4,4) =0.001;
      kk =0;
      do i=1,nest;
        pf(i,i) =pad(i)**2;
      enddo;

      xti(:) =xt(:);
      xei(:) =xf(:);
      xoi(:) =xo(:,1);
      if (first) then;
        write (11,1000) 3*(mmax)/10+4;    !
 1000   format ('ZONE T="own_targ", I=',i4,', F=POINT')
        write (11,'(4g13.5)') xoi(:2);
      endif;
      itp =0;
      t =0.d0;
!      write (10,'(6g13.5)') t,xf(:)-xt(:)

      do it=1,mmax;
        t =it*dt;
        xti(:) =xt(:) +(/ t*xt(3), t*xt(4), 0.d0,  0.d0 /);     !true model
        xa(:) =xf(:) +(/ dt*xf(3), dt*xf(4), 0.d0, 0.d0 /);     !estimate model
        xref(:) =xei(:) +(/ t*xei(3), t*xei(4), 0.d0, 0.d0 /);     !reference based on initial estimate
        if (t <= tturn) then;
          xoi(3:4) =xo(3:4,1);
        else;
          xoi(3:4) =xo(3:4,2);
        endif;
        xoi(1:2) =xoi(1:2) +dt*xoi(3:4);
        phi(:,:) =0.d0;
        forall (i=1:n) phi(i,i) =1.d0;
        phi(1,3) =dt;
        phi(2,4) =dt;
        pa(:,:) =matmul(matmul(phi,pf),transpose(phi)) +q(:,:);

        !** compute meas and partials
        call get_yh (y(it),h,  xti,xoi,n);
        y(it) =y(it) +dyr(it);
        call get_yh (ye(it),h,  xa,xoi,n);
        if (method == 3) then;
          !** recompute h for linearized filter
          call get_yh (ytmp,h,  xref,xoi,n);
        endif;

        hp(:) =matmul(h,pa);
        g(:) =hp(:)/(dot_product(h,hp)+rn**2);
        dy =y(it)-ye(it);     !/rn
        xf(:) =xa(:) +g(:)*dy;
        ytmp =ye(it);

        if (method == 2) then;
          xfi(:) =xf(:);
          xfil(:) =1.d20;
          do k=1,100;
            call get_yh (ye(it),h,  xfi,xoi,n);
            hp(:) =matmul(h,pa);
            g(:) =hp(:)/(dot_product(h,hp)+rn**2);
            dy =y(it)-ye(it) -dot_product(h,xa(:)-xfi(:));     !/rn
            xfi(:) =xa(:) +g(:)*dy;
            if (sqrt(sum((xfil -xfi)**2)) < 0.001) exit;
            xfil(:) =xfi(:);
            if (k > 80) then;
              !** if more than 80 iterations, use original ekf estimate
              xfi(:) =xf(:);
              call get_yh (ye(it),h,  xa,xoi,n);
              hp(:) =matmul(h,pa);
              g(:) =hp(:)/(dot_product(h,hp)+rn**2);
              write (10,'("more than 80 iter: ",g13.5)') ye(it)-ytmp;
              exit;
            endif;
!            write (6,'("iekf x =",g13.5)') xfi
          enddo;
          xf(:) =xfi(:);
        endif;

        forall (i=1:nest) pf(:,i) =pa(:,i) -g(:)*hp(i);
        do i=2,nest;
          do j=1,nest-1;
            pf(j,i) =0.5d0*(pf(j,i)+pf(i,j));
            pf(i,j) =pf(j,i);
          enddo;
        enddo;

        write (10,'(6g13.5)') t,xf(:)-xti(:);

        if (first) then;
          write (*,'(i3,6g13.5)') it,y(it),y(it)-ye(it),h(:);
          if (mod(it-1,10) == 0) then;
            if (it > 1) then;
              write (11,'(4g13.5)') xpl(:2);
            endif;
            write (11,'(3g13.5)') xti(:2);
            write (11,'(3g13.5)') xoi(:2);
            xpl(:) =xti(:2);
            itp =itp+1;
            xop(:,itp) =xoi(:2);
            xtp(:,itp) =xti(:2);
          endif;
        endif;
      enddo;    !end time loop

      nmeas =mmax;

      if (first) then;
        write (11,'(3g13.5)') xtp(:,itp);
        write (11,1010) itp;
 1010   format ('ZONE T="ownp", I=',i4,', F=POINT')
        do it=1,itp;
          write (11,'(4g13.5)') xop(:,it);
        enddo;
        write (11,1020) itp;
 1020   format ('ZONE T="target", I=',i4,', F=POINT')
        do it=1,itp;
          write (11,'(4g13.5)') xtp(:,it);
        enddo;
      endif;

      return;
      end subroutine ekf;
!***************************************************************
      subroutine get_yh (y,h,  x,xoi,n);
      !** computes measurements of bearing and biased velocity and partials
      implicit none;

      integer(4),intent(in) :: n;
      real(8),intent(in) :: x(n),xoi(4);
      real(8),intent(out) :: y,h(n);

      real(8) dxe(2),r2;
      !_______________________________________________

!            dxe(:) =xfi(:2) -xoi(:2)
!            r2 =sum(dxe(:2)**2)
!            ye(it) =atan2(dxe(1),dxe(2))
!            h(:) =(/ dxe(2), -dxe(1), 0.d0, 0.d0 /)/r2

      dxe(:) =x(:2) -xoi(:2);
      r2 =dxe(1)**2 +dxe(2)**2;!sum(dxe(:2)**2)
      y =atan2(dxe(1),dxe(2));

!      if (firstIter) then
!        !### alternate using bearing: 2nd iteration is almost as good as 4th iter above, but then never converges
!        dydx(1,:) =(/ cos(y(it)), -sin(y(it)), 0.d0, 0.d0 /)/sqrt(r2)
!      else
        h(:) =(/ dxe(2), -dxe(1), 0.d0, 0.d0 /)/r2;
!      endif

      return;
      end subroutine get_yh;
!*******************************************************************
      subroutine funct (cost,pf,au,bu,dxn,nmeas,                        &
     &                  nmeas1,partial,xest,nest,debugp); !(cost,au,bu,  xe,n)
      use modelDat, only:xt,xo,tturn,xp,pad,dyr,mmax,dt,rn,optlPath,    &
     &              firstIter,h,res;
      use optlDat,only:fstat;
      implicit none;

      integer(4),intent(in) :: nest,nmeas1;
      logical(4),intent(in) :: partial,debugp;
      real(8),intent(in) :: xest(nest);
      real(8),intent(out) :: pf(nest*(nest+1)/2);
      real(8),intent(out) :: cost,au(nest*(nest+1)/2),bu(nest),dxn(nest);
      integer(4),intent(out) :: nmeas;

      integer(4) it,j,k,kk,i,ier,itp,n;
      real(8) t,hnorm(mmax),phi(nest,nest),xti(nest),xei(nest);
      real(8) dxt(2),dxe(2),eps,errm,xop(2,40),xtp(2,40);
      real(8) y(mmax),ye(mmax),dydx(nest),dy,xpl(2),xoi(4),r2;    !d2ydx2(n)
      logical(4),save :: first =.true.;
      !_________________________________________

      if (nest /= 4) stop 'nest /= 4';
      if (partial) au(:) =0.d0;
      bu(:) =0.d0;
      h(:,:) =0.d0;
      cost =0.d0;
      n =nest;

      !** add prior
      kk =0;
      do i=1,nest;
        cost =cost +0.5d0*((xp(i)-xest(i))/pad(i))**2;
        bu(i) =bu(i) +(xp(i)-xest(i))/pad(i)**2;
        res(i) =(xp(i)-xest(i))/pad(i);
        h(i,i) =1.d0/pad(i);
        if (partial) then;
          kk =kk+i;
          au(kk) =1.d0/pad(i)**2;
        endif;
      enddo;

      xti(:) =xt(:);
      xei(:) =xest(:);
      xoi(:) =xo(:,1);
      if (first) then;
        write (11,1000) 3*(mmax)/10+4;    !
 1000   format ('ZONE T="own_targ", I=',i4,', F=POINT')
        write (11,'(4g13.5)') xoi(:2);
      endif;
      itp =0;

      do it=1,nmeas1;
        t =(it-1)*dt;
        xti(:) =xt(:) +(/ t*xt(3), t*xt(4), 0.d0,  0.d0 /);     !true model
        xei(:) =xest(:) +(/ t*xest(3), t*xest(4), 0.d0, 0.d0 /);     !estimate model
        if (t <= tturn) then;
          xoi(3:4) =xo(3:4,1);
        else;
          xoi(3:4) =xo(3:4,2);
        endif;
        xoi(1:2) =xoi(1:2) +dt*xoi(3:4);
        phi(1,:) =(/ 1.d0, 0.d0, t, 0.d0 /);
        phi(2,:) =(/ 0.d0, 1.d0, 0.d0, t /);
        phi(3,:) =(/ 0.d0, 0.d0, 1.d0, 0.d0 /);
        phi(4,:) =(/ 0.d0, 0.d0, 0.d0, 1.d0 /);

        dxt(:) =xti(:2) -xoi(:2);
        y(it) =atan2(dxt(1),dxt(2)) +dyr(it);!bearing         !sqrt(sum(dxt(:)**2)) +dyr(it)
!        y(it) =atan2(dxt(2),dxt(1)) +dyr(it) !bearing

        dxe(:) =xei(:2) -xoi(:2);
        r2 =sum(dxe(:2)**2);
        ye(it) =atan2(dxe(1),dxe(2));
!        ye(it) =atan2(dxe(2),dxe(1))
        if (firstIter) then;
          !### alternate using bearing: 2nd iteration is almost as good as 4th iter above, but then never converges
          dydx(:) =(/ cos(y(it)), -sin(y(it)), 0.d0, 0.d0 /)/sqrt(r2);
!          dydx(:) =(/ -sin(y(it)), cos(y(it)), 0.d0, 0.d0 /)/sqrt(r2)
        else;
          dydx(:) =(/ dxe(2), -dxe(1), 0.d0, 0.d0 /)/r2;
!          dydx(:) =(/ -dxe(2), dxe(1), 0.d0, 0.d0 /)/r2
        endif;

        dy =(y(it)-ye(it))/rn;
        res(it+n) =dy;
        cost =cost +0.5d0*dy**2;

        if (partial) then;
          h(it+n,:nest) =matmul(dydx,phi)/rn;
          bu(:) =bu(:) +h(it+n,:nest)*dy;
          kk =0;
          do j=1,nest;
            do k=1,j;
              kk =kk+1;
              au(kk) =au(kk) +h(it+n,j)*h(it+n,k);
            enddo;
          enddo;

          hnorm(it) =sqrt(sum(h(it+n,1:nest)**2));
          if (first) then;
            write (*,'(i3,6g13.5)') it,y(it),y(it)-ye(it),h(it+n,:nest);
            if (mod(it-1,10) == 0) then;
              if (it > 1) then;
                write (11,'(4g13.5)') xpl(:2);
              endif;
              write (11,'(3g13.5)') xti(:2);
              write (11,'(3g13.5)') xoi(:2);
              xpl(:) =xti(:2);
              itp =itp+1;
              xop(:,itp) =xoi(:2);
              xtp(:,itp) =xti(:2);
            endif;
          endif;
        else;
          bu(:) =bu(:) +h(it+n,:nest)*dy;    !use last GN iteration value of h
        endif;
      enddo;    !end time loop

      if (first) then;
        write (11,'(3g13.5)') xtp(:,itp);
        write (11,1010) itp;
 1010   format ('ZONE T="ownp", I=',i4,', F=POINT')
        do it=1,itp;
          write (11,'(4g13.5)') xop(:,it);
        enddo;
        write (11,1020) itp;
 1020   format ('ZONE T="target", I=',i4,', F=POINT')
        do it=1,itp;
          write (11,'(4g13.5)') xtp(:,it);
        enddo;
      endif;

      first =.false.;
      nmeas =mmax;

      if (partial) then;
        !** compute GN iteration
        pf(:) =au(:);
        eps =1.d-4;
        call sinv (ier,errm,  pf,  nest,eps,.true.);
        if (ier < 0) stop '"*** sinv error';
        call symult (dxn,  pf,bu,nest);
        fstat%testc =sqrt(dot_product(bu,dxn)/nest);
        if (optlPath) then;
          write (10,'(7g14.6)') xest(:)-xt(:),cost, fstat%testc, 0.;
          write (12,'(7g14.6)') xest(:)-xt(:),cost, fstat%testc, 0.;
        endif;
      else;
        fstat%testc =0.;
!        if (optlPath)     !normally not printed for path plots ###
!     &    write (10,'(7g14.6)') xest(:)-xt(:),cost, fstat%testc, 0.
      endif;
      bu(:) =-bu(:);    !make gradient

      fstat%rrms =sqrt(2.d0*cost/mmax);
      fstat%cost =cost;
      fstat%nmeas =nmeas;

      return;
      end subroutine funct;

