!** covar_Anal.f is a driver for linear covariance error analysis of the effects of
!**  unadjusted analyze biases, consider biases, initial condition errors,
!**  meas noise, and process noise.  Covar_Anal is structured to be general-purpose and
!**  may be used for many systems, but some write statements are designed for the 27-state IMU model

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort covar_Anal.f90 /c
!**************************************

      module covarMod;
      !** contains global variables used in defining Kalman filter covariance error analysis

      implicit none;
      integer(4) nstate;        !total number of states in model (set by model)
      integer(4) ncore;         !number of core states in model
      integer(4) nMarkov;       !number of markov states in model
      integer(4) nbias;         !number of bias states in model

      integer(4) nest;          !total number of estimated states
      integer(4) nConsider;     !total number of consider parameters
      integer(4) nUnadj;        !total number of unadjusted-analyze parameters
      integer(4) ncoreE;        !number of estimated core states
      integer(4) nMarkE;        !number of estimated Markov states
      integer(4) nbiasE;        !number of estimated bias states
      integer(4) nDynE;         !number of estimated core+markov states

      integer(4) nmeas;         !total number of vector meas in input file

      integer(4),parameter :: mmaxi =10;  !maximum number of allowed scalar measurements at one time point (10)

      integer(4),allocatable :: mapxe(:);!mapping from estimated states (nest) to total model states
      integer(4),allocatable :: mapxu(:);!mapping from unadjusted analyze parameter (nUnadj) to total model states
      integer(4),allocatable :: mapxc(:);!mapping from consider parameters (nconsider) to total model states
      integer(4),allocatable :: mapxei(:);!mapping from total model states (nstate) to estimated states
      integer(4),allocatable :: mapxui(:);!mapping from total model states (nstate) to estimated states
      integer(4),allocatable :: mapxci(:);!mapping from total model states (nstate) to estimated states

!      character(80) setupFile   !name of file containing data for structure xstate
!      character(80) measFile    !name of file containing measurement data
      character(20),allocatable :: xnames(:); !names of total states as set by model routines

      real(8) tepoch;           !epoch time of x0 and p0
      real(8) tend;             !end time for filter processing


!      real(8),allocatable :: phit(:,:)  !total model state transition matrix (in model module)
!      real(8),allocatable :: qt(:,:)    !total model state process noise matrix (in model module)
      real(8),allocatable :: phiee(:,:);!state transition matrix for estimated-estimated states (nest,nest)
      real(8),allocatable :: phieu(:,:);!state transition matrix for estimated-unadjust states (nest,nUnadj)
      real(8),allocatable :: phiec(:,:);!state transition matrix for estimated-consider states (nest,nConsider)
      real(8),allocatable :: qe(:,:);   !process noise matrix for estimated-estimated states (nest,nest)

      real(8),allocatable :: pae(:,:);  !a priori covariance of estimated-estimated states (nest,nest)
      real(8),allocatable :: par(:,:);  !component of a priori covariance due to meas noise (nest,nest)
      real(8),allocatable :: paq(:,:);  !component of a priori covariance due to process noise (nest,nest)
      real(8),allocatable :: pfe(:,:);  !filter a posteriori covariance of estimated states (nest,nest)
      real(8),allocatable :: pf0(:,:);  !component of a posteriori state covar due to initial errors (nest,nest)
      real(8),allocatable :: pfu(:,:);  !component of a posteriori state covar due to unadjust parm (nest,nest)
      real(8),allocatable :: pfc(:,:);  !component of a posteriori state covar due to consider param (nest,nest)
      real(8),allocatable :: pfr(:,:);  !component of a posteriori state error covariance due to meas noise
      real(8),allocatable :: pfq(:,:);  !component of a posteriori state error covariance due to process noise

      real(8),allocatable :: su(:,:);   !sensitivity matrix of state wrt error in unadjusted-analyze parameters
      real(8),allocatable :: sc(:,:);   !sensitivity matrix of state wrt error in consider parameters
      real(8),allocatable :: sa(:,:);   !sensitivity matrix of state wrt error in a priori estimate

      real(8),allocatable :: xfa(:);    !a priori values of estimated states (nest)
      real(8),allocatable :: xfe(:);    !a posteriori values of estimated states (nest)
      real(8),allocatable :: xtot(:);   !total state as used in model subroutines (nstate)
      real(8),allocatable :: xtota(:);  !predicted total state as used in model subroutines (nstate)

      real(8),allocatable :: pu(:,:);   !covariance of unadjust parameters (nUnadj)
      real(8),allocatable :: pc(:,:);   !covariance of consider parameters (nConsider)
      real(8),allocatable :: p0(:,:);   !covariance of initial state x0 (nest)
      real(8),allocatable :: xu(:);     !assumed value of unadjust parameters (nUnadj)
      real(8),allocatable :: xc(:);     !assumed value of consider parameters (nConsider)
      real(8),allocatable :: x0(:);     !assumed initial value of estimated states (nest)

!      real(8),allocatable :: h(:,:)       !total model meas partial matrix (mmaxi,nstate)  (in model module)
      real(8),allocatable :: he(:,:);     !model meas partial matrix (mmaxi,nest)
      real(8),allocatable :: hu(:,:);     !model meas partial matrix (mmaxi,nUnadj)
      real(8),allocatable :: hc(:,:);     !model meas partial matrix (mmaxi,nConsider)

      real(8) rnoise(mmaxi,mmaxi);          !meas noise covariance

      type states;
        integer(4) status;      !0=ignore, 1=unadjusted analyze, 2=consider, 3=estimate
        integer(4) xtype;       !1=core state, 2=1 or 2 order Markov, 3=bias
        character(20) xname;    !name of state that must match name in hard-coded model
        real(8) x0;             !a priori estimate at epoch time
        real(8) sig0;           !1-sigma uncertainty of x0 at epoch time
        real(8) qs;             !PSD of driving white noise
        real(8) const(2);       !known constants associated with this state in dynamic or measurement model
      end type;
      type (states),allocatable :: xstate(:);

      type meass;
        integer(4) mtype;       !meas type (must match types in model)
        integer(4) m;           !number of scalar meas included in this meas
        character(20) mname;    !name of meas for output (may include sensor name)
        real(8) t;              !meas time
        real(8) y(mmaxi);        !up to mmax scalar meas values
        real(8) sigr(mmaxi);     !1-sigma meas noise for mmax scalar meas
      end type;
      type (meass),allocatable :: meas(:);

      contains;
!*****************************************************
      subroutine KFcovar1 (phit,qt,ht,y,sigr,t,tlst,xtr,m,mdim,n,ndim);
      !** KFcovar1 is called to process each measurement set
      implicit none;
      integer(4),intent(in) :: m;      !number of measurements in this set
      integer(4),intent(in) :: mdim;   !measurement dimension of ht
      integer(4),intent(in) :: n;      !number of states in total model
      integer(4),intent(in) :: ndim;   !state dimension of phit, qt
      real(8),intent(in) :: t;         !current measurement time
      real(8),intent(in) :: tlst;      !time of last measurement
      real(8),intent(in) :: xtr(9);    !true (simulated) values of 9 states (for error print when used with ins_error.f)
      real(8),intent(in) :: phit(ndim,n);!total state transition matrix
      real(8),intent(in) :: qt(ndim,n);!total state process covariance matrix
      real(8),intent(in) :: ht(mdim,n);!partial derivatives of measurements wrt states
      real(8),intent(in) :: y(m);      !measurement vector
      real(8),intent(in) :: sigr(m);   !expected standard deviation of measurements

      integer(4) j,jm;
      real(8) dt,ypre(m);
      logical(4),save :: testf =.true.;!### if true, state & covariance of error analysis filter
                                       ! are compared with standard filter (tst_KF or tst_udKF)
      character(40) pltFile;
      !_______________________________

      if (ndim /=nstate) call errhand ("KFcovar1 error: ndim /= nstate");
      if (n /= nstate) call errhand ("KFcovar1 error: n /= nstate");
      if (mdim > mmaxi) call errhand ("KFcovar1 error: mdim > mmaxi");

      !** filter time update
      dt =t-tlst;

      !** predicted total state (full model). xtot is current total state
      xtota(:n) =matmul(phit(:n,:n),xtot(:n));
      call filt_time_update (xtota,xtot,phit,qt,t,tlst,ndim);

      !** filter measurement update
      ypre(:m) =matmul(ht(:m,:n),xtota(:n));
      call filt_meas_update (ht,y,ypre,sigr,t,m,mdim);
      write (12,1000) t, 304.8*(xfe(:9) -xtr(:9)),                      & !pos/vel in mm
     &                   xfe(7:9) -xtr(7:9);       !tilt in arc-sec
 1000 format (50g13.5)

      if (testf) then;
        !** run standard filter for comparison of state & covariance
!        call tst_udKF (phit,qt,ht,y,sigr,t,xtr,m,mdim,n,ndim)    !###  UD filter
        call tst_KF (phit,qt,ht,y,sigr,t,xtr,m,mdim,n,ndim); !standard covariance Kalman filter
      endif;

      pltFile ="filterSig.plt";
      call write_results (t,pltFile);

      return;
      end subroutine KFcovar1;
!*****************************************************
      subroutine tst_KF (phit,qt,ht,y,sigr,t,xtr,m,mdim,n,ndim);
      !** standard Kalman filter without separate calculation of error terms
      !** (for testing covariance filter).  This filter only works correctly
      !** when all estimated states are located at the beginning of the state vector
      !** with no gaps.  If consider states are present, they must immediately follow
      !** the estimated states.
      implicit none;
      integer(4),intent(in) :: m,mdim,n,ndim;
      real(8),intent(in) :: t,xtr(9);
      real(8),intent(in) :: phit(ndim,n),qt(ndim,n);
      real(8),intent(in) :: ht(mdim,n),y(m),sigr(m);

      integer(4) i,j,k,ier,nf1;
      integer(4),save :: nf;
      integer(4) :: method =2;      !###
      real(8) eps,epsm;
      real(8),allocatable,save :: pf(:,:),pa(:,:),xf(:),xa(:),phip(:,:);
      real(8),allocatable,save :: g(:,:),hp(:,:),dy(:),py(:),py2(:,:);
      real(8),allocatable,save :: a(:,:),b(:,:);
      logical(4),save :: first =.true.;
      !_______________________________

      if (ndim /=nstate) call errhand ("tst_KF error: ndim /= nstate");
      if (n /= nstate) call errhand ("tst_KF error: n /= nstate");
      if (mdim > mmaxi) call errhand ("tst_KF error: mdim > mmaxi");

      if (first) then;
        first =.false.;
        allocate (pf(nstate,nstate),pa(nstate,nstate),                  &
     &       xf(nstate),xa(nstate),g(nstate,mmaxi),phip(nstate,nstate), &
     &       hp(mmaxi,nstate),dy(mmaxi),py(mmaxi*(mmaxi+1)/2),          &
     &       py2(mmaxi,mmaxi),a(nstate,nstate),b(nstate,mmaxi),         &
     &       stat=ier);
        if (ier /= 0) call errhand ("KFcovar1 error: allocate");
        pf(:,:) =0.d0;
        forall (i=1:nstate) pf(i,i) =xstate(i)%sig0**2;
        xf(:) =0.d0;
        nf =0;
        do i=1,n;
          if (xstate(i)%status >= 2) nf =nf+1;
        enddo;
      endif;

      phip(:nf,:nf) =matmul(phit(:nf,:nf),pf(:nf,:nf));
      pa(:,:) =matmul(phip(:nf,:nf),transpose(phit(:nf,:nf)))           &
     &          +qt(:nf,:nf);
      xa(:) =matmul(phit(:nf,:nf),xf(:nf));
!     call prtmx (pa,nstate,nf,nf,"pa");

      !** filter time update
      nf1 =ubound(pae,1);
      write (6,'("pa diff =",10(/10g13.5))') (pa(i,i)-pae(i,i),i=1,nf1);
      write (6,'("xa diff =",10(/10g13.5))') (xa(i)-xfa(i),i=1,nf1);
      hp(:m,:nf) =matmul(ht(:m,:nf),pa(:nf,:nf));
!     call prtmx (hp,mmaxi,m,nf,"hp_tst");

      !** form residual covariance
      k =0;
      do i=1,m;
        do j=1,i;
          k=k+1;
          py(k) =dot_product(hp(i,:nf),ht(j,:nf));
        enddo;
        py(k) =py(k) +sigr(i)**2;
      enddo;
!     call prtsmx (py,m,"py");

      eps =1.d-4;
      call sinv (ier,epsm,  py,  m,eps,.true.);
      if (ier /= 0) call errhand ("KFcovar1 error: sinv failed");
      k =0;
      do i=1,m;
        do j=1,i;
          k=k+1;
          py2(i,j) =py(k);
          py2(j,i) =py(k);
        enddo;
      enddo;

      g(:nf,:m) =matmul(transpose(hp(:m,:nf)),py2(:m,:m));
      do i=1,nf;
        if (xstate(i)%status == 2) g(i,:m) =0.d0;     !zero gain rows for consider states
      enddo;
      dy(:m) =y(:m) -matmul(ht(:m,:nf),xa(:nf));
      xf(:nf) =xa(:nf) +matmul(g(:nf,:m),dy(:m));

      !** measurement update
      if (method == 1) then;
        !** update state covariance using short Kalman form
        !** Pf =(I-G*H)*Pa
        pf(:nf,:nf) =pa(:nf,:nf) -matmul(g(:nf,:m),hp(:m,:nf));   !doesn't match filt_meas_update (diff states 7-9 are 0.8)

      else if (method == 2) then;
        !** bierman's version of joseph
        a(:nf,:nf) =pa(:,:) -matmul(g(:nf,:m),hp(:m,:nf));
        b(:nf,:m) =-matmul(a(:nf,:nf),transpose(ht(:m,:nf)));
!        forall (j=1:m) b(:nf,j) =b(:nf,j) +g(:nf,j)*sigr(j)**2
        pf(:nf,:nf) =a(:nf,:nf) +matmul(b(:nf,:m),transpose(g(:nf,:m))); !matches filt_meas_update within 1e-7
        forall (j=1:m) b(:nf,j) =g(:nf,j)*sigr(j)**2;
        pf(:nf,:nf) =pf(:nf,:nf) +matmul(b(:nf,:m),transpose(g(:nf,:m)));

      else if (method == 3) then;
        !** full joseph for !** update state covariance
        !** Pf =(I-G*H)*Pa*(I-G*H)^T +G*Rmeas*G^T
        !**    =A +[-(A*H^T) +G*Rmeas]*G^T    where A=(I-G*H)*Pa
        a(:nf,:nf) = -matmul(g(:nf,:m),ht(:m,:nf));
        forall (i=1:nf) a(i,i) =a(i,i) +1.d0;
        pf(:nf,:nf) =matmul(matmul(a(:nf,:nf),pa(:nf,:nf)),             &
     &              transpose(a(:nf,:nf)));
        forall (j=1:m) b(:nf,j) =g(:nf,j)*sigr(j)**2;
        pf(:nf,:nf) =pf(:nf,:nf) +matmul(b(:nf,:m),transpose(g(:nf,:m)));   !matches filt_meas_update within 1e-7
      endif;

      !** make symmetric: averaging matches filt_meas_update within 1e-6
      !** making symmetric (without avg) give about the same accuracy: 1e-6
      do i=2,nf;
        do j=1,i-1;
!          pf(i,j) =0.5d0*(pf(i,j)+pf(j,i))
!          pf(j,i) =pf(i,j)
          pf(i,j) =pf(j,i); !works for consider states
        enddo;
      enddo;

!     call prtmx (pf,nstate,nf,nf,"pf");

      write (6,'("pf diff =",10(/10g13.5))') (pf(i,i)-pfe(i,i),i=1,nf1);
      write (6,'("xf diff =",10(/10g13.5))') (xf(i)-xfe(i),i=1,nf1);
      write (20,'(10g13.5)') t, 304.8*(xf(:6)-xtr(:6)),xf(7:9)-xtr(7:9);

      return;
      end subroutine tst_kf;
!*****************************************************
      subroutine tst_udKF (phit,qt,ht,y,sigr,t,xtr,m,mdim,n,ndim);
      !** u-d version of Kalman filter without separate calculation of error terms
      !** (for testing covariance filter). This filter only works correctly
      !** when all model states are estimated.
      implicit none;
      integer(4),intent(in) :: m,mdim,n,ndim;
      real(8),intent(in) :: t;
      real(8),intent(in) :: xtr(9);    !true (simulated) values of 9 states (for error print when used with ins_error.f)
      real(8),intent(in) :: phit(ndim,n),qt(ndim,n);
      real(8),intent(in) :: ht(mdim,n),y(m),sigr(m);

      integer(4) i,j,l,jm,ier,ns2;
      real(8) z,Pz,dz,zres,rvar,vf(n);
      real(8),allocatable,save :: ud(:),uda(:),xf(:),xa(:),phip(:,:);
      real(8),allocatable,save :: g(:),dy(:),pf(:),psym(:);
      real(8),allocatable,save :: d(:),w(:,:),dw(:),v(:),vd(:),f(:),h(:);
      logical(4),save :: first =.true.;
      !_______________________________

      if (ndim /=nstate) call errhand ("tst_KF error: ndim /= nstate");
      if (n /= nstate) call errhand ("tst_KF error: n /= nstate");
      if (mdim > mmaxi) call errhand ("tst_KF error: mdim > mmaxi");

      if (first) then;
        first =.false.;
        ns2 =nstate*(nstate+1)/2;
        allocate (ud(ns2+nstate),uda(ns2+nstate),pf(ns2),               &
     &       xf(nstate),xa(nstate),g(nstate),phip(nstate,nstate),       &
     &       dy(mmaxi),d(nstate),psym(ns2),f(nstate),h(nstate),         &
     &       v(2*nstate),vd(nstate),w(nstate,2*nstate),dw(2*nstate),    &
     &       stat=ier);
        if (ier /= 0) call errhand ("tst_udKF error: allocate");
        ud(:) =0.d0;
        do i=1,nstate;
          ud(i*(i+1)/2) =xstate(i)%sig0**2;
        enddo;
        xf(:) =0.d0;
      endif;

!     write (6,'("starting U-D time update")');
      !** UD filter time update
      call phiu_m (phip,  phit,ndim,n,n,ud,n,nstate);
      l=0;
      do i=1,n;
        l=l+i;
        d(i) =ud(l);
      enddo;

      !** weighted gram-schmidt time update
      w(:n,:n) =phip(:n,:n);
      dw(:n) =d(:n);

      l =0;
      do i=1,n;
        do j=1,i;
          l=l+1;
          psym(l) =qt(j,i);
        enddo;
      enddo;
!      call prtsmx (psym,n,"sqrt-qd")

      call cov2ud_m (psym,n);
      w(:n,n+1:2*n) =0.d0;
      l =0;
      do i=1,n;
        do j=1,i;
          l=l+1;
          w(j,i+n) =psym(l);
        enddo;
        w(i,i+n) =1.d0; !replaces diagonal
        dw(i+n) =psym(l);
      enddo;
!      call prtmx (w,n,n,2*n,"w")
!      call prtmx (dw,1,1,2*n,"dw")

      !*** weighted gram-schmidt orthogonalization
      call wgs_m (ud,v,  w,  n,n,2*n,dw,.false.);

!      call prtmx (w,n,n,2*n,"wgs-w")
!      call prtsmx (ud,n,"wgs-ud")
      call ud2cov_m (pf,  ud,n);

      xa(:) =matmul(phit(:n,:n),xf(:n));
      xf(:) =xa(:);
!      write (6,'("pf,pae ubound =",3i5)') ubound(pf),ubound(pae)
      write (6,'("pa diff =",10(/10g13.5))')                            &
     &      (pf(i*(i+1)/2)-pae(i,i),i=1,nest);


      !** process meas one at a time
      do i=1,m;
        !** UD filter measurement update
        z =0.;
        rvar =sigr(i)**2;
        h(:n) =ht(i,:n);
        call udmes1 (f,g,Pz,dz,  ud,n,rvar,h,xf,z);

!        chiSq =zres**2/Pz
!        if (chiSq > editThresh**2) then
!          write (msgtxt,1100) tobsUTCstr, monStaF(ista)%staNo,zres,chiSq
!          call logmsg ('E230',SVNe,trim(msgtxt))
!          lEdit =2
!          MSresid(iSta,iSat)%edit =2
!          return
!        endif

        call udmes2 (g,ud,xf,  n,Rvar,f,zres);
!        kg(:) =g(:)/Pz
!        hp(:) =g(:)     !save for ??

        !** compute covariance (this is very slow computationally)
        call ud2cov_m (pf,  ud,n);
!        forall (i=1:n) dxsig(i) =sqrt(kg(i)*hp(i))

!        xf(:) =xa(:) -matmul(g(:n,:m),dy(:m))
      enddo;

      do i=1,nest;
        jm =mapxe(i);
        vf(i) =pf(jm*(jm+1)/2);
      enddo;
      write (6,'("pf diff =",10(/10g13.5))')                            &
     &        (vf(i)-pfe(i,i),i=1,nest);   !matches filt_meas_update within 1e-13
      write (6,'("xf diff =",10(/10g13.5))') (xf(i)-xfe(i),i=1,nest);
      write (20,'(10g13.5)') t, 304.8*(xf(:6)-xtr(:6)),xf(7:9)-xtr(7:9);

      return;
      end subroutine tst_udkf;
!*****************************************************
      subroutine KF_init (setupFile,nstateM,ncoreM,nMarkovM,nbiasM,     &
     &                     sig0,xnameM);
      implicit none;
      integer(4),intent(in) :: nstateM,ncoreM,nMarkovM,nbiasM;
      character(*),intent(in) :: setupFile, xnameM(nstateM);
      real(8),intent(in) :: sig0(nstateM);

      integer(4) ier;
!      real(8) t
      !_______________________________

      !** get model size
      nstate =nstateM;
      ncore =ncoreM;
      nmarkov =nMarkovM;
      nbias =nbiasM;
      allocate (mapxe(nState),mapxu(nState),mapxc(nState),              &
     &          mapxei(nState),mapxui(nState),mapxci(nState),           &
     &          xtot(nstate),xtota(nstate),xnames(nstate),              &
     &          xstate(nstate),stat=ier);
      if (ier /= 0) call errhand ("KF_init error: allocate error");
      xnames(:) =xnameM(:);

      !** read setup file to allocate and initialize xstate, phi
      call read_setup (setupFile);
!      call read_meas (nmeas,  measFile)

!      xstate(:nstate)%sig0 =sig0(:nstate)    !### test reset from driver
      call init_filt;

      return;
      end subroutine KF_init;
!*****************************************************
      subroutine read_setup (setupFile);

      implicit none;
      character(*),intent(in) :: setupFile;

      integer(4) ns,ier;
!      real(8)
      character(15) keyword;
      character(80) aline;
      !_______________________________

      open (10,file=setupFile,status='old',iostat=ier);
      if (ier /= 0) call errhand ('error opening setup file');

      do;
        read (10,'(a12,a80)',err=999) keyword,aline;
!        write (6,'(1x,a15,a60)') keyword,aline;
        if (keyword(1:1) == '#') cycle;     !comment

        select case(keyword);
          case ('TEpoch');
            read (aline,*) tepoch;
          case ('TEnd');
            read (aline,*) tEnd;
          case ('StateInit');
            ns =0;
            nest =0;          !total number of estimated states
            nConsider =0;     !total number of consider parameters
            nUnadj =0;        !total number of unadjusted-analyze parameters
            ncoreE =0;       !total number of estimated core states
            nMarkE =0;       !total number of estimated Markov states
            nbiasE =0;        !total number of estimated bias states

            !** read info on states
            do;
              read (10,'(a)',err=999,end=999) aline;
              ns =ns+1;
              if (ns > nstate) exit;
              write (*,'(i3,1x,a)') ns,trim(aline);
              read (aline,*,err=999) xstate(ns)%status,                 &
     &             xstate(ns)%xtype, xstate(ns)%xname, xstate(ns)%x0,   &
     &             xstate(ns)%sig0, xstate(ns)%qs, xstate(ns)%const(2);
              if (xstate(ns)%xname /= xnames(ns)) call errhand          &
     &             ('read_setup error: states names do not match');

              if (xstate(ns)%status == 0) cycle; !skip ignore states
              if (xstate(ns)%status == 1) then;  !unadjusted-analyze state
                if (ns <= ncore+NMarkov) call errhand                   &
     &    ('read_setup error: unadjusted-analyze states must be biases');
                nUnadj =nUnadj +1;
                mapxu(nUnadj) =ns;
                mapxui(ns) =nUnadj;
!                if (xstate(ns)%xtype == 1) then  !core state  !or get from model ??
              else if (xstate(ns)%status == 2) then;  !consider state
                if (ns <= ncore+NMarkov) call errhand                   &
     &             ('read_setup error: consider states must be biases');
                nConsider =nConsider +1;
                mapxc(nConsider) =ns;
                mapxci(ns) =nConsider;
              else if (xstate(ns)%status == 3) then;  !estimated state
                nest =nest +1;
                mapxe(nest) =ns;
                mapxei(ns) =nest;
                if (ns <= ncore+nMarkov) nDynE =nest;
!                if (xstate(ns)%xtype == 1) then  !core state  !or get from model ??
              endif;

            enddo;  !___ end read states loop

            exit;
        end select;

      enddo;

      !** set values and allocate arrays

      close (10);
      write (6,'("nest,nunadj,nconsider =",3i5)') nest,nunadj,nconsider;

      allocate (phiee(nest,nest),phieu(nest,nUnadj),                    &
     &          phiec(nest,nConsider),qe(nDynE,nDynE),stat=ier);
      if (ier /= 0) call errhand                                        &
     &     ('read_setup error: allocate phi error');

      allocate (pae(nest,nest),su(nest,nUnadj), sc(nest,nConsider),     &
     &     sa(nest,nest),pfe(nest,nest),pf0(nest,nest),                 &
     &     pfu(nest,nest), pfc(nest,nest), pfr(nest,nest),              &
     &     pfq(nest,nest),xfa(nest),                                    &
     &     xfe(nest), pu(nUnadj,nUnadj),par(nest,nest),paq(nest,nest),  &
     &     pc(nConsider,nConsider), p0(nest,nest), xu(nUnadj),          &
     &     xc(nConsider), x0(nest),stat=ier);
      if (ier /= 0) call errhand                                        &
     &     ('read_setup error: allocate p,x error');

      allocate (he(mmaxi,nest),hu(mmaxi,nUnadj),hc(mmaxi,nConsider),    &
     &      stat=ier);
      if (ier /= 0) call errhand                                        &
     &     ('read_setup error: allocate h error');
      return;

  999 call errhand ('read_setup error: read setup file');
      end subroutine read_setup;
!*****************************************************
      subroutine read_meas (nmeas,  measFile);

      implicit none;
      character(*),intent(in) :: measFile;
      integer(4),intent(out) :: nmeas;
      integer(4) m,ier;
      character(80) aline;
      !_______________________________

      open (10,file=measFile,status='old',iostat=ier);
      if (ier /= 0) call errhand ('error opening measurement file');

      nmeas =0;
      !** read file once to determine # meas
      do;
        read (10,'(a80)',err=100) aline;
        nmeas =nmeas +1;
      enddo;

  100 if (nmeas == 0) call errhand ('read_meas error: nmeas =0');
      allocate (meas(nmeas),stat=ier);
      if (ier /= 0) call errhand                                        &
     &     ('read_meas error: allocate meas error');

      !** loop to read meas
      nmeas =0;
      do;
        read (10,'(a80)',err=100) aline;
        nmeas =nmeas +1;
!        write (6,'(1x,a15,a60)') keyword,aline;

        read (aline,*,end=999) meas(nmeas)%mtype, m, meas(nmeas)%mname, &
     &       meas(nmeas)%t, meas(nmeas)%y(:m), meas(nmeas)%sigr(:m);
        meas(nmeas)%m =m;
      enddo;

  999 close (10);
      return;
      end subroutine read_meas;
!*****************************************************
      subroutine init_filt;

      implicit none;
      integer(4) i,im;
!      real(8)
      !_______________________________

      pfe(:,:) =0.d0;  !a posteriori covariance of estimated-estimated states (nest,nest)
      pu(:,:) =0.d0;
      pc(:,:) =0.d0;
      p0(:,:) =0.d0;
      par(:,:) =0.d0;
      paq(:,:) =0.d0;
      pfr(:,:) =0.d0;
      pfq(:,:) =0.d0;
      pfu(:,:) =0.d0;
      pfc(:,:) =0.d0;
      xu(:) =0.d0;
      xc(:) =0.d0;
      x0(:) =0.d0;
      sa(:,:)=0.d0;
      su(:,:)=0.d0;
      sc(:,:)=0.d0;
      xtot(:) =0.d0;

      do i=1,nest;
        im =mapxe(i);
        p0(i,i) =xstate(im)%sig0**2;
        pfe(i,i) =p0(i,i);
        x0(i) =xstate(im)%x0;
        xtot(im) =x0(i);
        xfe(i) =x0(i);
        sa(i,i) =1.d0;
      enddo;

      do i=1,nUnadj;
        im =mapxu(i);
        pu(i,i) =xstate(im)%sig0**2;
        xu(i) =xstate(im)%x0;
      enddo;

      do i=1,nConsider;
        im =mapxc(i);
        pc(i,i) =xstate(im)%sig0**2;
        xc(i) =xstate(im)%x0;
      enddo;

      return;
      end subroutine init_filt;
!*****************************************************
      subroutine filt_time_update (xtota,xtot,phit,qt,t,tlst,ndim);
!***    filter time update, including error sensitivity eqns
!      use model
      implicit none;
      integer(4),intent(in) :: ndim;
      real(8),intent(in) :: t,tlst;     !current and last meas time
      real(8),intent(in) :: xtot(nstate); !full state at last meas time
      real(8),intent(in) :: xtota(nstate);!full state at current meas time
      real(8),intent(in) :: phit(ndim,nstate);
      real(8),intent(in) :: qt(ndim,nstate);

      integer(4) i,j,im,jm;
      real(8) dt;
      !_______________________________

      dt =t-tlst;
      call prtmx (phit,ndim,nstate,nstate,"phit");
      call prtmx (qt,ndim,nstate,nstate,"qt");

      do i=1,nest;
        im =mapxe(i);
        xfa(i) =xtota(im);   !set predicted filter state
        do j=1,nest;
          jm =mapxe(j);
          phiee(i,j) =phit(im,jm);
          if (i <= ndynE .and. j <= ndynE) qe(i,j) =qt(im,jm);
        enddo;
        do j=1,nunadj;
          jm =mapxu(j);
          phieu(i,j) =phit(im,jm);
        enddo;
        do j=1,nConsider;
          jm =mapxc(j);
          phiec(i,j) =phit(im,jm);
        enddo;
      enddo;

      call prtmx (sa,nest,nest,nest,"sa");
      call prtmx (phiee,nest,nest,nest,"phiee");

      !** update state covariance partitions
      if (nUnadj > 0)    su(:,:) =matmul(phiee,su) +phieu(:,:);
      if (nConsider > 0) sc(:,:) =matmul(phiee,sc) +phiec(:,:);
      sa(:,:) =matmul(phiee,sa);
      par(:,:) =matmul(matmul(phiee,pfr),transpose(phiee));
      paq(:,:) =matmul(matmul(phiee,pfq),transpose(phiee));
      paq(:nDynE,:nDynE) =paq(:nDynE,:nDynE) +qe(:,:);
      pae(:,:) =matmul(matmul(sa,p0),transpose(sa)) +par(:,:) +paq(:,:);     !filter model without consider errors

      !** form total a priori state covariance for prior, r, and q
!      pfa(:,:) =matmul(matmul(sa,p0),transpose(sa)) +
!     &          matmul(matmul(su,pu),transpose(su)) +
!     &          matmul(matmul(sc,pc),transpose(sc)) +
!     &          par(:,:) +paq(:,:)

      call prtmx (sa,nest,nest,nest,"phi*sa");

      return;
      end subroutine filt_time_update;
!*****************************************************
      subroutine filt_meas_update (ht,y,ypre,rsig,t,m,mdim);
!***    filter measurement update, including error sensitivity eqns
      implicit none;
      integer(4),intent(in) :: m,mdim;
      real(8),intent(in) :: ht(mdim,nstate);
      real(8),intent(in) :: y(m),ypre(m),rsig(m);
      real(8),intent(in) :: t;

      integer(4) i,j,k,jm,ier;
      real(8) eps,epsm,rmeas(m);
      real(8),allocatable :: dy(:),he(:,:),hp(:,:),hu(:,:),hc(:,:);
      real(8),allocatable :: ta(:,:),tc(:,:),tu(:,:),py(:,:),pys(:);
      real(8),allocatable :: tpa(:,:),tpc(:,:),tpu(:,:),g(:,:);
      real(8),allocatable :: hepr(:,:),hepq(:,:),a(:,:),b(:,:);
      !_______________________________

      if (m <= 0) call errhand ("filt_meas_update error: m <= 0");

      allocate (dy(m),he(m,nest),hp(m,nest),hu(m,nUnadj),g(nest,m),     &
     &     hc(m,nConsider),ta(m,nest),tc(m,nConsider),tu(m,nUnadj),     &
     &     py(m,m),pys(m*(m+1)/2),tpa(m,nest),tpc(m,nConsider),         &
     &     tpu(m,nUnadj),hepr(m,nest),hepq(m,nest),                     &
     &     a(nest,nest),b(nest,m),stat=ier);
      if (ier /= 0) call errhand                                        &
     &      ("filt_meas_update error: allocate failed");

      dy(:) =y(:) -ypre(:);
      rmeas(:) =rsig(:)**2;      !assume uncorrelated meas
!      t =meas(imeas)%t
!      rmeas(:m) =meas(imeas)%sigr(:m)**2     !assume uncorrelated meas

      !** map meas partials to estimated, consider and unadjusted
      do i=1,m;
        do j=1,nest;
          jm =mapxe(j);
          he(i,j) =ht(i,jm);
        enddo;
        do j=1,nunadj;
          jm =mapxu(j);
          hu(i,j) =ht(i,jm);
        enddo;
        do j=1,nConsider;
          jm =mapxc(j);
          hc(i,j) =ht(i,jm);
        enddo;
      enddo;

      !** now process measurements

      call prtmx (sa,nest,nest,nest,"sa");
      if (nUnadj > 0) call prtmx (su,nest,nest,nUnadj,"su");
      if (nConsider > 0) call prtmx (sc,nest,nest,nConsider,"sc");

      !** First compute sensitivity of meas residual to constant parameters
      ta(:,:) =-matmul(he,sa);
      tpa(:,:) =matmul(ta,p0);   !residual-state covariances
      hepr(:,:) =matmul(he,par);
      hepq(:,:) =matmul(he,paq);
      call prtmx (he,m,m,nest,"he");
      call prtmx (tpa,m,m,nest,"tpa");

      if (nConsider > 0) then;
        tc(:,:) =-(matmul(he,sc) +hc(:,:));
        tpc(:,:) =matmul(tc,pc);
!       call prtmx (tpc,m,m,nConsider,"tpc");
      endif;
      if (nUnadj > 0) then;
        tu(:,:) =-(matmul(he,su) +hu(:,:));
        tpu(:,:) =matmul(tu,pu);
!       call prtmx (tpu,m,m,nUnadj,"tpu");
      endif;

!     call prtmx (hepr,m,m,nest,"hepr");
!     call prtmx (hepq,m,m,nest,"hepq");
!     call prtmx (-matmul(tpa,transpose(sa)),m,m,nest,"hep0");

      !** form  hp =he*pfa +hc*pc^T*sc^T   where pfa =sa*p0*sa^T +sc*pc*sc^T +par +paq
      hp(:,:) =-matmul(tpa,transpose(sa)) +hepr(:,:) +hepq(:,:);

      !** form meas residual covariance (without pu contribution)
      py(:,:) =matmul(tpa,transpose(ta)) +                              &
     &         matmul((hepr+hepq),transpose(he));
      forall (i=1:m) py(i,i) =py(i,i) +rmeas(i);
      call prtmx (rmeas,1,1,m,"rmeas");

      if (nConsider > 0) then;
        hp(:,:) =hp(:,:) -matmul(tpc,transpose(sc));
        py(:,:) =py(:,:) +matmul(tpc,transpose(tc));
      endif;
!     call prtmx (hp,m,m,nest,"hp");

      !** To invert residual covariance, first copy to symmetric storage
      k =0;
      do i=1,m;
        do j=1,i;
          k =k+1;
!          pys(k) =py(i,j)
          pys(k) =dot_product(tpa(i,:),ta(j,:))                         &
     &           +dot_product(hepr(i,:)+hepq(i,:),he(j,:));
          if (nConsider > 0)                                            &
     &            pys(k) =pys(k) +dot_product(tpc(i,:),tc(j,:));
        enddo;
        pys(k) =pys(k) +rmeas(i);
      enddo;
      call prtsmx (pys,m,"pys");

      !** invert pys
      eps =1.d-5;
      call sinv (ier,epsm,  pys,  m,eps,.true.);
      if (ier /= 0) call errhand                                        &
     &       ("filt_meas_update error: sinv failed");
!     call prtsmx (pys,m,"pys-inv");

      !** g = hp * inv(pfs)
      do i=1,nest;
        call symult (g(i,:),  pys,hp(:,i),m);
      enddo;
      call prtmx (g,nest,nest,m,"g");

      a(:,:m) =transpose(hp)-matmul(g,py);   !checks
      call prtmx (a,nest,nest,m,"phT-g*py");

      !** update sensitivity arrays: sa =(I-g*h)*sa, ...
      sa(:,:) =sa(:,:) +matmul(g(:,:),ta(:,:));    !ta, tc, tu already have (-) sign
      if (nConsider > 0) sc(:,:) =sc(:,:) +matmul(g(:,:),tc(:,:));
      if (nUnadj > 0)    su(:,:) =su(:,:) +matmul(g(:,:),tu(:,:));
!     call prtmx (sa,nest,nest,nest,"(I-kh)*sa");

      !** update state estimate
      xfe(:nest) =xfa(:nest) +matmul(g,dy);
      do i=1,nest;
        jm =mapxe(i);
        xtot(jm) =xfe(i);
      enddo;

      !** update state covariance partitions
      !** Pr =(I-G*H)*Par*(I-G*H)^T +G*Rmeas*G^T
      !**    =A +[-(A*H^T) +G*Rmeas]*G^T    where A=(I-G*H)*Par

      a(:,:) =par(:,:) -matmul(g(:,:),hepr(:,:));
      b(:,:) =-matmul(a,transpose(he));
      forall (j=1:m) b(:,j) =b(:,j) +g(:,j)*rMeas(j);
      pfr(:,:) =a(:,:) +matmul(b,transpose(g));

      !** Pq =(I-G*H)*Paq*(I-G*H)^T
      !**    =A -(A*H^T)*G^T           where A=(I-G*H)*Paq

      a(:,:) =paq(:,:) -matmul(g(:,:),hepq(:,:));
      b(:,:) =-matmul(a,transpose(he));
      pfq(:,:) =a(:,:) +matmul(b,transpose(g));

      !** form total state covariance
      pf0(:,:) =matmul(matmul(sa,p0),transpose(sa));
      if (nConsider > 0) pfc(:,:) =matmul(matmul(sc,pc),transpose(sc));
      if (nUnadj > 0)    pfu(:,:) =matmul(matmul(su,pu),transpose(su));

      pfe(:,:) =pf0(:,:) +pfr(:,:) +pfq(:,:);
      !** add unadjusted error to get total (true) error covariance,
      !** but do not include for comparison in tst_kf (tst_kf only computes filter covariance)
!      if (nunadj > 0) pfe(:,:) =pfe(:,:) +pfu(:,:)    !###
      if (nConsider > 0) pfe(:,:) =pfe(:,:) +pfc(:,:);

      deallocate (dy,he,hp,g,ta,py,pys,tpa);
      deallocate (hepr,hepq,a,b);
      deallocate (hu,tu,tpu,hc,tc,tpc);

      write (11,1000) t, 0.3048*xfe(:6),xfe(7:nest); !convert pos/vel to meters
 1000 format (50g13.5)

      return;
      end subroutine filt_meas_update;
!*****************************************************
      subroutine write_results (t,pltFileName);

      implicit none;
      real(8),intent(in) :: t;   !meas time
      character(*),intent(in) :: pltFileName;

      integer(4) i,j;
      logical(4),save :: first =.true.;
      character(20) :: labels(5,nest);
      real(8) sigx(5,nest);
      !_______________________________

      if (first) then;
        open (10,file=pltFileName,status='unknown');
        first =.false.;
        !**
        do i=1,nest;
          j =mapxe(i);
          labels(1,i) ="sig0-"//xstate(j)%xname(:15);
          labels(2,i) ="sigu-"//xstate(j)%xname(:15);
          labels(3,i) ="sigc-"//xstate(j)%xname(:15);
          labels(4,i) ="sigr-"//xstate(j)%xname(:15);
          labels(5,i) ="sigq-"//xstate(j)%xname(:15);
        enddo;

        write (10,1000) "t (s)", (trim(labels(1,i)),trim(labels(2,i)),  &
     &   trim(labels(3,i)),trim(labels(4,i)),trim(labels(5,i)),i=1,nest);
 1000   format('variables =',200('"',a,'", '))
      endif;

      do i=1,nest;
        sigx(1,i) =sqrt(pf0(i,i));
        sigx(2,i) =sqrt(pfu(i,i));
        sigx(3,i) =sqrt(pfc(i,i));
        sigx(4,i) =sqrt(pfr(i,i));
        sigx(5,i) =sqrt(pfq(i,i));
!        sigx(:,i) =(/ pf0(i,i), pfu(i,i), pfc(i,i), pfr(i,i), pfq(i,i)/)  !### test
      enddo;
      sigx(:,1:3) =0.3048*1.d3*sigx(:,1:3);    !change units from ft to mm
      sigx(:,4:6) =0.3048*1.d3*sigx(:,4:6);    !change units from ft to microm/s
      write (10,1100) t,(sigx(:,i),i=1,nest);
 1100 format (200g13.5)

      return;
      end subroutine write_results;
!*****************************************************
      subroutine KFcovar (setupFile,measFile);
      implicit none;
      character(*),intent(in) :: setupFile, measFile;
      integer(4) i,j,jm,ier,m;
      real(8) t,tlst,dt;
      real(8),allocatable :: ht(:,:), ypre(:),phit(:,:),qt(:,:);
      !_______________________________

      !** get model size
      call modelInit (nstate,ncore,nMarkov,nbias);

      allocate (mapxe(nState),mapxu(nState),mapxc(nState),              &
     &          mapxei(nState),mapxui(nState),mapxci(nState),           &
     &          xtot(nstate),xtota(nstate),xnames(nstate),stat=ier);
      if (ier /= 0) call errhand ("KFcovar eror: allocate error");

      allocate (ht(mmaxi,nstate),ypre(mmaxi),                           &
     &         phit(nstate,nstate),qt(nstate,nstate),stat=ier);
      if (ier /= 0) call errhand ("KFcovar eror: allocate y error");


      !** get state names
      call stateNames (xnames);

      !** read setup file to allocate and initialize xstate, phi
      call read_setup (setupFile);

      call read_meas (nmeas,  measFile);

      call init_filt;

      t =tepoch;
      do i=1,nmeas;
        tlst =t;
        t =meas(i)%t;
        if (t > tend) exit;
        dt =t-tlst;
        m =meas(i)%m;

        call get_phiq (phit,qt,xtota,  xtot,t,tlst,nstate);

        !** filter time update
        call filt_time_update (xtota,xtot,phit,qt,t,tlst,nstate); !??? check if xt or xf
        do j=1,nest;
          jm =mapxe(j);
          xfe(j) =xtota(jm);
        enddo;

        call get_hy (ht,ypre,t,m,mmaxi,nstate);

        !** filter measurement update
        call filt_meas_update (ht,meas(i)%y,ypre,meas(i)%sigr,t,m,mmaxi);
        call write_results (t,"filterSig.plt");
      enddo;

      return;
      end subroutine KFcovar;
!*******************************************************
      subroutine get_phiq (phit,qt,xtota,  xtot,t,tlst,nst);
      !** dummy routine to avoid link error
      implicit none;
      integer(4),intent(in) :: nst;
      real(8),intent(in) :: t,tlst;       !current and last meas time
      real(8),intent(in) :: xtot(nst);      !state at last meas time
      real(8),intent(out) :: phit(nst,nst);
      real(8),intent(out) :: xtota(nst);    !state at current meas time
      real(8),intent(out) :: qt(nst,nst);
      !_________________________________________

      xtota(:) =0.d0;
      phit(:,:) =0.d0;
      qt(:,:) =0.d0;

      return;
      end subroutine get_phiq;
!*******************************************************
      subroutine get_hy (ht,ypre,  t,m,mdim,nst);
      !** dummy routine to avoid link error
      implicit none;
      integer(4),intent(in) :: m;
      integer(4),intent(in) :: mdim,nst;
      real(8),intent(in) :: t;
      real(8),intent(out) :: ht(mdim,nst);
      real(8),intent(out) :: ypre(m);
      !___________________________________________

      ht =0.d0;
      ypre(:) =0.d0;

      return;
      end subroutine get_hy;
!************************************************************
      subroutine modelInit (nstate,ncore,nMarkov,nbias);
      !** dummy routine to avoid link error
      implicit none;
      integer(4),intent(out) :: nstate,ncore,nMarkov,nbias;
      !___________________________________________

      nstate =0;
      ncore =0;
      nMarkov =0;
      nbias =0;

      return;
      end subroutine modelInit;
!************************************************************
      subroutine stateNames (xnames);
      !** dummy routine to avoid link error
      implicit none;
      character(*),intent(out) :: xnames(nstate);
      !___________________________________________

      xnames ="";
      return;
      end subroutine stateNames;

      end module;
!************************************************************
      subroutine errhand (message);
      character(*),intent(in) :: message;
      write (6,'(a)') trim(message);
      write (*,'(a)') trim(message);
      stop "error";
      end subroutine errhand;
