!** Program INS_error analyzes the effects of five error sources on Kalman
!** filter estimates using the approach described in Section 9.3.1.

!** The five error sources are initial condition errors, un-estimated bias
!** parameters, consider bias parameters, measurement noise and process noise.
!** The simulated system is the 27-state model for errors of an inertial
!** navigation system as described in Section 3.3.   The input measurement
!** file taimes.dat is generated for a simulated missile flight using
!** program sim_INS.

!** Text file taiState.ctl defines the state names, initial values,
!** initial 1-s uncertainty, process noise and status (unadjusted, consider, adjusted).

!** The output plot files are
!**  "filtErr.plt" (filter estimate errors versus time as generated by a standard Kalman filter),
!**  "KFcovarEst.plt" (error analysis filter state estimates),
!**  "KFcovarErr.plt" (error in error analysis filter state estimates), and
!**  "filterSig.plt" (breakdown of 1-s filter uncertainty for each of the five error sources.)

!** Program INS_error is used in Example 9.3-1.
!** It calls module covarAnal which contains various subroutines used for the
!** error analysis.  Covar_Anal is structured to be general-purpose and thus
!** may be used for other systems.

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**        in ifort.cfg
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort INS_error.f90 covar_Anal.obj    (libkf_bnd_ifort.lib should be linked by default)
!**   INS_error insState_e15c3.ctl >INSerror_e15c3.lis
!**************************************

      module modelDat;
      implicit none;
      integer(4),parameter :: ndim=27, n2dim=(ndim*(ndim+1))/2, mdim=3;

      real(8) phi(ndim,ndim),q(n2dim),h(mdim,ndim),sigr(mdim),y(mdim);
      real(8) x(ndim),sig0(ndim),q1(45),q2(ndim,ndim);
      real(8) xt(9),at(3),f(3,ndim);

      end module modelDat;
!*****************************************************
      program INS_error;
      use modelDat;
      use covarMod,only: kf_init,KFcovar1;
      implicit none;
      integer(4),parameter :: maxjtim=90,maxjmp=20,nparm=1230,maxbias=30;
      character(80) setupFile;
      integer(4) ncore,nMarkov,nbias,nmeas;
      integer(4) i,j,k,j1,j2,l,n,nx,nt,m,nq,nimub,ier;
      integer(4) lenferr,nargs;
      real(8) :: tmax =800.d0;
      real(8) :: t,tlst,rsum;
      character(80) line;
      character(20) :: xname(ndim) = (/ "delPx", "delPy", "delPz",      &
     &  "delVx", "delVy", "delVz", "tiltx", "tilty", "tiltz",           &
     &  "accBx", "accBy","accBz", "accSFEx", "accSFEy", "accSFEz",      &
     &  "GBdrftx", "GBdrfty","GBdrftz", "GSFEx", "GSFEy", "GSFEz",      &
     &  "Gsenx", "Gseny", "Gsenz", "G2senx", "G2seny", "G2senz" /);
      !_____________________________________________________________

      setupFile ="insState_e15c3.ctl";       !"insState_e15.ctl", "insState_e18.ctl", "insState_e27.ctl"

      lenferr =nargs()-1;        !get number of command line arguments
! ....  Get run control file name
      if (lenferr == 0) then;
        write (6,'(" Input run state control file:")');
        read (5,'(a40)') setupFile;
      else;
        call getarg(1,line,ier);   !get command line argument
        read (line,*) setupFile;
        if (ier < 0) stop 'ERROR IN RETRIEVING setupFile';
      endif;
      write (6,'(" Input run state control file =",a)') trim(setupFile);

!**     open output plot files

      open (20,file="filtErr.plt",status='unknown');
      write (20,1010);
 1010 format ('variables ="time", "delP-x", "delP-y", "delP-z", '       &
     &    '"delV-x", "delV-y", "delV-z", "delT-x", "delT-y", "delT-z"')
      open (11,file="KFcovarEst.plt",status='unknown');
      write (11,1020);
      open (12,file="KFcovarErr.plt",status='unknown');
      write (12,1020);
 1020 format ('variables ="time", "delP-x", "delP-y", "delP-z", '        &
     &    '"delV-x", "delV-y", "delV-z", "delT-x", "delT-y", "delT-z"')

!**     input file of measurement data
      open(30,file='taimes.dat',status='old', form='unformatted');
      phi(:,:) =0.d0;
      q(:) =0.d0;
      t =0.d0;
      nmeas =0;

!**      read header of initial standard deviations
      read(30) sig0;
      write (6,'("sigx0 =",5(/10g13.5))') sig0(:ndim);

      ncore =9;
      nMarkov =0;       !no Markov stets in model
      nbias =0;         !filter does not separate biases
      call kf_init (setupFile,ndim,ncore,nMarkov,nbias,                 &
     &               sig0,xname);

!****     time loop for reading and filtering measurement data
      do while (t < tmax);
        read(30,end=999) t,nt,m,nq,y(1:m),sigr(1:m);
!        print *,t,nt,m
        !**  set for bias case
        n =nt-nbias;       !number of dynamic states
        nx =n;
        nbias =nt-nx;      !reset nbias
        nmeas =nmeas+m;
        if (nt > ndim .or. m > mdim) stop 'invalid nt or m';

        read(30) xt(:),at(:),(f(1:3,j),j=7,ndim);  !at is accel error, not total
        h(:,:) =0.d0;
        do i=1,m;
          read(30) j1,j2,h(i,j1:j2);
          if (j1 < 1 .or. j1 > nt .or. j2 < 1 .or. j2 > nt) stop 40;
        enddo;

        q(:) =0.d0;
        do i=1,nq;
          read(30) l,q(l);
          if (l < 1 .or. l > n2dim) stop 43;
        enddo;

        phi(:,:) =0.d0;
        do i=1,nt;
          read(30) j1,j2,(phi(i,j),j=j1,j2);
          if (j1 < 1 .or. j1 > nt .or. j2 < 1 .or. j2 > nt) stop 41;
        enddo;

!        write(6,*) 't =',t
!        call prtmx(y,1,1,m,'y       ')
!        call prtmx(r,1,1,m,'rsig    ')
!        call prtmx(h,mdim,m,n,'h       ')
!        call prtmx(phi,ndim,n,n,'phi     ')
!        call prtsmx(q,n,'q       ')

        !** standard filter
        call filter (t,phi,q,h,sigr,y,sig0,n,nt,nx,m,ndim,mdim,rsum);

        !** error analysis filter
        k =0;
        do i=1,ndim;
          do j=1,i;
            k =k+1;
            q2(i,j) =q(k);
            q2(j,i) =q(k);
          enddo;
        enddo;
        call KFcovar1 (phi,q2,h,y,sigr,t,tlst,xt,m,mdim,ndim,ndim);
      enddo;   !___  end of filter loop

  999 write(6,'(" rsum =",g12.5,", nmeas =",i4)')  rsum,nmeas;
      nimub=0;

      stop;
      end program;
! *****************************************************************
      subroutine filter (t,phi,q,h,r,y,sig0,n,nt,nx,m,ndim,mdim,  rsum);
!         kalman filter subroutine with auxilliary eqns to compute
!         sensitivity to jump states

      implicit none;
      integer(4),intent(in) :: n,nt,nx,m,ndim,mdim;
      real(8),intent(in) :: t,sig0(n),phi(ndim,nt),q(n*(n+1)/2);
      real(8),intent(in) :: h(mdim,nt),r(m),y(m);
      real(8),intent(out) :: rsum;

      integer(4) i,j,k,l,im;
      integer(4),save :: ifirst =1;
      real*4 temp;
      real(8) sum,c,yp,dz;
      integer(4),parameter :: nn=27, n2dim=(nn*(nn+1))/2, mm=3;

      real(8),save :: p(n2dim),x(nn);
      real(8),save :: d1(nn),d2(nn,nn),d3(nn),hp(nn),g(nn);
      !_____________________________________

      if (ifirst > 0) then;
        ifirst=0;
        rsum=0.;
        x(:) =0.d0;
        p(:) =0.d0;
        !** initialize p using sig0
        l=0;
        do i=1,n;
          l=l+i;
          p(l)=sig0(i)**2;
        enddo;
      endif;

!         time update
      call smult (d1,  phi,x, n,n,1, ndim,n,n);
      x(:n) =d1(:n);

!         p=phi*p*phi**t+q
      do i=1,n;
        d1(:n) =phi(i,:n);
        call symult (d2(1,i),  p,d1,n);
      enddo;

      l=0;
      do i=1,n;
        do j=1,i;
          l=l+1;
          p(l) =q(l) +dot_product(phi(i,:n),d2(:n,j));
        enddo;
      enddo;

!          propagate state sensitivity to jump parameters
!          add new parameters at current time
!      call jumptime (phi,t,nt,nx,m,ndim)    !###

!         measurement update (scalar)
      do im=1,m;
        yp=0.;
        d1(:nt) =h(im,:nt);
        yp =dot_product(h(im,:n),x(:n));
        dz =y(im)-yp;

        call symult (hp,  p,d1,n);
        c =dot_product(hp(:n),h(im,:n)) +r(im)**2;
        rsum =rsum+dz**2/c;
!        write(6,'(" t, rsum = ",f10.2,g12.5)') t, rsum

        l=0;
        do i=1,n;
          g(i) =hp(i)/c;
          x(i) =x(i)+g(i)*dz;
          do j=1,i;
            l=l+1;
            p(l) =p(l)-g(i)*hp(j);
          enddo;
        enddo;

!           update jump sensitivities
!        call jumpmes (d1,g,c,dz,im,nt,nx)  !###
      enddo;

      return;
      end;

