!** filter_smoother.f is a general-purpose nonlinear filter-smoother that uses sparse matrix
!** operators (for large problems), implements pass-disposable bias parameters and

!**  Module modelMod contains the model-specific information and filter_smoother is the driver for
!**  the filter/smoother calculations.

!**  Notes:
!**    1) This smoother version has not been fully tested for ndynam < nestf when
!**       dynamic model (xdot) is a function of the biases.

!** A "keyword-type" control input text file sets control parameters, sets initial state values.
!** The acceptable keyword options are:

!  1) MeasFile   - defines the input file name of measurements
!  2) TEpoch     - sets the start time of measurement processing
!  3) Tend       - sets the end time of measurement processing
!  4) Rsig       - defines the measurement noise 1-sigma for a given sensor (first input #)
!  5) editThresh - measurement residual editing threshold (e.g., 5)
!  6) Debug      - sets debug print flags for up to 10 different options (only index 1, 2, and 4 are currently used)
!  7) StateInit  - denotes that the following records will input state initial values:
!                   index = state index
!                   name = state name as hard-coded in modelMod.
!                   x0 = initial state value at TEpoch
!                   sig0 = 1-sigma uncertainty
!                   Qs = power spectral density of process noise
!                   const = constant associated with the state (e.g, Markov process feedback constant)
!  8) endState   - terminates StateInit input

!Sample input file (acTrk.ctl) for testing aircraft tracking

!# state setup file for 8-state (x, y, z, xdot, ydot, zdot, tanphi, a-acc)
!#  to be used for testing the filter_smoother program.
!#  Measurements are radar range, azimuth and elevation.
!# Notes: 1) Lines beginning with # are ignored
!#        2) First 12 characters of lines (except stateInit data) are a keyword.
!#           Accepted values are TEpoch, TEnd, MeasFile, Rsig, editThresh, Debug, StateInit,
!#           endState. Other inputs are read as freeform (blanks or comma separator).
!#
!#
!TEpoch       0.
!TEnd         125.
!MeasFile     acMeas.txt
!Rsig          1    1.0   0.0002   0.0002       !RAE noise
!#Debug          T T T T T T T T T T
!StateInit    (index  name   x0    sig0    Qs      const)
!   1                 x      1100.  100.   0.0       0.
!   2                 y      10.0   100.   0.0       0.
!   3                 z      330.0  100.   0.0       0.
!   4                 xdot   20.0    20.   .001      0.
!   5                 ydot   0.0     20.   .001      0.
!   6                 zdot   0.0     0.6   .08       0.
!   7                 tanphi 0.0     0.5   .01       0.
!   8                 a-acc  0.0     0.63  .40       0.
!endState
!   6                 zdot   0.0     0.6   .144      5.   for Markov process option
!   7                 tanphi 0.0     0.5   .05      10.
!   8                 a-acc  0.0     0.63  .08      10.

!**  Search for "###" to find parameters that are typically changed
!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort filter_smoother.f90 modelMod.obj      (libkf_bnd_ifort.lib should be linked by default)
!**   filter_smoother acTrk.ctl >filter_smoother.lis
!**   filter_smoother acTrk_markov.ctl >filter_smoother.lis    (when using markov process model)
!**********************************************************************
      module filtSmoMod
      !** contains global variables used in filter-smoother

      implicit none;
      integer(4) irecs;                   !direct access file record number for smoothing
      logical(4) :: lprint(10) =.false.   !controls debug print
      logical(4) :: udfilt =.false.       !if true, use factored UD filter rather than covariance:
                                          !can be reset in read_setup
      logical(4) :: integCov =.true.      !if true, smoothed covariance is computed (can be slow)
      real(8) tstart,tend;                !start and end times for filter processing
      real(8) :: editThresh =5.           !Kalman filter editing threshold

      end module filtSmoMod
!**************************************
      program filter_smoother;
      !** general purpose filter smoother
      use modelMod;
      use filtSmoMod, only:irecs,tend,integCov
      use filterPrec, only: real_kind1;   !precision (4 or 8)
      implicit none;

      integer(4) i,nd,im,isensor,nrec,ibias,lenrec,ierr;
      integer(4) lenferr,nargs,ier,ledit;
      character(100) setupfile,measFile,line;
      real(8) ym(mmeas),tobs,tLst,dt,zsig,chiSq,logLike;
      real(8) Rvar;
      !____________________________________________________

      !** filter_smoother will only work in double precision
      if (real_kind1 /= 8) call errhand ('error: real_kind1 /= 8')

! ....  Read run control file
      lenferr =nargs()-1;         !get number of command line arguments
      if (lenferr == 0) then;
        write (6,'(" Input setup file:")');
        read (5,'(a40)') setupFile;
      else;
        call getarg(1,line,ier);    !get command line argument
        read (line,*) setupFile;
        if (ier < 0) call errhand ('ERROR IN RETRIEVING setupFile');
      endif;

      !** Read text control file to set the xstate structure.  Note that variables
      !** nEstf and xnames must have been set in module modelMod before calling setup.
      call read_setup (measFile,  setupFile);
      write (6,'("t0 =",g13.5,", tend =",g13.5)') t0,tend;

      open (10,file=measFile,status='old')
      open (20,file='Filt_X.plt',status='unknown');
      write (20,1100);
 1100 format('VARIABLES ="t", "x-error", "y-error", "z-error", ',       &
     &   '"xdot-error", "ydot-error", "zdot-error", "tanphi-error", ',  &
     &   '"accel-a-error"')
      open (21,file='Filt_sig.plt',status='unknown');
      write (21,1110);
 1110 format('VARIABLES ="t", "sig-x", "sig-y", "sig-z", "sig-xd", ',   &
     &   '"sig-yd", "sig-zd", "sig-tanphi", "sig-acca"')

      !** smoothRTS
      nd =ndynamEst
      lenrec =2*(nstate +(nd+2)*nestf +(nestf*(nestf+1))/2              &
     &       +(nd*(nd+1))/2 +2);  !record length in 4-byte wordsxref,xa,xf,pf,phi,qd;
      open (42,form='unformatted',access='direct',recl=lenrec);
      irecs =0;

      !** scratch direct access for AC track
      open (43,form='unformatted',access='direct',recl=2+2*nEstf);

      !** sequential smoothed track file
      open (44,file='smoothTrk.dat',form='binary',status='unknown',     &
     &        iostat=ierr);
      if (ierr /= 0)                                                    &
     &        call errhand('error opening smoothed track file');
      write (44) nEstf;

      tobs =0.d0;
      tLst =tobs;
      logLike =0.d0;

      !** initialize Kalman filter xf, pf
      call init_filter

      !*** process measurements in filter
      do while (tobs < tend);
        read (10,*,end=999) tobs,isensor,ym(:)
        ibias =0;
        dt =tobs-tLst

        if (dt > 0.d0) then;
          !** filter time update
          call KF_timeUp (xa,pa,  xf,pf,tobs,tLst);
          irecs =irecs+1;
          write (42,rec=irecs) tobs,tLst,xref,xa,xf,pf,phi,qd;
          xf(:) =xa(:)
          pf(:) =pa(:)
        endif

        !** now measurement update: one component-at-a-time
        do im=1,mmeas;
          Rvar =Rsig(im,isensor)**2
          call KF_measUP (zsig,chiSq,ledit,   logLike,xf,pf,             &
     &                     ym(im),Rvar,im,isensor,tobs,ibias);
        enddo;   !___ end measurement loop

        tLst =tobs;
        write (20,'(11g13.5)') tobs,xf(:);
        write (21,'(11g13.5)') tobs,(sqrt(pf(i*(i+1)/2)),i=1,nEstf);
      enddo;   !__ end measurement read loop

  999 continue;

      !***  execute RTS smoother
      IntegCov =.true.;         !compute smoothed covariance
      nrec =irecs;
      call RTSsmoother (tobs,nrec,integCov);

      stop;
      end program;
!*******************************************************************
      subroutine KF_timeUp (xa,pa,  xf,pf,tobs,tKLst);
!**    Kalman filter time update
      use modelMod, only:ud,psym,xref,nEstF,nF3,nd3,nDynam,phi,qd,       &
     &                  get_phiq;
      use filtSmoMod, only:lprint,udfilt;
      implicit none;

      real(8),intent(in) :: tobs,tKLst;      !current & last time
      real(8),intent(in) :: xf(nEstF)        !a posteriori filter state vector at tLst
      real(8),intent(in) :: pf(nF3);         !a posteriori filter state error covariance
                                             ! at tLst (stored as a vector)
      real(8),intent(out) :: xa(nEstF)       !a priori filter state vector at tobs
      real(8),intent(out) :: pa(nF3);        !a priori filter state error covariance at tobs

      !** local variables
      integer(4) i,j,k,l,ierr;
      integer(4),parameter :: method =2;  !UD time update method  ###
      real(8) delt;
      real(8),allocatable,save :: phip(:,:),pp(:,:),phipt(:,:);
      real(8),allocatable,save :: d(:),d2(:,:),v(:),vd(:),w(:,:),dw(:);
      !_______________________________________________________

      if (.not.allocated(phip)) then;
        write (6,'("KF_timeUp allocate: nestf =",i4,", ndynam =",i4,    &
     &              ", tobs =",g13.5)') nestf,ndynam,tobs
        allocate (phip(nEstF,nEstF),pp(nEstF,nEstF),          &
     &          phipt(nEstf,nEstf),d(nEstF),d2(ndynam,nEstF),           &
     &          v(2*nEstF),vd(nEstF),w(nEstF,2*nEstF),                  &
     &          dw(2*nEstF),stat=ierr);
        if (ierr > 0) call errhand ('KF_timeUp allocate failed');
      endif;

      delt =tobs -tKlst;
      if (lprint(1)) write (6,'(/"****************"/"KF_timeUp: tobs =", &
     &          f15.4,", delt (sec) =",g13.5)') tobs, delt;

      !***  time update

!***   Compute xa (predicted state), phi and q.
!***   The following code (or equivalent) should appear in get_phiq
!      call smult (xa,  phi,xf, ndynam,nEstf,1, ndynam,nEstf,nEstF);   !xa =phi*xf
!      if (nEstf > ndynam) xa(ndynam+1:) =xf(ndynam+1:);

      call get_phiq (phi,qd,xa,   xf,delt,nEstf,ndynam);

      if (lprint(2)) then;
        call prtmx (phi,ndynam,ndynam,nEstF,"phi");
        call prtsmx (qd,ndynam,"qd");
      endif

      if (lprint(4)) then;
        call prtmx (xa,1,1,nEstF,"xf_i-1");
        if (udFilt) then
          call ud2cov_m (pf,  ud,nEstF);
          call prtsmx (pf,nEstF,"pf_i-1");
        else
          call prtsmx (pf,nEstF,"pf_i-1");
        endif
      endif;

      if (udFilt) then;
        !** UD filter time update
        call phiu_m (phip,  phi,ndynam,ndynam,nEstF, ud,nEstF,nEstf);

        if (nEstf > ndynam) then;
          phip(ndynam+1:nEstf,:) =0.d0;
          do i=ndynam+1,nEstf;
            k =(i*(i-1))/2 +ndynam;
            do j=ndynam+1,i;
              k=k+1;
              phip(j,i) =ud(k);
            enddo;
            phip(i,i) =1.d0;
          enddo;
        endif;

        l=0;
        do i=1,nEstF;
          l=l+i;
          d(i) =ud(l);
        enddo;

        if (method == 1) then;
          !** compute phi*u, square ud factors, add qd, and re-factor
          !** (squaring is not as accurate as WGS)
          call triphu (ud,v,  phip,vd,   nEstF,nEstF,d);
          call ud2cov_m (psym,  ud,nEstF);
          l =0;
          do i=1,nEstF;
            do j=1,i;
              l=l+1;
              if (i <= ndynam) psym(l) =psym(l) +qd(l);
              ud(l) =psym(l);
            enddo;
          enddo;
          call cov2ud_m (ud,nEstF);

        else;
          !** weighted gram-schmidt  (could modify to work with ndynam partition)
          w(:,1:nEstF) =phip(:,:);
          dw(1:nEstF) =d(:);
          psym(:) =0.d0;
          psym(:nd3) =qd(:);
!          call prtsmx (psym,nEstf,"sqrt-qd")

          call cov2ud_m (psym,nEstf);
          w(:,nEstF+1:2*nEstF) =0.d0;
          l =0;
          do i=1,nEstF;
            do j=1,i;
              l=l+1;
              w(j,i+nEstf) =psym(l);
            enddo;
            w(i,i+nEstf) =1.d0; !replaces diagonal
            dw(i+nEstf) =psym(l);
          enddo;
!          call prtmx (w,nEstF,nEstF,2*nEstF,"w")
!          call prtmx (dw,1,1,2*nEstF,"dw")

          !*** weighted gram-schmidt orthogonalization
          call wgs_m (ud,v,  w,  nEstF,nEstF,2*nEstF,dw,.true.);

!          call prtmx (w,nEstF,nEstF,2*nEstF,"wgs-w")
!          call prtsmx (ud,nEstf,"wgs-ud")
          call ud2cov_m (pa,  ud,nEstF);
        endif;

      else;
        !** Kalman filter covariance time update
        psym(:) =pf(:);     !save a posteriori from last step
        k =0;
        do i=1,nEstF;
          do j=1,i;
            k =k+1;
            pp(i,j) =pf(k);
            pp(j,i) =pf(k);
          enddo;
        enddo;

        !**  phip =phi * pp
        call smult (phip,  phi,pp,                                      &
     &              ndynam,nEstf,nEstF,  ndynam,nEstF,nEstF);
        !** phipt =pp * phi^T
        phipt(:nEstf,:ndynam) =transpose(phip(:ndynam,:nEstf));
        if (nEstF > ndynam) then;
          phipt(:nEstf,ndynam+1:nEstf)                                  &
     &                        =transpose(pp(ndynam+1:nEstf,:nEstf));
        endif
        !** w =phi * phipt = phi * pp * phi^T
        call smult (w,  phi,phipt, ndynam,nEstf,nEstf,                  &
     &                             ndynam,nEstF,nEstF);
        if (nEstF > ndynam) then;
          !** copy bias portion to w
          w(ndynam+1:nEstF,ndynam+1:nEstF) =                            &
     &               phipt(ndynam+1:nEstF,ndynam+1:nEstF);
        endif;

        !**   pa = phi * pp * phi^T +qd
        k =0;
        do i=1,nEstF;
          !** make symmetric using upper triangular partition of phip and qd
          do j=1,i;
            k =k+1;
            pa(k) =w(j,i);
            if (k <= nd3) pa(k) =pa(k) +qd(k);
          enddo;
        enddo;

      endif;

      if (lprint(4)) then;
        call prtmx (xa,1,1,nEstF,"xa-");
        if (udFilt) then
          call ud2cov_m (pa,  ud,nEstF);
          call prtsmx (pa,nEstF,"pa-");
        else
          call prtsmx (pa,nEstF,"pa-");
        endif
      endif;

      return;
      end subroutine KF_timeUp;
!*******************************************************************
      subroutine KF_measUP (zsig,chiSq,ledit,   logLike,xf,pf,          &
     &                     zmeas,Rvar,meastyp,isensor,tobs,ibias);
!****   Kalman filter scalar measurement update

      use modelMod, only:ud,nEstF,nF3,phi,qd,get_yh;
      use filtSmoMod, only:lprint,udfilt,editThresh;
      implicit none;

      integer(4),intent(in) :: meastyp;    !measurement type for given sensor
      integer(4),intent(in) :: isensor;    !measurement sensor
      integer(4),intent(in) :: ibias;      !index in xf of meas bias (not used)
      real(8),intent(in) :: zmeas;         !measurement
      real(8),intent(in) :: Rvar;          !measurement noise variance
      real(8),intent(in) :: tobs;          !obs time
      real(8),intent(inout) :: xf(nEstF)     !filter state vector
      real(8),intent(inout) :: pf(nF3);      !filter state error covariance matrix stored as a vector
      real(8),intent(inout) :: logLike;    !-log(likelihood)
      real(8),intent(out) :: zsig;         !1-sigma of measurement residual
      real(8),intent(out) :: chiSq;        !measurement residual Chi-squared
      integer(4),intent(out) :: ledit;     !measurement editing flag: 0=accept, 1=below elev cutoff, 2=chisq edit

      !** local variables
      integer(4) i,j,k,im;
      logical(4) :: covar =.true.;     !slow, but required for smoothing
      real(8) hp(nEstF),h(nEstF),kg(nEstF),Pz;
      real(8) f(nEstf),g(nEstF),dxSig(nEstf),z,zp,dz,zres;
      !_______________________________________________________

      im =meastyp;
      lEdit =0;   !default is accept meas

!     call prtmx (xf,1,1,nEstF,"xp-");
!     write (6,'(" obs time=",f17.7)') tobs;

!***   get predicted meas., residual and partials (h)
      call get_yh (zp,h,  xf,isensor,meastyp);
      zres =zmeas -zp;

!***    measurement update
      lEdit =0;

      if (udFilt) then;
        !** UD filter measurement update
        z =0.;
        call udmes1 (f,g,Pz,dz,  ud,nEstf,Rvar,h,xf,z);

        chiSq =zres**2/Pz;
        if (chiSq > editThresh**2) then;
          write (6,1100) tobs, isensor,meastyp,zres,chiSq;
          lEdit =2;
          return;
        endif;

        call udmes2 (g,ud,xf,  nEstF,Rvar,f,zres);
        kg(:) =g(:)/Pz;
        hp(:) =g(:);    !save for ??

        if (covar .or. lprint(4)) then;
          !** compute covariance (this is very slow computationally)
          call ud2cov_m (pf,  ud,nEstF);
          do i=1,nEstF;
            dxsig(i) =sqrt(kg(i)*hp(i));
          enddo;
        endif;

      else;
        !** covariance Kalman filter update
        call sysmult (hp,  pf,h,nEstF);         !sparse symmetric mult: hp =pf*h
        Pz =Rvar +dot_product(hp,h);
        kg(:) =hp(:)/Pz;
        chiSq =zres**2/Pz;
!        fstat%rrms =fstat%rrms +chiSq;

        if (chiSq > editThresh**2) then;
          write (6,1100) tobs, isensor,meastyp,zres,chiSq;
 1100     format("edit large meas residual at ",g16.8,                      &
     &           ", sensor ",i3,", meas ",i3,", zres =",g13.5,              &
     &            ", chi-sq =",g13.5)
          lEdit =2;
          return;
        endif;

        xf(:) =xf(:) +kg(:)*zres;
        k =0;
        do i=1,nEstF;
          dxsig(i) =kg(i)*hp(i);
          if (dxsig(i) >= 0.d0) then;
            dxsig(i) =sqrt(dxsig(i));
          else;
            write (6,'("error: dxvar(",i4,") =",g13.5)') i,dxsig(i);
            dxsig(i) =0.d0;
          endif;
          do j=1,i;
            k =k+1;
            pf(k) =pf(k) -kg(i)*hp(j);
          enddo;
        enddo;
      endif;

      zsig =sqrt(Pz);
      chiSq =zres**2/Pz;
      logLike =logLike +0.5d0*(chiSq +log(Pz) +1.837877d0);   !(1.837877 =ln(2*pi))

      if (lprint(1)) write (6,'(" meastyp ",i3,", residual =",g13.5,    &
     &         ", resid sigma =",g13.5,", Rsig=",g13.5)')    &
     &         isensor*10+meastyp, zres, zsig, sqrt(Rvar);
!     call prtmx (xf,1,1,nEstF,"xp+");

      if (lprint(4)) then;
        call prtmx (h,1,1,nEstF,"h");
        call prtmx (xf,1,1,nEstF,"xf+");
        call prtsmx (pf,nEstF,"pf+");
      endif;

      return;
      end subroutine KF_measUP;
!*****************************************************
      subroutine read_setup (measFile, setupFile);
!***   Reads text file of model and state configuration data.
!**    Additional keywords may be added for additional features.

      use modelMod, only:nEstf,ndynam,nd3,t0,xnames,                    &
     &                   xstate,defaultp,rsig,nsensor,mmeas;
      use filtSmoMod, only:lprint,udfilt,editThresh,tend;
      implicit none;
      character(*),intent(in) :: setupFile; !name of input setup file
      character(*),intent(out) :: measFile;  !name of measurement file

      integer(4) i,ns,nst,ier,isensor;
      character(15) keyword;
      character(80) aline;
      !_______________________________

      open (10,file=setupFile,status='old',iostat=ier);
      if (ier /= 0) call errhand ('error opening setup file');
      measFile ='';
      lprint(:) =.false.
      editThresh =5.

      !** set default parameters for filter
      call defaultp

      do;
        read (10,'(a12,a80)',end=990,err=999) keyword,aline;
        if (keyword(1:1) == '#' .or. keyword == '') cycle;     !comment
        write (*,'(a12,a)') keyword,trim(aline);

        select case(keyword);
          case ('TEpoch');
            read (aline,*) t0;
          case ('TEnd');
            read (aline,*) tEnd;
          case ('MeasFile');
            read (aline,*) measFile;
          case ('editThresh');
            read (aline,*) editThresh;
          case ("Rnoise")
            read (aline,*,err=999) isensor, rsig(:,isensor);
          case ("Debug");
            read (aline,*) lprint(:)
          case ("UDfilt")
            UDfilt =.true.;
          case ('StateInit');
            ns =0;
            write (6,'("Input state initialization from file ",a)')      &
     &         trim(setupFile);

            !** read info on states
            do;
              read (10,'(a)',err=999,end=997) aline;
              if (index(aline,"endState") > 0) exit;
              write (*,'(a)') trim(aline);
              ns =ns+1;
              if (ns > nEstF) exit;
              read (aline,*,err=999) i, xstate(ns)%xname, xstate(ns)%x0, &  !xstate(ns)%status, xtype
     &             xstate(ns)%sig0, xstate(ns)%Qs,xstate(ns)%const(1);
              if (xstate(ns)%xname /= xnames(ns)) then;
                call errhand                                            &
     &             ('read_setup error: states names do not match');
              endif;
            enddo;  !___ end read states loop

          case default;
            write (6,'("invalid input name ",a)') keyword;
        end select;

      enddo;

  990 continue
      if (ns /= nEstf) then
        write (6,'(/"input error: number of input states =",i4,           &
     &         ", model has ",i4)') ns, nEstf
        call errhand ('read_setup error: # states wrong');
      else if (nst > 0 .and. nst /= nEstf) then
        write (6,'(/"input error: number of input true states =",i4,      &
     &         ", model has ",i4)') nst, nEstf
        call errhand ('read_setup error: # states wrong');
      endif

  997 close (10);
      write (6,'(/"Finished reading setup file:"/"# states =",i4)')       &
     &       nestF;
      write (6,'(/)')
      return;

  999 call errhand ('read_setup error: read setup file');
      end subroutine read_setup;
!*********************************************************
      subroutine integx (xa,  xf,dt,n);
!***    integrates state in small intervals for improved accuracy
      use modelMod, only:nState,ndynam,get_f;
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(in) :: dt;         !time step (s)
      real(8),intent(out) :: xa(n);     !predicted state at t+dt

      real(8) xdot(nState),f(n,n),dti,t;
      real(8) :: dtmax =0.2;        !### will be problem dependent (make shorter than dominant time constants)
      !_______________________________________

      if (n /= nState) stop 'integx error: n /= nState';

      t =0.d0;
      if (dt <= 0.) stop 'integx error: dt <= 0';
      xa(:) =xf(:);

      do while (t < dt);
        dti =min(dtmax,dt-t);
        t =t+dti;
        call get_F (f,xdot,   xa,n,ndynam);
        xa(:) =xa(:) +dti*xdot(:);
      enddo;
!      write (6,'("integx t =",g13.5)') t

      return;
      end subroutine integx;
!***********************************************************************
      subroutine RTSsmoother (tFinal,nrec,IntegCov);
!***   RTS fixed-interval smoother.

      use modelMod, only:xref,xa,xf,pf,phi,qd,mmax,                    &
     &              nState,ndynam,nestF,nF3,nDynamEst;
      use filtSmoMod, only:irecs;
      implicit none;

!  .... passed variables
      integer(4),intent(in) :: nrec;       !index of last record in unit 40
      logical(4),intent(in) :: IntegCov;   !if true, smoothed covariance is computed and
                                           ! printed in file sigout.dat
      real(8),intent(inout) :: tFinal;     !time of 1st ob (min from J2k)

!  ....  local
      integer(4),parameter :: maxbias =100;     !max # of reconstructed biases allowed in this step
      integer(4) i,j,k,ier,nd;    !,irow,icol,nb,isat,i1,i2
      real(8),allocatable :: phid(:,:),g(:,:),pbias(:,:);
      real(8),allocatable :: t1(:,:),t2(:,:),t3(:),dx(:),qds(:),t4(:,:);
      real(8),allocatable :: pbs(:),xb(:),ps(:),pas(:),sigx(:);
      real(8) tobsK,tKLst;
      !__________________________________________________________

      nd =nDynamEst;
      allocate (phid(nDynamEst,nEstF),g(nDynamEst,nestF),           &
     &         t1(nestF,nestF),t2(nestF,nestF),t3(nestF),dx(nestF),     &
     &         pbs(nF3),xb(nestF),sigx(nestF),ps(nF3),pas(nF3),         &
     &         qds(nDynamEst*(nDynamEst+1)/2),pbias(nestF,maxbias),     &
     &         t4(nestF,nestF),stat=ier);
      if (ier > 0) call errhand ('RTSsmoother allocate failed');

      !** initialize smoother state & cov to last filter state & cov
      xb(:) =xf(:);
      pbs(:) =pf(:);

!      write (6,'(/"smoother residuals"/
!     &           "  time       station   mtype     filter N/E",
!     &            "              smoother  N/E")')
      write (6,1070);
 1070 format ('TITLE ="smoother output"'/'VARIABLES ="t", '              &
     &   '"xb-xref", "yb-yref", "zb-zref", "xdb-xrefd", ',               &
     &   '"ydb-yrefd", "zdb-zrefd", "tanphib-tanphit", "accab-accaref"')

!***   open plot file for saving state estimates
      open (24,file='smo_x.plt',status='unknown');
      write (24,1130);
 1130 format('VARIABLES ="t", "x-error", "y-error", "z-error", ',       &
     &   '"xdot-error", "ydot-error", "zdot-error", "tanphi-error", ',  &
     &   '"accel-a-error"')

      if (IntegCov) then;
        open (22,file='Smo_sig.plt',status='unknown');
        write (22,1100);
 1100   format('TITLE="smoothed state 1sigma"')
          write (22,1120);
 1120   format('VARIABLES ="t","sig-xb", "sig-yb", "sig-zb", ',         &
     &   '"sig-xdb", "sig-ydb", "sig-zdb", "sig-tanphib", "sig-accab"')
      endif;

      !*** process filter data backward in time
      irecs =nrec+1;

      do;
        !** read filter variables
        irecs =irecs-1;
        if (irecs <= 0) exit;
        read (42,rec=irecs) tobsK,tKLst,xref,xa,xf,pf,phi,qd;

        !** check that phi(:ndynam,nd+1:nest) =0: As written, RTS smoother code does not
        !** allow phi(:ndynam,nd+1:nest) /=0 ,but can be easily modified to work.  ###
!        do i=nd+1,nEstf
!          do j=1,nd
!            if (phi(j,i) /= 0.d0) call errhand                          &
!     &        ("RTSsmoother error: phi(:ndynam,nd+1:nest) =0")
!          enddo
!        enddo

        !** copy a posteriori covariance to 2-d array for multiply
        k=0
        do i=1,nEstf
          do j=1,i
            k =k+1
            t1(j,i) =pf(k);
            t1(i,j) =pf(k);
          enddo
        enddo

        !**  form pa =pf(i+1/i) using sparse arrays
        call smult (t4,    phi,t1,  nd,nEstF,nEstf,  nd,nEstf,nEstf)  !t4(:nd,:nestF) =matmul(phi(:nd,:nestF),pf(:nestF,:nestF));
        t1(:nEstf,:nd) =transpose(t4(:nd,:nEstf))
        call smult (t2,    phi,t1,  nd,nEstF,nd, nd,nEstf,nEstf)      !t2(:nd,:nd) =matmul(phi(:nd,:nest),t1(:nest,:nd))));

        k=0;
        do i=1,nEstF;
          do j=1,i;
            k =k+1;
            if (i <= nd) then;
              qds(k) =qd(k);
              pas(k) =t2(j,i) +qds(k);
            else if (j <= nd) then;
              pas(k) =t4(j,i);
            else;
              pas(k) =pf(k);
            endif;
          enddo;
        enddo;
        call prtmx (phi,ndynamEst,ndynamEst,nEstF,'phi');
!        call prtsmx (pf,nEstF,'pf')
!        call prtsmx (pas,nEstF,'pas')

        if (abs(tobsK-tKlst) <= 1.d-6) then;    !should not happen
          stop 'error: tobsK =tKlst';
          !** xa =xf, pa =pf
          k =0;
          do i=1,nestF;
            if (xa(i) /= xf(i)) stop 'xa(i) /= xf(i)';
            k =k+i;
            if (pas(k) /= pf(k)) stop 'pas(l) /= pf(l)';
          enddo;
          cycle;     !bypass xb & pb computations and write to files
        endif;

!        call prtsmx (pas,nEstF,'pas+')
!        call prtmx (pbias,nEstF,nEstF,nb,'pbias')
        pbs(:) =pbs(:) -pas(:); !compute before pas inversion

        !*** compute smoother gain g  (same results with RTSgain0 or RTSgain1,
        !*** but RTSgain1 run time is much smaller than RTSgain0 for large problems)
        phid(:nd,:nEstf) =phi(:nd,:nEstf);
        call RTSgain1 (g,  pas,t1,t2,  phid,qds);

!        write (6,'("xa =",13g13.5)') xa(:)
!        write (6,'("xf =",13g13.5)') xf(:)
!        write (6,'("xb =",13g13.5)') xb(:)
        dx(:) =xb(:)-xa(:);
        xb(1:nd) =xf(1:nd) +matmul(g,dx);
        !** xb(nDynamEst+1:) is unchanged (not smoothable)
        write (6,'(f14.5,5(/10g13.5))') tKlst/1440.d0, xb(1:nd);

        !** write smoothed states: compute total state estimate
!        xref(:nestF) =xref(:nestF) +xb(:nestF)     !need to use mapx if missing states
        write (43,rec=irecs) tKlst, xref(:nestF);
        write (24,'(11g13.5)') tKlst, xb(:nestF) -xref(:nestF);

        if (IntegCov) then;
          !** compute smoother covariance
          !** t2(:,:) =(pb-pa)*transpose(g)
          !** Note: pbs(:) =pbs(:) -pas(:) from above
          t1(1:nd,:) =g(:nd,:);
          t1(nd+1:,:) =0.d0;
          forall (i=nd+1:nEstF) t1(i,i) =1.d0;

          !** t2 =pbs*t1^T
          do i=1,nestF;
            t3(:) =t1(i,:);
            call sysmult (t2(:,i),  pbs,t3,nestF);
          enddo;

          !** pbs =pf +t1*t2
          k =0;
          do i=1,nestF;
            t3(:) =t1(i,:);
            call smult (dx,  t3,t2, 1,nEstF,i, 1,nestF,1);
            do j=1,i;
              k =k+1;
              pbs(k) =pf(k) +dx(j);
            enddo;
            sigx(i) =sqrt(pbs(k));
          enddo;
!          call prtsmx (pbs,nEstF,'pbs+')

          write (6,'(40x,13g13.5)') sigx(1:nestF);
          !*** write error sigma to file sigout.dat
          write (22,'(11g13.5)') tKlst,sigx(:);
        endif;

      enddo;   !____ end backward time loop

      close (42,status='delete');
      if (IntegCov) close (22);

      !** now write track file in forward order
      do irecs=1,nrec;
        read (43,rec=irecs) tKlst, xref(:);
        write (44) tKlst, xref(:);
      enddo;
      close (43,status='delete');
      close (44);

      xf(:) =xb(:);
      tFinal =tKLst;

      deallocate (phid,g,t1,t2,t3,dx,pbs,xb,sigx,ps,pas,qds,            &
     &            pbias);

      return;
      end subroutine RTSsmoother;
!*****************************************************************
      subroutine RTSgain1 (g,  pas,t1,t2,  phid,qds);
!***   computes RTS smoother gain by computing
!***   G =inv(phi)*[I-Q*inv(P(i+1/i))].  This version gives the same results
!***   as the version using G =P(i/i)*phi^T*inv(P(i+1/i)

      use modelMod, only:nEstF,nF3,nDynamEst;
      implicit none;

      real(8),intent(in) :: phid(nDynamEst,nEstF);      !state transition matrix (assumes that there is no
                                                            ! coupling between biases and dynamic states in phi)
      real(8),intent(in) :: qds(nDynamEst*(nDynamEst+1)/2); !process noise matrix
      real(8),intent(inout) :: pas(nF3);                    !p(i+1/i) [inverse on output]
      real(8),intent(inout) :: t1(nDynamEst,nEstF);         !scratch array (actually dimensioned as nEstF,nEstF, but we don't need all)
      real(8),intent(inout) :: t2(nDynamEst,nDynamEst);     !scratch array (actually dimensioned as nEstF,nEstF, but we don't need all)
      real(8),intent(out) :: g(nDynamEst,nestF);            !RTS gain for dynamic states

      !** local
      logical(4) :: check =.false.;
      integer(4) i,j,k,nbias,nr,nd,ier;
      real(8) t5(nDynamEst,nEstF),qd(nDynamEst,nDynamEst);   !explicitly allocate or pass  !!!
      real(8) eps,err,epsm;
      !_______________________________________________________

      !*** invert pas = filter P(i+1/i)
      eps =1.d-5;
      call sinv (ier,epsm,  pas,  nestF,eps,.false.);
      if (ier < 0) stop 'pas inversion error';

      !** temporarily set g =inv(pas) and q =-qds
      k =0;
      do i=1,nestF;
        do j=1,i;
          k=k+1;
          if (j <= nDynamEst) g(j,i) =pas(k);
          if (i <= nDynamEst) then;
            g(i,j) =pas(k);
            qd(i,j) =-qds(k);
            qd(j,i) =-qds(k);
          endif;
        enddo;
      enddo;

      !** t1 =I+qd*g =I-qds*inv(pas)
      call smult (t1,  qd,g, nDynamEst,nDynamEst,nEstF,                  &
     &                    nDynamEst,nDynamEst,nDynamEst);
      forall (i=1:nDynamEst) t1(i,i) =t1(i,i) +1.d0;

      !** RTS gain gn =inv(phid)*(I-qd*inv(pp))
      !** first compute t2 =inv(phid(1,1))            !SLOW: should invert independent diagonal blocks separately !!!!
      !** and t5 =inv(phid(1,1))*phid(1,2)
      t2(:,:) =phid(:,:nd);

      nd =nDynamEst;
      nbias =nEstf-nd;
      if (nbias > 0) then;
        !** t5 =phid(1,2)
        t5(:,:nbias) =phid(:,nd+1:nEstf);
        nr =nbias
      else
        t5(:,1) =0.d0
        nr =1
      endif
      call matInvGJ (ier,  t2,t5, nDynamEst,nDynamEst,nr);
      if (ier > 0) stop 'inversion error on phid';

      if (check) then;
        !** check inversion
        qd(:,:nd) =matmul(t2,phid(:,:nd));
        do i=1,ndynamEst;
          do j=1,nDynamEst;
            err =qd(i,j);
            if (j == i) err =err-1.d0;
            if (abs(err) > 1.d-13)                                      &
     &        write(6,'("phid inversion error: i,j =",2i4,              &
     &              ", phid*inv(phid) =",g13.5)') i,j,err;
          enddo;
        enddo;
!        call prtmx (t2,nDynamEst,nDynamEst,nDynamEst,"inv(phid)")
      endif;

      !** finally g =t2*t1 =inv(phi)*(I-qd*inv(pp))
      call smult (g,  t2,t1, nDynamEst,nDynamEst,nEstF,                 &
     &                     nDynamEst,nDynamEst,nDynamEst);
      if (nbias > 0) then
        g(:,nd+1:nEstf) =g(:,nd+1:nEstf) +t5(:,:nbias)
      endif
      if (check) call prtmx (g,nDynamEst,nDynamEst,nEstF,"g");

      return;
      end subroutine RTSgain1;
!***************************************************************
      subroutine zeroRowCol (v,  p, j,n);
!**     sets the j-th row & column of symmetric, upper-triangular stored array p to 0
!**     and returns the original row/colum in vector v
      implicit none;
      integer(4),intent(in) :: n;       !dimension of p
      integer(4),intent(in) :: j;       !row/column to zero
      real(8),intent(inout) :: p(n*(n+1)/2);    !symmetric array p
      real(8),intent(out) :: v(n);      !original j-th row/column

      !** local
      integer(4) i,k;
      !___________________________________________

      k =((j-1)*j)/2;
      v(1:j) =p(k+1:k+j);
      p(k+1:k+j) =0.d0;
      k =k+j;
      do i=j+1,n;
        v(i) =p(k+j);
        p(k+j) =0.d0;
        k =k+i;
      enddo;

      return;
      end subroutine zeroRowCol;
