!**  Module modelMod contains the model-specific information for an 8-state aircraft
!**  tracking problem (states  x, y, z, xdot, ydot, zdot, tanphi, a-acc).
!**  Units are m and m/s.

!**  Modify this module as necessary for other problems and models.

!**  Search for "###" to find parameters that are typically changed

!**  The filter is coded with the following assumptions:
!**    1) filter state is initialized with a diagonal covariance
!**    2) measurement noise is assumed independent (R is diagonal)

!**************************************
! The author grants the user a non-exclusive, worldwide, royalty-free copyright license to

! 1. reproduce and modify the software for your own purposes, and
! 2. to distribute the software provided that you give credit to the author,
!    do not make claims against the author or Wiley-Interscience,
!    and mark any modifications as your own.  If the software is incorporated in
!    a commercial product, you agree to defend and indemnify this author and
!    Wiley-Interscience against any losses, damages and costs arising from claims,
!    lawsuits and other legal actions brought by a third party.

! The software is provided on an as is basis, without warranties or conditions
! of any kind, either express or implied (including any warranties or conditions
! of title, non-infringement, merchantability or fitness for a particular purpose).
! The user is solely responsible for determining the appropriateness of using and
! distributing the program and assumes all risks associated with its exercise of
! rights under this agreement.
!**************************************

!** sample compile and execute using Intel compiler:

!**     Include in ifort.cfg:
!**   /check:bounds /traceback /include:\kf_book\distribSWm\library /link /nodefaultlib:msvcrt.lib
!**     \kf_book\distribSWm\library\libkf_bnd_ifort.lib

!**     To compile and execute:
!**   set ifortcfg=\kf_book\distribSWm\ifort.cfg
!**   ifort modelMod.f90 /c
!**   ifort filter_smoother.f90 modelMod.obj
!**   filter_smoother acTrk.ctl >filtSmo.lis
!**************************************
      module modelMod;
      !** contains model-specific information for filter_smoother
      !**
      integer(4),parameter :: nState =8;    !number of states in full model
!      integer(4),parameter :: nState =10;    !number of states in full model  !### w biases
      integer(4),parameter :: nEstf =nState;!number of filter states
      integer(4),parameter :: ndynam =8;    !number of dynamic states in full model (must be located before biases, ndynam <= nState)
      integer(4),parameter :: ndynamEst =ndynam;  !number of estimated dynamic states (must be located before biases)
      integer(4),parameter :: nf3 =(nEstf*(nEstf+1))/2;
      integer(4),parameter :: nd3 =(ndynam*(ndynam+1))/2;
      integer(4),parameter :: nsensor =2;  !number of possible sensors
      integer(4),parameter :: mmeas =3;    !number of measurements at one time
      integer(4),parameter :: mmax =3000;  !maximum total number of meas time samples

      real(8) phi(ndynam,nEstf);          !state transition matrix
      real(8) qd(ndynam*(ndynam+1)/2);    !discrete state noise matrix
      real(8) xf(nEstf);                  !a posteriori filter state
      real(8) xref(nState);               !reference state (set to 0 in most problems)
      real(8) xa(nEstf);                  !a priori filter state
      real(8) pf(nf3);                    !a posteriorifilter state covariance
      real(8) pa(nf3);                    !a priori filter state covariance
      real(8) psym(nf3);                  !filter state covariance
      real(8) ud(nf3);                    !filter state covariance   ??? check if n*(n+3)/2 ??

      real(8) xt0(nEstf);                 !true initial state
      real(8) tau(nEstf);                 !time constants (s) for 1st-order markov time constants
      real(8) :: t0 =0.d0;                !initial state epoch (sec)

      character(20) :: xnames(nEstf) =(/"x", "y", "z", "xdot", "ydot",  &  !filter state names
     &                                  "zdot", "tanphi", "a-acc"/);

!      character(20) :: xnames(nEstf) =(/"x", "y", "z", "xdot", "ydot",  &  !filter state names ### w biases
!     &                  "zdot", "tanphi", "a-acc", "bias1", "bias2"/);

      logical(4) :: markov =.false.;      !true denotes that zdot, tanphi and a-acc states      !###
                                          ! are modeled as 1st-ordeer Markov processes
      integer(4) :: nmeas;                !actual number of measurements

      real(8) :: sig0(nEstf)              !initial filter state 1-sigma
      real(8) :: x0(nEstf)                !initial filter state estimate
      real(8) :: Rsig(mmeas,nsensor)      !meas noise 1-sigma for range and bearing
      real(8) :: xsensor(3,nsensor)       !sensor positions
      real(8) :: gacc = 9.80d0;           !m/s^2

      type measDats
        integer(4) isensor       !antenna, instrument, etc
!        integer(4) mtype         !meas type for given sensor
        real(8)  time            !meas time
        real(8)  y(mmeas)        !meas value
        real(8)  sigm(mmeas)     !meas noise 1-sigma
      end type
      type (measDats) measDat(mmax)

      type states;
!        integer(4) xtype;       !1=core state, 2=bias  (not used in filter_smoother)
!        character(8) status;    !must be "estimate"  (not used in filter_smoother)
        character(20) xname;    !name of state that must match name in hard-coded model
        real(8) x0;             !a priori estimate at epoch time
        real(8) sig0;           !1-sigma uncertainty of x0 at epoch time
        real(8) Qs;             !power spectral density of process noise
        real(8) const(2);       !known constants associated with this state in dynamic or measurement model
      end type;
      type (states) :: xstate(nEstf);  !state data used for filter
      type (states) :: xstateTmp(nEstf); !temporary copy of state data

      contains;
!*********************************************************************
      subroutine defaultp
      !** sets default parameters for filter
      implicit none
      !_______________________________________________

      xsensor(:,1) = (/ 0.d0, 0.d0, 0.d0 /)    !sensor 1 position
      xsensor(:,2) = (/ 5.d0,  3.d0, 0.d0 /)   !sensor 2 position

      Rsig(:,1) =(/ 1., 0.2e-3, 0.2e-3 /);  !true meas noise 1-sigma (RAE), sensor 1
      Rsig(:,2) =Rsig(:,1)                 !true meas noise 1-sigma, sensor 2

      !** set xstate values for this case ###
      xstate(:)%x0 =(/ 1100.d0, 10.d0, 330.d0,  20.d0, 0.d0, 0.d0,      &
     &                 0.d0, 0.d0/)
      xstate(:)%sig0 =(/ 100.d0, 100.d0, 100.d0, 20.d0, 20.d0, 0.6d0,   &
     &                0.50d0, 0.63d0 /); !prior 1-sigma
      xstate(:)%Qs =(/0.d0, 0.d0, 0.d0, 0.001d0, 0.001d0, 0.08d0, 0.01d0, 0.4d0 /)

!      xstate(:)%x0 =(/ 1100.d0, 10.d0, 330.d0,  20.d0, 0.d0, 0.d0,      & !### w biases
!     &                 0.d0, 0.d0, 0.d0, 0.d0/)
!      xstate(:)%sig0 =(/ 100.d0, 100.d0, 100.d0, 20.d0, 20.d0, 0.6d0,   & !###
!     &                0.50d0, 0.63d0, 2.d-3, 2.d-3 /); !prior 1-sigma
!      xstate(:)%Qs =(/0.d0, 0.d0, 0.d0, 0.001d0, 0.001d0, 0.08d0,       & !###
!     &                0.01d0, 0.4d0, 0.d0, 0.d0 /)

      xstate(:)%const(1) =0.d0

      if (markov) xstate(6:8)%const(1) =(/ 5., 10., 10. /);
      xref(:) =0.d0

      return
      end subroutine defaultp
!**********************************************************************
      subroutine init_filter
      !** initializes Kalman filter state and covariance
      implicit none

      integer(4) i,k
      !_____________________________________________________________

      pf(:) =0.d0;
      xf(:) =xstate(:)%x0;
      k =0
      do i=1,nEstf;
        k =k+i
        pf(k) =xstate(i)%sig0**2;
!        qd(i) =xstate(i)%Qs
        tau(i) =xstate(i)%const(1)
        if (markov .and. i >= 6) then
          if (tau(i) <= 0.d0)                                           &
     &         call errhand ("tau(6:8) = 0 when using markov option")
          xstate(i)%Qs =2.d0*pf(k)/tau(i);
        endif
      enddo;

      ud(:) =pf(:);

      end subroutine init_filter
!********************************************************************
      subroutine get_yh (yp,h,  xf,isensor,im);
      !** compute measurements, residuals, partials and measurement noise variance
      !** from given filter state vector, xparm vector and actual measurement (if available)
      implicit none;
      integer(4),intent(in) :: im;     !meas index (1 or 2)
      integer(4),intent(in) :: isensor;!sensor (1 to 2)
      real(8),intent(in) :: xf(nEstf);  !filter target state vector
      real(8),intent(out) :: yp;       !model-predicted measurement
      real(8),intent(out) :: h(nEstf); !measurement partials wrt x

      real(8) dx(3),r,r2,az,el,rh,rh2;
      !_______________________________________

      dx(:) =xf(:3) -xsensor(:,isensor);
      r2 =sum(dx(1:3)**2);
      r =sqrt(r2);
      rh2 =dx(1)**2 +dx(2)**2;
      rh =sqrt(rh2);
      az =atan2(dx(1),dx(2));
      el =asin(dx(3)/r);

      h(:) =0.d0;
      select case(im)
        case (1)     !range
          yp =r;
          h(:3) =(/ dx(1), dx(2), dx(3) /)/r;
!          yp =r +xf(9);          !### w biases
!          h(9) =1.d0             !###
        case (2)     !azimuth
          yp =az;
          h(:3) =(/ dx(2), -dx(1), 0.d0 /)/rh2;
        case (3)     !elevation
          yp =el;
          h(:3) =(/ -dx(1)*dx(3), -dx(2)*dx(3), rh2 /)/(rh*r2);
      end select

      return;
      end subroutine get_yh;
!***********************************************************
      subroutine get_meas (measFile);
!** reads measurements from file measFile and loads structure measDat
      implicit none
      character(*),intent(in) :: measFile;

      real(8) t,yp(mmeas);
      !____________________________________________

      open (10,file=measFile,status='old');
      nmeas =0;

      !** time loop
      do;
        read (10,*,end=999,err=999) t,yp(:);
        nmeas =nmeas+1;
        measDat(nmeas)%time =t;
        measDat(nmeas)%isensor =1;
        measDat(nmeas)%y(:) =yp(:);
        measDat(nmeas)%sigm(:) =(/ 1., 0.2e-3, 0.2e-3 /);  !RAE noise; or rnoise
      enddo;

  999 close (10);
      return;
      end subroutine get_meas;
!**********************************************************************
      subroutine get_F (F,xdot,   xf,n,nd);
!***    computes state transition and process noise matrix and integrates state
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      integer(4),intent(in) :: nd;      !# dynamic states (ros of F)
      real(8),intent(in) :: xf(n);     !target state
      real(8),intent(out) :: F(nd,n); !state transition matrix
      real(8),intent(out) :: xdot(nd);     !steta derivative

      integer(4) i;
      real(8) ac,aa,vh,vh2;
      !_______________________________________

      if (n /= nState) stop 'get_f error: n /= nState';

      ac =gacc*xf(7);   !crosstrack accel
      aa =xf(8);        !alongtrack accel
      vh2 =xf(4)**2 +xf(5)**2;
      vh =sqrt(vh2);
      xdot(:3) =xf(4:6);
      xdot(4) =(ac*xf(5) +aa*xf(4))/vh;
      xdot(5) =(-ac*xf(4) +aa*xf(5))/vh;
      xdot(6:8) =0.d0;
      if (markov) forall (i=6:8) xdot(i) =-xf(i)/tau(i);

      f(:,:) =0.d0;
      forall (i=1:3) f(i,i+3) =1.d0;
      f(4,4:8) =(/  aa -xdot(4)*xf(4)/vh, ac -xdot(4)*xf(5)/vh,         &
     &            0.d0,  gacc*xf(5), xf(4) /) /vh;
      f(5,4:8) =(/ -ac -xdot(5)*xf(4)/vh, aa -xdot(5)*xf(5)/vh,         &
     &            0.d0, -gacc*xf(4), xf(5) /) /vh;
      if (markov) forall (i=6:8) f(i,i) =-1.d0/tau(i);

      return;
      end subroutine get_F;
!*********************************************************
      subroutine get_phiq (phi,qd,xa,   xf,dt,n,nd);
!***    computes state transition and process noise matrix and integrates state
      implicit none;
      integer(4),intent(in) :: n;       !state dimension
      integer(4),intent(in) :: nd;      !number of dynamic parameters (row dimension of phi and qd)
      real(8),intent(in) :: xf(n);      !filter state
      real(8),intent(in) :: dt;         !time step (s)
      real(8),intent(out) :: xa(n);     !predicted state at t+dt
      real(8),intent(out) :: phi(nd,n); !state transition matrix
      real(8),intent(out) :: qd(nd*(nd+1)/2);   !state process noise matrix

      integer(4) i,j,k;
      real(8) xdot(nd),f(nd,n),phii(nd,n),qd2(nd,nd),dti,t;
!      real(8) :: dtmax =1.1  !###
      real(8) :: dtmax =0.2;     !little difference between 0.1 and 0.2 for AC tracking problem
      !_______________________________________

      if (n /= nState) stop 'get_phiq error: n /= nState';
!      write (6,'("get_phiq dt =",g13.5/"xf =",10g13.5)') dt,xf(:);

      t =0.d0;
      if (dt <= 0.) stop 'get_phix error: dt <= 0';
      xa(:) =xf(:);
      phi(:,:) =0.d0;
      forall (i=1:nd) phi(i,i) =1.d0;
      qd2(:,:) =0.d0

      do while (t < dt);
        dti =min(dtmax,dt-t);
        t =t+dti;
        call get_F (f,xdot,   xa,n,nd);
!        call prtmx (f,nd,nd,n,"F")
        phii(:,:) =dti*f(:,:);
!        phii(:,:) =dti*(f(:,:) +0.5*dti*matmul(f,f))        !### same results for AC tracking
        forall (i=1:nd) phii(i,i) =phii(i,i) +1.d0;
        phi(:,:) =matmul(phii(:,:nd),phi(:,:));
        if (n > nd) then
          phi(:,nd+1:n) =phi(:,nd+1:n) +phii(:,nd+1:n);
        endif
!        call prtmx (phi,n,n,n,"phii")
        qd2(:,:)=matmul(matmul(phii(:,:nd),qd2),transpose(phii(:,:nd)));
        forall (i=1:nd) qd2(i,i) =qd2(i,i) +xstate(i)%Qs*dti;
        xa(:nd) =xa(:nd) +dti*xdot(:);
      enddo;

      !** copy qd2 to qd
      k =0;
      do i=1,nd;
        do j=1,i;
          k=k+1;
          qd(k) =qd2(j,i);
        enddo
      enddo

      return;
      end subroutine get_phiq;
!************************************************************8
      end module modelMod;